/*
 * Copyright (C) 2013 Michael Pippig
 *
 * This file is part of ScaFaCoS.
 * 
 * ScaFaCoS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ScaFaCoS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *	
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include "FCSCommon.h"
#include "part_derive_one_over_norm_x.h"

/* This code was generated with Mathematica. */
static void sort(
    fcs_int *i, fcs_int *j,
    fcs_float *x, fcs_float *y
    )
{
  if(*i < *j){
    fcs_int   tmpi = *i; *i = *j; *j = tmpi;
    fcs_float tmpx = *x; *x = *y; *y = tmpx;
  }
}

/* This code was generated with Mathematica. */
fcs_float ifcs_p2nfft_part_derive_one_over_norm_x(
    fcs_int i, fcs_int j, fcs_int k,
    fcs_float x, fcs_float y, fcs_float z
    )
{
  /* Canonicalize to i >= j >= k due to symmetry */ 
  sort(&i, &j, &x, &y);
  sort(&i, &k, &x, &z);
  sort(&j, &k, &y, &z);

  fcs_float t = fcs_sqrt(1/(x*x+y*y+z*z));
  fcs_float u = t*x;
  fcs_float v = t*y;
  fcs_float w = t*z;

  fcs_float u2 = u*u;
  fcs_float v2 = v*v;
  fcs_float w2 = w*w;

  switch(i){
    case 0 : switch(j){
      case 0 : switch(k){
        case 0 : return t;
      }
    }
    case 1 : switch(j){
      case 0 : switch(k){
        case 0 : return -1.*fcs_pow(t,2)*u;
      }
      case 1 : switch(k){
        case 0 : return 3.*fcs_pow(t,3)*u*v;
        case 1 : return -15.*fcs_pow(t,4)*u*v*w;
      }
    }
    case 2 : switch(j){
      case 0 : switch(k){
        case 0 : return fcs_pow(t,3)*(-1. + 3.*u2);
      }
      case 1 : switch(k){
        case 0 : return 3.*fcs_pow(t,4)*(1. - 5.*u2)*v;
        case 1 : return 15.*fcs_pow(t,5)*(-1. + 7.*u2)*v*w;
      }
      case 2 : switch(k){
        case 0 : return 3.*fcs_pow(t,5)*(1. - 5.*v2 + 5.*u2*(-1. + 7.*v2));
        case 1 : return -15.*fcs_pow(t,6)*(1. - 7.*v2 + 7.*u2*(-1. + 9.*v2))*w;
        case 2 : return 15.*fcs_pow(t,7)*(-1. + 7.*w2 + v2*(7. - 63.*w2) + 7.*u2*(1. - 9.*w2 + v2*(-9. + 99.*w2)));
      }
    }
    case 3 : switch(j){
      case 0 : switch(k){
        case 0 : return 3.*fcs_pow(t,4)*u*(3. - 5.*u2);
      }
      case 1 : switch(k){
        case 0 : return 15.*fcs_pow(t,5)*u*(-3. + 7.*u2)*v;
        case 1 : return 315.*fcs_pow(t,6)*u*(1. - 3.*u2)*v*w;
      }
      case 2 : switch(k){
        case 0 : return -15.*fcs_pow(t,6)*u*(3. - 21.*v2 + 7.*u2*(-1. + 9.*v2));
        case 1 : return 315.*fcs_pow(t,7)*u*(1. - 9.*v2 + u2*(-3. + 33.*v2))*w;
        case 2 : return -315.*fcs_pow(t,8)*u*(-1. + 9.*w2 + v2*(9. - 99.*w2) + u2*(3. - 33.*w2 + 33.*v2*(-1. + 13.*w2)));
      }
      case 3 : switch(k){
        case 0 : return 945.*fcs_pow(t,7)*u*v*(1. - 3.*v2 + u2*(-3. + 11.*v2));
        case 1 : return -945.*fcs_pow(t,8)*u*v*(9. - 33.*v2 + 11.*u2*(-3. + 13.*v2))*w;
        case 2 : return 945.*fcs_pow(t,9)*u*v*(-9. + 99.*w2 + v2*(33. - 429.*w2) + 11.*u2*(3. - 39.*w2 + 13.*v2*(-1. + 15.*w2)));
        case 3 : return -31185.*fcs_pow(t,10)*u*v*w*(-9. + 39.*w2 - 39.*v2*(-1. + 5.*w2) + 13.*u2*(3. - 15.*w2 + 5.*v2*(-3. + 17.*w2)));
      }
    }
    case 4 : switch(j){
      case 0 : switch(k){
        case 0 : return 3.*fcs_pow(t,5)*(3. - 30.*u2 + 35.*fcs_pow(u,4));
      }
      case 1 : switch(k){
        case 0 : return -45.*fcs_pow(t,6)*(1. - 14.*u2 + 21.*fcs_pow(u,4))*v;
        case 1 : return 315.*fcs_pow(t,7)*(1. - 18.*u2 + 33.*fcs_pow(u,4))*v*w;
      }
      case 2 : switch(k){
        case 0 : return 45.*fcs_pow(t,7)*(-1. + 7.*v2 - 14.*u2*(-1. + 9.*v2) + 21.*fcs_pow(u,4)*(-1. + 11.*v2));
        case 1 : return -315.*fcs_pow(t,8)*(-1. + 9.*v2 + u2*(18. - 198.*v2) + 33.*fcs_pow(u,4)*(-1. + 13.*v2))*w;
        case 2 : return 315.*fcs_pow(t,9)*(1. - 9.*w2 + v2*(-9. + 99.*w2) - 18.*u2*(1. - 11.*w2 + 11.*v2*(-1. + 13.*w2)) + 33.*fcs_pow(u,4)*(1. - 13.*w2 + 13.*v2*(-1. + 15.*w2)));
      }
      case 3 : switch(k){
        case 0 : return -945.*fcs_pow(t,8)*v*(-1. + 3.*v2 + u2*(18. - 66.*v2) + 11.*fcs_pow(u,4)*(-3. + 13.*v2));
        case 1 : return 2835.*fcs_pow(t,9)*v*(-3. + 11.*v2 + u2*(66. - 286.*v2) + 143.*fcs_pow(u,4)*(-1. + 5.*v2))*w;
        case 2 : return -2835.*fcs_pow(t,10)*v*(3. - 33.*w2 + 11.*v2*(-1. + 13.*w2) - 22.*u2*(3. - 39.*w2 + 13.*v2*(-1. + 15.*w2)) + 143.*fcs_pow(u,4)*(1. - 15.*w2 + v2*(-5. + 85.*w2)));
        case 3 : return 31185.*fcs_pow(t,11)*v*w*(9. - 39.*w2 + 39.*v2*(-1. + 5.*w2) - 78.*u2*(3. - 15.*w2 + 5.*v2*(-3. + 17.*w2)) + 65.*fcs_pow(u,4)*(9. - 51.*w2 + 17.*v2*(-3. + 19.*w2)));
      }
      case 4 : switch(k){
        case 0 : return 945.*fcs_pow(t,9)*(1. - 18.*v2 + 33.*fcs_pow(v,4) + 33.*fcs_pow(u,4)*(1. - 26.*v2 + 65.*fcs_pow(v,4)) - 6.*u2*(3. - 66.*v2 + 143.*fcs_pow(v,4)));
        case 1 : return -2835.*fcs_pow(t,10)*(3. - 66.*v2 + 143.*fcs_pow(v,4) - 66.*u2*(1. - 26.*v2 + 65.*fcs_pow(v,4)) + 143.*fcs_pow(u,4)*(1. - 30.*v2 + 85.*fcs_pow(v,4)))*w;
        case 2 : return 2835.*fcs_pow(t,11)*(-3. + 33.*w2 + v2*(66. - 858.*w2) + 143.*fcs_pow(v,4)*(-1. + 15.*w2) - 66.*u2*(-1. + 13.*w2 + v2*(26. - 390.*w2) + 65.*fcs_pow(v,4)*(-1. + 17.*w2)) + 143.*fcs_pow(u,4)*(-1. + 15.*w2 + v2*(30. - 510.*w2) + 85.*fcs_pow(v,4)*(-1. + 19.*w2)));
        case 3 : return -93555.*fcs_pow(t,12)*w*(-3. + 13.*w2 - 78.*v2*(-1. + 5.*w2) + 65.*fcs_pow(v,4)*(-3. + 17.*w2) + 65.*fcs_pow(u,4)*(-3. + 17.*w2 + 323.*fcs_pow(v,4)*(-1. + 7.*w2) - 34.*v2*(-3. + 19.*w2)) - 26.*u2*(-3. + 15.*w2 + v2*(90. - 510.*w2) + 85.*fcs_pow(v,4)*(-3. + 19.*w2)));
        case 4 : return 93555.*fcs_pow(t,13)*(3. - 78.*w2 + 195.*fcs_pow(w,4) - 78.*v2*(1. - 30.*w2 + 85.*fcs_pow(w,4)) + 65.*fcs_pow(v,4)*(3. - 102.*w2 + 323.*fcs_pow(w,4)) + 65.*fcs_pow(u,4)*(3. - 102.*w2 + 323.*fcs_pow(w,4) - 102.*v2*(1. - 38.*w2 + 133.*fcs_pow(w,4)) + 323.*fcs_pow(v,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4))) - 78.*u2*(1. - 30.*w2 + 85.*fcs_pow(w,4) + 85.*fcs_pow(v,4)*(1. - 38.*w2 + 133.*fcs_pow(w,4)) - 10.*v2*(3. - 102.*w2 + 323.*fcs_pow(w,4))));
      }
    }
    case 5 : switch(j){
      case 0 : switch(k){
        case 0 : return -15.*fcs_pow(t,6)*u*(15. - 70.*u2 + 63.*fcs_pow(u,4));
      }
      case 1 : switch(k){
        case 0 : return 315.*fcs_pow(t,7)*u*(5. - 30.*u2 + 33.*fcs_pow(u,4))*v;
        case 1 : return -945.*fcs_pow(t,8)*u*(15. - 110.*u2 + 143.*fcs_pow(u,4))*v*w;
      }
      case 2 : switch(k){
        case 0 : return -315.*fcs_pow(t,8)*u*(-5. + 45.*v2 + u2*(30. - 330.*v2) + 33.*fcs_pow(u,4)*(-1. + 13.*v2));
        case 1 : return 945.*fcs_pow(t,9)*u*(-15. + 165.*v2 - 110.*u2*(-1. + 13.*v2) + 143.*fcs_pow(u,4)*(-1. + 15.*v2))*w;
        case 2 : return -945.*fcs_pow(t,10)*u*(15.*(1. - 11.*w2 + 11.*v2*(-1. + 13.*w2)) - 110.*u2*(1. - 13.*w2 + 13.*v2*(-1. + 15.*w2)) + 143.*fcs_pow(u,4)*(1. - 15.*w2 + 15.*v2*(-1. + 17.*w2)));
      }
      case 3 : switch(k){
        case 0 : return 945.*fcs_pow(t,9)*u*v*(-45. + 165.*v2 + 429.*fcs_pow(u,4)*(-1. + 5.*v2) - 110.*u2*(-3. + 13.*v2));
        case 1 : return -155925.*fcs_pow(t,10)*u*v*(-3. + 13.*v2 - 26.*u2*(-1. + 5.*v2) + 13.*fcs_pow(u,4)*(-3. + 17.*v2))*w;
        case 2 : return 155925.*fcs_pow(t,11)*u*v*(3. - 39.*w2 + 13.*v2*(-1. + 15.*w2) + 13.*fcs_pow(u,4)*(3. - 51.*w2 + 17.*v2*(-1. + 19.*w2)) - 26.*u2*(1. - 15.*w2 + v2*(-5. + 85.*w2)));
        case 3 : return -2.027025e6*fcs_pow(t,12)*u*v*w*(9. - 45.*w2 + 15.*v2*(-3. + 17.*w2) + 51.*fcs_pow(u,4)*(3. - 19.*w2 + 19.*v2*(-1. + 7.*w2)) - 10.*u2*(9. - 51.*w2 + 17.*v2*(-3. + 19.*w2)));
      }
      case 4 : switch(k){
        case 0 : return -2835.*fcs_pow(t,10)*u*(-110.*u2*(1. - 26.*v2 + 65.*fcs_pow(v,4)) + 143.*fcs_pow(u,4)*(1. - 30.*v2 + 85.*fcs_pow(v,4)) + 5.*(3. - 66.*v2 + 143.*fcs_pow(v,4)));
        case 1 : return 155925.*fcs_pow(t,11)*u*(3. - 78.*v2 + 195.*fcs_pow(v,4) - 26.*u2*(1. - 30.*v2 + 85.*fcs_pow(v,4)) + 13.*fcs_pow(u,4)*(3. - 102.*v2 + 323.*fcs_pow(v,4)))*w;
        case 2 : return -155925.*fcs_pow(t,12)*u*(3.*(-1. + 13.*w2 + v2*(26. - 390.*w2) + 65.*fcs_pow(v,4)*(-1. + 17.*w2)) - 26.*u2*(-1. + 15.*w2 + v2*(30. - 510.*w2) + 85.*fcs_pow(v,4)*(-1. + 19.*w2)) + 13.*fcs_pow(u,4)*(-3. + 51.*w2 - 102.*v2*(-1. + 19.*w2) + 323.*fcs_pow(v,4)*(-1. + 21.*w2)));
        case 3 : return 6.081075e6*fcs_pow(t,13)*u*w*(-3. + 15.*w2 + v2*(90. - 510.*w2) + 85.*fcs_pow(v,4)*(-3. + 19.*w2) - 10.*u2*(-3. + 17.*w2 + 323.*fcs_pow(v,4)*(-1. + 7.*w2) - 34.*v2*(-3. + 19.*w2)) + 17.*fcs_pow(u,4)*(-3. + 19.*w2 - 114.*v2*(-1. + 7.*w2) + 133.*fcs_pow(v,4)*(-3. + 23.*w2)));
        case 4 : return -6.081075e6*fcs_pow(t,14)*u*(-10.*u2*(3. - 102.*w2 + 323.*fcs_pow(w,4) - 102.*v2*(1. - 38.*w2 + 133.*fcs_pow(w,4)) + 323.*fcs_pow(v,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4))) + 3.*(1. - 30.*w2 + 85.*fcs_pow(w,4) + 85.*fcs_pow(v,4)*(1. - 38.*w2 + 133.*fcs_pow(w,4)) - 10.*v2*(3. - 102.*w2 + 323.*fcs_pow(w,4))) + 17.*fcs_pow(u,4)*(3. - 114.*w2 + 399.*fcs_pow(w,4) - 114.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return 155925.*fcs_pow(t,11)*u*v*(-130.*u2*(1. - 10.*v2 + 17.*fcs_pow(v,4)) + 5.*(3. - 26.*v2 + 39.*fcs_pow(v,4)) + 13.*fcs_pow(u,4)*(15. - 170.*v2 + 323.*fcs_pow(v,4)));
        case 1 : return -2.027025e6*fcs_pow(t,12)*u*v*(15.*(1. - 10.*v2 + 17.*fcs_pow(v,4)) - 10.*u2*(15. - 170.*v2 + 323.*fcs_pow(v,4)) + 17.*fcs_pow(u,4)*(15. - 190.*v2 + 399.*fcs_pow(v,4)))*w;
        case 2 : return 2.027025e6*fcs_pow(t,13)*u*v*(15.*(-1. + 15.*w2 + v2*(10. - 170.*w2) + 17.*fcs_pow(v,4)*(-1. + 19.*w2)) - 10.*u2*(-15. + 255.*w2 - 170.*v2*(-1. + 19.*w2) + 323.*fcs_pow(v,4)*(-1. + 21.*w2)) + 17.*fcs_pow(u,4)*(-15. + 285.*w2 - 190.*v2*(-1. + 21.*w2) + 399.*fcs_pow(v,4)*(-1. + 23.*w2)));
        case 3 : return -6.081075e6*fcs_pow(t,14)*u*v*w*(5.*(-45. + 255.*w2 + 969.*fcs_pow(v,4)*(-1. + 7.*w2) - 170.*v2*(-3. + 19.*w2)) - 170.*u2*(-15. + 95.*w2 - 190.*v2*(-1. + 7.*w2) + 133.*fcs_pow(v,4)*(-3. + 23.*w2)) + 323.*fcs_pow(u,4)*(15.*(-1. + 7.*w2) - 70.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)));
        case 4 : return 6.081075e6*fcs_pow(t,15)*u*v*(15.*(-170.*v2*(1. - 38.*w2 + 133.*fcs_pow(w,4)) + 323.*fcs_pow(v,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 5.*(3. - 102.*w2 + 323.*fcs_pow(w,4))) + 323.*fcs_pow(u,4)*(15.*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 70.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) - 170.*u2*(15.*(1. - 38.*w2 + 133.*fcs_pow(w,4)) - 190.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4))));
        case 5 : return -2.584456875e9*fcs_pow(t,16)*u*v*w*(-38.*u2*(-70.*v2*(3. - 46.*w2 + 115.*fcs_pow(w,4)) + 161.*fcs_pow(v,4)*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 3.*(5. - 70.*w2 + 161.*fcs_pow(w,4))) + 3.*(15. - 190.*w2 + 399.*fcs_pow(w,4) + 133.*fcs_pow(v,4)*(3. - 46.*w2 + 115.*fcs_pow(w,4)) - 38.*v2*(5. - 70.*w2 + 161.*fcs_pow(w,4))) + 133.*fcs_pow(u,4)*(9. - 138.*w2 + 345.*fcs_pow(w,4) - 46.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 69.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4))));
      }
    }
    case 6 : switch(j){
      case 0 : switch(k){
        case 0 : return 45.*fcs_pow(t,7)*(-5. + 105.*u2 - 315.*fcs_pow(u,4) + 231.*fcs_pow(u,6));
      }
      case 1 : switch(k){
        case 0 : return -315.*fcs_pow(t,8)*(-5. + 135.*u2 - 495.*fcs_pow(u,4) + 429.*fcs_pow(u,6))*v;
        case 1 : return 14175.*fcs_pow(t,9)*(-1. + 33.*u2 - 143.*fcs_pow(u,4) + 143.*fcs_pow(u,6))*v*w;
      }
      case 2 : switch(k){
        case 0 : return 315.*fcs_pow(t,9)*(5. - 45.*v2 + 135.*u2*(-1. + 11.*v2) - 495.*fcs_pow(u,4)*(-1. + 13.*v2) + 429.*fcs_pow(u,6)*(-1. + 15.*v2));
        case 1 : return -14175.*fcs_pow(t,10)*(1. - 11.*v2 + 33.*u2*(-1. + 13.*v2) - 143.*fcs_pow(u,4)*(-1. + 15.*v2) + 143.*fcs_pow(u,6)*(-1. + 17.*v2))*w;
        case 2 : return 14175.*fcs_pow(t,11)*(-1. + 11.*w2 + v2*(11. - 143.*w2) + 33.*u2*(1. - 13.*w2 + 13.*v2*(-1. + 15.*w2)) - 143.*fcs_pow(u,4)*(1. - 15.*w2 + 15.*v2*(-1. + 17.*w2)) + 143.*fcs_pow(u,6)*(1. - 17.*w2 + 17.*v2*(-1. + 19.*w2)));
      }
      case 3 : switch(k){
        case 0 : return -14175.*fcs_pow(t,10)*v*(3. - 11.*v2 - 429.*fcs_pow(u,4)*(-1. + 5.*v2) + 33.*u2*(-3. + 13.*v2) + 143.*fcs_pow(u,6)*(-3. + 17.*v2));
        case 1 : return 155925.*fcs_pow(t,11)*v*(3. - 13.*v2 + 117.*u2*(-1. + 5.*v2) - 195.*fcs_pow(u,4)*(-3. + 17.*v2) + 221.*fcs_pow(u,6)*(-3. + 19.*v2))*w;
        case 2 : return -155925.*fcs_pow(t,12)*v*(-3. + 39.*w2 + v2*(13. - 195.*w2) - 195.*fcs_pow(u,4)*(3. - 51.*w2 + 17.*v2*(-1. + 19.*w2)) + 221.*fcs_pow(u,6)*(3. - 57.*w2 + 19.*v2*(-1. + 21.*w2)) + 117.*u2*(1. - 15.*w2 + v2*(-5. + 85.*w2)));
        case 3 : return 6.081075e6*fcs_pow(t,13)*v*w*(-3. + 15.*w2 + v2*(15. - 85.*w2) - 255.*fcs_pow(u,4)*(3. - 19.*w2 + 19.*v2*(-1. + 7.*w2)) + 15.*u2*(9. - 51.*w2 + 17.*v2*(-3. + 19.*w2)) + 323.*fcs_pow(u,6)*(3. - 21.*w2 + 7.*v2*(-3. + 23.*w2)));
      }
      case 4 : switch(k){
        case 0 : return 14175.*fcs_pow(t,11)*(-3. + 66.*v2 - 143.*fcs_pow(v,4) + 99.*u2*(1. - 26.*v2 + 65.*fcs_pow(v,4)) - 429.*fcs_pow(u,4)*(1. - 30.*v2 + 85.*fcs_pow(v,4)) + 143.*fcs_pow(u,6)*(3. - 102.*v2 + 323.*fcs_pow(v,4)));
        case 1 : return -467775.*fcs_pow(t,12)*(-1. + 26.*v2 - 65.*fcs_pow(v,4) + 39.*u2*(1. - 30.*v2 + 85.*fcs_pow(v,4)) + 221.*fcs_pow(u,6)*(1. - 38.*v2 + 133.*fcs_pow(v,4)) - 65.*fcs_pow(u,4)*(3. - 102.*v2 + 323.*fcs_pow(v,4)))*w;
        case 2 : return 467775.*fcs_pow(t,13)*(1. - 13.*w2 + 26.*v2*(-1. + 15.*w2) - 65.*fcs_pow(v,4)*(-1. + 17.*w2) + 39.*u2*(-1. + 15.*w2 + v2*(30. - 510.*w2) + 85.*fcs_pow(v,4)*(-1. + 19.*w2)) - 65.*fcs_pow(u,4)*(-3. + 51.*w2 - 102.*v2*(-1. + 19.*w2) + 323.*fcs_pow(v,4)*(-1. + 21.*w2)) + 221.*fcs_pow(u,6)*(-1. + 19.*w2 + v2*(38. - 798.*w2) + 133.*fcs_pow(v,4)*(-1. + 23.*w2)));
        case 3 : return -6.081075e6*fcs_pow(t,14)*w*(3. - 15.*w2 + 30.*v2*(-3. + 17.*w2) - 85.*fcs_pow(v,4)*(-3. + 19.*w2) + 45.*u2*(-3. + 17.*w2 + 323.*fcs_pow(v,4)*(-1. + 7.*w2) - 34.*v2*(-3. + 19.*w2)) - 255.*fcs_pow(u,4)*(-3. + 19.*w2 - 114.*v2*(-1. + 7.*w2) + 133.*fcs_pow(v,4)*(-3. + 23.*w2)) + 323.*fcs_pow(u,6)*(-3. + 21.*w2 - 42.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)));
        case 4 : return 1.8243225e7*fcs_pow(t,15)*(-1. + 30.*w2 - 85.*fcs_pow(w,4) - 85.*fcs_pow(v,4)*(1. - 38.*w2 + 133.*fcs_pow(w,4)) + 10.*v2*(3. - 102.*w2 + 323.*fcs_pow(w,4)) + 15.*u2*(3. - 102.*w2 + 323.*fcs_pow(w,4) - 102.*v2*(1. - 38.*w2 + 133.*fcs_pow(w,4)) + 323.*fcs_pow(v,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4))) + 323.*fcs_pow(u,6)*(1. - 42.*w2 + 161.*fcs_pow(w,4) + 161.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 14.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) - 85.*fcs_pow(u,4)*(3. - 114.*w2 + 399.*fcs_pow(w,4) - 114.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return -155925.*fcs_pow(t,12)*v*(585.*u2*(1. - 10.*v2 + 17.*fcs_pow(v,4)) - 5.*(3. - 26.*v2 + 39.*fcs_pow(v,4)) - 195.*fcs_pow(u,4)*(15. - 170.*v2 + 323.*fcs_pow(v,4)) + 221.*fcs_pow(u,6)*(15. - 190.*v2 + 399.*fcs_pow(v,4)));
        case 1 : return 6.081075e6*fcs_pow(t,13)*v*(-5. + 50.*v2 - 85.*fcs_pow(v,4) + 323.*fcs_pow(u,6)*(5. - 70.*v2 + 161.*fcs_pow(v,4)) + 15.*u2*(15. - 170.*v2 + 323.*fcs_pow(v,4)) - 85.*fcs_pow(u,4)*(15. - 190.*v2 + 399.*fcs_pow(v,4)))*w;
        case 2 : return -6.081075e6*fcs_pow(t,14)*v*(5. - 75.*w2 + 50.*v2*(-1. + 17.*w2) - 85.*fcs_pow(v,4)*(-1. + 19.*w2) + 15.*u2*(-15. + 255.*w2 - 170.*v2*(-1. + 19.*w2) + 323.*fcs_pow(v,4)*(-1. + 21.*w2)) - 85.*fcs_pow(u,4)*(-15. + 285.*w2 - 190.*v2*(-1. + 21.*w2) + 399.*fcs_pow(v,4)*(-1. + 23.*w2)) + 323.*fcs_pow(u,6)*(-5. + 105.*w2 - 70.*v2*(-1. + 23.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)));
        case 3 : return 3.0405375e7*fcs_pow(t,15)*v*w*(45. - 255.*w2 - 969.*fcs_pow(v,4)*(-1. + 7.*w2) + 170.*v2*(-3. + 19.*w2) + 153.*u2*(-15. + 95.*w2 - 190.*v2*(-1. + 7.*w2) + 133.*fcs_pow(v,4)*(-3. + 23.*w2)) + 2261.*fcs_pow(u,6)*(-9. + 69.*w2 + 345.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) - 969.*fcs_pow(u,4)*(15.*(-1. + 7.*w2) - 70.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)));
        case 4 : return -9.1216125e7*fcs_pow(t,16)*v*(170.*v2*(1. - 38.*w2 + 133.*fcs_pow(w,4)) - 323.*fcs_pow(v,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4)) - 5.*(3. - 102.*w2 + 323.*fcs_pow(w,4)) + 2261.*fcs_pow(u,6)*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 46.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 115.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) - 323.*fcs_pow(u,4)*(15.*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 70.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 51.*u2*(15.*(1. - 38.*w2 + 133.*fcs_pow(w,4)) - 190.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4))));
        case 5 : return 7.753370625e9*fcs_pow(t,17)*v*w*(-15. + 190.*w2 - 399.*fcs_pow(w,4) - 133.*fcs_pow(v,4)*(3. - 46.*w2 + 115.*fcs_pow(w,4)) + 38.*v2*(5. - 70.*w2 + 161.*fcs_pow(w,4)) + 57.*u2*(-70.*v2*(3. - 46.*w2 + 115.*fcs_pow(w,4)) + 161.*fcs_pow(v,4)*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 3.*(5. - 70.*w2 + 161.*fcs_pow(w,4))) - 665.*fcs_pow(u,4)*(9. - 138.*w2 + 345.*fcs_pow(w,4) - 46.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 69.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4))) + 3059.*fcs_pow(u,6)*(3. - 50.*w2 + 135.*fcs_pow(w,4) - 10.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 9.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4))));
      }
      case 6 : switch(k){
        case 0 : return 467775.*fcs_pow(t,13)*(5. - 195.*v2 + 975.*fcs_pow(v,4) - 1105.*fcs_pow(v,6) + 195.*u2*(-1. + 45.*v2 - 255.*fcs_pow(v,4) + 323.*fcs_pow(v,6)) - 195.*fcs_pow(u,4)*(-5. + 255.*v2 - 1615.*fcs_pow(v,4) + 2261.*fcs_pow(v,6)) + 221.*fcs_pow(u,6)*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)));
        case 1 : return -3.0405375e7*fcs_pow(t,14)*(1. - 45.*v2 + 255.*fcs_pow(v,4) - 323.*fcs_pow(v,6) + 323.*fcs_pow(u,6)*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) + 9.*u2*(-5. + 255.*v2 - 1615.*fcs_pow(v,4) + 2261.*fcs_pow(v,6)) - 51.*fcs_pow(u,4)*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)))*w;
        case 2 : return 3.0405375e7*fcs_pow(t,15)*(-1. + 15.*w2 + v2*(45. - 765.*w2) + 255.*fcs_pow(v,4)*(-1. + 19.*w2) - 323.*fcs_pow(v,6)*(-1. + 21.*w2) + 9.*u2*(5. - 85.*w2 + 255.*v2*(-1. + 19.*w2) - 1615.*fcs_pow(v,4)*(-1. + 21.*w2) + 2261.*fcs_pow(v,6)*(-1. + 23.*w2)) - 51.*fcs_pow(u,4)*(5. - 95.*w2 + 285.*v2*(-1. + 21.*w2) - 1995.*fcs_pow(v,4)*(-1. + 23.*w2) + 3059.*fcs_pow(v,6)*(-1. + 25.*w2)) + 323.*fcs_pow(u,6)*(1. - 21.*w2 + 63.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 805.*fcs_pow(v,6)*(-1. + 27.*w2)));
        case 3 : return -9.1216125e7*fcs_pow(t,16)*w*(-15. + 85.*w2 + 4845.*fcs_pow(v,4)*(-1. + 7.*w2) - 255.*v2*(-3. + 19.*w2) - 2261.*fcs_pow(v,6)*(-3. + 23.*w2) - 4845.*fcs_pow(u,4)*(1. - 7.*w2 + 805.*fcs_pow(v,6)*(-1. + 9.*w2) + 21.*v2*(-3. + 23.*w2) - 161.*fcs_pow(v,4)*(-3. + 25.*w2)) + 51.*u2*(15. - 95.*w2 + 855.*v2*(-1. + 7.*w2) - 1995.*fcs_pow(v,4)*(-3. + 23.*w2) + 3059.*fcs_pow(v,6)*(-3. + 25.*w2)) + 2261.*fcs_pow(u,6)*(3. - 23.*w2 - 1725.*fcs_pow(v,4)*(-1. + 9.*w2) + 69.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)));
        case 4 : return 9.1216125e7*fcs_pow(t,17)*(-765.*v2*(1. - 38.*w2 + 133.*fcs_pow(w,4)) + 4845.*fcs_pow(v,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 5.*(3. - 102.*w2 + 323.*fcs_pow(w,4)) - 2261.*fcs_pow(v,6)*(3. - 138.*w2 + 575.*fcs_pow(w,4)) - 4845.*fcs_pow(u,4)*(-1. + 42.*w2 - 161.*fcs_pow(w,4) - 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 805.*fcs_pow(v,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 21.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 153.*u2*(-5. + 190.*w2 - 665.*fcs_pow(w,4) + 285.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 3059.*fcs_pow(v,6)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 665.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 2261.*fcs_pow(u,6)*(-3. + 138.*w2 - 575.*fcs_pow(w,4) + 207.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 1725.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
        case 5 : return -7.753370625e9*fcs_pow(t,18)*w*(15. - 190.*w2 + 399.*fcs_pow(w,4) + 1995.*fcs_pow(v,4)*(3. - 46.*w2 + 115.*fcs_pow(w,4)) - 3059.*fcs_pow(v,6)*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 171.*v2*(5. - 70.*w2 + 161.*fcs_pow(w,4)) + 171.*u2*(-5. + 70.*w2 - 161.*fcs_pow(w,4) + 105.*v2*(3. - 46.*w2 + 115.*fcs_pow(w,4)) - 805.*fcs_pow(v,4)*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 805.*fcs_pow(v,6)*(5. - 90.*w2 + 261.*fcs_pow(w,4))) - 1995.*fcs_pow(u,4)*(-3. + 46.*w2 - 115.*fcs_pow(w,4) + 69.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 345.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 207.*fcs_pow(v,6)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) + 3059.*fcs_pow(u,6)*(-3. + 50.*w2 - 135.*fcs_pow(w,4) + 45.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 135.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))));
        case 6 : return 2.3260111875e10*fcs_pow(t,19)*(-5. + 285.*w2 - 1995.*fcs_pow(w,4) + 3059.*fcs_pow(w,6) - 285.*v2*(-1. + 63.*w2 - 483.*fcs_pow(w,4) + 805.*fcs_pow(w,6)) + 1995.*fcs_pow(v,4)*(-1. + 69.*w2 - 575.*fcs_pow(w,4) + 1035.*fcs_pow(w,6)) - 3059.*fcs_pow(v,6)*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) + 285.*u2*(1. - 63.*w2 + 483.*fcs_pow(w,4) - 805.*fcs_pow(w,6) + 63.*v2*(-1. + 69.*w2 - 575.*fcs_pow(w,4) + 1035.*fcs_pow(w,6)) - 483.*fcs_pow(v,4)*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) + 161.*fcs_pow(v,6)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6))) + 3059.*fcs_pow(u,6)*(1. - 75.*w2 + 675.*fcs_pow(w,4) - 1305.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 15.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 135.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 1995.*fcs_pow(u,4)*(1. - 69.*w2 + 575.*fcs_pow(w,4) - 1035.*fcs_pow(w,6) + 69.*v2*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 115.*fcs_pow(v,4)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 207.*fcs_pow(v,6)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))));
      }
    }
    case 7 : switch(j){
      case 0 : switch(k){
        case 0 : return -315.*fcs_pow(t,8)*u*(-35. + 315.*u2 - 693.*fcs_pow(u,4) + 429.*fcs_pow(u,6));
      }
      case 1 : switch(k){
        case 0 : return 2835.*fcs_pow(t,9)*u*(-35. + 385.*u2 - 1001.*fcs_pow(u,4) + 715.*fcs_pow(u,6))*v;
        case 1 : return -155925.*fcs_pow(t,10)*u*(-7. + 91.*u2 - 273.*fcs_pow(u,4) + 221.*fcs_pow(u,6))*v*w;
      }
      case 2 : switch(k){
        case 0 : return -2835.*fcs_pow(t,10)*u*(35. - 385.*v2 + 385.*u2*(-1. + 13.*v2) - 1001.*fcs_pow(u,4)*(-1. + 15.*v2) + 715.*fcs_pow(u,6)*(-1. + 17.*v2));
        case 1 : return 155925.*fcs_pow(t,11)*u*(7. - 91.*v2 + 91.*u2*(-1. + 15.*v2) - 273.*fcs_pow(u,4)*(-1. + 17.*v2) + 221.*fcs_pow(u,6)*(-1. + 19.*v2))*w;
        case 2 : return -155925.*fcs_pow(t,12)*u*(-7. + 91.*w2 - 91.*v2*(-1. + 15.*w2) + 91.*u2*(1. - 15.*w2 + 15.*v2*(-1. + 17.*w2)) - 273.*fcs_pow(u,4)*(1. - 17.*w2 + 17.*v2*(-1. + 19.*w2)) + 221.*fcs_pow(u,6)*(1. - 19.*w2 + 19.*v2*(-1. + 21.*w2)));
      }
      case 3 : switch(k){
        case 0 : return 155925.*fcs_pow(t,11)*u*v*(21. - 91.*v2 + 273.*u2*(-1. + 5.*v2) - 273.*fcs_pow(u,4)*(-3. + 17.*v2) + 221.*fcs_pow(u,6)*(-3. + 19.*v2));
        case 1 : return -6.081075e6*fcs_pow(t,12)*u*v*(7. - 35.*v2 + 323.*fcs_pow(u,6)*(-1. + 7.*v2) + 35.*u2*(-3. + 17.*v2) - 119.*fcs_pow(u,4)*(-3. + 19.*v2))*w;
        case 2 : return 6.081075e6*fcs_pow(t,13)*u*v*(7.*(-1. + 15.*w2 + v2*(5. - 85.*w2)) + 35.*u2*(3. - 51.*w2 + 17.*v2*(-1. + 19.*w2)) - 119.*fcs_pow(u,4)*(3. - 57.*w2 + 19.*v2*(-1. + 21.*w2)) + 323.*fcs_pow(u,6)*(1. - 21.*w2 + 7.*v2*(-1. + 23.*w2)));
        case 3 : return -4.2567525e7*fcs_pow(t,14)*u*v*w*(255.*u2*(3. - 19.*w2 + 19.*v2*(-1. + 7.*w2)) - 5.*(9. - 51.*w2 + 17.*v2*(-3. + 19.*w2)) - 969.*fcs_pow(u,4)*(3. - 21.*w2 + 7.*v2*(-3. + 23.*w2)) + 323.*fcs_pow(u,6)*(9. - 69.*w2 + 23.*v2*(-3. + 25.*w2)));
      }
      case 4 : switch(k){
        case 0 : return -467775.*fcs_pow(t,12)*u*(-7.*(1. - 26.*v2 + 65.*fcs_pow(v,4)) + 91.*u2*(1. - 30.*v2 + 85.*fcs_pow(v,4)) + 221.*fcs_pow(u,6)*(1. - 38.*v2 + 133.*fcs_pow(v,4)) - 91.*fcs_pow(u,4)*(3. - 102.*v2 + 323.*fcs_pow(v,4)));
        case 1 : return 6.081075e6*fcs_pow(t,13)*u*(-7.*(1. - 30.*v2 + 85.*fcs_pow(v,4)) - 357.*fcs_pow(u,4)*(1. - 38.*v2 + 133.*fcs_pow(v,4)) + 323.*fcs_pow(u,6)*(1. - 42.*v2 + 161.*fcs_pow(v,4)) + 35.*u2*(3. - 102.*v2 + 323.*fcs_pow(v,4)))*w;
        case 2 : return -6.081075e6*fcs_pow(t,14)*u*(-7.*(-1. + 15.*w2 + v2*(30. - 510.*w2) + 85.*fcs_pow(v,4)*(-1. + 19.*w2)) + 35.*u2*(-3. + 51.*w2 - 102.*v2*(-1. + 19.*w2) + 323.*fcs_pow(v,4)*(-1. + 21.*w2)) - 357.*fcs_pow(u,4)*(-1. + 19.*w2 + v2*(38. - 798.*w2) + 133.*fcs_pow(v,4)*(-1. + 23.*w2)) + 323.*fcs_pow(u,6)*(-1. + 21.*w2 + v2*(42. - 966.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)));
        case 3 : return 1.27702575e8*fcs_pow(t,15)*u*w*(-5.*(-3. + 17.*w2 + 323.*fcs_pow(v,4)*(-1. + 7.*w2) - 34.*v2*(-3. + 19.*w2)) + 85.*u2*(-3. + 19.*w2 - 114.*v2*(-1. + 7.*w2) + 133.*fcs_pow(v,4)*(-3. + 23.*w2)) + 323.*fcs_pow(u,6)*(-3. + 23.*w2 + 575.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) - 323.*fcs_pow(u,4)*(-3. + 21.*w2 - 42.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)));
        case 4 : return -1.27702575e8*fcs_pow(t,16)*u*(-5.*(3. - 102.*w2 + 323.*fcs_pow(w,4) - 102.*v2*(1. - 38.*w2 + 133.*fcs_pow(w,4)) + 323.*fcs_pow(v,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4))) + 323.*fcs_pow(u,6)*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 138.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) - 969.*fcs_pow(u,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4) + 161.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 14.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 85.*u2*(3. - 114.*w2 + 399.*fcs_pow(w,4) - 114.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return 6.081075e6*fcs_pow(t,13)*u*v*(-35.*(1. - 10.*v2 + 17.*fcs_pow(v,4)) + 323.*fcs_pow(u,6)*(5. - 70.*v2 + 161.*fcs_pow(v,4)) + 35.*u2*(15. - 170.*v2 + 323.*fcs_pow(v,4)) - 119.*fcs_pow(u,4)*(15. - 190.*v2 + 399.*fcs_pow(v,4)));
        case 1 : return -4.2567525e7*fcs_pow(t,14)*u*v*(-75. + 850.*v2 - 1615.*fcs_pow(v,4) + 1615.*fcs_pow(u,6)*(3. - 46.*v2 + 115.*fcs_pow(v,4)) - 969.*fcs_pow(u,4)*(5. - 70.*v2 + 161.*fcs_pow(v,4)) + 85.*u2*(15. - 190.*v2 + 399.*fcs_pow(v,4)))*w;
        case 2 : return 4.2567525e7*fcs_pow(t,15)*u*v*(-5.*(-15. + 255.*w2 - 170.*v2*(-1. + 19.*w2) + 323.*fcs_pow(v,4)*(-1. + 21.*w2)) + 85.*u2*(-15. + 285.*w2 - 190.*v2*(-1. + 21.*w2) + 399.*fcs_pow(v,4)*(-1. + 23.*w2)) - 969.*fcs_pow(u,4)*(-5. + 105.*w2 - 70.*v2*(-1. + 23.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) + 1615.*fcs_pow(u,6)*(-3. + 69.*w2 - 46.*v2*(-1. + 25.*w2) + 115.*fcs_pow(v,4)*(-1. + 27.*w2)));
        case 3 : return -1.0854718875e10*fcs_pow(t,16)*u*v*w*(15. - 95.*w2 + 190.*v2*(-1. + 7.*w2) - 133.*fcs_pow(v,4)*(-3. + 23.*w2) - 133.*fcs_pow(u,4)*(-9. + 69.*w2 + 345.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) + 19.*u2*(15.*(-1. + 7.*w2) - 70.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)) + 437.*fcs_pow(u,6)*(-3. + 25.*w2 - 50.*v2*(-1. + 9.*w2) + 45.*fcs_pow(v,4)*(-3. + 29.*w2)));
        case 4 : return 1.0854718875e10*fcs_pow(t,17)*u*v*(-15.*(1. - 38.*w2 + 133.*fcs_pow(w,4)) + 190.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) - 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4)) - 399.*fcs_pow(u,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 46.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 115.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) + 19.*u2*(15.*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 70.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 437.*fcs_pow(u,6)*(3. - 150.*w2 + 675.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 45.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
        case 5 : return -1.031198293125e12*fcs_pow(t,18)*u*v*w*(70.*v2*(3. - 46.*w2 + 115.*fcs_pow(w,4)) - 161.*fcs_pow(v,4)*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 3.*(5. - 70.*w2 + 161.*fcs_pow(w,4)) + 35.*u2*(9. - 138.*w2 + 345.*fcs_pow(w,4) - 46.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 69.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4))) - 483.*fcs_pow(u,4)*(3. - 50.*w2 + 135.*fcs_pow(w,4) - 10.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 9.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) + 69.*fcs_pow(u,6)*(5.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 30.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 87.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))));
      }
      case 6 : switch(k){
        case 0 : return -6.081075e6*fcs_pow(t,14)*u*(-35.*(-1. + 45.*v2 - 255.*fcs_pow(v,4) + 323.*fcs_pow(v,6)) + 1615.*fcs_pow(u,6)*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) + 105.*u2*(-5. + 255.*v2 - 1615.*fcs_pow(v,4) + 2261.*fcs_pow(v,6)) - 357.*fcs_pow(u,4)*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)));
        case 1 : return 6.38512875e8*fcs_pow(t,15)*u*(5. - 255.*v2 + 1615.*fcs_pow(v,4) - 2261.*fcs_pow(v,6) - 323.*fcs_pow(u,4)*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) + 323.*fcs_pow(u,6)*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 17.*u2*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)))*w;
        case 2 : return -6.38512875e8*fcs_pow(t,16)*u*(-5. + 85.*w2 - 255.*v2*(-1. + 19.*w2) + 1615.*fcs_pow(v,4)*(-1. + 21.*w2) - 2261.*fcs_pow(v,6)*(-1. + 23.*w2) + 17.*u2*(5. - 95.*w2 + 285.*v2*(-1. + 21.*w2) - 1995.*fcs_pow(v,4)*(-1. + 23.*w2) + 3059.*fcs_pow(v,6)*(-1. + 25.*w2)) - 323.*fcs_pow(u,4)*(1. - 21.*w2 + 63.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 805.*fcs_pow(v,6)*(-1. + 27.*w2)) + 323.*fcs_pow(u,6)*(1. - 23.*w2 + 69.*v2*(-1. + 25.*w2) - 575.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)));
        case 3 : return 1.0854718875e10*fcs_pow(t,17)*u*w*(-15. + 95.*w2 - 855.*v2*(-1. + 7.*w2) + 1995.*fcs_pow(v,4)*(-3. + 23.*w2) - 3059.*fcs_pow(v,6)*(-3. + 25.*w2) + 285.*u2*(1. - 7.*w2 + 805.*fcs_pow(v,6)*(-1. + 9.*w2) + 21.*v2*(-3. + 23.*w2) - 161.*fcs_pow(v,4)*(-3. + 25.*w2)) - 399.*fcs_pow(u,4)*(3. - 23.*w2 - 1725.*fcs_pow(v,4)*(-1. + 9.*w2) + 69.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)) + 437.*fcs_pow(u,6)*(3. - 25.*w2 + 225.*v2*(-1. + 9.*w2) - 675.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)));
        case 4 : return -3.2564156625e10*fcs_pow(t,18)*u*(5. - 190.*w2 + 665.*fcs_pow(w,4) - 285.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) - 3059.*fcs_pow(v,6)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 665.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4)) + 95.*u2*(-1. + 42.*w2 - 161.*fcs_pow(w,4) - 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 805.*fcs_pow(v,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 21.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 437.*fcs_pow(u,6)*(-1. + 50.*w2 - 225.*fcs_pow(w,4) + 75.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 225.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 133.*fcs_pow(u,4)*(-3. + 138.*w2 - 575.*fcs_pow(w,4) + 207.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 1725.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
        case 5 : return 3.093594879375e12*fcs_pow(t,19)*u*w*(5. - 70.*w2 + 161.*fcs_pow(w,4) - 105.*v2*(3. - 46.*w2 + 115.*fcs_pow(w,4)) + 805.*fcs_pow(v,4)*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 805.*fcs_pow(v,6)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 35.*u2*(-3. + 46.*w2 - 115.*fcs_pow(w,4) + 69.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 345.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 207.*fcs_pow(v,6)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) + 115.*fcs_pow(u,6)*(-5. + 90.*w2 - 261.*fcs_pow(w,4) + 8091.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 27.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 261.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 161.*fcs_pow(u,4)*(-3. + 50.*w2 - 135.*fcs_pow(w,4) + 45.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 135.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))));
        case 6 : return -3.093594879375e12*fcs_pow(t,20)*u*(-5.*(1. - 63.*w2 + 483.*fcs_pow(w,4) - 805.*fcs_pow(w,6) + 63.*v2*(-1. + 69.*w2 - 575.*fcs_pow(w,4) + 1035.*fcs_pow(w,6)) - 483.*fcs_pow(v,4)*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) + 161.*fcs_pow(v,6)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6))) + 115.*fcs_pow(u,6)*(5. - 405.*w2 + 3915.*fcs_pow(w,4) - 8091.*fcs_pow(w,6) - 3915.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 81.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 483.*fcs_pow(u,4)*(1. - 75.*w2 + 675.*fcs_pow(w,4) - 1305.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 15.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 135.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 105.*u2*(1. - 69.*w2 + 575.*fcs_pow(w,4) - 1035.*fcs_pow(w,6) + 69.*v2*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 115.*fcs_pow(v,4)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 207.*fcs_pow(v,6)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))));
      }
      case 7 : switch(k){
        case 0 : return 1.27702575e8*fcs_pow(t,15)*u*v*(-35.*(-5. + 85.*v2 - 323.*fcs_pow(v,4) + 323.*fcs_pow(v,6)) + 595.*u2*(-5. + 95.*v2 - 399.*fcs_pow(v,4) + 437.*fcs_pow(v,6)) - 2261.*fcs_pow(u,4)*(-5. + 105.*v2 - 483.*fcs_pow(v,4) + 575.*fcs_pow(v,6)) + 1615.*fcs_pow(u,6)*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)));
        case 1 : return -1.0854718875e10*fcs_pow(t,16)*u*v*(-7.*(-5. + 95.*v2 - 399.*fcs_pow(v,4) + 437.*fcs_pow(v,6)) + 133.*u2*(-5. + 105.*v2 - 483.*fcs_pow(v,4) + 575.*fcs_pow(v,6)) - 399.*fcs_pow(u,4)*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 437.*fcs_pow(u,6)*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)))*w;
        case 2 : return 1.0854718875e10*fcs_pow(t,17)*u*v*(-7.*(5. - 95.*w2 + 95.*v2*(-1. + 21.*w2) - 399.*fcs_pow(v,4)*(-1. + 23.*w2) + 437.*fcs_pow(v,6)*(-1. + 25.*w2)) + 133.*u2*(5. - 105.*w2 + 105.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 575.*fcs_pow(v,6)*(-1. + 27.*w2)) - 399.*fcs_pow(u,4)*(7. - 161.*w2 + 161.*v2*(-1. + 25.*w2) - 805.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) + 437.*fcs_pow(u,6)*(7. - 175.*w2 + 175.*v2*(-1. + 27.*w2) - 945.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)));
        case 3 : return -6.18718975875e11*fcs_pow(t,18)*u*v*w*(-7.*(5. - 35.*w2 + 575.*fcs_pow(v,6)*(-1. + 9.*w2) + 35.*v2*(-3. + 23.*w2) - 161.*fcs_pow(v,4)*(-3. + 25.*w2)) + 35.*u2*(21. - 161.*w2 - 2415.*fcs_pow(v,4)*(-1. + 9.*w2) + 161.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)) + 115.*fcs_pow(u,6)*(35. - 315.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 1827.*fcs_pow(v,4)*(-3. + 31.*w2)) - 161.*fcs_pow(u,4)*(21. - 175.*w2 + 525.*v2*(-1. + 9.*w2) - 945.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)));
        case 4 : return 6.18718975875e11*fcs_pow(t,19)*u*v*(-7.*(-5. + 210.*w2 - 805.*fcs_pow(w,4) - 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 35.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 115.*fcs_pow(u,6)*(-35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 5481.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 483.*fcs_pow(u,4)*(-7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 175.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 315.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 35.*u2*(483.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 2415.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 7.*(3. - 138.*w2 + 575.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
        case 5 : return -3.093594879375e12*fcs_pow(t,20)*u*v*w*(-35.*(-7.*(3. - 46.*w2 + 115.*fcs_pow(w,4)) + 161.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 483.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 207.*fcs_pow(v,6)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) + 1035.*fcs_pow(u,6)*(-18879.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 203.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 483.*fcs_pow(u,4)*(40455.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 35.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 315.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 1827.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 805.*u2*(-7.*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 105.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 189.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))));
        case 6 : return 4.6403923190625e13*fcs_pow(t,21)*u*v*(207.*fcs_pow(u,6)*(1015.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 6293.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 7.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 161.*fcs_pow(u,4)*(-5481.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 189.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 161.*u2*(7. - 525.*w2 + 4725.*fcs_pow(w,4) - 9135.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 35.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 189.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 7.*(7. - 483.*w2 + 4025.*fcs_pow(w,4) - 7245.*fcs_pow(w,6) + 161.*v2*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 161.*fcs_pow(v,4)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 207.*fcs_pow(v,6)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))));
        case 7 : return -7.471031633690625e15*fcs_pow(t,22)*u*v*w*(-189.*fcs_pow(u,4)*(35. - 1015.*w2 + 6293.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 203.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6))) + 7.*u2*(40455.*fcs_pow(v,6)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 5481.*fcs_pow(v,4)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 5.*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 135.*v2*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 7.*(7. - 175.*w2 + 945.*fcs_pow(w,4) - 1305.*fcs_pow(w,6) + 261.*fcs_pow(v,6)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 5.*v2*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 27.*fcs_pow(v,4)*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 261.*fcs_pow(u,6)*(7.*(5. - 155.*w2 + 1023.*fcs_pow(w,4) - 1705.*fcs_pow(w,6)) + 1085.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 7161.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))));
      }
    }
    case 8 : switch(j){
      case 0 : switch(k){
        case 0 : return 315.*fcs_pow(t,9)*(35. - 1260.*u2 + 6930.*fcs_pow(u,4) - 12012.*fcs_pow(u,6) + 6435.*fcs_pow(u,8));
      }
      case 1 : switch(k){
        case 0 : return -14175.*fcs_pow(t,10)*(7. - 308.*u2 + 2002.*fcs_pow(u,4) - 4004.*fcs_pow(u,6) + 2431.*fcs_pow(u,8))*v;
        case 1 : return 155925.*fcs_pow(t,11)*(7. - 364.*u2 + 2730.*fcs_pow(u,4) - 6188.*fcs_pow(u,6) + 4199.*fcs_pow(u,8))*v*w;
      }
      case 2 : switch(k){
        case 0 : return 14175.*fcs_pow(t,11)*(-7. + 77.*v2 - 308.*u2*(-1. + 13.*v2) + 2002.*fcs_pow(u,4)*(-1. + 15.*v2) - 4004.*fcs_pow(u,6)*(-1. + 17.*v2) + 2431.*fcs_pow(u,8)*(-1. + 19.*v2));
        case 1 : return -155925.*fcs_pow(t,12)*(-7. + 91.*v2 - 364.*u2*(-1. + 15.*v2) + 2730.*fcs_pow(u,4)*(-1. + 17.*v2) - 6188.*fcs_pow(u,6)*(-1. + 19.*v2) + 4199.*fcs_pow(u,8)*(-1. + 21.*v2))*w;
        case 2 : return 155925.*fcs_pow(t,13)*(7. - 91.*w2 + 91.*v2*(-1. + 15.*w2) - 364.*u2*(1. - 15.*w2 + 15.*v2*(-1. + 17.*w2)) + 2730.*fcs_pow(u,4)*(1. - 17.*w2 + 17.*v2*(-1. + 19.*w2)) - 6188.*fcs_pow(u,6)*(1. - 19.*w2 + 19.*v2*(-1. + 21.*w2)) + 4199.*fcs_pow(u,8)*(1. - 21.*w2 + 21.*v2*(-1. + 23.*w2)));
      }
      case 3 : switch(k){
        case 0 : return -155925.*fcs_pow(t,12)*v*(-21. + 91.*v2 - 1092.*u2*(-1. + 5.*v2) + 12597.*fcs_pow(u,8)*(-1. + 7.*v2) + 2730.*fcs_pow(u,4)*(-3. + 17.*v2) - 6188.*fcs_pow(u,6)*(-3. + 19.*v2));
        case 1 : return 4.2567525e7*fcs_pow(t,13)*v*(-1. + 5.*v2 + u2*(60. - 340.*v2) - 1292.*fcs_pow(u,6)*(-1. + 7.*v2) + 170.*fcs_pow(u,4)*(-3. + 19.*v2) + 323.*fcs_pow(u,8)*(-3. + 23.*v2))*w;
        case 2 : return -4.2567525e7*fcs_pow(t,14)*v*(1. - 15.*w2 + v2*(-5. + 85.*w2) - 20.*u2*(3. - 51.*w2 + 17.*v2*(-1. + 19.*w2)) + 170.*fcs_pow(u,4)*(3. - 57.*w2 + 19.*v2*(-1. + 21.*w2)) - 1292.*fcs_pow(u,6)*(1. - 21.*w2 + 7.*v2*(-1. + 23.*w2)) + 323.*fcs_pow(u,8)*(3. - 69.*w2 + 23.*v2*(-1. + 25.*w2)));
        case 3 : return 4.2567525e7*fcs_pow(t,15)*v*w*(-1020.*u2*(3. - 19.*w2 + 19.*v2*(-1. + 7.*w2)) + 22287.*fcs_pow(u,8)*(3. - 25.*w2 + 25.*v2*(-1. + 9.*w2)) + 5.*(9. - 51.*w2 + 17.*v2*(-3. + 19.*w2)) + 9690.*fcs_pow(u,4)*(3. - 21.*w2 + 7.*v2*(-3. + 23.*w2)) - 9044.*fcs_pow(u,6)*(9. - 69.*w2 + 23.*v2*(-3. + 25.*w2)));
      }
      case 4 : switch(k){
        case 0 : return 467775.*fcs_pow(t,13)*(7.*(1. - 26.*v2 + 65.*fcs_pow(v,4)) - 364.*u2*(1. - 30.*v2 + 85.*fcs_pow(v,4)) - 6188.*fcs_pow(u,6)*(1. - 38.*v2 + 133.*fcs_pow(v,4)) + 4199.*fcs_pow(u,8)*(1. - 42.*v2 + 161.*fcs_pow(v,4)) + 910.*fcs_pow(u,4)*(3. - 102.*v2 + 323.*fcs_pow(v,4)));
        case 1 : return -4.2567525e7*fcs_pow(t,14)*(1. - 30.*v2 + 85.*fcs_pow(v,4) + 510.*fcs_pow(u,4)*(1. - 38.*v2 + 133.*fcs_pow(v,4)) - 1292.*fcs_pow(u,6)*(1. - 42.*v2 + 161.*fcs_pow(v,4)) - 20.*u2*(3. - 102.*v2 + 323.*fcs_pow(v,4)) + 323.*fcs_pow(u,8)*(3. - 138.*v2 + 575.*fcs_pow(v,4)))*w;
        case 2 : return 4.2567525e7*fcs_pow(t,15)*(-1. + 15.*w2 + v2*(30. - 510.*w2) + 85.*fcs_pow(v,4)*(-1. + 19.*w2) - 20.*u2*(-3. + 51.*w2 - 102.*v2*(-1. + 19.*w2) + 323.*fcs_pow(v,4)*(-1. + 21.*w2)) + 510.*fcs_pow(u,4)*(-1. + 19.*w2 + v2*(38. - 798.*w2) + 133.*fcs_pow(v,4)*(-1. + 23.*w2)) - 1292.*fcs_pow(u,6)*(-1. + 21.*w2 + v2*(42. - 966.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) + 323.*fcs_pow(u,8)*(-3. + 69.*w2 - 138.*v2*(-1. + 25.*w2) + 575.*fcs_pow(v,4)*(-1. + 27.*w2)));
        case 3 : return -1.27702575e8*fcs_pow(t,16)*w*(5.*(-3. + 17.*w2 + 323.*fcs_pow(v,4)*(-1. + 7.*w2) - 34.*v2*(-3. + 19.*w2)) - 340.*u2*(-3. + 19.*w2 - 114.*v2*(-1. + 7.*w2) + 133.*fcs_pow(v,4)*(-3. + 23.*w2)) - 9044.*fcs_pow(u,6)*(-3. + 23.*w2 + 575.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) + 3230.*fcs_pow(u,4)*(-3. + 21.*w2 - 42.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)) + 7429.*fcs_pow(u,8)*(-3. + 25.*w2 - 150.*v2*(-1. + 9.*w2) + 225.*fcs_pow(v,4)*(-3. + 29.*w2)));
        case 4 : return 1.27702575e8*fcs_pow(t,17)*(5.*(3. - 102.*w2 + 323.*fcs_pow(w,4) - 102.*v2*(1. - 38.*w2 + 133.*fcs_pow(w,4)) + 323.*fcs_pow(v,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4))) - 9044.*fcs_pow(u,6)*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 138.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) + 9690.*fcs_pow(u,4)*(1. - 42.*w2 + 161.*fcs_pow(w,4) + 161.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 14.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) - 340.*u2*(3. - 114.*w2 + 399.*fcs_pow(w,4) - 114.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 22287.*fcs_pow(u,8)*(1. - 50.*w2 + 225.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 75.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return -4.2567525e7*fcs_pow(t,14)*v*(5. - 50.*v2 + 85.*fcs_pow(v,4) + 1615.*fcs_pow(u,8)*(3. - 46.*v2 + 115.*fcs_pow(v,4)) - 1292.*fcs_pow(u,6)*(5. - 70.*v2 + 161.*fcs_pow(v,4)) - 20.*u2*(15. - 170.*v2 + 323.*fcs_pow(v,4)) + 170.*fcs_pow(u,4)*(15. - 190.*v2 + 399.*fcs_pow(v,4)));
        case 1 : return 2.12837625e8*fcs_pow(t,15)*v*(15. - 170.*v2 + 323.*fcs_pow(v,4) - 9044.*fcs_pow(u,6)*(3. - 46.*v2 + 115.*fcs_pow(v,4)) + 7429.*fcs_pow(u,8)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) + 1938.*fcs_pow(u,4)*(5. - 70.*v2 + 161.*fcs_pow(v,4)) - 68.*u2*(15. - 190.*v2 + 399.*fcs_pow(v,4)))*w;
        case 2 : return -2.12837625e8*fcs_pow(t,16)*v*(-15. + 255.*w2 - 170.*v2*(-1. + 19.*w2) + 323.*fcs_pow(v,4)*(-1. + 21.*w2) - 68.*u2*(-15. + 285.*w2 - 190.*v2*(-1. + 21.*w2) + 399.*fcs_pow(v,4)*(-1. + 23.*w2)) + 1938.*fcs_pow(u,4)*(-5. + 105.*w2 - 70.*v2*(-1. + 23.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) - 9044.*fcs_pow(u,6)*(-3. + 69.*w2 - 46.*v2*(-1. + 25.*w2) + 115.*fcs_pow(v,4)*(-1. + 27.*w2)) + 7429.*fcs_pow(u,8)*(-3. + 75.*w2 - 50.*v2*(-1. + 27.*w2) + 135.*fcs_pow(v,4)*(-1. + 29.*w2)));
        case 3 : return 1.0854718875e10*fcs_pow(t,17)*v*w*(-15. + 95.*w2 - 190.*v2*(-1. + 7.*w2) + 133.*fcs_pow(v,4)*(-3. + 23.*w2) + 1330.*fcs_pow(u,4)*(-9. + 69.*w2 + 345.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) - 76.*u2*(15.*(-1. + 7.*w2) - 70.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)) - 12236.*fcs_pow(u,6)*(-3. + 25.*w2 - 50.*v2*(-1. + 9.*w2) + 45.*fcs_pow(v,4)*(-3. + 29.*w2)) + 6555.*fcs_pow(u,8)*(-5. + 45.*w2 + v2*(90. - 870.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)));
        case 4 : return -1.0854718875e10*fcs_pow(t,18)*v*(15.*(1. - 38.*w2 + 133.*fcs_pow(w,4)) - 190.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4)) + 3990.*fcs_pow(u,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 46.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 115.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) - 76.*u2*(15.*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 70.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 6555.*fcs_pow(u,8)*(5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 30.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 12236.*fcs_pow(u,6)*(3. - 150.*w2 + 675.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 45.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
        case 5 : return 1.031198293125e12*fcs_pow(t,19)*v*w*(-70.*v2*(3. - 46.*w2 + 115.*fcs_pow(w,4)) + 161.*fcs_pow(v,4)*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 3.*(5. - 70.*w2 + 161.*fcs_pow(w,4)) - 140.*u2*(9. - 138.*w2 + 345.*fcs_pow(w,4) - 46.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 69.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4))) + 4830.*fcs_pow(u,4)*(3. - 50.*w2 + 135.*fcs_pow(w,4) - 10.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 9.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) + 1035.*fcs_pow(u,8)*(45. - 870.*w2 + 2697.*fcs_pow(w,4) + 2697.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 58.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 1932.*fcs_pow(u,6)*(5.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 30.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 87.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))));
      }
      case 6 : switch(k){
        case 0 : return 2.12837625e8*fcs_pow(t,15)*(-1. + 45.*v2 - 255.*fcs_pow(v,4) + 323.*fcs_pow(v,6) - 1292.*fcs_pow(u,6)*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) + 969.*fcs_pow(u,8)*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 12.*u2*(-5. + 255.*v2 - 1615.*fcs_pow(v,4) + 2261.*fcs_pow(v,6)) + 102.*fcs_pow(u,4)*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)));
        case 1 : return -6.38512875e8*fcs_pow(t,16)*(-5. + 255.*v2 - 1615.*fcs_pow(v,4) + 2261.*fcs_pow(v,6) + 3230.*fcs_pow(u,4)*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) - 9044.*fcs_pow(u,6)*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 7429.*fcs_pow(u,8)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 68.*u2*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)))*w;
        case 2 : return 6.38512875e8*fcs_pow(t,17)*(5. - 85.*w2 + 255.*v2*(-1. + 19.*w2) - 1615.*fcs_pow(v,4)*(-1. + 21.*w2) + 2261.*fcs_pow(v,6)*(-1. + 23.*w2) - 68.*u2*(5. - 95.*w2 + 285.*v2*(-1. + 21.*w2) - 1995.*fcs_pow(v,4)*(-1. + 23.*w2) + 3059.*fcs_pow(v,6)*(-1. + 25.*w2)) + 3230.*fcs_pow(u,4)*(1. - 21.*w2 + 63.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 805.*fcs_pow(v,6)*(-1. + 27.*w2)) - 9044.*fcs_pow(u,6)*(1. - 23.*w2 + 69.*v2*(-1. + 25.*w2) - 575.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) + 7429.*fcs_pow(u,8)*(1. - 25.*w2 + 75.*v2*(-1. + 27.*w2) - 675.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)));
        case 3 : return -1.0854718875e10*fcs_pow(t,18)*w*(15. - 95.*w2 + 855.*v2*(-1. + 7.*w2) - 1995.*fcs_pow(v,4)*(-3. + 23.*w2) + 3059.*fcs_pow(v,6)*(-3. + 25.*w2) - 1140.*u2*(1. - 7.*w2 + 805.*fcs_pow(v,6)*(-1. + 9.*w2) + 21.*v2*(-3. + 23.*w2) - 161.*fcs_pow(v,4)*(-3. + 25.*w2)) + 3990.*fcs_pow(u,4)*(3. - 23.*w2 - 1725.*fcs_pow(v,4)*(-1. + 9.*w2) + 69.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)) + 6555.*fcs_pow(u,8)*(5. - 45.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 135.*v2*(-3. + 29.*w2) - 1305.*fcs_pow(v,4)*(-3. + 31.*w2)) - 12236.*fcs_pow(u,6)*(3. - 25.*w2 + 225.*v2*(-1. + 9.*w2) - 675.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)));
        case 4 : return 3.2564156625e10*fcs_pow(t,19)*(-5. + 190.*w2 - 665.*fcs_pow(w,4) + 285.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 3059.*fcs_pow(v,6)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 665.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4)) - 380.*u2*(-1. + 42.*w2 - 161.*fcs_pow(w,4) - 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 805.*fcs_pow(v,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 21.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 2185.*fcs_pow(u,8)*(-5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 3915.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 135.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 12236.*fcs_pow(u,6)*(-1. + 50.*w2 - 225.*fcs_pow(w,4) + 75.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 225.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 1330.*fcs_pow(u,4)*(-3. + 138.*w2 - 575.*fcs_pow(w,4) + 207.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 1725.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
        case 5 : return -3.093594879375e12*fcs_pow(t,20)*w*(-5. + 70.*w2 - 161.*fcs_pow(w,4) + 105.*v2*(3. - 46.*w2 + 115.*fcs_pow(w,4)) - 805.*fcs_pow(v,4)*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 805.*fcs_pow(v,6)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 140.*u2*(-3. + 46.*w2 - 115.*fcs_pow(w,4) + 69.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 345.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 207.*fcs_pow(v,6)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) + 1035.*fcs_pow(u,8)*(-15. + 290.*w2 - 899.*fcs_pow(w,4) - 13485.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 87.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 3220.*fcs_pow(u,6)*(-5. + 90.*w2 - 261.*fcs_pow(w,4) + 8091.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 27.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 261.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 1610.*fcs_pow(u,4)*(-3. + 50.*w2 - 135.*fcs_pow(w,4) + 45.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 135.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))));
        case 6 : return 1.5467974396875e13*fcs_pow(t,21)*(1. - 63.*w2 + 483.*fcs_pow(w,4) - 805.*fcs_pow(w,6) + 63.*v2*(-1. + 69.*w2 - 575.*fcs_pow(w,4) + 1035.*fcs_pow(w,6)) - 483.*fcs_pow(v,4)*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) + 161.*fcs_pow(v,6)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 621.*fcs_pow(u,8)*(5. - 435.*w2 + 4495.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) + 435.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 4495.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6))) - 644.*fcs_pow(u,6)*(5. - 405.*w2 + 3915.*fcs_pow(w,4) - 8091.*fcs_pow(w,6) - 3915.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 81.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 966.*fcs_pow(u,4)*(1. - 75.*w2 + 675.*fcs_pow(w,4) - 1305.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 15.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 135.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 84.*u2*(1. - 69.*w2 + 575.*fcs_pow(w,4) - 1035.*fcs_pow(w,6) + 69.*v2*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 115.*fcs_pow(v,4)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 207.*fcs_pow(v,6)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))));
      }
      case 7 : switch(k){
        case 0 : return -6.38512875e8*fcs_pow(t,16)*v*(7.*(-5. + 85.*v2 - 323.*fcs_pow(v,4) + 323.*fcs_pow(v,6)) - 476.*u2*(-5. + 95.*v2 - 399.*fcs_pow(v,4) + 437.*fcs_pow(v,6)) + 4522.*fcs_pow(u,4)*(-5. + 105.*v2 - 483.*fcs_pow(v,4) + 575.*fcs_pow(v,6)) - 9044.*fcs_pow(u,6)*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 7429.*fcs_pow(u,8)*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)));
        case 1 : return 1.0854718875e10*fcs_pow(t,17)*v*(7.*(-5. + 95.*v2 - 399.*fcs_pow(v,4) + 437.*fcs_pow(v,6)) - 532.*u2*(-5. + 105.*v2 - 483.*fcs_pow(v,4) + 575.*fcs_pow(v,6)) + 3990.*fcs_pow(u,4)*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 12236.*fcs_pow(u,6)*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 2185.*fcs_pow(u,8)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)))*w;
        case 2 : return -1.0854718875e10*fcs_pow(t,18)*v*(7.*(5. - 95.*w2 + 95.*v2*(-1. + 21.*w2) - 399.*fcs_pow(v,4)*(-1. + 23.*w2) + 437.*fcs_pow(v,6)*(-1. + 25.*w2)) - 532.*u2*(5. - 105.*w2 + 105.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 575.*fcs_pow(v,6)*(-1. + 27.*w2)) + 3990.*fcs_pow(u,4)*(7. - 161.*w2 + 161.*v2*(-1. + 25.*w2) - 805.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) - 12236.*fcs_pow(u,6)*(7. - 175.*w2 + 175.*v2*(-1. + 27.*w2) - 945.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) + 2185.*fcs_pow(u,8)*(35. - 945.*w2 + 945.*v2*(-1. + 29.*w2) - 5481.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)));
        case 3 : return 6.18718975875e11*fcs_pow(t,19)*v*w*(7.*(5. - 35.*w2 + 575.*fcs_pow(v,6)*(-1. + 9.*w2) + 35.*v2*(-3. + 23.*w2) - 161.*fcs_pow(v,4)*(-3. + 25.*w2)) - 140.*u2*(21. - 161.*w2 - 2415.*fcs_pow(v,4)*(-1. + 9.*w2) + 161.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)) - 3220.*fcs_pow(u,6)*(35. - 315.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 1827.*fcs_pow(v,4)*(-3. + 31.*w2)) + 1610.*fcs_pow(u,4)*(21. - 175.*w2 + 525.*v2*(-1. + 9.*w2) - 945.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)) + 1035.*fcs_pow(u,8)*(35.*(3. - 29.*w2) - 18879.*fcs_pow(v,4)*(-1. + 11.*w2) + 1015.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)));
        case 4 : return -6.18718975875e11*fcs_pow(t,20)*v*(7.*(-5. + 210.*w2 - 805.*fcs_pow(w,4) - 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 35.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) - 3220.*fcs_pow(u,6)*(-35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 5481.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 4830.*fcs_pow(u,4)*(-7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 175.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 315.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 140.*u2*(483.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 2415.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 7.*(3. - 138.*w2 + 575.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 1035.*fcs_pow(u,8)*(3045.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18879.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 35.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return 1.08275820778125e14*fcs_pow(t,21)*v*w*(-7.*(3. - 46.*w2 + 115.*fcs_pow(w,4)) + 161.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 483.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 207.*fcs_pow(v,6)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 828.*fcs_pow(u,6)*(-18879.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 203.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 138.*fcs_pow(u,4)*(40455.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 35.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 315.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 1827.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 92.*u2*(-7.*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 105.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 189.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 6003.*fcs_pow(u,8)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 465.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 1023.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 341.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))));
        case 6 : return -3.24827462334375e14*fcs_pow(t,22)*v*(7. - 483.*w2 + 4025.*fcs_pow(w,4) - 7245.*fcs_pow(w,6) + 161.*v2*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 161.*fcs_pow(v,4)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 207.*fcs_pow(v,6)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) - 828.*fcs_pow(u,6)*(1015.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 6293.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 7.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 230.*fcs_pow(u,4)*(-5481.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 189.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 92.*u2*(7. - 525.*w2 + 4725.*fcs_pow(w,4) - 9135.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 35.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 189.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 6003.*fcs_pow(u,8)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 155.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 1023.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))));
        case 7 : return 7.471031633690625e15*fcs_pow(t,23)*v*w*(1890.*fcs_pow(u,4)*(35. - 1015.*w2 + 6293.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 203.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6))) - 28.*u2*(40455.*fcs_pow(v,6)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 5481.*fcs_pow(v,4)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 5.*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 135.*v2*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 7.*(7. - 175.*w2 + 945.*fcs_pow(w,4) - 1305.*fcs_pow(w,6) + 261.*fcs_pow(v,6)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 5.*v2*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 27.*fcs_pow(v,4)*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 7308.*fcs_pow(u,6)*(7.*(5. - 155.*w2 + 1023.*fcs_pow(w,4) - 1705.*fcs_pow(w,6)) + 1085.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 7161.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 8091.*fcs_pow(u,8)*(-35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 1155.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
      }
      case 8 : switch(k){
        case 0 : return 6.38512875e8*fcs_pow(t,17)*(22610.*fcs_pow(u,4)*(1. - 84.*v2 + 966.*fcs_pow(v,4) - 3220.*fcs_pow(v,6) + 3105.*fcs_pow(v,8)) + 7.*(5. - 340.*v2 + 3230.*fcs_pow(v,4) - 9044.*fcs_pow(v,6) + 7429.*fcs_pow(v,8)) - 476.*u2*(5. - 380.*v2 + 3990.*fcs_pow(v,4) - 12236.*fcs_pow(v,6) + 10925.*fcs_pow(v,8)) - 9044.*fcs_pow(u,6)*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) + 7429.*fcs_pow(u,8)*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)));
        case 1 : return -1.0854718875e10*fcs_pow(t,18)*(-2660.*u2*(1. - 84.*v2 + 966.*fcs_pow(v,4) - 3220.*fcs_pow(v,6) + 3105.*fcs_pow(v,8)) + 7.*(5. - 380.*v2 + 3990.*fcs_pow(v,4) - 12236.*fcs_pow(v,6) + 10925.*fcs_pow(v,8)) + 3990.*fcs_pow(u,4)*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) - 12236.*fcs_pow(u,6)*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) + 2185.*fcs_pow(u,8)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)))*w;
        case 2 : return 1.0854718875e10*fcs_pow(t,19)*(7.*(-5. + 95.*w2 - 380.*v2*(-1. + 21.*w2) + 3990.*fcs_pow(v,4)*(-1. + 23.*w2) - 12236.*fcs_pow(v,6)*(-1. + 25.*w2) + 10925.*fcs_pow(v,8)*(-1. + 27.*w2)) - 2660.*u2*(-1. + 21.*w2 - 84.*v2*(-1. + 23.*w2) + 966.*fcs_pow(v,4)*(-1. + 25.*w2) - 3220.*fcs_pow(v,6)*(-1. + 27.*w2) + 3105.*fcs_pow(v,8)*(-1. + 29.*w2)) + 3990.*fcs_pow(u,4)*(-7. + 161.*w2 - 644.*v2*(-1. + 25.*w2) + 8050.*fcs_pow(v,4)*(-1. + 27.*w2) - 28980.*fcs_pow(v,6)*(-1. + 29.*w2) + 30015.*fcs_pow(v,8)*(-1. + 31.*w2)) - 12236.*fcs_pow(u,6)*(-7. + 175.*w2 - 700.*v2*(-1. + 27.*w2) + 9450.*fcs_pow(v,4)*(-1. + 29.*w2) - 36540.*fcs_pow(v,6)*(-1. + 31.*w2) + 40455.*fcs_pow(v,8)*(-1. + 33.*w2)) + 2185.*fcs_pow(u,8)*(-35. + 945.*w2 - 3780.*v2*(-1. + 29.*w2) + 54810.*fcs_pow(v,4)*(-1. + 31.*w2) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2)));
        case 3 : return -2.1655164155625e13*fcs_pow(t,20)*w*(-1. + 7.*w2 + v2*(84. - 644.*w2) - 3220.*fcs_pow(v,6)*(-1. + 9.*w2) + 322.*fcs_pow(v,4)*(-3. + 25.*w2) + 1035.*fcs_pow(v,8)*(-3. + 29.*w2) + 46.*fcs_pow(u,4)*(-2100.*v2*(-1. + 9.*w2) + 121365.*fcs_pow(v,8)*(-1. + 11.*w2) + 7.*(-3. + 25.*w2) + 9450.*fcs_pow(v,4)*(-3. + 29.*w2) - 36540.*fcs_pow(v,6)*(-3. + 31.*w2)) - 4.*u2*(24150.*fcs_pow(v,4)*(-1. + 9.*w2) + 7.*(-3. + 23.*w2) - 644.*v2*(-3. + 25.*w2) - 28980.*fcs_pow(v,6)*(-3. + 29.*w2) + 30015.*fcs_pow(v,8)*(-3. + 31.*w2)) - 92.*fcs_pow(u,6)*(35.*(-1. + 9.*w2) - 226548.*fcs_pow(v,6)*(-1. + 11.*w2) - 1260.*v2*(-3. + 29.*w2) + 18270.*fcs_pow(v,4)*(-3. + 31.*w2) + 89001.*fcs_pow(v,8)*(-3. + 35.*w2)) + 207.*fcs_pow(u,8)*(26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 5.*(-3. + 29.*w2) - 580.*v2*(-3. + 31.*w2) - 39556.*fcs_pow(v,6)*(-3. + 35.*w2) + 49445.*fcs_pow(v,8)*(-3. + 37.*w2)));
        case 4 : return 2.1655164155625e13*fcs_pow(t,21)*(1. - 42.*w2 + 161.*fcs_pow(w,4) + 966.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 3220.*fcs_pow(v,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 28.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4)) + 1035.*fcs_pow(v,8)*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 138.*fcs_pow(u,4)*(7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 700.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 36540.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 40455.*fcs_pow(v,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 3150.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 4.*u2*(-1932.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 24150.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 90045.*fcs_pow(v,8)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 7.*(3. - 138.*w2 + 575.*fcs_pow(w,4)) - 28980.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 207.*fcs_pow(u,8)*(-1740.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 148335.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 92.*fcs_pow(u,6)*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 54810.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 226548.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 1260.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return -1.08275820778125e14*fcs_pow(t,22)*w*(7.*(3. - 46.*w2 + 115.*fcs_pow(w,4)) - 644.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 4830.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 5796.*fcs_pow(v,6)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 6003.*fcs_pow(v,8)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 690.*fcs_pow(u,4)*(-226548.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 7.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 252.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 3654.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 92.*u2*(121365.*fcs_pow(v,8)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 7.*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 420.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 1890.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 7308.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 6003.*fcs_pow(u,8)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 10230.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 37851.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 5796.*fcs_pow(u,6)*(15. - 290.*w2 + 899.*fcs_pow(w,4) + 26970.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 116.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))));
        case 6 : return 3.24827462334375e14*fcs_pow(t,23)*(7.*(-1. + 69.*w2 - 575.*fcs_pow(w,4) + 1035.*fcs_pow(w,6)) - 644.*v2*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) + 30015.*fcs_pow(v,8)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 1610.*fcs_pow(v,4)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 5796.*fcs_pow(v,6)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 230.*fcs_pow(u,4)*(54810.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 226548.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 267003.*fcs_pow(v,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 756.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 92.*u2*(7.*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 36540.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 40455.*fcs_pow(v,8)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 140.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 1890.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 5796.*fcs_pow(u,6)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) - 580.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8990.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 39556.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 6003.*fcs_pow(u,8)*(5.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 620.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return -7.471031633690625e15*fcs_pow(t,24)*w*(-140.*u2*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 10962.*fcs_pow(v,4)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 108.*v2*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 7.*(-7. + 175.*w2 - 945.*fcs_pow(w,4) + 1305.*fcs_pow(w,6) + 40455.*fcs_pow(v,8)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 7308.*fcs_pow(v,6)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 20.*v2*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 270.*fcs_pow(v,4)*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 1890.*fcs_pow(u,4)*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) + 62930.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 276892.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 812.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 7308.*fcs_pow(u,6)*(-4340.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 71610.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 7.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 8091.*fcs_pow(u,8)*(35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 15873.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
        case 8 : return 7.471031633690625e15*fcs_pow(t,25)*(-140.*u2*(35. - 3780.*w2 + 54810.*fcs_pow(w,4) - 226548.*fcs_pow(w,6) + 267003.*fcs_pow(w,8) + 54810.*fcs_pow(v,4)*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 226548.*fcs_pow(v,6)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 267003.*fcs_pow(v,8)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 756.*v2*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8))) + 7.*(7. - 700.*w2 + 9450.*fcs_pow(w,4) - 36540.*fcs_pow(w,6) + 40455.*fcs_pow(w,8) - 36540.*fcs_pow(v,6)*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 40455.*fcs_pow(v,8)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 1890.*fcs_pow(v,4)*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) - 20.*v2*(35. - 3780.*w2 + 54810.*fcs_pow(w,4) - 226548.*fcs_pow(w,6) + 267003.*fcs_pow(w,8))) + 1890.*fcs_pow(u,4)*(-4060.*v2*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 62930.*fcs_pow(v,4)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 276892.*fcs_pow(v,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 7.*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) + 9889.*fcs_pow(v,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 8091.*fcs_pow(u,8)*(35.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 79365.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 11396.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 7308.*fcs_pow(u,6)*(35.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 4340.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 9548.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
      }
    }
    case 9 : switch(j){
      case 0 : switch(k){
        case 0 : return -2835.*fcs_pow(t,10)*u*(315. - 4620.*u2 + 18018.*fcs_pow(u,4) - 25740.*fcs_pow(u,6) + 12155.*fcs_pow(u,8));
      }
      case 1 : switch(k){
        case 0 : return 155925.*fcs_pow(t,11)*u*(63. - 1092.*u2 + 4914.*fcs_pow(u,4) - 7956.*fcs_pow(u,6) + 4199.*fcs_pow(u,8))*v;
        case 1 : return -6.081075e6*fcs_pow(t,12)*u*(21. - 420.*u2 + 2142.*fcs_pow(u,4) - 3876.*fcs_pow(u,6) + 2261.*fcs_pow(u,8))*v*w;
      }
      case 2 : switch(k){
        case 0 : return -155925.*fcs_pow(t,12)*u*(-63. + 819.*v2 - 1092.*u2*(-1. + 15.*v2) + 4914.*fcs_pow(u,4)*(-1. + 17.*v2) - 7956.*fcs_pow(u,6)*(-1. + 19.*v2) + 4199.*fcs_pow(u,8)*(-1. + 21.*v2));
        case 1 : return 6.081075e6*fcs_pow(t,13)*u*(-21. + 315.*v2 - 420.*u2*(-1. + 17.*v2) + 2142.*fcs_pow(u,4)*(-1. + 19.*v2) - 3876.*fcs_pow(u,6)*(-1. + 21.*v2) + 2261.*fcs_pow(u,8)*(-1. + 23.*v2))*w;
        case 2 : return -6.081075e6*fcs_pow(t,14)*u*(21.*(1. - 15.*w2 + 15.*v2*(-1. + 17.*w2)) - 420.*u2*(1. - 17.*w2 + 17.*v2*(-1. + 19.*w2)) + 2142.*fcs_pow(u,4)*(1. - 19.*w2 + 19.*v2*(-1. + 21.*w2)) - 3876.*fcs_pow(u,6)*(1. - 21.*w2 + 21.*v2*(-1. + 23.*w2)) + 2261.*fcs_pow(u,8)*(1. - 23.*w2 + 23.*v2*(-1. + 25.*w2)));
      }
      case 3 : switch(k){
        case 0 : return 6.081075e6*fcs_pow(t,13)*u*v*(63.*(-1. + 5.*v2) - 11628.*fcs_pow(u,6)*(-1. + 7.*v2) - 420.*u2*(-3. + 17.*v2) + 2142.*fcs_pow(u,4)*(-3. + 19.*v2) + 2261.*fcs_pow(u,8)*(-3. + 23.*v2));
        case 1 : return -4.2567525e7*fcs_pow(t,14)*u*v*(17442.*fcs_pow(u,4)*(-1. + 7.*v2) + 45.*(-3. + 17.*v2) - 1020.*u2*(-3. + 19.*v2) - 11628.*fcs_pow(u,6)*(-3. + 23.*v2) + 7429.*fcs_pow(u,8)*(-3. + 25.*v2))*w;
        case 2 : return 4.2567525e7*fcs_pow(t,15)*u*v*(45.*(3. - 51.*w2 + 17.*v2*(-1. + 19.*w2)) - 1020.*u2*(3. - 57.*w2 + 19.*v2*(-1. + 21.*w2)) + 17442.*fcs_pow(u,4)*(1. - 21.*w2 + 7.*v2*(-1. + 23.*w2)) - 11628.*fcs_pow(u,6)*(3. - 69.*w2 + 23.*v2*(-1. + 25.*w2)) + 7429.*fcs_pow(u,8)*(3. - 75.*w2 + 25.*v2*(-1. + 27.*w2)));
        case 3 : return -6.512831325e9*fcs_pow(t,16)*u*v*w*(15.*(3. - 19.*w2 + 19.*v2*(-1. + 7.*w2)) - 5244.*fcs_pow(u,6)*(3. - 25.*w2 + 25.*v2*(-1. + 9.*w2)) - 380.*u2*(3. - 21.*w2 + 7.*v2*(-3. + 23.*w2)) + 798.*fcs_pow(u,4)*(9. - 69.*w2 + 23.*v2*(-3. + 25.*w2)) + 10925.*fcs_pow(u,8)*(1. - 9.*w2 + v2*(-9. + 87.*w2)));
      }
      case 4 : switch(k){
        case 0 : return -6.081075e6*fcs_pow(t,14)*u*(63.*(1. - 30.*v2 + 85.*fcs_pow(v,4)) + 6426.*fcs_pow(u,4)*(1. - 38.*v2 + 133.*fcs_pow(v,4)) - 11628.*fcs_pow(u,6)*(1. - 42.*v2 + 161.*fcs_pow(v,4)) - 420.*u2*(3. - 102.*v2 + 323.*fcs_pow(v,4)) + 2261.*fcs_pow(u,8)*(3. - 138.*v2 + 575.*fcs_pow(v,4)));
        case 1 : return 1.27702575e8*fcs_pow(t,15)*u*(-1020.*u2*(1. - 38.*v2 + 133.*fcs_pow(v,4)) + 5814.*fcs_pow(u,4)*(1. - 42.*v2 + 161.*fcs_pow(v,4)) + 7429.*fcs_pow(u,8)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) + 15.*(3. - 102.*v2 + 323.*fcs_pow(v,4)) - 3876.*fcs_pow(u,6)*(3. - 138.*v2 + 575.*fcs_pow(v,4)))*w;
        case 2 : return -1.27702575e8*fcs_pow(t,16)*u*(15.*(-3. + 51.*w2 - 102.*v2*(-1. + 19.*w2) + 323.*fcs_pow(v,4)*(-1. + 21.*w2)) - 1020.*u2*(-1. + 19.*w2 + v2*(38. - 798.*w2) + 133.*fcs_pow(v,4)*(-1. + 23.*w2)) + 5814.*fcs_pow(u,4)*(-1. + 21.*w2 + v2*(42. - 966.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) - 3876.*fcs_pow(u,6)*(-3. + 69.*w2 - 138.*v2*(-1. + 25.*w2) + 575.*fcs_pow(v,4)*(-1. + 27.*w2)) + 7429.*fcs_pow(u,8)*(-1. + 25.*w2 - 50.*v2*(-1. + 27.*w2) + 225.*fcs_pow(v,4)*(-1. + 29.*w2)));
        case 3 : return 6.512831325e9*fcs_pow(t,17)*u*w*(15.*(-3. + 19.*w2 - 114.*v2*(-1. + 7.*w2) + 133.*fcs_pow(v,4)*(-3. + 23.*w2)) + 2394.*fcs_pow(u,4)*(-3. + 23.*w2 + 575.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) - 380.*u2*(-3. + 21.*w2 - 42.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)) - 5244.*fcs_pow(u,6)*(-3. + 25.*w2 - 150.*v2*(-1. + 9.*w2) + 225.*fcs_pow(v,4)*(-3. + 29.*w2)) + 10925.*fcs_pow(u,8)*(-1. + 9.*w2 + v2*(54. - 522.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)));
        case 4 : return -6.512831325e9*fcs_pow(t,18)*u*(2394.*fcs_pow(u,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 138.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) - 1140.*u2*(1. - 42.*w2 + 161.*fcs_pow(w,4) + 161.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 14.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 15.*(3. - 114.*w2 + 399.*fcs_pow(w,4) - 114.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 10925.*fcs_pow(u,8)*(1. - 54.*w2 + 261.*fcs_pow(w,4) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 15732.*fcs_pow(u,6)*(1. - 50.*w2 + 225.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 75.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return 4.2567525e7*fcs_pow(t,15)*u*v*(-58140.*fcs_pow(u,6)*(3. - 46.*v2 + 115.*fcs_pow(v,4)) + 37145.*fcs_pow(u,8)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) + 17442.*fcs_pow(u,4)*(5. - 70.*v2 + 161.*fcs_pow(v,4)) + 45.*(15. - 170.*v2 + 323.*fcs_pow(v,4)) - 1020.*u2*(15. - 190.*v2 + 399.*fcs_pow(v,4)));
        case 1 : return -1.0854718875e10*fcs_pow(t,16)*u*v*(45. - 570.*v2 + 1197.*fcs_pow(v,4) + 2394.*fcs_pow(u,4)*(3. - 46.*v2 + 115.*fcs_pow(v,4)) - 5244.*fcs_pow(u,6)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) - 228.*u2*(5. - 70.*v2 + 161.*fcs_pow(v,4)) + 2185.*fcs_pow(u,8)*(5. - 90.*v2 + 261.*fcs_pow(v,4)))*w;
        case 2 : return 1.0854718875e10*fcs_pow(t,17)*u*v*(3.*(-15. + 285.*w2 - 190.*v2*(-1. + 21.*w2) + 399.*fcs_pow(v,4)*(-1. + 23.*w2)) - 228.*u2*(-5. + 105.*w2 - 70.*v2*(-1. + 23.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) + 2394.*fcs_pow(u,4)*(-3. + 69.*w2 - 46.*v2*(-1. + 25.*w2) + 115.*fcs_pow(v,4)*(-1. + 27.*w2)) - 5244.*fcs_pow(u,6)*(-3. + 75.*w2 - 50.*v2*(-1. + 27.*w2) + 135.*fcs_pow(v,4)*(-1. + 29.*w2)) + 2185.*fcs_pow(u,8)*(-5. + 135.*w2 - 90.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)));
        case 3 : return -6.18718975875e11*fcs_pow(t,18)*u*v*w*(-140.*u2*(-9. + 69.*w2 + 345.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) + 3.*(15.*(-1. + 7.*w2) - 70.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)) + 2898.*fcs_pow(u,4)*(-3. + 25.*w2 - 50.*v2*(-1. + 9.*w2) + 45.*fcs_pow(v,4)*(-3. + 29.*w2)) + 345.*fcs_pow(u,8)*(-45. + 435.*w2 + 2697.*fcs_pow(v,4)*(-1. + 11.*w2) - 290.*v2*(-3. + 31.*w2)) - 4140.*fcs_pow(u,6)*(-5. + 45.*w2 + v2*(90. - 870.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)));
        case 4 : return 1.856156927625e12*fcs_pow(t,19)*u*v*(15.*(1. - 42.*w2 + 161.*fcs_pow(w,4)) + 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 70.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4)) - 140.*u2*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 46.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 115.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) + 345.*fcs_pow(u,8)*(-290.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 1380.*fcs_pow(u,6)*(5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 30.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 966.*fcs_pow(u,4)*(3. - 150.*w2 + 675.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 45.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
        case 5 : return -9.280784638125e12*fcs_pow(t,20)*u*v*w*(10005.*fcs_pow(u,8)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 310.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4))) + 35.*(9. - 138.*w2 + 345.*fcs_pow(w,4) - 46.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 69.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4))) - 3220.*u2*(3. - 50.*w2 + 135.*fcs_pow(w,4) - 10.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 9.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) - 4140.*fcs_pow(u,6)*(45. - 870.*w2 + 2697.*fcs_pow(w,4) + 2697.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 58.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 2898.*fcs_pow(u,4)*(5.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 30.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 87.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))));
      }
      case 6 : switch(k){
        case 0 : return -6.38512875e8*fcs_pow(t,16)*u*(5814.*fcs_pow(u,4)*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) - 11628.*fcs_pow(u,6)*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 7429.*fcs_pow(u,8)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 9.*(-5. + 255.*v2 - 1615.*fcs_pow(v,4) + 2261.*fcs_pow(v,6)) - 204.*u2*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)));
        case 1 : return 1.0854718875e10*fcs_pow(t,17)*u*(-1140.*u2*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) + 7182.*fcs_pow(u,4)*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 15732.*fcs_pow(u,6)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 9.*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)) + 2185.*fcs_pow(u,8)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)))*w;
        case 2 : return -1.0854718875e10*fcs_pow(t,18)*u*(9.*(5. - 95.*w2 + 285.*v2*(-1. + 21.*w2) - 1995.*fcs_pow(v,4)*(-1. + 23.*w2) + 3059.*fcs_pow(v,6)*(-1. + 25.*w2)) - 1140.*u2*(1. - 21.*w2 + 63.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 805.*fcs_pow(v,6)*(-1. + 27.*w2)) + 7182.*fcs_pow(u,4)*(1. - 23.*w2 + 69.*v2*(-1. + 25.*w2) - 575.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) - 15732.*fcs_pow(u,6)*(1. - 25.*w2 + 75.*v2*(-1. + 27.*w2) - 675.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) + 2185.*fcs_pow(u,8)*(5. - 135.*w2 + 405.*v2*(-1. + 29.*w2) - 3915.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)));
        case 3 : return 1.856156927625e12*fcs_pow(t,19)*u*w*(15.*(1. - 7.*w2 + 805.*fcs_pow(v,6)*(-1. + 9.*w2) + 21.*v2*(-3. + 23.*w2) - 161.*fcs_pow(v,4)*(-3. + 25.*w2)) - 140.*u2*(3. - 23.*w2 - 1725.*fcs_pow(v,4)*(-1. + 9.*w2) + 69.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)) - 1380.*fcs_pow(u,6)*(5. - 45.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 135.*v2*(-3. + 29.*w2) - 1305.*fcs_pow(v,4)*(-3. + 31.*w2)) + 966.*fcs_pow(u,4)*(3. - 25.*w2 + 225.*v2*(-1. + 9.*w2) - 675.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)) + 345.*fcs_pow(u,8)*(15. - 145.*w2 - 13485.*fcs_pow(v,4)*(-1. + 11.*w2) + 435.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)));
        case 4 : return -1.856156927625e12*fcs_pow(t,20)*u*(15.*(-1. + 42.*w2 - 161.*fcs_pow(w,4) - 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 805.*fcs_pow(v,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 21.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) - 1380.*fcs_pow(u,6)*(-5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 3915.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 135.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 2898.*fcs_pow(u,4)*(-1. + 50.*w2 - 225.*fcs_pow(w,4) + 75.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 225.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 140.*u2*(-3. + 138.*w2 - 575.*fcs_pow(w,4) + 207.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 1725.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 345.*fcs_pow(u,8)*(1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 13485.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return 4.6403923190625e13*fcs_pow(t,21)*u*w*(21.*(-3. + 46.*w2 - 115.*fcs_pow(w,4) + 69.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 345.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 207.*fcs_pow(v,6)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) - 2484.*fcs_pow(u,6)*(-15. + 290.*w2 - 899.*fcs_pow(w,4) - 13485.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 87.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 2898.*fcs_pow(u,4)*(-5. + 90.*w2 - 261.*fcs_pow(w,4) + 8091.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 27.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 261.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 644.*u2*(-3. + 50.*w2 - 135.*fcs_pow(w,4) + 45.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 135.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 2001.*fcs_pow(u,8)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 1395.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 5115.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))));
        case 6 : return -1.39211769571875e14*fcs_pow(t,22)*u*(-2484.*fcs_pow(u,6)*(5. - 435.*w2 + 4495.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) + 435.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 4495.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6))) + 966.*fcs_pow(u,4)*(5. - 405.*w2 + 3915.*fcs_pow(w,4) - 8091.*fcs_pow(w,6) - 3915.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 81.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 644.*u2*(1. - 75.*w2 + 675.*fcs_pow(w,4) - 1305.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 15.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 135.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 21.*(1. - 69.*w2 + 575.*fcs_pow(w,4) - 1035.*fcs_pow(w,6) + 69.*v2*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 115.*fcs_pow(v,4)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 207.*fcs_pow(v,6)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 2001.*fcs_pow(u,8)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 5115.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 2387.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))));
      }
      case 7 : switch(k){
        case 0 : return 1.0854718875e10*fcs_pow(t,17)*u*v*(63.*(-5. + 95.*v2 - 399.*fcs_pow(v,4) + 437.*fcs_pow(v,6)) - 1596.*u2*(-5. + 105.*v2 - 483.*fcs_pow(v,4) + 575.*fcs_pow(v,6)) + 7182.*fcs_pow(u,4)*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 15732.*fcs_pow(u,6)*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 2185.*fcs_pow(u,8)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)));
        case 1 : return -1.856156927625e12*fcs_pow(t,18)*u*v*(7.*(-5. + 105.*v2 - 483.*fcs_pow(v,4) + 575.*fcs_pow(v,6)) - 140.*u2*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 966.*fcs_pow(u,4)*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 460.*fcs_pow(u,6)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 345.*fcs_pow(u,8)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)))*w;
        case 2 : return 1.856156927625e12*fcs_pow(t,19)*u*v*(7.*(5. - 105.*w2 + 105.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 575.*fcs_pow(v,6)*(-1. + 27.*w2)) - 140.*u2*(7. - 161.*w2 + 161.*v2*(-1. + 25.*w2) - 805.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) + 966.*fcs_pow(u,4)*(7. - 175.*w2 + 175.*v2*(-1. + 27.*w2) - 945.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) - 460.*fcs_pow(u,6)*(35. - 945.*w2 + 945.*v2*(-1. + 29.*w2) - 5481.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) + 345.*fcs_pow(u,8)*(35. - 1015.*w2 + 1015.*v2*(-1. + 31.*w2) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)));
        case 3 : return -9.280784638125e12*fcs_pow(t,20)*u*v*w*(21.*(21. - 161.*w2 - 2415.*fcs_pow(v,4)*(-1. + 9.*w2) + 161.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)) + 2898.*fcs_pow(u,4)*(35. - 315.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 1827.*fcs_pow(v,4)*(-3. + 31.*w2)) - 644.*u2*(21. - 175.*w2 + 525.*v2*(-1. + 9.*w2) - 945.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)) - 2484.*fcs_pow(u,6)*(35.*(3. - 29.*w2) - 18879.*fcs_pow(v,4)*(-1. + 11.*w2) + 1015.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) + 14007.*fcs_pow(u,8)*(15. - 155.*w2 + 465.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 1705.*fcs_pow(v,6)*(-3. + 37.*w2)));
        case 4 : return 2.7842353914375e13*fcs_pow(t,21)*u*v*(966.*fcs_pow(u,4)*(-35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 5481.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 644.*u2*(-7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 175.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 315.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 7.*(483.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 2415.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 7.*(3. - 138.*w2 + 575.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 14007.*fcs_pow(u,8)*(-5.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 155.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 1705.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 828.*fcs_pow(u,6)*(3045.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18879.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 35.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return -2.241309490107187e16*fcs_pow(t,22)*u*v*w*(-7.*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 105.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 189.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 54.*fcs_pow(u,4)*(-18879.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 203.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 4.*u2*(40455.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 35.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 315.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 1827.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 2697.*fcs_pow(u,8)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 55.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 1044.*fcs_pow(u,6)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 465.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 1023.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 341.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))));
        case 6 : return 2.241309490107187e16*fcs_pow(t,23)*u*v*(162.*fcs_pow(u,4)*(1015.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 6293.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 7.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 20.*u2*(-5481.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 189.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 3.*(7. - 525.*w2 + 4725.*fcs_pow(w,4) - 9135.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 35.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 189.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 3132.*fcs_pow(u,6)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 155.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 1023.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 2697.*fcs_pow(u,8)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return -6.723928470321562e16*fcs_pow(t,24)*u*v*w*(-1260.*u2*(35. - 1015.*w2 + 6293.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 203.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6))) + 7.*(40455.*fcs_pow(v,6)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 5481.*fcs_pow(v,4)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 5.*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 135.*v2*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 10962.*fcs_pow(u,4)*(7.*(5. - 155.*w2 + 1023.*fcs_pow(w,4) - 1705.*fcs_pow(w,6)) + 1085.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 7161.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 32364.*fcs_pow(u,6)*(-35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 1155.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 29667.*fcs_pow(u,8)*(-35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 35.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 481.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
      }
      case 8 : switch(k){
        case 0 : return -1.0854718875e10*fcs_pow(t,18)*u*(-7980.*u2*(1. - 84.*v2 + 966.*fcs_pow(v,4) - 3220.*fcs_pow(v,6) + 3105.*fcs_pow(v,8)) + 63.*(5. - 380.*v2 + 3990.*fcs_pow(v,4) - 12236.*fcs_pow(v,6) + 10925.*fcs_pow(v,8)) + 7182.*fcs_pow(u,4)*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) - 15732.*fcs_pow(u,6)*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) + 2185.*fcs_pow(u,8)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)));
        case 1 : return 1.856156927625e12*fcs_pow(t,19)*u*(35.*(1. - 84.*v2 + 966.*fcs_pow(v,4) - 3220.*fcs_pow(v,6) + 3105.*fcs_pow(v,8)) - 140.*u2*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) + 966.*fcs_pow(u,4)*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) + 2415.*fcs_pow(u,8)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) - 460.*fcs_pow(u,6)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)))*w;
        case 2 : return -1.856156927625e12*fcs_pow(t,20)*u*(35.*(-1. + 21.*w2 - 84.*v2*(-1. + 23.*w2) + 966.*fcs_pow(v,4)*(-1. + 25.*w2) - 3220.*fcs_pow(v,6)*(-1. + 27.*w2) + 3105.*fcs_pow(v,8)*(-1. + 29.*w2)) - 140.*u2*(-7. + 161.*w2 - 644.*v2*(-1. + 25.*w2) + 8050.*fcs_pow(v,4)*(-1. + 27.*w2) - 28980.*fcs_pow(v,6)*(-1. + 29.*w2) + 30015.*fcs_pow(v,8)*(-1. + 31.*w2)) + 966.*fcs_pow(u,4)*(-7. + 175.*w2 - 700.*v2*(-1. + 27.*w2) + 9450.*fcs_pow(v,4)*(-1. + 29.*w2) - 36540.*fcs_pow(v,6)*(-1. + 31.*w2) + 40455.*fcs_pow(v,8)*(-1. + 33.*w2)) - 460.*fcs_pow(u,6)*(-35. + 945.*w2 - 3780.*v2*(-1. + 29.*w2) + 54810.*fcs_pow(v,4)*(-1. + 31.*w2) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2)) + 2415.*fcs_pow(u,8)*(-5. + 145.*w2 - 580.*v2*(-1. + 31.*w2) + 8990.*fcs_pow(v,4)*(-1. + 33.*w2) - 39556.*fcs_pow(v,6)*(-1. + 35.*w2) + 49445.*fcs_pow(v,8)*(-1. + 37.*w2)));
        case 3 : return 6.4965492466875e13*fcs_pow(t,21)*u*w*(-92.*u2*(-2100.*v2*(-1. + 9.*w2) + 121365.*fcs_pow(v,8)*(-1. + 11.*w2) + 7.*(-3. + 25.*w2) + 9450.*fcs_pow(v,4)*(-3. + 29.*w2) - 36540.*fcs_pow(v,6)*(-3. + 31.*w2)) + 3.*(24150.*fcs_pow(v,4)*(-1. + 9.*w2) + 7.*(-3. + 23.*w2) - 644.*v2*(-3. + 25.*w2) - 28980.*fcs_pow(v,6)*(-3. + 29.*w2) + 30015.*fcs_pow(v,8)*(-3. + 31.*w2)) + 414.*fcs_pow(u,4)*(35.*(-1. + 9.*w2) - 226548.*fcs_pow(v,6)*(-1. + 11.*w2) - 1260.*v2*(-3. + 29.*w2) + 18270.*fcs_pow(v,4)*(-3. + 31.*w2) + 89001.*fcs_pow(v,8)*(-3. + 35.*w2)) + 10005.*fcs_pow(u,8)*(-3. + 31.*w2 - 372.*v2*(-1. + 11.*w2) + 37851.*fcs_pow(v,8)*(-1. + 13.*w2) + 2046.*fcs_pow(v,4)*(-3. + 35.*w2) - 9548.*fcs_pow(v,6)*(-3. + 37.*w2)) - 2484.*fcs_pow(u,6)*(26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 5.*(-3. + 29.*w2) - 580.*v2*(-3. + 31.*w2) - 39556.*fcs_pow(v,6)*(-3. + 35.*w2) + 49445.*fcs_pow(v,8)*(-3. + 37.*w2)));
        case 4 : return -1.94896477400625e14*fcs_pow(t,22)*u*(-1932.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 24150.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 90045.*fcs_pow(v,8)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 7.*(3. - 138.*w2 + 575.*fcs_pow(w,4)) - 28980.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 92.*u2*(7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 700.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 36540.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 40455.*fcs_pow(v,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 3150.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 10005.*fcs_pow(u,8)*(1. - 62.*w2 + 341.*fcs_pow(w,4) - 124.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 682.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 828.*fcs_pow(u,6)*(-1740.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 148335.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 138.*fcs_pow(u,4)*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 54810.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 226548.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 1260.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return 2.241309490107187e16*fcs_pow(t,23)*u*w*(121365.*fcs_pow(v,8)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 7.*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 420.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 1890.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 7308.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) - 20.*u2*(-226548.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 7.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 252.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 3654.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 1044.*fcs_pow(u,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 10230.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 37851.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 378.*fcs_pow(u,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4) + 26970.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 116.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 2697.*fcs_pow(u,8)*(5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 220.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 770.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return -2.241309490107187e16*fcs_pow(t,24)*u*(-20.*u2*(54810.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 226548.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 267003.*fcs_pow(v,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 756.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 3.*(7.*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 36540.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 40455.*fcs_pow(v,8)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 140.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 1890.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 1134.*fcs_pow(u,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) - 580.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8990.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 39556.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 2697.*fcs_pow(u,8)*(5.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 79365.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 3132.*fcs_pow(u,6)*(5.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 620.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return 6.723928470321562e16*fcs_pow(t,25)*u*w*(35.*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 10962.*fcs_pow(v,4)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 108.*v2*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 1260.*u2*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) + 62930.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 276892.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 812.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 10962.*fcs_pow(u,4)*(-4340.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 71610.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 7.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 29667.*fcs_pow(u,8)*(35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 140.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2590.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 13468.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 32364.*fcs_pow(u,6)*(35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 15873.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
        case 8 : return -6.723928470321562e16*fcs_pow(t,26)*u*(35.*(35. - 3780.*w2 + 54810.*fcs_pow(w,4) - 226548.*fcs_pow(w,6) + 267003.*fcs_pow(w,8) + 54810.*fcs_pow(v,4)*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 226548.*fcs_pow(v,6)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 267003.*fcs_pow(v,8)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 756.*v2*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8))) - 1260.*u2*(-4060.*v2*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 62930.*fcs_pow(v,4)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 276892.*fcs_pow(v,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 7.*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) + 9889.*fcs_pow(v,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 1.038345e6*fcs_pow(u,8)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8) + 19721.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 1924.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 4.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 74.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 32364.*fcs_pow(u,6)*(35.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 79365.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 11396.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 10962.*fcs_pow(u,4)*(35.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 4340.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 9548.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
      }
      case 9 : switch(k){
        case 0 : return 5.568470782875e12*fcs_pow(t,19)*u*v*(21.*(5. - 140.*v2 + 966.*fcs_pow(v,4) - 2300.*fcs_pow(v,6) + 1725.*fcs_pow(v,8)) - 140.*u2*(21. - 644.*v2 + 4830.*fcs_pow(v,4) - 12420.*fcs_pow(v,6) + 10005.*fcs_pow(v,8)) + 966.*fcs_pow(u,4)*(21. - 700.*v2 + 5670.*fcs_pow(v,4) - 15660.*fcs_pow(v,6) + 13485.*fcs_pow(v,8)) - 1380.*fcs_pow(u,6)*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) + 115.*fcs_pow(u,8)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)));
        case 1 : return -2.7842353914375e13*fcs_pow(t,20)*u*v*(21.*(21. - 644.*v2 + 4830.*fcs_pow(v,4) - 12420.*fcs_pow(v,6) + 10005.*fcs_pow(v,8)) - 644.*u2*(21. - 700.*v2 + 5670.*fcs_pow(v,4) - 15660.*fcs_pow(v,6) + 13485.*fcs_pow(v,8)) + 2898.*fcs_pow(u,4)*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) + 4669.*fcs_pow(u,8)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) - 828.*fcs_pow(u,6)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)))*w;
        case 2 : return 2.7842353914375e13*fcs_pow(t,21)*u*v*(21.*(-21. + 483.*w2 - 644.*v2*(-1. + 25.*w2) + 4830.*fcs_pow(v,4)*(-1. + 27.*w2) - 12420.*fcs_pow(v,6)*(-1. + 29.*w2) + 10005.*fcs_pow(v,8)*(-1. + 31.*w2)) - 644.*u2*(-21. + 525.*w2 - 700.*v2*(-1. + 27.*w2) + 5670.*fcs_pow(v,4)*(-1. + 29.*w2) - 15660.*fcs_pow(v,6)*(-1. + 31.*w2) + 13485.*fcs_pow(v,8)*(-1. + 33.*w2)) + 2898.*fcs_pow(u,4)*(-35. + 945.*w2 - 1260.*v2*(-1. + 29.*w2) + 10962.*fcs_pow(v,4)*(-1. + 31.*w2) - 32364.*fcs_pow(v,6)*(-1. + 33.*w2) + 29667.*fcs_pow(v,8)*(-1. + 35.*w2)) - 828.*fcs_pow(u,6)*(315.*(-1. + 29.*w2) - 12180.*v2*(-1. + 31.*w2) + 113274.*fcs_pow(v,4)*(-1. + 33.*w2) - 356004.*fcs_pow(v,6)*(-1. + 35.*w2) + 346115.*fcs_pow(v,8)*(-1. + 37.*w2)) + 4669.*fcs_pow(u,8)*(45.*(-1. + 31.*w2) - 1860.*v2*(-1. + 33.*w2) + 18414.*fcs_pow(v,4)*(-1. + 35.*w2) - 61380.*fcs_pow(v,6)*(-1. + 37.*w2) + 63085.*fcs_pow(v,8)*(-1. + 39.*w2)));
        case 3 : return -1.344785694064312e16*fcs_pow(t,22)*u*v*w*(3.*(-700.*v2*(-1. + 9.*w2) + 13485.*fcs_pow(v,8)*(-1. + 11.*w2) + 7.*(-3. + 25.*w2) + 1890.*fcs_pow(v,4)*(-3. + 29.*w2) - 5220.*fcs_pow(v,6)*(-3. + 31.*w2)) - 60.*u2*(35.*(-1. + 9.*w2) - 32364.*fcs_pow(v,6)*(-1. + 11.*w2) - 420.*v2*(-3. + 29.*w2) + 3654.*fcs_pow(v,4)*(-3. + 31.*w2) + 9889.*fcs_pow(v,8)*(-3. + 35.*w2)) - 1044.*fcs_pow(u,6)*(-45. + 465.*w2 - 1860.*v2*(-1. + 11.*w2) + 63085.*fcs_pow(v,8)*(-1. + 13.*w2) + 6138.*fcs_pow(v,4)*(-3. + 35.*w2) - 20460.*fcs_pow(v,6)*(-3. + 37.*w2)) + 18.*fcs_pow(u,4)*(339822.*fcs_pow(v,4)*(-1. + 11.*w2) + 315.*(-3. + 29.*w2) - 12180.*v2*(-3. + 31.*w2) - 356004.*fcs_pow(v,6)*(-3. + 35.*w2) + 346115.*fcs_pow(v,8)*(-3. + 37.*w2)) + 4495.*fcs_pow(u,8)*(-9. + 99.*w2 - 14652.*fcs_pow(v,6)*(-1. + 13.*w2) - 132.*v2*(-3. + 35.*w2) + 1386.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)));
        case 4 : return 1.344785694064312e16*fcs_pow(t,23)*u*v*(3.*(21.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 700.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 15660.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 13485.*fcs_pow(v,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 1890.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 1044.*fcs_pow(u,6)*(45.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 1860.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 61380.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 54.*fcs_pow(u,4)*(-12180.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 113274.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 346115.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 105.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 118668.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 60.*u2*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 10962.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 32364.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 420.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 4495.*fcs_pow(u,8)*(9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 4158.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 132.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -6.723928470321562e16*fcs_pow(t,24)*u*v*w*(9.*(-161820.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 49445.*fcs_pow(v,8)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 35.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 420.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 3654.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 10962.*fcs_pow(u,4)*(45. - 930.*w2 + 3069.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 4092.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 60.*u2*(339822.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 356004.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 63.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 2436.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 69223.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 9889.*fcs_pow(u,8)*(45.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 13986.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 420.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 17316.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 32364.*fcs_pow(u,6)*(45.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 660.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1386.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return 2.017178541096469e17*fcs_pow(t,25)*u*v*(15.*(10962.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 32364.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 29667.*fcs_pow(v,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 252.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 60.*u2*(-12180.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 113274.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 356004.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 63.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 69223.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 9889.*fcs_pow(u,8)*(45.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 86580.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 32364.*fcs_pow(u,6)*(15.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 3654.*fcs_pow(u,4)*(45.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 1860.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 18414.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 12276.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return -2.017178541096469e17*fcs_pow(t,26)*u*v*w*(315.*(113274.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 356004.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 2436.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 9.*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 12180.*u2*(-13020.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 128898.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 63.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 12276.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 9889.*fcs_pow(u,8)*(2.9680105e7*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 3.54978e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 15540.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 181818.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 356004.*fcs_pow(u,6)*(315.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 17316.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 339822.*fcs_pow(u,4)*(105.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 5291.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
        case 8 : return 2.017178541096469e17*fcs_pow(t,27)*u*v*(315.*(-12180.*v2*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 113274.*fcs_pow(v,4)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 356004.*fcs_pow(v,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 63.*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) + 9889.*fcs_pow(v,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 346115.*fcs_pow(u,8)*(-709956.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 25974.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 9.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 444.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 356004.*fcs_pow(u,6)*(315.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 690235.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 86580.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 420.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 4662.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 339822.*fcs_pow(u,4)*(105.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 26455.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 1386.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 4884.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 12180.*u2*(315.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 13020.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 128898.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 12276.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -5.264835992261783e19*fcs_pow(t,28)*u*v*w*(441595.*fcs_pow(u,8)*(-825084.*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 9594.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 156.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 12276.*fcs_pow(u,6)*(2.9680105e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 1.18326e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 15540.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 60606.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 42966.*fcs_pow(u,4)*(105.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 98605.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 13986.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 17316.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 140.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 13020.*u2*(105.*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) - 1540.*v2*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 14652.*fcs_pow(v,6)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5291.*fcs_pow(v,8)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 462.*fcs_pow(v,4)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 315.*(-4340.*v2*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) + 14322.*fcs_pow(v,4)*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 7.*(45. - 1860.*w2 + 18414.*fcs_pow(w,4) - 61380.*fcs_pow(w,6) + 63085.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1364.*fcs_pow(v,6)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))));
      }
    }
    case 10 : switch(j){
      case 0 : switch(k){
        case 0 : return 14175.*fcs_pow(t,11)*(-63. + 3465.*u2 - 30030.*fcs_pow(u,4) + 90090.*fcs_pow(u,6) - 109395.*fcs_pow(u,8) + 46189.*fcs_pow(u,10));
      }
      case 1 : switch(k){
        case 0 : return -467775.*fcs_pow(t,12)*(-21. + 1365.*u2 - 13650.*fcs_pow(u,4) + 46410.*fcs_pow(u,6) - 62985.*fcs_pow(u,8) + 29393.*fcs_pow(u,10))*v;
        case 1 : return 4.2567525e7*fcs_pow(t,13)*(-3. + 225.*u2 - 2550.*fcs_pow(u,4) + 9690.*fcs_pow(u,6) - 14535.*fcs_pow(u,8) + 7429.*fcs_pow(u,10))*v*w;
      }
      case 2 : switch(k){
        case 0 : return 467775.*fcs_pow(t,13)*(21. - 273.*v2 + 1365.*u2*(-1. + 15.*v2) - 13650.*fcs_pow(u,4)*(-1. + 17.*v2) + 46410.*fcs_pow(u,6)*(-1. + 19.*v2) - 62985.*fcs_pow(u,8)*(-1. + 21.*v2) + 29393.*fcs_pow(u,10)*(-1. + 23.*v2));
        case 1 : return -4.2567525e7*fcs_pow(t,14)*(3. - 45.*v2 + 225.*u2*(-1. + 17.*v2) - 2550.*fcs_pow(u,4)*(-1. + 19.*v2) + 9690.*fcs_pow(u,6)*(-1. + 21.*v2) - 14535.*fcs_pow(u,8)*(-1. + 23.*v2) + 7429.*fcs_pow(u,10)*(-1. + 25.*v2))*w;
        case 2 : return 4.2567525e7*fcs_pow(t,15)*(-3. + 45.*w2 + v2*(45. - 765.*w2) + 225.*u2*(1. - 17.*w2 + 17.*v2*(-1. + 19.*w2)) - 2550.*fcs_pow(u,4)*(1. - 19.*w2 + 19.*v2*(-1. + 21.*w2)) + 9690.*fcs_pow(u,6)*(1. - 21.*w2 + 21.*v2*(-1. + 23.*w2)) - 14535.*fcs_pow(u,8)*(1. - 23.*w2 + 23.*v2*(-1. + 25.*w2)) + 7429.*fcs_pow(u,10)*(1. - 25.*w2 + 25.*v2*(-1. + 27.*w2)));
      }
      case 3 : switch(k){
        case 0 : return -4.2567525e7*fcs_pow(t,14)*v*(9. - 45.*v2 + 29070.*fcs_pow(u,6)*(-1. + 7.*v2) + 225.*u2*(-3. + 17.*v2) - 2550.*fcs_pow(u,4)*(-3. + 19.*v2) - 14535.*fcs_pow(u,8)*(-3. + 23.*v2) + 7429.*fcs_pow(u,10)*(-3. + 25.*v2));
        case 1 : return 6.38512875e8*fcs_pow(t,15)*v*(9. - 51.*v2 - 9690.*fcs_pow(u,4)*(-1. + 7.*v2) + 37145.*fcs_pow(u,10)*(-1. + 9.*v2) + 255.*u2*(-3. + 19.*v2) + 13566.*fcs_pow(u,6)*(-3. + 23.*v2) - 22287.*fcs_pow(u,8)*(-3. + 25.*v2))*w;
        case 2 : return -6.38512875e8*fcs_pow(t,16)*v*(-9. + 153.*w2 + v2*(51. - 969.*w2) + 255.*u2*(3. - 57.*w2 + 19.*v2*(-1. + 21.*w2)) - 9690.*fcs_pow(u,4)*(1. - 21.*w2 + 7.*v2*(-1. + 23.*w2)) + 13566.*fcs_pow(u,6)*(3. - 69.*w2 + 23.*v2*(-1. + 25.*w2)) - 22287.*fcs_pow(u,8)*(3. - 75.*w2 + 25.*v2*(-1. + 27.*w2)) + 37145.*fcs_pow(u,10)*(1. - 27.*w2 + 9.*v2*(-1. + 29.*w2)));
        case 3 : return 3.2564156625e10*fcs_pow(t,17)*v*w*(-9. + 57.*w2 - 57.*v2*(-1. + 7.*w2) + 18354.*fcs_pow(u,6)*(3. - 25.*w2 + 25.*v2*(-1. + 9.*w2)) + 285.*u2*(3. - 21.*w2 + 7.*v2*(-3. + 23.*w2)) - 1330.*fcs_pow(u,4)*(9. - 69.*w2 + 23.*v2*(-3. + 25.*w2)) + 6555.*fcs_pow(u,10)*(9. - 87.*w2 + 29.*v2*(-3. + 31.*w2)) - 98325.*fcs_pow(u,8)*(1. - 9.*w2 + v2*(-9. + 87.*w2)));
      }
      case 4 : switch(k){
        case 0 : return 1.27702575e8*fcs_pow(t,15)*(-3. + 90.*v2 - 255.*fcs_pow(v,4) - 2550.*fcs_pow(u,4)*(1. - 38.*v2 + 133.*fcs_pow(v,4)) + 9690.*fcs_pow(u,6)*(1. - 42.*v2 + 161.*fcs_pow(v,4)) + 7429.*fcs_pow(u,10)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) + 75.*u2*(3. - 102.*v2 + 323.*fcs_pow(v,4)) - 4845.*fcs_pow(u,8)*(3. - 138.*v2 + 575.*fcs_pow(v,4)));
        case 1 : return -6.38512875e8*fcs_pow(t,16)*(-9. + 306.*v2 - 969.*fcs_pow(v,4) + 765.*u2*(1. - 38.*v2 + 133.*fcs_pow(v,4)) - 9690.*fcs_pow(u,4)*(1. - 42.*v2 + 161.*fcs_pow(v,4)) - 66861.*fcs_pow(u,8)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) + 37145.*fcs_pow(u,10)*(1. - 54.*v2 + 261.*fcs_pow(v,4)) + 13566.*fcs_pow(u,6)*(3. - 138.*v2 + 575.*fcs_pow(v,4)))*w;
        case 2 : return 6.38512875e8*fcs_pow(t,17)*(9. - 153.*w2 + 306.*v2*(-1. + 19.*w2) - 969.*fcs_pow(v,4)*(-1. + 21.*w2) + 765.*u2*(-1. + 19.*w2 + v2*(38. - 798.*w2) + 133.*fcs_pow(v,4)*(-1. + 23.*w2)) - 9690.*fcs_pow(u,4)*(-1. + 21.*w2 + v2*(42. - 966.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) + 13566.*fcs_pow(u,6)*(-3. + 69.*w2 - 138.*v2*(-1. + 25.*w2) + 575.*fcs_pow(v,4)*(-1. + 27.*w2)) - 66861.*fcs_pow(u,8)*(-1. + 25.*w2 - 50.*v2*(-1. + 27.*w2) + 225.*fcs_pow(v,4)*(-1. + 29.*w2)) + 37145.*fcs_pow(u,10)*(-1. + 27.*w2 - 54.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)));
        case 3 : return -9.7692469875e10*fcs_pow(t,18)*w*(3. - 19.*w2 + 114.*v2*(-1. + 7.*w2) - 133.*fcs_pow(v,4)*(-3. + 23.*w2) - 1330.*fcs_pow(u,4)*(-3. + 23.*w2 + 575.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) + 95.*u2*(-3. + 21.*w2 - 42.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)) + 6118.*fcs_pow(u,6)*(-3. + 25.*w2 - 150.*v2*(-1. + 9.*w2) + 225.*fcs_pow(v,4)*(-3. + 29.*w2)) + 6555.*fcs_pow(u,10)*(-3. + 29.*w2 + 899.*fcs_pow(v,4)*(-1. + 11.*w2) - 58.*v2*(-3. + 31.*w2)) - 32775.*fcs_pow(u,8)*(-1. + 9.*w2 + v2*(54. - 522.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)));
        case 4 : return 9.7692469875e10*fcs_pow(t,19)*(-3. + 114.*w2 - 399.*fcs_pow(w,4) + 114.*v2*(1. - 42.*w2 + 161.*fcs_pow(w,4)) - 133.*fcs_pow(v,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4)) - 1330.*fcs_pow(u,4)*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 138.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) + 6555.*fcs_pow(u,10)*(3. - 174.*w2 + 899.*fcs_pow(w,4) - 174.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4))) + 285.*u2*(1. - 42.*w2 + 161.*fcs_pow(w,4) + 161.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 14.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) - 32775.*fcs_pow(u,8)*(1. - 54.*w2 + 261.*fcs_pow(w,4) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 18354.*fcs_pow(u,6)*(1. - 50.*w2 + 225.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 75.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return -6.38512875e8*fcs_pow(t,16)*v*(-45. + 510.*v2 - 969.*fcs_pow(v,4) + 67830.*fcs_pow(u,6)*(3. - 46.*v2 + 115.*fcs_pow(v,4)) - 111435.*fcs_pow(u,8)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) - 9690.*fcs_pow(u,4)*(5. - 70.*v2 + 161.*fcs_pow(v,4)) + 37145.*fcs_pow(u,10)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) + 255.*u2*(15. - 190.*v2 + 399.*fcs_pow(v,4)));
        case 1 : return 3.2564156625e10*fcs_pow(t,17)*v*(-15. + 190.*v2 - 399.*fcs_pow(v,4) - 6650.*fcs_pow(u,4)*(3. - 46.*v2 + 115.*fcs_pow(v,4)) + 30590.*fcs_pow(u,6)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) + 285.*u2*(5. - 70.*v2 + 161.*fcs_pow(v,4)) - 32775.*fcs_pow(u,8)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) + 6555.*fcs_pow(u,10)*(15. - 290.*v2 + 899.*fcs_pow(v,4)))*w;
        case 2 : return -3.2564156625e10*fcs_pow(t,18)*v*(15. - 285.*w2 + 190.*v2*(-1. + 21.*w2) - 399.*fcs_pow(v,4)*(-1. + 23.*w2) + 285.*u2*(-5. + 105.*w2 - 70.*v2*(-1. + 23.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) - 6650.*fcs_pow(u,4)*(-3. + 69.*w2 - 46.*v2*(-1. + 25.*w2) + 115.*fcs_pow(v,4)*(-1. + 27.*w2)) + 30590.*fcs_pow(u,6)*(-3. + 75.*w2 - 50.*v2*(-1. + 27.*w2) + 135.*fcs_pow(v,4)*(-1. + 29.*w2)) - 32775.*fcs_pow(u,8)*(-5. + 135.*w2 - 90.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)) + 6555.*fcs_pow(u,10)*(-15. + 435.*w2 - 290.*v2*(-1. + 31.*w2) + 899.*fcs_pow(v,4)*(-1. + 33.*w2)));
        case 3 : return 1.856156927625e12*fcs_pow(t,19)*v*w*(15. - 105.*w2 + 70.*v2*(-3. + 23.*w2) - 161.*fcs_pow(v,4)*(-3. + 25.*w2) + 175.*u2*(-9. + 69.*w2 + 345.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) - 8050.*fcs_pow(u,4)*(-3. + 25.*w2 - 50.*v2*(-1. + 9.*w2) + 45.*fcs_pow(v,4)*(-3. + 29.*w2)) - 5175.*fcs_pow(u,8)*(-45. + 435.*w2 + 2697.*fcs_pow(v,4)*(-1. + 11.*w2) - 290.*v2*(-3. + 31.*w2)) + 24150.*fcs_pow(u,6)*(-5. + 45.*w2 + v2*(90. - 870.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)) + 10005.*fcs_pow(u,10)*(-310.*v2*(-1. + 11.*w2) + 5.*(-3. + 31.*w2) + 341.*fcs_pow(v,4)*(-3. + 35.*w2)));
        case 4 : return -1.856156927625e12*fcs_pow(t,20)*v*(-15.*(1. - 42.*w2 + 161.*fcs_pow(w,4)) - 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 70.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4)) + 525.*u2*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 46.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 115.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) - 15525.*fcs_pow(u,8)*(-290.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 24150.*fcs_pow(u,6)*(5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 30.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 8050.*fcs_pow(u,4)*(3. - 150.*w2 + 675.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 45.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 10005.*fcs_pow(u,10)*(15.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 310.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return 4.6403923190625e13*fcs_pow(t,21)*v*w*(-90045.*fcs_pow(u,8)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 310.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4))) - 7.*(9. - 138.*w2 + 345.*fcs_pow(w,4) - 46.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 69.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4))) + 2415.*u2*(3. - 50.*w2 + 135.*fcs_pow(w,4) - 10.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 9.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) + 14490.*fcs_pow(u,6)*(45. - 870.*w2 + 2697.*fcs_pow(w,4) + 2697.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 58.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 4830.*fcs_pow(u,4)*(5.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 30.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 87.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 62031.*fcs_pow(u,10)*(15.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 110.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))));
      }
      case 6 : switch(k){
        case 0 : return 6.38512875e8*fcs_pow(t,17)*(-48450.*fcs_pow(u,4)*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) + 203490.*fcs_pow(u,6)*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 334305.*fcs_pow(u,8)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 9.*(-5. + 255.*v2 - 1615.*fcs_pow(v,4) + 2261.*fcs_pow(v,6)) + 765.*u2*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)) + 37145.*fcs_pow(u,10)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)));
        case 1 : return -9.7692469875e10*fcs_pow(t,18)*(5. - 285.*v2 + 1995.*fcs_pow(v,4) - 3059.*fcs_pow(v,6) + 475.*u2*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) - 6650.*fcs_pow(u,4)*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 30590.*fcs_pow(u,6)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 10925.*fcs_pow(u,8)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 6555.*fcs_pow(u,10)*(-5. + 435.*v2 - 4495.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)))*w;
        case 2 : return 9.7692469875e10*fcs_pow(t,19)*(-5. + 95.*w2 - 285.*v2*(-1. + 21.*w2) + 1995.*fcs_pow(v,4)*(-1. + 23.*w2) - 3059.*fcs_pow(v,6)*(-1. + 25.*w2) + 475.*u2*(1. - 21.*w2 + 63.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 805.*fcs_pow(v,6)*(-1. + 27.*w2)) - 6650.*fcs_pow(u,4)*(1. - 23.*w2 + 69.*v2*(-1. + 25.*w2) - 575.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) + 30590.*fcs_pow(u,6)*(1. - 25.*w2 + 75.*v2*(-1. + 27.*w2) - 675.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) - 10925.*fcs_pow(u,8)*(5. - 135.*w2 + 405.*v2*(-1. + 29.*w2) - 3915.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) + 6555.*fcs_pow(u,10)*(5. - 145.*w2 + 435.*v2*(-1. + 31.*w2) - 4495.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)));
        case 3 : return -9.280784638125e12*fcs_pow(t,20)*w*(-3.*(1. - 7.*w2 + 805.*fcs_pow(v,6)*(-1. + 9.*w2) + 21.*v2*(-3. + 23.*w2) - 161.*fcs_pow(v,4)*(-3. + 25.*w2)) + 105.*u2*(3. - 23.*w2 - 1725.*fcs_pow(v,4)*(-1. + 9.*w2) + 69.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)) + 4830.*fcs_pow(u,6)*(5. - 45.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 135.*v2*(-3. + 29.*w2) - 1305.*fcs_pow(v,4)*(-3. + 31.*w2)) - 1610.*fcs_pow(u,4)*(3. - 25.*w2 + 225.*v2*(-1. + 9.*w2) - 675.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)) - 3105.*fcs_pow(u,8)*(15. - 145.*w2 - 13485.*fcs_pow(v,4)*(-1. + 11.*w2) + 435.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) + 10005.*fcs_pow(u,10)*(3. - 31.*w2 + 279.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 2387.*fcs_pow(v,6)*(-3. + 37.*w2)));
        case 4 : return 2.7842353914375e13*fcs_pow(t,21)*(1. - 42.*w2 + 161.*fcs_pow(w,4) + 483.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 805.*fcs_pow(v,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 21.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4)) + 1610.*fcs_pow(u,6)*(-5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 3915.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 135.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 1610.*fcs_pow(u,4)*(-1. + 50.*w2 - 225.*fcs_pow(w,4) + 75.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 225.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 35.*u2*(-3. + 138.*w2 - 575.*fcs_pow(w,4) + 207.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 1725.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 10005.*fcs_pow(u,10)*(-1. + 62.*w2 - 341.*fcs_pow(w,4) + 93.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 1035.*fcs_pow(u,8)*(1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 13485.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return -1.39211769571875e14*fcs_pow(t,22)*w*(-7.*(-3. + 46.*w2 - 115.*fcs_pow(w,4) + 69.*v2*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 345.*fcs_pow(v,4)*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 207.*fcs_pow(v,6)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) + 14490.*fcs_pow(u,6)*(-15. + 290.*w2 - 899.*fcs_pow(w,4) - 13485.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 87.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 8050.*fcs_pow(u,4)*(-5. + 90.*w2 - 261.*fcs_pow(w,4) + 8091.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 27.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 261.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 805.*u2*(-3. + 50.*w2 - 135.*fcs_pow(w,4) + 45.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 135.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 62031.*fcs_pow(u,10)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 165.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2849.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 385.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 30015.*fcs_pow(u,8)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 1395.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 5115.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))));
        case 6 : return 1.39211769571875e14*fcs_pow(t,23)*(43470.*fcs_pow(u,6)*(5. - 435.*w2 + 4495.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) + 435.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 4495.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6))) - 8050.*fcs_pow(u,4)*(5. - 405.*w2 + 3915.*fcs_pow(w,4) - 8091.*fcs_pow(w,6) - 3915.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 81.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 2415.*u2*(1. - 75.*w2 + 675.*fcs_pow(w,4) - 1305.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 15.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 135.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 21.*(1. - 69.*w2 + 575.*fcs_pow(w,4) - 1035.*fcs_pow(w,6) + 69.*v2*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 115.*fcs_pow(v,4)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 207.*fcs_pow(v,6)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 90045.*fcs_pow(u,8)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 5115.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 2387.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 62031.*fcs_pow(u,10)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 1155.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2849.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
      }
      case 7 : switch(k){
        case 0 : return -9.7692469875e10*fcs_pow(t,18)*v*(-7.*(-5. + 95.*v2 - 399.*fcs_pow(v,4) + 437.*fcs_pow(v,6)) + 665.*u2*(-5. + 105.*v2 - 483.*fcs_pow(v,4) + 575.*fcs_pow(v,6)) - 6650.*fcs_pow(u,4)*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 30590.*fcs_pow(u,6)*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 10925.*fcs_pow(u,8)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 6555.*fcs_pow(u,10)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)));
        case 1 : return 1.856156927625e12*fcs_pow(t,19)*v*(7.*(5. - 105.*v2 + 483.*fcs_pow(v,4) - 575.*fcs_pow(v,6)) + 525.*u2*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 8050.*fcs_pow(u,4)*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 70035.*fcs_pow(u,10)*(-5. + 155.*v2 - 1023.*fcs_pow(v,4) + 1705.*fcs_pow(v,6)) + 8050.*fcs_pow(u,6)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) - 15525.*fcs_pow(u,8)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)))*w;
        case 2 : return -1.856156927625e12*fcs_pow(t,20)*v*(-7.*(5. - 105.*w2 + 105.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 575.*fcs_pow(v,6)*(-1. + 27.*w2)) + 525.*u2*(7. - 161.*w2 + 161.*v2*(-1. + 25.*w2) - 805.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) - 8050.*fcs_pow(u,4)*(7. - 175.*w2 + 175.*v2*(-1. + 27.*w2) - 945.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) + 8050.*fcs_pow(u,6)*(35. - 945.*w2 + 945.*v2*(-1. + 29.*w2) - 5481.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) - 15525.*fcs_pow(u,8)*(35. - 1015.*w2 + 1015.*v2*(-1. + 31.*w2) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)) + 70035.*fcs_pow(u,10)*(5. - 155.*w2 + 155.*v2*(-1. + 33.*w2) - 1023.*fcs_pow(v,4)*(-1. + 35.*w2) + 1705.*fcs_pow(v,6)*(-1. + 37.*w2)));
        case 3 : return 1.94896477400625e14*fcs_pow(t,21)*v*w*(2415.*fcs_pow(v,4)*(-1. + 9.*w2) + 7.*(-3. + 23.*w2) - 161.*v2*(-3. + 25.*w2) - 1035.*fcs_pow(v,6)*(-3. + 29.*w2) - 1150.*fcs_pow(u,4)*(35. - 315.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 1827.*fcs_pow(v,4)*(-3. + 31.*w2)) + 115.*u2*(21. - 175.*w2 + 525.*v2*(-1. + 9.*w2) - 945.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)) + 2070.*fcs_pow(u,6)*(35.*(3. - 29.*w2) - 18879.*fcs_pow(v,4)*(-1. + 11.*w2) + 1015.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) + 310155.*fcs_pow(u,10)*(1. - 11.*w2 + 407.*fcs_pow(v,6)*(-1. + 13.*w2) + 11.*v2*(-3. + 35.*w2) - 77.*fcs_pow(v,4)*(-3. + 37.*w2)) - 30015.*fcs_pow(u,8)*(15. - 155.*w2 + 465.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 1705.*fcs_pow(v,6)*(-3. + 37.*w2)));
        case 4 : return -1.94896477400625e14*fcs_pow(t,22)*v*(-483.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 2415.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 7.*(3. - 138.*w2 + 575.*fcs_pow(w,4)) - 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 1150.*fcs_pow(u,4)*(-35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 5481.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 345.*u2*(-7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 175.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 315.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 310155.*fcs_pow(u,10)*(-1. + 66.*w2 - 385.*fcs_pow(w,4) - 231.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 11.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 90045.*fcs_pow(u,8)*(-5.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 155.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 1705.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 2070.*fcs_pow(u,6)*(3045.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18879.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 35.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return 2.241309490107187e16*fcs_pow(t,23)*v*w*(7.*(3. - 50.*w2 + 135.*fcs_pow(w,4)) - 105.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 189.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) - 450.*fcs_pow(u,4)*(-18879.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 203.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 15.*u2*(40455.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 35.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 315.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 1827.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 121365.*fcs_pow(u,8)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 55.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 18270.*fcs_pow(u,6)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 465.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 1023.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 341.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 29667.*fcs_pow(u,10)*(-5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 777.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 35.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 481.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return -6.723928470321562e16*fcs_pow(t,24)*v*(7.*(-1. + 75.*w2 - 675.*fcs_pow(w,4) + 1305.*fcs_pow(w,6)) - 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 35.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 189.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) - 450.*fcs_pow(u,4)*(1015.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 6293.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 7.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 25.*u2*(-5481.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 189.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 18270.*fcs_pow(u,6)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 155.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 1023.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 29667.*fcs_pow(u,10)*(5. - 525.*w2 + 6475.*fcs_pow(w,4) - 16835.*fcs_pow(w,6) + 2405.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 35.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 40455.*fcs_pow(u,8)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return 6.723928470321562e16*fcs_pow(t,25)*v*w*(4725.*u2*(35. - 1015.*w2 + 6293.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 203.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6))) - 7.*(40455.*fcs_pow(v,6)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 5481.*fcs_pow(v,4)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 5.*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 135.*v2*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 91350.*fcs_pow(u,4)*(7.*(5. - 155.*w2 + 1023.*fcs_pow(w,4) - 1705.*fcs_pow(w,6)) + 1085.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 7161.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 566370.*fcs_pow(u,6)*(-35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 1155.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 29667.*fcs_pow(u,10)*(98605.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 35.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 10101.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 1.335015e6*fcs_pow(u,8)*(-35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 35.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 481.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
      }
      case 8 : switch(k){
        case 0 : return 9.7692469875e10*fcs_pow(t,19)*(3325.*u2*(1. - 84.*v2 + 966.*fcs_pow(v,4) - 3220.*fcs_pow(v,6) + 3105.*fcs_pow(v,8)) - 7.*(5. - 380.*v2 + 3990.*fcs_pow(v,4) - 12236.*fcs_pow(v,6) + 10925.*fcs_pow(v,8)) - 6650.*fcs_pow(u,4)*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) + 30590.*fcs_pow(u,6)*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) + 45885.*fcs_pow(u,10)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) - 10925.*fcs_pow(u,8)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)));
        case 1 : return -6.4965492466875e13*fcs_pow(t,20)*(-1. + 84.*v2 - 966.*fcs_pow(v,4) + 3220.*fcs_pow(v,6) - 3105.*fcs_pow(v,8) + 10005.*fcs_pow(u,10)*(1. - 124.*v2 + 2046.*fcs_pow(v,4) - 9548.*fcs_pow(v,6) + 12617.*fcs_pow(v,8)) + 15.*u2*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) - 230.*fcs_pow(u,4)*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) - 3105.*fcs_pow(u,8)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) + 230.*fcs_pow(u,6)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)))*w;
        case 2 : return 6.4965492466875e13*fcs_pow(t,21)*(1. - 21.*w2 + 84.*v2*(-1. + 23.*w2) - 966.*fcs_pow(v,4)*(-1. + 25.*w2) + 3220.*fcs_pow(v,6)*(-1. + 27.*w2) - 3105.*fcs_pow(v,8)*(-1. + 29.*w2) + 15.*u2*(-7. + 161.*w2 - 644.*v2*(-1. + 25.*w2) + 8050.*fcs_pow(v,4)*(-1. + 27.*w2) - 28980.*fcs_pow(v,6)*(-1. + 29.*w2) + 30015.*fcs_pow(v,8)*(-1. + 31.*w2)) - 230.*fcs_pow(u,4)*(-7. + 175.*w2 - 700.*v2*(-1. + 27.*w2) + 9450.*fcs_pow(v,4)*(-1. + 29.*w2) - 36540.*fcs_pow(v,6)*(-1. + 31.*w2) + 40455.*fcs_pow(v,8)*(-1. + 33.*w2)) + 230.*fcs_pow(u,6)*(-35. + 945.*w2 - 3780.*v2*(-1. + 29.*w2) + 54810.*fcs_pow(v,4)*(-1. + 31.*w2) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2)) - 3105.*fcs_pow(u,8)*(-5. + 145.*w2 - 580.*v2*(-1. + 31.*w2) + 8990.*fcs_pow(v,4)*(-1. + 33.*w2) - 39556.*fcs_pow(v,6)*(-1. + 35.*w2) + 49445.*fcs_pow(v,8)*(-1. + 37.*w2)) + 10005.*fcs_pow(u,10)*(-1. + 31.*w2 - 124.*v2*(-1. + 33.*w2) + 2046.*fcs_pow(v,4)*(-1. + 35.*w2) - 9548.*fcs_pow(v,6)*(-1. + 37.*w2) + 12617.*fcs_pow(v,8)*(-1. + 39.*w2)));
        case 3 : return -1.94896477400625e14*fcs_pow(t,22)*w*(21. - 161.*w2 - 24150.*fcs_pow(v,4)*(-1. + 9.*w2) + 644.*v2*(-3. + 25.*w2) + 28980.*fcs_pow(v,6)*(-3. + 29.*w2) - 30015.*fcs_pow(v,8)*(-3. + 31.*w2) + 115.*u2*(-2100.*v2*(-1. + 9.*w2) + 121365.*fcs_pow(v,8)*(-1. + 11.*w2) + 7.*(-3. + 25.*w2) + 9450.*fcs_pow(v,4)*(-3. + 29.*w2) - 36540.*fcs_pow(v,6)*(-3. + 31.*w2)) - 1150.*fcs_pow(u,4)*(35.*(-1. + 9.*w2) - 226548.*fcs_pow(v,6)*(-1. + 11.*w2) - 1260.*v2*(-3. + 29.*w2) + 18270.*fcs_pow(v,4)*(-3. + 31.*w2) + 89001.*fcs_pow(v,8)*(-3. + 35.*w2)) - 150075.*fcs_pow(u,8)*(-3. + 31.*w2 - 372.*v2*(-1. + 11.*w2) + 37851.*fcs_pow(v,8)*(-1. + 13.*w2) + 2046.*fcs_pow(v,4)*(-3. + 35.*w2) - 9548.*fcs_pow(v,6)*(-3. + 37.*w2)) + 14490.*fcs_pow(u,6)*(26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 5.*(-3. + 29.*w2) - 580.*v2*(-3. + 31.*w2) - 39556.*fcs_pow(v,6)*(-3. + 35.*w2) + 49445.*fcs_pow(v,8)*(-3. + 37.*w2)) + 310155.*fcs_pow(u,10)*(-1. + 11.*w2 - 11396.*fcs_pow(v,6)*(-1. + 13.*w2) - 44.*v2*(-3. + 35.*w2) + 770.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)));
        case 4 : return 1.94896477400625e14*fcs_pow(t,23)*(1932.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 24150.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 90045.*fcs_pow(v,8)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 7.*(3. - 138.*w2 + 575.*fcs_pow(w,4)) + 28980.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 345.*u2*(7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 700.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 36540.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 40455.*fcs_pow(v,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 3150.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 450225.*fcs_pow(u,8)*(1. - 62.*w2 + 341.*fcs_pow(w,4) - 124.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 682.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 14490.*fcs_pow(u,6)*(-1740.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 148335.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 1150.*fcs_pow(u,4)*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 54810.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 226548.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 1260.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 310155.*fcs_pow(u,10)*(1. - 66.*w2 + 385.*fcs_pow(w,4) + 2310.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 44.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -2.241309490107187e16*fcs_pow(t,24)*w*(-121365.*fcs_pow(v,8)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 7.*(3. - 50.*w2 + 135.*fcs_pow(w,4)) + 420.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 1890.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 7308.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 75.*u2*(-226548.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 7.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 252.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 3654.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 18270.*fcs_pow(u,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 10230.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 37851.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 3150.*fcs_pow(u,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4) + 26970.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 116.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 29667.*fcs_pow(u,10)*(5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 140.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 13468.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 121365.*fcs_pow(u,8)*(5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 220.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 770.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return 6.723928470321562e16*fcs_pow(t,25)*(7. - 525.*w2 + 4725.*fcs_pow(w,4) - 9135.*fcs_pow(w,6) + 36540.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 40455.*fcs_pow(v,8)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 140.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 1890.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 25.*u2*(54810.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 226548.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 267003.*fcs_pow(v,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 756.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 3150.*fcs_pow(u,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) - 580.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8990.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 39556.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 148335.*fcs_pow(u,10)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6) - 13468.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 28.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 518.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 40455.*fcs_pow(u,8)*(5.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 79365.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 18270.*fcs_pow(u,6)*(5.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 620.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return -3.361964235160781e17*fcs_pow(t,26)*w*(-7.*(-35. + 945.*w2 - 5481.*fcs_pow(w,4) + 8091.*fcs_pow(w,6) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 10962.*fcs_pow(v,4)*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 108.*v2*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 945.*u2*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) + 62930.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 276892.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 812.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 18270.*fcs_pow(u,4)*(-4340.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 71610.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 7.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 207669.*fcs_pow(u,10)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 78884.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 148.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 2886.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 267003.*fcs_pow(u,8)*(35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 140.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2590.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 13468.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 113274.*fcs_pow(u,6)*(35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 15873.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
        case 8 : return 2.353374964612547e18*fcs_pow(t,27)*(-35. + 3780.*w2 - 54810.*fcs_pow(w,4) + 226548.*fcs_pow(w,6) - 267003.*fcs_pow(w,8) - 54810.*fcs_pow(v,4)*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 226548.*fcs_pow(v,6)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 267003.*fcs_pow(v,8)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 756.*v2*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) + 135.*u2*(-4060.*v2*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 62930.*fcs_pow(v,4)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 276892.*fcs_pow(v,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 7.*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) + 9889.*fcs_pow(v,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 29667.*fcs_pow(u,10)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8) - 552188.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 14430.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 148.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 1.335015e6*fcs_pow(u,8)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8) + 19721.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 1924.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 4.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 74.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 16182.*fcs_pow(u,6)*(35.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 79365.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 11396.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 2610.*fcs_pow(u,4)*(35.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 4340.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 9548.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
      }
      case 9 : switch(k){
        case 0 : return -5.568470782875e12*fcs_pow(t,20)*v*(-21.*(5. - 140.*v2 + 966.*fcs_pow(v,4) - 2300.*fcs_pow(v,6) + 1725.*fcs_pow(v,8)) + 525.*u2*(21. - 644.*v2 + 4830.*fcs_pow(v,4) - 12420.*fcs_pow(v,6) + 10005.*fcs_pow(v,8)) - 8050.*fcs_pow(u,4)*(21. - 700.*v2 + 5670.*fcs_pow(v,4) - 15660.*fcs_pow(v,6) + 13485.*fcs_pow(v,8)) + 24150.*fcs_pow(u,6)*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) + 23345.*fcs_pow(u,10)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) - 5175.*fcs_pow(u,8)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)));
        case 1 : return 5.84689432201875e14*fcs_pow(t,21)*v*(-21. + 644.*v2 - 4830.*fcs_pow(v,4) + 12420.*fcs_pow(v,6) - 10005.*fcs_pow(v,8) + 103385.*fcs_pow(u,10)*(3. - 132.*v2 + 1386.*fcs_pow(v,4) - 4884.*fcs_pow(v,6) + 5291.*fcs_pow(v,8)) + 115.*u2*(21. - 700.*v2 + 5670.*fcs_pow(v,4) - 15660.*fcs_pow(v,6) + 13485.*fcs_pow(v,8)) - 1150.*fcs_pow(u,4)*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) - 10005.*fcs_pow(u,8)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) + 690.*fcs_pow(u,6)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)))*w;
        case 2 : return -5.84689432201875e14*fcs_pow(t,22)*v*(21. - 483.*w2 + 644.*v2*(-1. + 25.*w2) - 4830.*fcs_pow(v,4)*(-1. + 27.*w2) + 12420.*fcs_pow(v,6)*(-1. + 29.*w2) - 10005.*fcs_pow(v,8)*(-1. + 31.*w2) + 115.*u2*(-21. + 525.*w2 - 700.*v2*(-1. + 27.*w2) + 5670.*fcs_pow(v,4)*(-1. + 29.*w2) - 15660.*fcs_pow(v,6)*(-1. + 31.*w2) + 13485.*fcs_pow(v,8)*(-1. + 33.*w2)) - 1150.*fcs_pow(u,4)*(-35. + 945.*w2 - 1260.*v2*(-1. + 29.*w2) + 10962.*fcs_pow(v,4)*(-1. + 31.*w2) - 32364.*fcs_pow(v,6)*(-1. + 33.*w2) + 29667.*fcs_pow(v,8)*(-1. + 35.*w2)) + 690.*fcs_pow(u,6)*(315.*(-1. + 29.*w2) - 12180.*v2*(-1. + 31.*w2) + 113274.*fcs_pow(v,4)*(-1. + 33.*w2) - 356004.*fcs_pow(v,6)*(-1. + 35.*w2) + 346115.*fcs_pow(v,8)*(-1. + 37.*w2)) - 10005.*fcs_pow(u,8)*(45.*(-1. + 31.*w2) - 1860.*v2*(-1. + 33.*w2) + 18414.*fcs_pow(v,4)*(-1. + 35.*w2) - 61380.*fcs_pow(v,6)*(-1. + 37.*w2) + 63085.*fcs_pow(v,8)*(-1. + 39.*w2)) + 103385.*fcs_pow(u,10)*(-3. + 99.*w2 - 132.*v2*(-1. + 35.*w2) + 1386.*fcs_pow(v,4)*(-1. + 37.*w2) - 4884.*fcs_pow(v,6)*(-1. + 39.*w2) + 5291.*fcs_pow(v,8)*(-1. + 41.*w2)));
        case 3 : return 1.344785694064312e16*fcs_pow(t,23)*v*w*(-3.*(-700.*v2*(-1. + 9.*w2) + 13485.*fcs_pow(v,8)*(-1. + 11.*w2) + 7.*(-3. + 25.*w2) + 1890.*fcs_pow(v,4)*(-3. + 29.*w2) - 5220.*fcs_pow(v,6)*(-3. + 31.*w2)) + 225.*u2*(35.*(-1. + 9.*w2) - 32364.*fcs_pow(v,6)*(-1. + 11.*w2) - 420.*v2*(-3. + 29.*w2) + 3654.*fcs_pow(v,4)*(-3. + 31.*w2) + 9889.*fcs_pow(v,8)*(-3. + 35.*w2)) + 18270.*fcs_pow(u,6)*(-45. + 465.*w2 - 1860.*v2*(-1. + 11.*w2) + 63085.*fcs_pow(v,8)*(-1. + 13.*w2) + 6138.*fcs_pow(v,4)*(-3. + 35.*w2) - 20460.*fcs_pow(v,6)*(-3. + 37.*w2)) - 150.*fcs_pow(u,4)*(339822.*fcs_pow(v,4)*(-1. + 11.*w2) + 315.*(-3. + 29.*w2) - 12180.*v2*(-3. + 31.*w2) - 356004.*fcs_pow(v,6)*(-3. + 35.*w2) + 346115.*fcs_pow(v,8)*(-3. + 37.*w2)) - 202275.*fcs_pow(u,8)*(-9. + 99.*w2 - 14652.*fcs_pow(v,6)*(-1. + 13.*w2) - 132.*v2*(-3. + 35.*w2) + 1386.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)) + 49445.*fcs_pow(u,10)*(13986.*fcs_pow(v,4)*(-1. + 13.*w2) + 9.*(-3. + 35.*w2) - 420.*v2*(-3. + 37.*w2) - 17316.*fcs_pow(v,6)*(-3. + 41.*w2) + 19721.*fcs_pow(v,8)*(-3. + 43.*w2)));
        case 4 : return -4.034357082192937e16*fcs_pow(t,24)*v*(-21.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 700.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 15660.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 13485.*fcs_pow(v,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 1890.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 6090.*fcs_pow(u,6)*(45.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 1860.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 61380.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 150.*fcs_pow(u,4)*(-12180.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 113274.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 346115.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 105.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 118668.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 75.*u2*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 10962.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 32364.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 420.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 49445.*fcs_pow(u,10)*(9. - 630.*w2 + 3885.*fcs_pow(w,4) - 420.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 4662.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 5772.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 67425.*fcs_pow(u,8)*(9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 4158.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 132.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return 2.017178541096469e17*fcs_pow(t,25)*v*w*(-3.*(-161820.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 49445.*fcs_pow(v,8)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 35.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 420.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 3654.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 30450.*fcs_pow(u,4)*(45. - 930.*w2 + 3069.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 4092.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 75.*u2*(339822.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 356004.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 63.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 2436.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 69223.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 9889.*fcs_pow(u,10)*(4.240015e6*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 1.18326e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 15540.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 60606.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 148335.*fcs_pow(u,8)*(45.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 13986.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 420.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 17316.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 188790.*fcs_pow(u,6)*(45.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 660.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1386.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return -1.008589270548234e18*fcs_pow(t,26)*v*(-3.*(10962.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 32364.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 29667.*fcs_pow(v,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 252.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 45.*u2*(-12180.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 113274.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 356004.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 63.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 69223.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 9889.*fcs_pow(u,10)*(181818.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 709956.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 63.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 3108.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 89001.*fcs_pow(u,8)*(45.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 86580.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 113274.*fcs_pow(u,6)*(15.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 6090.*fcs_pow(u,4)*(45.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 1860.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 18414.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 12276.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return 9.077303434934109e18*fcs_pow(t,27)*v*w*(-7.*(113274.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 356004.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 2436.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 9.*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 1015.*u2*(-13020.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 128898.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 63.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 12276.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 2.561251e6*fcs_pow(u,10)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6) - 91676.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 22919.*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 3198.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 52.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 9889.*fcs_pow(u,8)*(2.9680105e7*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 3.54978e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 15540.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 181818.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 138446.*fcs_pow(u,6)*(315.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 17316.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 62930.*fcs_pow(u,4)*(105.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 5291.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
        case 8 : return -6.354112404453877e19*fcs_pow(t,28)*v*(12180.*v2*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 113274.*fcs_pow(v,4)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 356004.*fcs_pow(v,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 63.*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) - 9889.*fcs_pow(v,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 365893.*fcs_pow(u,10)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8) + 22386.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 91676.*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 260.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8))) - 49445.*fcs_pow(u,8)*(-709956.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 25974.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 9.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 444.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 19778.*fcs_pow(u,6)*(315.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 690235.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 86580.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 420.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 4662.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 8990.*fcs_pow(u,4)*(105.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 26455.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 1386.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 4884.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 145.*u2*(315.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 13020.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 128898.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 12276.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return 1.842692597291624e21*fcs_pow(t,29)*v*w*(164021.*fcs_pow(u,10)*(666414.*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 2460.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 63468.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 82861.*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 9.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 567765.*fcs_pow(u,8)*(-825084.*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 9594.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 156.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 6138.*fcs_pow(u,6)*(2.9680105e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 1.18326e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 15540.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 60606.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 10230.*fcs_pow(u,4)*(105.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 98605.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 13986.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 17316.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 140.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 1395.*u2*(105.*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) - 1540.*v2*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 14652.*fcs_pow(v,6)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5291.*fcs_pow(v,8)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 462.*fcs_pow(v,4)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 9.*(-4340.*v2*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) + 14322.*fcs_pow(v,4)*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 7.*(45. - 1860.*w2 + 18414.*fcs_pow(w,4) - 61380.*fcs_pow(w,6) + 63085.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1364.*fcs_pow(v,6)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))));
      }
      case 10 : switch(k){
        case 0 : return 5.84689432201875e14*fcs_pow(t,21)*(1. - 105.*v2 + 1610.*fcs_pow(v,4) - 8050.*fcs_pow(v,6) + 15525.*fcs_pow(v,8) - 10005.*fcs_pow(v,10) - 230.*fcs_pow(u,4)*(-7. + 875.*v2 - 15750.*fcs_pow(v,4) + 91350.*fcs_pow(v,6) - 202275.*fcs_pow(v,8) + 148335.*fcs_pow(v,10)) + 3335.*fcs_pow(u,10)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) + 1150.*fcs_pow(u,6)*(-7. + 945.*v2 - 18270.*fcs_pow(v,4) + 113274.*fcs_pow(v,6) - 267003.*fcs_pow(v,8) + 207669.*fcs_pow(v,10)) + 5.*u2*(-21. + 2415.*v2 - 40250.*fcs_pow(v,4) + 217350.*fcs_pow(v,6) - 450225.*fcs_pow(v,8) + 310155.*fcs_pow(v,10)) - 1725.*fcs_pow(u,8)*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)));
        case 1 : return -5.84689432201875e14*fcs_pow(t,22)*(21. - 2415.*v2 + 40250.*fcs_pow(v,4) - 217350.*fcs_pow(v,6) + 450225.*fcs_pow(v,8) - 310155.*fcs_pow(v,10) + 345.*u2*(-7. + 875.*v2 - 15750.*fcs_pow(v,4) + 91350.*fcs_pow(v,6) - 202275.*fcs_pow(v,8) + 148335.*fcs_pow(v,10)) - 150075.*fcs_pow(u,8)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) - 5750.*fcs_pow(u,4)*(-7. + 945.*v2 - 18270.*fcs_pow(v,4) + 113274.*fcs_pow(v,6) - 267003.*fcs_pow(v,8) + 207669.*fcs_pow(v,10)) + 103385.*fcs_pow(u,10)*(-3. + 495.*v2 - 11550.*fcs_pow(v,4) + 85470.*fcs_pow(v,6) - 238095.*fcs_pow(v,8) + 216931.*fcs_pow(v,10)) + 24150.*fcs_pow(u,6)*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)))*w;
        case 2 : return 5.84689432201875e14*fcs_pow(t,23)*(-21. + 483.*w2 - 2415.*v2*(-1. + 25.*w2) + 40250.*fcs_pow(v,4)*(-1. + 27.*w2) - 217350.*fcs_pow(v,6)*(-1. + 29.*w2) + 450225.*fcs_pow(v,8)*(-1. + 31.*w2) - 310155.*fcs_pow(v,10)*(-1. + 33.*w2) + 345.*u2*(7. - 175.*w2 + 875.*v2*(-1. + 27.*w2) - 15750.*fcs_pow(v,4)*(-1. + 29.*w2) + 91350.*fcs_pow(v,6)*(-1. + 31.*w2) - 202275.*fcs_pow(v,8)*(-1. + 33.*w2) + 148335.*fcs_pow(v,10)*(-1. + 35.*w2)) - 5750.*fcs_pow(u,4)*(7. - 189.*w2 + 945.*v2*(-1. + 29.*w2) - 18270.*fcs_pow(v,4)*(-1. + 31.*w2) + 113274.*fcs_pow(v,6)*(-1. + 33.*w2) - 267003.*fcs_pow(v,8)*(-1. + 35.*w2) + 207669.*fcs_pow(v,10)*(-1. + 37.*w2)) + 24150.*fcs_pow(u,6)*(9. - 261.*w2 + 1305.*v2*(-1. + 31.*w2) - 26970.*fcs_pow(v,4)*(-1. + 33.*w2) + 178002.*fcs_pow(v,6)*(-1. + 35.*w2) - 445005.*fcs_pow(v,8)*(-1. + 37.*w2) + 365893.*fcs_pow(v,10)*(-1. + 39.*w2)) - 150075.*fcs_pow(u,8)*(3. - 93.*w2 + 465.*v2*(-1. + 33.*w2) - 10230.*fcs_pow(v,4)*(-1. + 35.*w2) + 71610.*fcs_pow(v,6)*(-1. + 37.*w2) - 189255.*fcs_pow(v,8)*(-1. + 39.*w2) + 164021.*fcs_pow(v,10)*(-1. + 41.*w2)) + 103385.*fcs_pow(u,10)*(3. - 99.*w2 + 495.*v2*(-1. + 35.*w2) - 11550.*fcs_pow(v,4)*(-1. + 37.*w2) + 85470.*fcs_pow(v,6)*(-1. + 39.*w2) - 238095.*fcs_pow(v,8)*(-1. + 41.*w2) + 216931.*fcs_pow(v,10)*(-1. + 43.*w2)));
        case 3 : return -4.034357082192937e16*fcs_pow(t,24)*w*(-2625.*v2*(-1. + 9.*w2) + 606825.*fcs_pow(v,8)*(-1. + 11.*w2) + 7.*(-3. + 25.*w2) + 15750.*fcs_pow(v,4)*(-3. + 29.*w2) - 91350.*fcs_pow(v,6)*(-3. + 31.*w2) - 148335.*fcs_pow(v,10)*(-3. + 35.*w2) - 5250.*fcs_pow(u,4)*(9. - 87.*w2 - 26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 365893.*fcs_pow(v,10)*(-1. + 13.*w2) + 435.*v2*(-3. + 31.*w2) + 59334.*fcs_pow(v,6)*(-3. + 35.*w2) - 148335.*fcs_pow(v,8)*(-3. + 37.*w2)) + 375.*u2*(7. - 63.*w2 + 113274.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 6090.*fcs_pow(v,4)*(-3. + 31.*w2) - 89001.*fcs_pow(v,8)*(-3. + 35.*w2) + 69223.*fcs_pow(v,10)*(-3. + 37.*w2)) + 30450.*fcs_pow(u,6)*(9. - 93.*w2 + 1395.*v2*(-1. + 11.*w2) - 567765.*fcs_pow(v,8)*(-1. + 13.*w2) - 10230.*fcs_pow(v,4)*(-3. + 35.*w2) + 71610.*fcs_pow(v,6)*(-3. + 37.*w2) + 164021.*fcs_pow(v,10)*(-3. + 41.*w2)) + 49445.*fcs_pow(u,10)*(9. - 105.*w2 - 38850.*fcs_pow(v,4)*(-1. + 13.*w2) + 848003.*fcs_pow(v,10)*(-1. + 15.*w2) + 525.*v2*(-3. + 37.*w2) + 101010.*fcs_pow(v,6)*(-3. + 41.*w2) - 295815.*fcs_pow(v,8)*(-3. + 43.*w2)) - 67425.*fcs_pow(u,8)*(9. - 99.*w2 + 256410.*fcs_pow(v,6)*(-1. + 13.*w2) + 495.*v2*(-3. + 35.*w2) - 11550.*fcs_pow(v,4)*(-3. + 37.*w2) - 238095.*fcs_pow(v,8)*(-3. + 41.*w2) + 216931.*fcs_pow(v,10)*(-3. + 43.*w2)));
        case 4 : return 4.034357082192937e16*fcs_pow(t,25)*(-5250.*fcs_pow(u,4)*(-9. + 522.*w2 - 2697.*fcs_pow(w,4) + 1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 445005.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 365893.*fcs_pow(v,10)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59334.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 375.*u2*(-7.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 18270.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 113274.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 207669.*fcs_pow(v,10)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 3.*(-7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 875.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 91350.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 202275.*fcs_pow(v,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 5250.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 49445.*fcs_pow(v,10)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 49445.*fcs_pow(u,10)*(-9. + 630.*w2 - 3885.*fcs_pow(w,4) + 1575.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 38850.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 887445.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 101010.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 202275.*fcs_pow(u,8)*(-3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 11550.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 85470.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 165.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 79365.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 30450.*fcs_pow(u,6)*(-9.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 1395.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 214830.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 567765.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 10230.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 164021.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -3.025767811644703e18*fcs_pow(t,26)*w*(-566370.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 445005.*fcs_pow(v,8)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 7.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 315.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 6090.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) - 69223.*fcs_pow(v,10)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 105.*u2*(-45. + 870.*w2 - 2697.*fcs_pow(w,4) - 134850.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 296670.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 365893.*fcs_pow(v,10)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 435.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) - 148335.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 148335.*fcs_pow(u,8)*(-9. + 210.*w2 - 777.*fcs_pow(w,4) + 848003.*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 295815.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 20202.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 62930.*fcs_pow(u,6)*(-9.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 495.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 51282.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 2310.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 47619.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 2030.*fcs_pow(u,4)*(-45. + 930.*w2 - 3069.*fcs_pow(w,4) + 6975.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 51150.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 567765.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 71610.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 164021.*fcs_pow(v,10)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 9889.*fcs_pow(u,10)*(-1.2720045e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1.38047e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3885.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 7.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 33670.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))));
        case 6 : return 3.025767811644703e18*fcs_pow(t,27)*(91350.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 566370.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 1.335015e6*fcs_pow(v,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 945.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) - 207669.*fcs_pow(v,10)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 148335.*fcs_pow(u,8)*(9. - 945.*w2 + 11655.*fcs_pow(w,4) - 30303.*fcs_pow(w,6) + 303030.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 315.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 188790.*fcs_pow(u,6)*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 238095.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 216931.*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 30450.*fcs_pow(u,4)*(3. - 279.*w2 + 3069.*fcs_pow(w,4) - 7161.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 164021.*fcs_pow(v,10)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 14322.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 37851.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 105.*u2*(6525.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 134850.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 890010.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) - 445005.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 365893.*fcs_pow(v,10)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 29667.*fcs_pow(u,10)*(-168350.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.38047e6*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.240015e6*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))));
        case 7 : return -6.354112404453877e19*fcs_pow(t,28)*w*(943950.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 6.23007e6*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 9135.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 9.*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 445005.*fcs_pow(v,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 365893.*fcs_pow(v,10)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 365893.*fcs_pow(u,10)*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 1.60433e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.031355e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 26650.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 195.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 445005.*fcs_pow(u,8)*(35. - 1295.*w2 + 10101.*fcs_pow(w,4) - 19721.*fcs_pow(w,6) - 4.240015e6*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 197210.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 185.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 4810.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 98890.*fcs_pow(u,6)*(-63.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 5.936021e6*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 60606.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 44950.*fcs_pow(u,4)*(21.*(1. - 33.*w2 + 231.*fcs_pow(w,4) - 407.*fcs_pow(w,6)) + 3465.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 216931.*fcs_pow(v,10)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 47619.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 435.*u2*(16275.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 358050.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 21.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 71610.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 189255.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 164021.*fcs_pow(v,10)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
        case 8 : return 6.354112404453877e19*fcs_pow(t,29)*(-45675.*v2*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 943950.*fcs_pow(v,4)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 6.23007e6*fcs_pow(v,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 63.*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) + 445005.*fcs_pow(v,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 365893.*fcs_pow(v,10)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) + 365893.*fcs_pow(u,10)*(-35. + 5460.*w2 - 111930.*fcs_pow(w,4) + 641732.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) - 186550.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.60433e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 975.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) - 445005.*fcs_pow(u,8)*(-35. + 5180.*w2 - 101010.*fcs_pow(w,4) + 552188.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 1.38047e6*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 4.240015e6*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 24050.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 185.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 692230.*fcs_pow(u,6)*(-9.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 887445.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 43290.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 45.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 1110.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 44950.*fcs_pow(u,4)*(-21.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 3465.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 1.518517e6*fcs_pow(v,10)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 238095.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 17094.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 2175.*u2*(-21.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 3255.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 164021.*fcs_pow(v,10)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 14322.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 37851.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -5.528077791874873e21*fcs_pow(t,30)*w*(-189255.*fcs_pow(u,8)*(1.443897e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 79950.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 585.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 71610.*fcs_pow(u,6)*(-315. + 15540.*w2 - 181818.*fcs_pow(w,4) + 709956.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) - 3.8160135e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 591630.*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1665.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 14430.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 51150.*fcs_pow(u,4)*(5.936021e6*fcs_pow(v,10)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 7.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 295815.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 7770.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 20202.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 2325.*u2*(-21.*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) + 1155.*v2*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 216931.*fcs_pow(v,10)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 51282.*fcs_pow(v,6)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 47619.*fcs_pow(v,8)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 770.*fcs_pow(v,4)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 3.*(16275.*v2*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) - 119350.*fcs_pow(v,4)*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 7.*(45. - 1860.*w2 + 18414.*fcs_pow(w,4) - 61380.*fcs_pow(w,6) + 63085.*fcs_pow(w,8)) - 567765.*fcs_pow(v,8)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 164021.*fcs_pow(v,10)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 23870.*fcs_pow(v,6)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 164021.*fcs_pow(u,10)*(-1.85115e6*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 3075.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 370230.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1.242915e6*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 3.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return 2.764038895937436e22*fcs_pow(t,31)*(-567765.*fcs_pow(u,8)*(21. - 4095.*w2 + 111930.*fcs_pow(w,4) - 962598.*fcs_pow(w,6) + 3.094065e6*fcs_pow(w,8) - 3.231579e6*fcs_pow(w,10) - 111930.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 962598.*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.077193e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 343785.*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 195.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 71610.*fcs_pow(u,6)*(2.484846e6*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 7.632027e6*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 9.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 1665.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 14430.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 71610.*fcs_pow(u,4)*(9. - 1575.*w2 + 38850.*fcs_pow(w,4) - 303030.*fcs_pow(w,6) + 887445.*fcs_pow(w,8) - 848003.*fcs_pow(w,10) - 887445.*fcs_pow(v,8)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 225.*v2*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 5550.*fcs_pow(v,4)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 14430.*fcs_pow(v,6)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 1395.*u2*(-7.*(-3. + 495.*w2 - 11550.*fcs_pow(w,4) + 85470.*fcs_pow(w,6) - 238095.*fcs_pow(w,8) + 216931.*fcs_pow(w,10)) + 1.518517e6*fcs_pow(v,10)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 385.*v2*(-9. + 1575.*w2 - 38850.*fcs_pow(w,4) + 303030.*fcs_pow(w,6) - 887445.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 11550.*fcs_pow(v,4)*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 85470.*fcs_pow(v,6)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 79365.*fcs_pow(v,8)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 3.*(-21.*(-3. + 465.*w2 - 10230.*fcs_pow(w,4) + 71610.*fcs_pow(w,6) - 189255.*fcs_pow(w,8) + 164021.*fcs_pow(w,10)) + 3255.*v2*(-3. + 495.*w2 - 11550.*fcs_pow(w,4) + 85470.*fcs_pow(w,6) - 238095.*fcs_pow(w,8) + 216931.*fcs_pow(w,10)) - 23870.*fcs_pow(v,4)*(-9. + 1575.*w2 - 38850.*fcs_pow(w,4) + 303030.*fcs_pow(w,6) - 887445.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 214830.*fcs_pow(v,6)*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 567765.*fcs_pow(v,8)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 164021.*fcs_pow(v,10)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 164021.*fcs_pow(u,10)*(3.*(21. - 4305.*w2 + 123410.*fcs_pow(w,4) - 1.11069e6*fcs_pow(w,6) + 3.728745e6*fcs_pow(w,8) - 4.060189e6*fcs_pow(w,10)) + 12915.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 370230.*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 3.728745e6*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 370230.*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))));
      }
    }
    case 11 : switch(j){
      case 0 : switch(k){
        case 0 : return -155925.*fcs_pow(t,12)*u*(-693. + 15015.*u2 - 90090.*fcs_pow(u,4) + 218790.*fcs_pow(u,6) - 230945.*fcs_pow(u,8) + 88179.*fcs_pow(u,10));
      }
      case 1 : switch(k){
        case 0 : return 6.081075e6*fcs_pow(t,13)*u*(-231. + 5775.*u2 - 39270.*fcs_pow(u,4) + 106590.*fcs_pow(u,6) - 124355.*fcs_pow(u,8) + 52003.*fcs_pow(u,10))*v;
        case 1 : return -2.12837625e8*fcs_pow(t,14)*u*(-99. + 2805.*u2 - 21318.*fcs_pow(u,4) + 63954.*fcs_pow(u,6) - 81719.*fcs_pow(u,8) + 37145.*fcs_pow(u,10))*v*w;
      }
      case 2 : switch(k){
        case 0 : return -6.081075e6*fcs_pow(t,14)*u*(231. - 3465.*v2 + 5775.*u2*(-1. + 17.*v2) - 39270.*fcs_pow(u,4)*(-1. + 19.*v2) + 106590.*fcs_pow(u,6)*(-1. + 21.*v2) - 124355.*fcs_pow(u,8)*(-1. + 23.*v2) + 52003.*fcs_pow(u,10)*(-1. + 25.*v2));
        case 1 : return 2.12837625e8*fcs_pow(t,15)*u*(99. - 1683.*v2 + 2805.*u2*(-1. + 19.*v2) - 21318.*fcs_pow(u,4)*(-1. + 21.*v2) + 63954.*fcs_pow(u,6)*(-1. + 23.*v2) - 81719.*fcs_pow(u,8)*(-1. + 25.*v2) + 37145.*fcs_pow(u,10)*(-1. + 27.*v2))*w;
        case 2 : return -2.12837625e8*fcs_pow(t,16)*u*(-99.*(1. - 17.*w2 + 17.*v2*(-1. + 19.*w2)) + 2805.*u2*(1. - 19.*w2 + 19.*v2*(-1. + 21.*w2)) - 21318.*fcs_pow(u,4)*(1. - 21.*w2 + 21.*v2*(-1. + 23.*w2)) + 63954.*fcs_pow(u,6)*(1. - 23.*w2 + 23.*v2*(-1. + 25.*w2)) - 81719.*fcs_pow(u,8)*(1. - 25.*w2 + 25.*v2*(-1. + 27.*w2)) + 37145.*fcs_pow(u,10)*(1. - 27.*w2 + 27.*v2*(-1. + 29.*w2)));
      }
      case 3 : switch(k){
        case 0 : return 2.12837625e8*fcs_pow(t,15)*u*v*(99.*(3. - 17.*v2) - 63954.*fcs_pow(u,4)*(-1. + 7.*v2) + 111435.*fcs_pow(u,10)*(-1. + 9.*v2) + 2805.*u2*(-3. + 19.*v2) + 63954.*fcs_pow(u,6)*(-3. + 23.*v2) - 81719.*fcs_pow(u,8)*(-3. + 25.*v2));
        case 1 : return -1.0854718875e10*fcs_pow(t,16)*u*v*(99. - 627.*v2 + 3135.*u2*(-1. + 7.*v2) - 120175.*fcs_pow(u,8)*(-1. + 9.*v2) - 8778.*fcs_pow(u,4)*(-3. + 23.*v2) + 28842.*fcs_pow(u,6)*(-3. + 25.*v2) + 19665.*fcs_pow(u,10)*(-3. + 29.*v2))*w;
        case 2 : return 1.0854718875e10*fcs_pow(t,17)*u*v*(-33.*(3. - 57.*w2 + 19.*v2*(-1. + 21.*w2)) + 3135.*u2*(1. - 21.*w2 + 7.*v2*(-1. + 23.*w2)) - 8778.*fcs_pow(u,4)*(3. - 69.*w2 + 23.*v2*(-1. + 25.*w2)) + 28842.*fcs_pow(u,6)*(3. - 75.*w2 + 25.*v2*(-1. + 27.*w2)) - 120175.*fcs_pow(u,8)*(1. - 27.*w2 + 9.*v2*(-1. + 29.*w2)) + 19665.*fcs_pow(u,10)*(3. - 87.*w2 + 29.*v2*(-1. + 31.*w2)));
        case 3 : return -6.18718975875e11*fcs_pow(t,18)*u*v*w*(-10626.*fcs_pow(u,4)*(3. - 25.*w2 + 25.*v2*(-1. + 9.*w2)) + 30015.*fcs_pow(u,10)*(3. - 31.*w2 + 31.*v2*(-1. + 11.*w2)) + 33.*(-3. + 21.*w2 - 7.*v2*(-3. + 23.*w2)) + 385.*u2*(9. - 69.*w2 + 23.*v2*(-3. + 25.*w2)) - 18975.*fcs_pow(u,8)*(9. - 87.*w2 + 29.*v2*(-3. + 31.*w2)) + 113850.*fcs_pow(u,6)*(1. - 9.*w2 + v2*(-9. + 87.*w2)));
      }
      case 4 : switch(k){
        case 0 : return -6.38512875e8*fcs_pow(t,16)*u*(2805.*u2*(1. - 38.*v2 + 133.*fcs_pow(v,4)) - 21318.*fcs_pow(u,4)*(1. - 42.*v2 + 161.*fcs_pow(v,4)) - 81719.*fcs_pow(u,8)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) + 37145.*fcs_pow(u,10)*(1. - 54.*v2 + 261.*fcs_pow(v,4)) - 33.*(3. - 102.*v2 + 323.*fcs_pow(v,4)) + 21318.*fcs_pow(u,6)*(3. - 138.*v2 + 575.*fcs_pow(v,4)));
        case 1 : return 1.0854718875e10*fcs_pow(t,17)*u*(-99.*(1. - 38.*v2 + 133.*fcs_pow(v,4)) + 3135.*u2*(1. - 42.*v2 + 161.*fcs_pow(v,4)) + 86526.*fcs_pow(u,6)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) - 120175.*fcs_pow(u,8)*(1. - 54.*v2 + 261.*fcs_pow(v,4)) - 8778.*fcs_pow(u,4)*(3. - 138.*v2 + 575.*fcs_pow(v,4)) + 19665.*fcs_pow(u,10)*(3. - 174.*v2 + 899.*fcs_pow(v,4)))*w;
        case 2 : return -1.0854718875e10*fcs_pow(t,18)*u*(-99.*(-1. + 19.*w2 + v2*(38. - 798.*w2) + 133.*fcs_pow(v,4)*(-1. + 23.*w2)) + 3135.*u2*(-1. + 21.*w2 + v2*(42. - 966.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) - 8778.*fcs_pow(u,4)*(-3. + 69.*w2 - 138.*v2*(-1. + 25.*w2) + 575.*fcs_pow(v,4)*(-1. + 27.*w2)) + 86526.*fcs_pow(u,6)*(-1. + 25.*w2 - 50.*v2*(-1. + 27.*w2) + 225.*fcs_pow(v,4)*(-1. + 29.*w2)) - 120175.*fcs_pow(u,8)*(-1. + 27.*w2 - 54.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)) + 19665.*fcs_pow(u,10)*(-3. + 87.*w2 - 174.*v2*(-1. + 31.*w2) + 899.*fcs_pow(v,4)*(-1. + 33.*w2)));
        case 3 : return 1.856156927625e12*fcs_pow(t,19)*u*w*(385.*u2*(-3. + 23.*w2 + 575.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) - 11.*(-3. + 21.*w2 - 42.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)) - 3542.*fcs_pow(u,4)*(-3. + 25.*w2 - 150.*v2*(-1. + 9.*w2) + 225.*fcs_pow(v,4)*(-3. + 29.*w2)) - 18975.*fcs_pow(u,8)*(-3. + 29.*w2 + 899.*fcs_pow(v,4)*(-1. + 11.*w2) - 58.*v2*(-3. + 31.*w2)) + 37950.*fcs_pow(u,6)*(-1. + 9.*w2 + v2*(54. - 522.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)) + 10005.*fcs_pow(u,10)*(-3. + 31.*w2 - 186.*v2*(-1. + 11.*w2) + 341.*fcs_pow(v,4)*(-3. + 35.*w2)));
        case 4 : return -1.856156927625e12*fcs_pow(t,20)*u*(385.*u2*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 138.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) - 18975.*fcs_pow(u,8)*(3. - 174.*w2 + 899.*fcs_pow(w,4) - 174.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4))) - 33.*(1. - 42.*w2 + 161.*fcs_pow(w,4) + 161.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 14.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) + 37950.*fcs_pow(u,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 10626.*fcs_pow(u,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 75.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 10005.*fcs_pow(u,10)*(3.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 186.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return 1.0854718875e10*fcs_pow(t,17)*u*v*(-43890.*fcs_pow(u,4)*(3. - 46.*v2 + 115.*fcs_pow(v,4)) + 144210.*fcs_pow(u,6)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) + 3135.*u2*(5. - 70.*v2 + 161.*fcs_pow(v,4)) - 120175.*fcs_pow(u,8)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) - 33.*(15. - 190.*v2 + 399.*fcs_pow(v,4)) + 19665.*fcs_pow(u,10)*(15. - 290.*v2 + 899.*fcs_pow(v,4)));
        case 1 : return -6.18718975875e11*fcs_pow(t,18)*u*v*(1925.*u2*(3. - 46.*v2 + 115.*fcs_pow(v,4)) - 17710.*fcs_pow(u,4)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) - 33.*(5. - 70.*v2 + 161.*fcs_pow(v,4)) + 37950.*fcs_pow(u,6)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) - 18975.*fcs_pow(u,8)*(15. - 290.*v2 + 899.*fcs_pow(v,4)) + 10005.*fcs_pow(u,10)*(15. - 310.*v2 + 1023.*fcs_pow(v,4)))*w;
        case 2 : return 6.18718975875e11*fcs_pow(t,19)*u*v*(-33.*(-5. + 105.*w2 - 70.*v2*(-1. + 23.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) + 1925.*u2*(-3. + 69.*w2 - 46.*v2*(-1. + 25.*w2) + 115.*fcs_pow(v,4)*(-1. + 27.*w2)) - 17710.*fcs_pow(u,4)*(-3. + 75.*w2 - 50.*v2*(-1. + 27.*w2) + 135.*fcs_pow(v,4)*(-1. + 29.*w2)) + 37950.*fcs_pow(u,6)*(-5. + 135.*w2 - 90.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)) - 18975.*fcs_pow(u,8)*(-15. + 435.*w2 - 290.*v2*(-1. + 31.*w2) + 899.*fcs_pow(v,4)*(-1. + 33.*w2)) + 10005.*fcs_pow(u,10)*(-15. + 465.*w2 - 310.*v2*(-1. + 33.*w2) + 1023.*fcs_pow(v,4)*(-1. + 35.*w2)));
        case 3 : return -9.280784638125e12*fcs_pow(t,20)*u*v*w*(-77.*(-9. + 69.*w2 + 345.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) + 8855.*u2*(-3. + 25.*w2 - 50.*v2*(-1. + 9.*w2) + 45.*fcs_pow(v,4)*(-3. + 29.*w2)) + 22770.*fcs_pow(u,6)*(-45. + 435.*w2 + 2697.*fcs_pow(v,4)*(-1. + 11.*w2) - 290.*v2*(-3. + 31.*w2)) - 53130.*fcs_pow(u,4)*(-5. + 45.*w2 + v2*(90. - 870.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)) - 110055.*fcs_pow(u,8)*(-310.*v2*(-1. + 11.*w2) + 5.*(-3. + 31.*w2) + 341.*fcs_pow(v,4)*(-3. + 35.*w2)) + 310155.*fcs_pow(u,10)*(-3. + 33.*w2 + v2*(66. - 770.*w2) + 77.*fcs_pow(v,4)*(-3. + 37.*w2)));
        case 4 : return 9.280784638125e12*fcs_pow(t,21)*u*v*(-231.*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 46.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 115.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) + 68310.*fcs_pow(u,6)*(-290.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 53130.*fcs_pow(u,4)*(5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 30.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 8855.*u2*(3. - 150.*w2 + 675.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 45.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 310155.*fcs_pow(u,10)*(3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 231.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 22.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 110055.*fcs_pow(u,8)*(15.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 310.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return -3.522057770168438e16*fcs_pow(t,22)*u*v*w*(2610.*fcs_pow(u,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 310.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4))) - 7.*(3. - 50.*w2 + 135.*fcs_pow(w,4) - 10.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 9.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) - 210.*fcs_pow(u,4)*(45. - 870.*w2 + 2697.*fcs_pow(w,4) + 2697.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 58.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 35.*u2*(5.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 30.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 87.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 899.*fcs_pow(u,10)*(15.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 777.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 70.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 4495.*fcs_pow(u,8)*(15.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 110.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))));
      }
      case 6 : switch(k){
        case 0 : return -1.0854718875e10*fcs_pow(t,18)*u*(15675.*u2*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) - 131670.*fcs_pow(u,4)*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 432630.*fcs_pow(u,6)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 99.*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)) - 120175.*fcs_pow(u,8)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 58995.*fcs_pow(u,10)*(-5. + 435.*v2 - 4495.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)));
        case 1 : return 9.280784638125e12*fcs_pow(t,19)*u*(11.*(1. - 63.*v2 + 483.*fcs_pow(v,4) - 805.*fcs_pow(v,6)) + 385.*u2*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 3542.*fcs_pow(u,4)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 10005.*fcs_pow(u,10)*(-1. + 93.*v2 - 1023.*fcs_pow(v,4) + 2387.*fcs_pow(v,6)) + 2530.*fcs_pow(u,6)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) - 3795.*fcs_pow(u,8)*(-5. + 435.*v2 - 4495.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)))*w;
        case 2 : return -9.280784638125e12*fcs_pow(t,20)*u*(-11.*(1. - 21.*w2 + 63.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 805.*fcs_pow(v,6)*(-1. + 27.*w2)) + 385.*u2*(1. - 23.*w2 + 69.*v2*(-1. + 25.*w2) - 575.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) - 3542.*fcs_pow(u,4)*(1. - 25.*w2 + 75.*v2*(-1. + 27.*w2) - 675.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) + 2530.*fcs_pow(u,6)*(5. - 135.*w2 + 405.*v2*(-1. + 29.*w2) - 3915.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) - 3795.*fcs_pow(u,8)*(5. - 145.*w2 + 435.*v2*(-1. + 31.*w2) - 4495.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)) + 10005.*fcs_pow(u,10)*(1. - 31.*w2 + 93.*v2*(-1. + 33.*w2) - 1023.*fcs_pow(v,4)*(-1. + 35.*w2) + 2387.*fcs_pow(v,6)*(-1. + 37.*w2)));
        case 3 : return 9.280784638125e12*fcs_pow(t,21)*u*w*(-231.*(3. - 23.*w2 - 1725.*fcs_pow(v,4)*(-1. + 9.*w2) + 69.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)) - 53130.*fcs_pow(u,4)*(5. - 45.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 135.*v2*(-3. + 29.*w2) - 1305.*fcs_pow(v,4)*(-3. + 31.*w2)) + 8855.*u2*(3. - 25.*w2 + 225.*v2*(-1. + 9.*w2) - 675.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)) + 68310.*fcs_pow(u,6)*(15. - 145.*w2 - 13485.*fcs_pow(v,4)*(-1. + 11.*w2) + 435.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) + 930465.*fcs_pow(u,10)*(1. - 11.*w2 + 2849.*fcs_pow(v,6)*(-1. + 13.*w2) + 33.*v2*(-3. + 35.*w2) - 385.*fcs_pow(v,4)*(-3. + 37.*w2)) - 550275.*fcs_pow(u,8)*(3. - 31.*w2 + 279.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 2387.*fcs_pow(v,6)*(-3. + 37.*w2)));
        case 4 : return -2.7842353914375e13*fcs_pow(t,22)*u*(-17710.*fcs_pow(u,4)*(-5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 3915.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 135.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 8855.*u2*(-1. + 50.*w2 - 225.*fcs_pow(w,4) + 75.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 225.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 77.*(-3. + 138.*w2 - 575.*fcs_pow(w,4) + 207.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 1725.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 310155.*fcs_pow(u,10)*(-1. + 66.*w2 - 385.*fcs_pow(w,4) - 1155.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 2849.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 33.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 550275.*fcs_pow(u,8)*(-1. + 62.*w2 - 341.*fcs_pow(w,4) + 93.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 22770.*fcs_pow(u,6)*(1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 13485.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return 3.522057770168438e16*fcs_pow(t,23)*u*w*(-630.*fcs_pow(u,4)*(-15. + 290.*w2 - 899.*fcs_pow(w,4) - 13485.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 87.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 175.*u2*(-5. + 90.*w2 - 261.*fcs_pow(w,4) + 8091.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 27.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 261.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 7.*(-3. + 50.*w2 - 135.*fcs_pow(w,4) + 45.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 135.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 13485.*fcs_pow(u,8)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 165.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2849.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 385.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 2610.*fcs_pow(u,6)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 1395.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 5115.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 2697.*fcs_pow(u,10)*(-5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 3885.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 3367.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return -3.522057770168438e16*fcs_pow(t,24)*u*(-1890.*fcs_pow(u,4)*(5. - 435.*w2 + 4495.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) + 435.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 4495.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6))) + 175.*u2*(5. - 405.*w2 + 3915.*fcs_pow(w,4) - 8091.*fcs_pow(w,6) - 3915.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 81.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 21.*(1. - 75.*w2 + 675.*fcs_pow(w,4) - 1305.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 15.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 135.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 7830.*fcs_pow(u,6)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 5115.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 2387.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 40455.*fcs_pow(u,10)*(1. - 105.*w2 + 1295.*fcs_pow(w,4) - 3367.*fcs_pow(w,6) + 3367.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 21.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 13485.*fcs_pow(u,8)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 1155.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2849.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
      }
      case 7 : switch(k){
        case 0 : return 1.856156927625e12*fcs_pow(t,19)*u*v*(-77.*(-5. + 105.*v2 - 483.*fcs_pow(v,4) + 575.*fcs_pow(v,6)) + 1925.*u2*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 17710.*fcs_pow(u,4)*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 70035.*fcs_pow(u,10)*(-5. + 155.*v2 - 1023.*fcs_pow(v,4) + 1705.*fcs_pow(v,6)) + 12650.*fcs_pow(u,6)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) - 18975.*fcs_pow(u,8)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)));
        case 1 : return -9.280784638125e12*fcs_pow(t,20)*u*v*(2.171085e6*fcs_pow(u,10)*(-1. + 33.*v2 - 231.*fcs_pow(v,4) + 407.*fcs_pow(v,6)) - 231.*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 8855.*u2*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 770385.*fcs_pow(u,8)*(-5. + 155.*v2 - 1023.*fcs_pow(v,4) + 1705.*fcs_pow(v,6)) - 17710.*fcs_pow(u,4)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 68310.*fcs_pow(u,6)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)))*w;
        case 2 : return 9.280784638125e12*fcs_pow(t,21)*u*v*(-231.*(7. - 161.*w2 + 161.*v2*(-1. + 25.*w2) - 805.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) + 8855.*u2*(7. - 175.*w2 + 175.*v2*(-1. + 27.*w2) - 945.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) - 17710.*fcs_pow(u,4)*(35. - 945.*w2 + 945.*v2*(-1. + 29.*w2) - 5481.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) + 68310.*fcs_pow(u,6)*(35. - 1015.*w2 + 1015.*v2*(-1. + 31.*w2) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)) - 770385.*fcs_pow(u,8)*(5. - 155.*w2 + 155.*v2*(-1. + 33.*w2) - 1023.*fcs_pow(v,4)*(-1. + 35.*w2) + 1705.*fcs_pow(v,6)*(-1. + 37.*w2)) + 2.171085e6*fcs_pow(u,10)*(1. - 33.*w2 + 33.*v2*(-1. + 35.*w2) - 231.*fcs_pow(v,4)*(-1. + 37.*w2) + 407.*fcs_pow(v,6)*(-1. + 39.*w2)));
        case 3 : return -4.930880878235812e16*fcs_pow(t,22)*u*v*w*(-525.*v2*(-1. + 9.*w2) + 7.*(-3. + 25.*w2) + 945.*fcs_pow(v,4)*(-3. + 29.*w2) - 1305.*fcs_pow(v,6)*(-3. + 31.*w2) + 25.*u2*(35. - 315.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 1827.*fcs_pow(v,4)*(-3. + 31.*w2)) - 90.*fcs_pow(u,4)*(35.*(3. - 29.*w2) - 18879.*fcs_pow(v,4)*(-1. + 11.*w2) + 1015.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) - 67425.*fcs_pow(u,8)*(1. - 11.*w2 + 407.*fcs_pow(v,6)*(-1. + 13.*w2) + 11.*v2*(-3. + 35.*w2) - 77.*fcs_pow(v,4)*(-3. + 37.*w2)) + 2610.*fcs_pow(u,6)*(15. - 155.*w2 + 465.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 1705.*fcs_pow(v,6)*(-3. + 37.*w2)) + 13485.*fcs_pow(u,10)*(3. - 35.*w2 - 777.*fcs_pow(v,4)*(-1. + 13.*w2) + 35.*v2*(-3. + 37.*w2) + 481.*fcs_pow(v,6)*(-3. + 41.*w2)));
        case 4 : return 4.930880878235812e16*fcs_pow(t,23)*u*v*(25.*u2*(-35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 5481.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 3.*(-7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 175.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 315.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 67425.*fcs_pow(u,8)*(-1. + 66.*w2 - 385.*fcs_pow(w,4) - 231.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 11.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 7830.*fcs_pow(u,6)*(-5.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 155.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 1705.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 90.*fcs_pow(u,4)*(3045.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18879.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 35.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 13485.*fcs_pow(u,10)*(-3. + 210.*w2 - 1295.*fcs_pow(w,4) + 105.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 777.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -7.396321317353719e17*fcs_pow(t,24)*u*v*w*(-40455.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 35.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 315.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 1827.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 75.*u2*(-18879.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 203.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 80910.*fcs_pow(u,6)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 55.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 6090.*fcs_pow(u,4)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 465.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 1023.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 341.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 899.*fcs_pow(u,10)*(98605.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3885.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 35.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 10101.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 49445.*fcs_pow(u,8)*(-5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 777.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 35.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 481.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return 3.698160658676859e18*fcs_pow(t,25)*u*v*(5481.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 189.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 45.*u2*(1015.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 6293.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 7.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 3654.*fcs_pow(u,4)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 155.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 1023.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 2697.*fcs_pow(u,10)*(-10101.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 259.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 29667.*fcs_pow(u,8)*(5. - 525.*w2 + 6475.*fcs_pow(w,4) - 16835.*fcs_pow(w,6) + 2405.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 35.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 16182.*fcs_pow(u,6)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return -1.109448197603058e19*fcs_pow(t,26)*u*v*w*(-63.*(35. - 1015.*w2 + 6293.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 203.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6))) + 3045.*u2*(7.*(5. - 155.*w2 + 1023.*fcs_pow(w,4) - 1705.*fcs_pow(w,6)) + 1085.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 7161.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 37758.*fcs_pow(u,4)*(-35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 1155.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 232841.*fcs_pow(u,10)*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 22919.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1599.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 39.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 9889.*fcs_pow(u,8)*(98605.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 35.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 10101.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 178002.*fcs_pow(u,6)*(-35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 35.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 481.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
      }
      case 8 : switch(k){
        case 0 : return -9.280784638125e12*fcs_pow(t,20)*u*(-77.*(1. - 84.*v2 + 966.*fcs_pow(v,4) - 3220.*fcs_pow(v,6) + 3105.*fcs_pow(v,8)) + 70035.*fcs_pow(u,10)*(1. - 124.*v2 + 2046.*fcs_pow(v,4) - 9548.*fcs_pow(v,6) + 12617.*fcs_pow(v,8)) + 385.*u2*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) - 3542.*fcs_pow(u,4)*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) - 26565.*fcs_pow(u,8)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) + 2530.*fcs_pow(u,6)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)));
        case 1 : return 6.4965492466875e13*fcs_pow(t,21)*u*(-550275.*fcs_pow(u,8)*(1. - 124.*v2 + 2046.*fcs_pow(v,4) - 9548.*fcs_pow(v,6) + 12617.*fcs_pow(v,8)) + 310155.*fcs_pow(u,10)*(1. - 132.*v2 + 2310.*fcs_pow(v,4) - 11396.*fcs_pow(v,6) + 15873.*fcs_pow(v,8)) - 33.*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) + 1265.*u2*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) + 68310.*fcs_pow(u,6)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) - 2530.*fcs_pow(u,4)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)))*w;
        case 2 : return -6.4965492466875e13*fcs_pow(t,22)*u*(-33.*(-7. + 161.*w2 - 644.*v2*(-1. + 25.*w2) + 8050.*fcs_pow(v,4)*(-1. + 27.*w2) - 28980.*fcs_pow(v,6)*(-1. + 29.*w2) + 30015.*fcs_pow(v,8)*(-1. + 31.*w2)) + 1265.*u2*(-7. + 175.*w2 - 700.*v2*(-1. + 27.*w2) + 9450.*fcs_pow(v,4)*(-1. + 29.*w2) - 36540.*fcs_pow(v,6)*(-1. + 31.*w2) + 40455.*fcs_pow(v,8)*(-1. + 33.*w2)) - 2530.*fcs_pow(u,4)*(-35. + 945.*w2 - 3780.*v2*(-1. + 29.*w2) + 54810.*fcs_pow(v,4)*(-1. + 31.*w2) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2)) + 68310.*fcs_pow(u,6)*(-5. + 145.*w2 - 580.*v2*(-1. + 31.*w2) + 8990.*fcs_pow(v,4)*(-1. + 33.*w2) - 39556.*fcs_pow(v,6)*(-1. + 35.*w2) + 49445.*fcs_pow(v,8)*(-1. + 37.*w2)) - 550275.*fcs_pow(u,8)*(-1. + 31.*w2 - 124.*v2*(-1. + 33.*w2) + 2046.*fcs_pow(v,4)*(-1. + 35.*w2) - 9548.*fcs_pow(v,6)*(-1. + 37.*w2) + 12617.*fcs_pow(v,8)*(-1. + 39.*w2)) + 310155.*fcs_pow(u,10)*(-1. + 33.*w2 - 132.*v2*(-1. + 35.*w2) + 2310.*fcs_pow(v,4)*(-1. + 37.*w2) - 11396.*fcs_pow(v,6)*(-1. + 39.*w2) + 15873.*fcs_pow(v,8)*(-1. + 41.*w2)));
        case 3 : return 4.930880878235812e16*fcs_pow(t,23)*u*w*(21. - 175.*w2 + 2100.*v2*(-1. + 9.*w2) - 121365.*fcs_pow(v,8)*(-1. + 11.*w2) - 9450.*fcs_pow(v,4)*(-3. + 29.*w2) + 36540.*fcs_pow(v,6)*(-3. + 31.*w2) + 25.*u2*(35.*(-1. + 9.*w2) - 226548.*fcs_pow(v,6)*(-1. + 11.*w2) - 1260.*v2*(-3. + 29.*w2) + 18270.*fcs_pow(v,4)*(-3. + 31.*w2) + 89001.*fcs_pow(v,8)*(-3. + 35.*w2)) + 13050.*fcs_pow(u,6)*(-3. + 31.*w2 - 372.*v2*(-1. + 11.*w2) + 37851.*fcs_pow(v,8)*(-1. + 13.*w2) + 2046.*fcs_pow(v,4)*(-3. + 35.*w2) - 9548.*fcs_pow(v,6)*(-3. + 37.*w2)) - 630.*fcs_pow(u,4)*(26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 5.*(-3. + 29.*w2) - 580.*v2*(-3. + 31.*w2) - 39556.*fcs_pow(v,6)*(-3. + 35.*w2) + 49445.*fcs_pow(v,8)*(-3. + 37.*w2)) - 67425.*fcs_pow(u,8)*(-1. + 11.*w2 - 11396.*fcs_pow(v,6)*(-1. + 13.*w2) - 44.*v2*(-3. + 35.*w2) + 770.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)) + 13485.*fcs_pow(u,10)*(-3. + 35.*w2 + 7770.*fcs_pow(v,4)*(-1. + 13.*w2) - 140.*v2*(-3. + 37.*w2) - 13468.*fcs_pow(v,6)*(-3. + 41.*w2) + 19721.*fcs_pow(v,8)*(-3. + 43.*w2)));
        case 4 : return -4.930880878235812e16*fcs_pow(t,24)*u*(-3.*(7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 700.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 36540.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 40455.*fcs_pow(v,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 3150.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 39150.*fcs_pow(u,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4) - 124.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 682.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 630.*fcs_pow(u,4)*(-1740.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 148335.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 25.*u2*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 54810.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 226548.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 1260.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 13485.*fcs_pow(u,10)*(3. - 210.*w2 + 1295.*fcs_pow(w,4) - 420.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 7770.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59163.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 13468.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 67425.*fcs_pow(u,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4) + 2310.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 44.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return 3.698160658676859e18*fcs_pow(t,25)*u*w*(226548.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 89001.*fcs_pow(v,8)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 252.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 3654.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) - 1218.*fcs_pow(u,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 10230.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 37851.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 105.*u2*(15. - 290.*w2 + 899.*fcs_pow(w,4) + 26970.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 116.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 899.*fcs_pow(u,10)*(2.544009e6*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 552188.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 3108.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 7.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 20202.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 9889.*fcs_pow(u,8)*(5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 140.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 13468.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 16182.*fcs_pow(u,6)*(5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 220.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 770.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return -3.698160658676859e18*fcs_pow(t,26)*u*(-54810.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 226548.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 267003.*fcs_pow(v,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 756.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 315.*u2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) - 580.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8990.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 39556.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 2697.*fcs_pow(u,10)*(101010.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 552188.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 1036.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 148335.*fcs_pow(u,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6) - 13468.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 28.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 518.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 16182.*fcs_pow(u,6)*(5.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 79365.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 3654.*fcs_pow(u,4)*(5.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 620.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return 1.109448197603058e19*fcs_pow(t,27)*u*w*(-63.*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) + 62930.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 276892.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 812.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 3045.*u2*(-4340.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 71610.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 7.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 232841.*fcs_pow(u,10)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6) - 641732.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 206271.*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 15990.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 156.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 346115.*fcs_pow(u,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 78884.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 148.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 2886.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 178002.*fcs_pow(u,6)*(35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 140.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2590.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 13468.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 37758.*fcs_pow(u,4)*(35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 15873.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
        case 8 : return -7.766137383221405e19*fcs_pow(t,28)*u*(33263.*fcs_pow(u,10)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8) + 111930.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 641732.*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 780.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 206271.*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8))) - 9.*(-4060.*v2*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 62930.*fcs_pow(v,4)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 276892.*fcs_pow(v,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 7.*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) + 9889.*fcs_pow(v,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 49445.*fcs_pow(u,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8) - 552188.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 14430.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 148.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 890010.*fcs_pow(u,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8) + 19721.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 1924.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 4.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 74.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 5394.*fcs_pow(u,4)*(35.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 79365.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 11396.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 435.*u2*(35.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 4340.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 9548.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
      }
      case 9 : switch(k){
        case 0 : return 2.7842353914375e13*fcs_pow(t,21)*u*v*(2.171085e6*fcs_pow(u,10)*(3. - 132.*v2 + 1386.*fcs_pow(v,4) - 4884.*fcs_pow(v,6) + 5291.*fcs_pow(v,8)) - 231.*(21. - 644.*v2 + 4830.*fcs_pow(v,4) - 12420.*fcs_pow(v,6) + 10005.*fcs_pow(v,8)) + 8855.*u2*(21. - 700.*v2 + 5670.*fcs_pow(v,4) - 15660.*fcs_pow(v,6) + 13485.*fcs_pow(v,8)) - 53130.*fcs_pow(u,4)*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) - 256795.*fcs_pow(u,8)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) + 22770.*fcs_pow(u,6)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)));
        case 1 : return -1.479264263470744e17*fcs_pow(t,22)*u*v*(-21. + 700.*v2 - 5670.*fcs_pow(v,4) + 15660.*fcs_pow(v,6) - 13485.*fcs_pow(v,8) - 22475.*fcs_pow(u,8)*(3. - 132.*v2 + 1386.*fcs_pow(v,4) - 4884.*fcs_pow(v,6) + 5291.*fcs_pow(v,8)) + 4495.*fcs_pow(u,10)*(9. - 420.*v2 + 4662.*fcs_pow(v,4) - 17316.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) + 25.*u2*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) + 870.*fcs_pow(u,6)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) - 30.*fcs_pow(u,4)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)))*w;
        case 2 : return 1.479264263470744e17*fcs_pow(t,23)*u*v*(21. - 525.*w2 + 700.*v2*(-1. + 27.*w2) - 5670.*fcs_pow(v,4)*(-1. + 29.*w2) + 15660.*fcs_pow(v,6)*(-1. + 31.*w2) - 13485.*fcs_pow(v,8)*(-1. + 33.*w2) + 25.*u2*(-35. + 945.*w2 - 1260.*v2*(-1. + 29.*w2) + 10962.*fcs_pow(v,4)*(-1. + 31.*w2) - 32364.*fcs_pow(v,6)*(-1. + 33.*w2) + 29667.*fcs_pow(v,8)*(-1. + 35.*w2)) - 30.*fcs_pow(u,4)*(315.*(-1. + 29.*w2) - 12180.*v2*(-1. + 31.*w2) + 113274.*fcs_pow(v,4)*(-1. + 33.*w2) - 356004.*fcs_pow(v,6)*(-1. + 35.*w2) + 346115.*fcs_pow(v,8)*(-1. + 37.*w2)) + 870.*fcs_pow(u,6)*(45.*(-1. + 31.*w2) - 1860.*v2*(-1. + 33.*w2) + 18414.*fcs_pow(v,4)*(-1. + 35.*w2) - 61380.*fcs_pow(v,6)*(-1. + 37.*w2) + 63085.*fcs_pow(v,8)*(-1. + 39.*w2)) - 22475.*fcs_pow(u,8)*(-3. + 99.*w2 - 132.*v2*(-1. + 35.*w2) + 1386.*fcs_pow(v,4)*(-1. + 37.*w2) - 4884.*fcs_pow(v,6)*(-1. + 39.*w2) + 5291.*fcs_pow(v,8)*(-1. + 41.*w2)) + 4495.*fcs_pow(u,10)*(-9. + 315.*w2 - 420.*v2*(-1. + 37.*w2) + 4662.*fcs_pow(v,4)*(-1. + 39.*w2) - 17316.*fcs_pow(v,6)*(-1. + 41.*w2) + 19721.*fcs_pow(v,8)*(-1. + 43.*w2)));
        case 3 : return -7.396321317353719e17*fcs_pow(t,24)*u*v*w*(-9.*(35.*(-1. + 9.*w2) - 32364.*fcs_pow(v,6)*(-1. + 11.*w2) - 420.*v2*(-3. + 29.*w2) + 3654.*fcs_pow(v,4)*(-3. + 31.*w2) + 9889.*fcs_pow(v,8)*(-3. + 35.*w2)) - 3654.*fcs_pow(u,4)*(-45. + 465.*w2 - 1860.*v2*(-1. + 11.*w2) + 63085.*fcs_pow(v,8)*(-1. + 13.*w2) + 6138.*fcs_pow(v,4)*(-3. + 35.*w2) - 20460.*fcs_pow(v,6)*(-3. + 37.*w2)) + 15.*u2*(339822.*fcs_pow(v,4)*(-1. + 11.*w2) + 315.*(-3. + 29.*w2) - 12180.*v2*(-3. + 31.*w2) - 356004.*fcs_pow(v,6)*(-3. + 35.*w2) + 346115.*fcs_pow(v,8)*(-3. + 37.*w2)) + 80910.*fcs_pow(u,6)*(-9. + 99.*w2 - 14652.*fcs_pow(v,6)*(-1. + 13.*w2) - 132.*v2*(-3. + 35.*w2) + 1386.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)) + 2697.*fcs_pow(u,10)*(-15540.*v2*(-1. + 13.*w2) + 848003.*fcs_pow(v,8)*(-1. + 15.*w2) + 105.*(-3. + 37.*w2) + 60606.*fcs_pow(v,4)*(-3. + 41.*w2) - 236652.*fcs_pow(v,6)*(-3. + 43.*w2)) - 49445.*fcs_pow(u,8)*(13986.*fcs_pow(v,4)*(-1. + 13.*w2) + 9.*(-3. + 35.*w2) - 420.*v2*(-3. + 37.*w2) - 17316.*fcs_pow(v,6)*(-3. + 41.*w2) + 19721.*fcs_pow(v,8)*(-3. + 43.*w2)));
        case 4 : return 2.218896395206116e18*fcs_pow(t,25)*u*v*(-1218.*fcs_pow(u,4)*(45.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 1860.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 61380.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 15.*u2*(-12180.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 113274.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 346115.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 105.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 118668.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 3.*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 10962.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 32364.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 420.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 899.*fcs_pow(u,10)*(315.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 15540.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 709956.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 60606.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 49445.*fcs_pow(u,8)*(9. - 630.*w2 + 3885.*fcs_pow(w,4) - 420.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 4662.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 5772.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 26970.*fcs_pow(u,6)*(9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 4158.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 132.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -1.109448197603058e19*fcs_pow(t,26)*u*v*w*(3045.*u2*(45. - 930.*w2 + 3069.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 4092.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 3.*(339822.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 356004.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 63.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 2436.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 69223.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 9889.*fcs_pow(u,8)*(4.240015e6*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 1.18326e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 15540.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 60606.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 59334.*fcs_pow(u,6)*(45.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 13986.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 420.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 17316.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 37758.*fcs_pow(u,4)*(45.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 660.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1386.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 99789.*fcs_pow(u,10)*(-275028.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 22386.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 21.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 364.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 22919.*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))));
        case 6 : return 1.109448197603058e19*fcs_pow(t,27)*u*v*(-9.*(-12180.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 113274.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 356004.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 63.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 69223.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 49445.*fcs_pow(u,8)*(181818.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 709956.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 63.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 3108.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 178002.*fcs_pow(u,6)*(45.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 86580.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 113274.*fcs_pow(u,4)*(15.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 3045.*u2*(45.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 1860.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 18414.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 12276.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 299367.*fcs_pow(u,10)*(-1820.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 22386.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 91676.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 22919.*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))));
        case 7 : return -2.895659795743981e21*fcs_pow(t,28)*u*v*w*(104377.*fcs_pow(u,10)*(-105. + 4305.*w2 - 37023.*fcs_pow(w,4) + 79335.*fcs_pow(w,6) + 74046.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 63468.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 820.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 82861.*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6))) - 7.*(-13020.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 128898.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 63.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 12276.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 441595.*fcs_pow(u,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6) - 91676.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 22919.*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 3198.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 52.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 682.*fcs_pow(u,6)*(2.9680105e7*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 3.54978e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 15540.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 181818.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 4774.*fcs_pow(u,4)*(315.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 17316.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 1085.*u2*(105.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 5291.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
        case 8 : return 2.026961857020787e22*fcs_pow(t,29)*u*v*(-315.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 13020.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 128898.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 12276.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) - 63085.*fcs_pow(u,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8) + 22386.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 91676.*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 260.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8))) + 14911.*fcs_pow(u,10)*(-5740.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 74046.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 15.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 63468.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 82861.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) + 3410.*fcs_pow(u,6)*(-709956.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 25974.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 9.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 444.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 682.*fcs_pow(u,4)*(315.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 690235.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 86580.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 420.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 4662.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 155.*u2*(105.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 26455.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 1386.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 4884.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -6.283581756764439e23*fcs_pow(t,30)*u*v*w*(-26455.*fcs_pow(u,8)*(666414.*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 2460.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 63468.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 82861.*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 9.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 36630.*fcs_pow(u,6)*(-825084.*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 9594.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 156.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 198.*fcs_pow(u,4)*(2.9680105e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 1.18326e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 15540.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 60606.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 165.*u2*(105.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 98605.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 13986.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 17316.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 140.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 9.*(105.*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) - 1540.*v2*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 14652.*fcs_pow(v,6)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5291.*fcs_pow(v,8)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 462.*fcs_pow(v,4)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 59163.*fcs_pow(u,10)*(-18060.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 15.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 5418.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 24252.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 14147.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
      }
      case 10 : switch(k){
        case 0 : return -5.84689432201875e14*fcs_pow(t,22)*u*(1265.*u2*(-7. + 875.*v2 - 15750.*fcs_pow(v,4) + 91350.*fcs_pow(v,6) - 202275.*fcs_pow(v,8) + 148335.*fcs_pow(v,10)) - 183425.*fcs_pow(u,8)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) - 12650.*fcs_pow(u,4)*(-7. + 945.*v2 - 18270.*fcs_pow(v,4) + 113274.*fcs_pow(v,6) - 267003.*fcs_pow(v,8) + 207669.*fcs_pow(v,10)) + 103385.*fcs_pow(u,10)*(-3. + 495.*v2 - 11550.*fcs_pow(v,4) + 85470.*fcs_pow(v,6) - 238095.*fcs_pow(v,8) + 216931.*fcs_pow(v,10)) - 11.*(-21. + 2415.*v2 - 40250.*fcs_pow(v,4) + 217350.*fcs_pow(v,6) - 450225.*fcs_pow(v,8) + 310155.*fcs_pow(v,10)) + 37950.*fcs_pow(u,6)*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)));
        case 1 : return 1.479264263470744e17*fcs_pow(t,23)*u*(3.*(7. - 875.*v2 + 15750.*fcs_pow(v,4) - 91350.*fcs_pow(v,6) + 202275.*fcs_pow(v,8) - 148335.*fcs_pow(v,10)) + 13050.*fcs_pow(u,6)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) + 125.*u2*(-7. + 945.*v2 - 18270.*fcs_pow(v,4) + 113274.*fcs_pow(v,6) - 267003.*fcs_pow(v,8) + 207669.*fcs_pow(v,10)) - 22475.*fcs_pow(u,8)*(-3. + 495.*v2 - 11550.*fcs_pow(v,4) + 85470.*fcs_pow(v,6) - 238095.*fcs_pow(v,8) + 216931.*fcs_pow(v,10)) - 1050.*fcs_pow(u,4)*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)) + 4495.*fcs_pow(u,10)*(-9. + 1575.*v2 - 38850.*fcs_pow(v,4) + 303030.*fcs_pow(v,6) - 887445.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)))*w;
        case 2 : return -1.479264263470744e17*fcs_pow(t,24)*u*(-3.*(7. - 175.*w2 + 875.*v2*(-1. + 27.*w2) - 15750.*fcs_pow(v,4)*(-1. + 29.*w2) + 91350.*fcs_pow(v,6)*(-1. + 31.*w2) - 202275.*fcs_pow(v,8)*(-1. + 33.*w2) + 148335.*fcs_pow(v,10)*(-1. + 35.*w2)) + 125.*u2*(7. - 189.*w2 + 945.*v2*(-1. + 29.*w2) - 18270.*fcs_pow(v,4)*(-1. + 31.*w2) + 113274.*fcs_pow(v,6)*(-1. + 33.*w2) - 267003.*fcs_pow(v,8)*(-1. + 35.*w2) + 207669.*fcs_pow(v,10)*(-1. + 37.*w2)) - 1050.*fcs_pow(u,4)*(9. - 261.*w2 + 1305.*v2*(-1. + 31.*w2) - 26970.*fcs_pow(v,4)*(-1. + 33.*w2) + 178002.*fcs_pow(v,6)*(-1. + 35.*w2) - 445005.*fcs_pow(v,8)*(-1. + 37.*w2) + 365893.*fcs_pow(v,10)*(-1. + 39.*w2)) + 13050.*fcs_pow(u,6)*(3. - 93.*w2 + 465.*v2*(-1. + 33.*w2) - 10230.*fcs_pow(v,4)*(-1. + 35.*w2) + 71610.*fcs_pow(v,6)*(-1. + 37.*w2) - 189255.*fcs_pow(v,8)*(-1. + 39.*w2) + 164021.*fcs_pow(v,10)*(-1. + 41.*w2)) - 22475.*fcs_pow(u,8)*(3. - 99.*w2 + 495.*v2*(-1. + 35.*w2) - 11550.*fcs_pow(v,4)*(-1. + 37.*w2) + 85470.*fcs_pow(v,6)*(-1. + 39.*w2) - 238095.*fcs_pow(v,8)*(-1. + 41.*w2) + 216931.*fcs_pow(v,10)*(-1. + 43.*w2)) + 4495.*fcs_pow(u,10)*(9. - 315.*w2 + 1575.*v2*(-1. + 37.*w2) - 38850.*fcs_pow(v,4)*(-1. + 39.*w2) + 303030.*fcs_pow(v,6)*(-1. + 41.*w2) - 887445.*fcs_pow(v,8)*(-1. + 43.*w2) + 848003.*fcs_pow(v,10)*(-1. + 45.*w2)));
        case 3 : return 1.109448197603058e19*fcs_pow(t,25)*u*w*(105.*u2*(9. - 87.*w2 - 26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 365893.*fcs_pow(v,10)*(-1. + 13.*w2) + 435.*v2*(-3. + 31.*w2) + 59334.*fcs_pow(v,6)*(-3. + 35.*w2) - 148335.*fcs_pow(v,8)*(-3. + 37.*w2)) - 3.*(7. - 63.*w2 + 113274.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 6090.*fcs_pow(v,4)*(-3. + 31.*w2) - 89001.*fcs_pow(v,8)*(-3. + 35.*w2) + 69223.*fcs_pow(v,10)*(-3. + 37.*w2)) - 1218.*fcs_pow(u,4)*(9. - 93.*w2 + 1395.*v2*(-1. + 11.*w2) - 567765.*fcs_pow(v,8)*(-1. + 13.*w2) - 10230.*fcs_pow(v,4)*(-3. + 35.*w2) + 71610.*fcs_pow(v,6)*(-3. + 37.*w2) + 164021.*fcs_pow(v,10)*(-3. + 41.*w2)) - 9889.*fcs_pow(u,8)*(9. - 105.*w2 - 38850.*fcs_pow(v,4)*(-1. + 13.*w2) + 848003.*fcs_pow(v,10)*(-1. + 15.*w2) + 525.*v2*(-3. + 37.*w2) + 101010.*fcs_pow(v,6)*(-3. + 41.*w2) - 295815.*fcs_pow(v,8)*(-3. + 43.*w2)) + 5394.*fcs_pow(u,6)*(9. - 99.*w2 + 256410.*fcs_pow(v,6)*(-1. + 13.*w2) + 495.*v2*(-3. + 35.*w2) - 11550.*fcs_pow(v,4)*(-3. + 37.*w2) - 238095.*fcs_pow(v,8)*(-3. + 41.*w2) + 216931.*fcs_pow(v,10)*(-3. + 43.*w2)) + 2697.*fcs_pow(u,10)*(21. - 259.*w2 + 3885.*v2*(-1. + 13.*w2) - 2.544009e6*fcs_pow(v,8)*(-1. + 15.*w2) - 33670.*fcs_pow(v,4)*(-3. + 41.*w2) + 276094.*fcs_pow(v,6)*(-3. + 43.*w2) + 848003.*fcs_pow(v,10)*(-3. + 47.*w2)));
        case 4 : return -1.109448197603058e19*fcs_pow(t,26)*u*(105.*u2*(-9. + 522.*w2 - 2697.*fcs_pow(w,4) + 1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 445005.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 365893.*fcs_pow(v,10)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59334.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 3.*(-7.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 18270.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 113274.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 207669.*fcs_pow(v,10)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 9889.*fcs_pow(u,8)*(-9. + 630.*w2 - 3885.*fcs_pow(w,4) + 1575.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 38850.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 887445.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 101010.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 16182.*fcs_pow(u,6)*(-3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 11550.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 85470.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 165.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 79365.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 1218.*fcs_pow(u,4)*(-9.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 1395.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 214830.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 567765.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 10230.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 164021.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 2697.*fcs_pow(u,10)*(-21.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 3885.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 828282.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 2.544009e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 33670.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return 3.328344592809173e19*fcs_pow(t,27)*u*w*(-21.*(-45. + 870.*w2 - 2697.*fcs_pow(w,4) - 134850.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 296670.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 365893.*fcs_pow(v,10)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 435.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) - 148335.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 296670.*fcs_pow(u,6)*(-9. + 210.*w2 - 777.*fcs_pow(w,4) + 848003.*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 295815.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 20202.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 62930.*fcs_pow(u,4)*(-9.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 495.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 51282.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 2310.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 47619.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 1015.*u2*(-45. + 930.*w2 - 3069.*fcs_pow(w,4) + 6975.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 51150.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 567765.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 71610.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 164021.*fcs_pow(v,10)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 49445.*fcs_pow(u,8)*(-1.2720045e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1.38047e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3885.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 7.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 33670.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 33263.*fcs_pow(u,10)*(4.81299e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 186550.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 21.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1365.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.031355e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return -9.98503377842752e19*fcs_pow(t,28)*u*(98890.*fcs_pow(u,6)*(9. - 945.*w2 + 11655.*fcs_pow(w,4) - 30303.*fcs_pow(w,6) + 303030.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 315.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 62930.*fcs_pow(u,4)*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 238095.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 216931.*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 5075.*u2*(3. - 279.*w2 + 3069.*fcs_pow(w,4) - 7161.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 164021.*fcs_pow(v,10)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 14322.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 37851.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 7.*(6525.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 134850.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 890010.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) - 445005.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 365893.*fcs_pow(v,10)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 49445.*fcs_pow(u,8)*(-168350.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.38047e6*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.240015e6*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 33263.*fcs_pow(u,10)*(6825.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 186550.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.60433e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.031355e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return 2.026961857020787e22*fcs_pow(t,29)*u*w*(-63085.*fcs_pow(u,8)*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 1.60433e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.031355e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 26650.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 195.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 30690.*fcs_pow(u,6)*(35. - 1295.*w2 + 10101.*fcs_pow(w,4) - 19721.*fcs_pow(w,6) - 4.240015e6*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 197210.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 185.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 4810.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 3410.*fcs_pow(u,4)*(-63.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 5.936021e6*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 60606.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 775.*u2*(21.*(1. - 33.*w2 + 231.*fcs_pow(w,4) - 407.*fcs_pow(w,6)) + 3465.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 216931.*fcs_pow(v,10)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 47619.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 3.*(16275.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 358050.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 21.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 71610.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 189255.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 164021.*fcs_pow(v,10)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 14911.*fcs_pow(u,10)*(105. - 4305.*w2 + 37023.*fcs_pow(w,4) - 79335.*fcs_pow(w,6) - 617050.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.11069e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 3075.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 3.728745e6*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 580027.*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return -1.013480928510393e23*fcs_pow(t,30)*u*(-12617.*fcs_pow(u,8)*(-35. + 5460.*w2 - 111930.*fcs_pow(w,4) + 641732.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) - 186550.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.60433e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 975.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) + 14911.*fcs_pow(u,10)*(-21. + 3444.*w2 - 74046.*fcs_pow(w,4) + 444276.*fcs_pow(w,6) - 745749.*fcs_pow(w,8) + 4305.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 123410.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 222138.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 745749.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) + 6138.*fcs_pow(u,6)*(-35. + 5180.*w2 - 101010.*fcs_pow(w,4) + 552188.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 1.38047e6*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 4.240015e6*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 24050.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 185.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 4774.*fcs_pow(u,4)*(-9.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 887445.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 43290.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 45.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 1110.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 155.*u2*(-21.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 3465.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 1.518517e6*fcs_pow(v,10)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 238095.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 17094.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 3.*(-21.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 3255.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 164021.*fcs_pow(v,10)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 14322.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 37851.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return 9.425372635146658e24*fcs_pow(t,31)*u*w*(2442.*fcs_pow(u,6)*(1.443897e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 79950.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 585.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 462.*fcs_pow(u,4)*(-315. + 15540.*w2 - 181818.*fcs_pow(w,4) + 709956.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) - 3.8160135e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 591630.*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1665.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 14430.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 165.*u2*(5.936021e6*fcs_pow(v,10)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 7.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 295815.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 7770.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 20202.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 3.*(-21.*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) + 1155.*v2*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 216931.*fcs_pow(v,10)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 51282.*fcs_pow(v,6)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 47619.*fcs_pow(v,8)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 770.*fcs_pow(v,4)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 19721.*fcs_pow(u,10)*(13545.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 3.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 9030.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 84882.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 240499.*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 127323.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 5291.*fcs_pow(u,8)*(-1.85115e6*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 3075.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 370230.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1.242915e6*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 3.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return -9.425372635146658e24*fcs_pow(t,32)*u*(36630.*fcs_pow(u,6)*(21. - 4095.*w2 + 111930.*fcs_pow(w,4) - 962598.*fcs_pow(w,6) + 3.094065e6*fcs_pow(w,8) - 3.231579e6*fcs_pow(w,10) - 111930.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 962598.*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.077193e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 343785.*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 195.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 2310.*fcs_pow(u,4)*(2.484846e6*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 7.632027e6*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 9.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 1665.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 14430.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 1155.*u2*(9. - 1575.*w2 + 38850.*fcs_pow(w,4) - 303030.*fcs_pow(w,6) + 887445.*fcs_pow(w,8) - 848003.*fcs_pow(w,10) - 887445.*fcs_pow(v,8)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 225.*v2*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 5550.*fcs_pow(v,4)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 14430.*fcs_pow(v,6)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 9.*(-7.*(-3. + 495.*w2 - 11550.*fcs_pow(w,4) + 85470.*fcs_pow(w,6) - 238095.*fcs_pow(w,8) + 216931.*fcs_pow(w,10)) + 1.518517e6*fcs_pow(v,10)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 385.*v2*(-9. + 1575.*w2 - 38850.*fcs_pow(w,4) + 303030.*fcs_pow(w,6) - 887445.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 11550.*fcs_pow(v,4)*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 85470.*fcs_pow(v,6)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 79365.*fcs_pow(v,8)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 26455.*fcs_pow(u,8)*(3.*(21. - 4305.*w2 + 123410.*fcs_pow(w,4) - 1.11069e6*fcs_pow(w,6) + 3.728745e6*fcs_pow(w,8) - 4.060189e6*fcs_pow(w,10)) + 12915.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 370230.*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 3.728745e6*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 370230.*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 138047.*fcs_pow(u,10)*(-9.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 1935.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 181890.*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 6450.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 90945.*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 34357.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
      }
      case 11 : switch(k){
        case 0 : return 1.479264263470744e17*fcs_pow(t,23)*u*v*(1375.*u2*(-7. + 315.*v2 - 3654.*fcs_pow(v,4) + 16182.*fcs_pow(v,6) - 29667.*fcs_pow(v,8) + 18879.*fcs_pow(v,10)) - 247225.*fcs_pow(u,8)*(-3. + 165.*v2 - 2310.*fcs_pow(v,4) + 12210.*fcs_pow(v,6) - 26455.*fcs_pow(v,8) + 19721.*fcs_pow(v,10)) - 11.*(-21. + 875.*v2 - 9450.*fcs_pow(v,4) + 39150.*fcs_pow(v,6) - 67425.*fcs_pow(v,8) + 40455.*fcs_pow(v,10)) + 47850.*fcs_pow(u,6)*(-9. + 465.*v2 - 6138.*fcs_pow(v,4) + 30690.*fcs_pow(v,6) - 63085.*fcs_pow(v,8) + 44733.*fcs_pow(v,10)) - 1650.*fcs_pow(u,4)*(-63. + 3045.*v2 - 37758.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 346115.*fcs_pow(v,8) + 232841.*fcs_pow(v,10)) + 4495.*fcs_pow(u,10)*(-99. + 5775.*v2 - 85470.*fcs_pow(v,4) + 476190.*fcs_pow(v,6) - 1.084655e6*fcs_pow(v,8) + 848003.*fcs_pow(v,10)));
        case 1 : return -3.698160658676859e18*fcs_pow(t,24)*u*v*(-33.*(-7. + 315.*v2 - 3654.*fcs_pow(v,4) + 16182.*fcs_pow(v,6) - 29667.*fcs_pow(v,8) + 18879.*fcs_pow(v,10)) + 178002.*fcs_pow(u,6)*(-3. + 165.*v2 - 2310.*fcs_pow(v,4) + 12210.*fcs_pow(v,6) - 26455.*fcs_pow(v,8) + 19721.*fcs_pow(v,10)) - 13398.*fcs_pow(u,4)*(-9. + 465.*v2 - 6138.*fcs_pow(v,4) + 30690.*fcs_pow(v,6) - 63085.*fcs_pow(v,8) + 44733.*fcs_pow(v,10)) + 165.*u2*(-63. + 3045.*v2 - 37758.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 346115.*fcs_pow(v,8) + 232841.*fcs_pow(v,10)) - 9889.*fcs_pow(u,8)*(-99. + 5775.*v2 - 85470.*fcs_pow(v,4) + 476190.*fcs_pow(v,6) - 1.084655e6*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 899.*fcs_pow(u,10)*(-693. + 42735.*v2 - 666666.*fcs_pow(v,4) + 3.904758e6*fcs_pow(v,6) - 9.328033e6*fcs_pow(v,8) + 7.632027e6*fcs_pow(v,10)))*w;
        case 2 : return 3.698160658676859e18*fcs_pow(t,25)*u*v*(-33.*(7. - 189.*w2 + 315.*v2*(-1. + 29.*w2) - 3654.*fcs_pow(v,4)*(-1. + 31.*w2) + 16182.*fcs_pow(v,6)*(-1. + 33.*w2) - 29667.*fcs_pow(v,8)*(-1. + 35.*w2) + 18879.*fcs_pow(v,10)*(-1. + 37.*w2)) + 165.*u2*(63. - 1827.*w2 + 3045.*v2*(-1. + 31.*w2) - 37758.*fcs_pow(v,4)*(-1. + 33.*w2) + 178002.*fcs_pow(v,6)*(-1. + 35.*w2) - 346115.*fcs_pow(v,8)*(-1. + 37.*w2) + 232841.*fcs_pow(v,10)*(-1. + 39.*w2)) - 13398.*fcs_pow(u,4)*(9. - 279.*w2 + 465.*v2*(-1. + 33.*w2) - 6138.*fcs_pow(v,4)*(-1. + 35.*w2) + 30690.*fcs_pow(v,6)*(-1. + 37.*w2) - 63085.*fcs_pow(v,8)*(-1. + 39.*w2) + 44733.*fcs_pow(v,10)*(-1. + 41.*w2)) + 178002.*fcs_pow(u,6)*(3. - 99.*w2 + 165.*v2*(-1. + 35.*w2) - 2310.*fcs_pow(v,4)*(-1. + 37.*w2) + 12210.*fcs_pow(v,6)*(-1. + 39.*w2) - 26455.*fcs_pow(v,8)*(-1. + 41.*w2) + 19721.*fcs_pow(v,10)*(-1. + 43.*w2)) - 9889.*fcs_pow(u,8)*(99. - 3465.*w2 + 5775.*v2*(-1. + 37.*w2) - 85470.*fcs_pow(v,4)*(-1. + 39.*w2) + 476190.*fcs_pow(v,6)*(-1. + 41.*w2) - 1.084655e6*fcs_pow(v,8)*(-1. + 43.*w2) + 848003.*fcs_pow(v,10)*(-1. + 45.*w2)) + 899.*fcs_pow(u,10)*(693.*(1. - 37.*w2) + 42735.*v2*(-1. + 39.*w2) - 666666.*fcs_pow(v,4)*(-1. + 41.*w2) + 3.904758e6*fcs_pow(v,6)*(-1. + 43.*w2) - 9.328033e6*fcs_pow(v,8)*(-1. + 45.*w2) + 7.632027e6*fcs_pow(v,10)*(-1. + 47.*w2)));
        case 3 : return -1.109448197603058e19*fcs_pow(t,26)*u*v*w*(-33.*(63.*(3. - 29.*w2) - 113274.*fcs_pow(v,4)*(-1. + 11.*w2) + 698523.*fcs_pow(v,10)*(-1. + 13.*w2) + 3045.*v2*(-3. + 31.*w2) + 178002.*fcs_pow(v,6)*(-3. + 35.*w2) - 346115.*fcs_pow(v,8)*(-3. + 37.*w2)) + 33495.*u2*(9. - 93.*w2 + 465.*v2*(-1. + 11.*w2) - 63085.*fcs_pow(v,8)*(-1. + 13.*w2) - 2046.*fcs_pow(v,4)*(-3. + 35.*w2) + 10230.*fcs_pow(v,6)*(-3. + 37.*w2) + 14911.*fcs_pow(v,10)*(-3. + 41.*w2)) + 59334.*fcs_pow(u,6)*(99.*(3. - 35.*w2) - 256410.*fcs_pow(v,4)*(-1. + 13.*w2) + 2.544009e6*fcs_pow(v,10)*(-1. + 15.*w2) + 5775.*v2*(-3. + 37.*w2) + 476190.*fcs_pow(v,6)*(-3. + 41.*w2) - 1.084655e6*fcs_pow(v,8)*(-3. + 43.*w2)) - 415338.*fcs_pow(u,4)*(9. - 99.*w2 + 36630.*fcs_pow(v,6)*(-1. + 13.*w2) + 165.*v2*(-3. + 35.*w2) - 2310.*fcs_pow(v,4)*(-3. + 37.*w2) - 26455.*fcs_pow(v,8)*(-3. + 41.*w2) + 19721.*fcs_pow(v,10)*(-3. + 43.*w2)) - 49445.*fcs_pow(u,8)*(693. - 8547.*w2 + 42735.*v2*(-1. + 13.*w2) - 9.328033e6*fcs_pow(v,8)*(-1. + 15.*w2) - 222222.*fcs_pow(v,4)*(-3. + 41.*w2) + 1.301586e6*fcs_pow(v,6)*(-3. + 43.*w2) + 2.544009e6*fcs_pow(v,10)*(-3. + 47.*w2)) + 99789.*fcs_pow(u,10)*(231. - 3003.*w2 + 1.512654e6*fcs_pow(v,6)*(-1. + 15.*w2) + 5005.*v2*(-3. + 41.*w2) - 82082.*fcs_pow(v,4)*(-3. + 43.*w2) - 1.260545e6*fcs_pow(v,8)*(-3. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-3. + 49.*w2)));
        case 4 : return 1.109448197603058e19*fcs_pow(t,27)*u*v*(-99.*(3045.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 37758.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 346115.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 232841.*fcs_pow(v,10)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 21.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 59334.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 178002.*fcs_pow(u,6)*(5775.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 85470.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 1.084655e6*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 33.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 158730.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 415338.*fcs_pow(u,4)*(-9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 6930.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 36630.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59163.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 165.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 26455.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 33495.*u2*(-9.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 465.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 30690.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 2046.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 14911.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 99789.*fcs_pow(u,10)*(-231.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 246246.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 1.512654e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 3.231579e6*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 5005.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 1.260545e6*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 49445.*fcs_pow(u,8)*(-693.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 42735.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 3.904758e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 9.328033e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 222222.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 2.544009e6*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return -9.652199319146603e20*fcs_pow(t,28)*u*v*w*(-23870.*fcs_pow(u,4)*(2.544009e6*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 99.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 1.084655e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 51282.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1155.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 95238.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 59675.*u2*(-9.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 165.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 19721.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7326.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 462.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 231.*(-45. + 930.*w2 - 3069.*fcs_pow(w,4) + 2325.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 10230.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 63085.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 10230.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 14911.*fcs_pow(v,10)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 10230.*fcs_pow(u,6)*(-4.6640165e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 6.50793e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 42735.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 231.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 222222.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 2.544009e6*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 14911.*fcs_pow(u,10)*(-4.07253e6*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 78925.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 1.2180567e7*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 231.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 1.74537e6*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 4.557355e6*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 63085.*fcs_pow(u,8)*(7.56327e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 410410.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 231.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 5005.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.260545e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return 1.44782989787199e22*fcs_pow(t,29)*u*v*(-4774.*fcs_pow(u,4)*(-99.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 476190.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 1.084655e6*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 1155.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 11935.*u2*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 462.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2442.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 77.*(9. - 837.*w2 + 9207.*fcs_pow(w,4) - 21483.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 6138.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 44733.*fcs_pow(v,10)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 6138.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 682.*fcs_pow(u,6)*(-3.33333e6*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.952379e7*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.6640165e7*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 693.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 7.632027e6*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 14911.*fcs_pow(u,10)*(-231.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 15785.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 271502.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 4.060189e6*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 349074.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 911471.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 12617.*fcs_pow(u,8)*(25025.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 410410.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 2.52109e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 77.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.260545e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return -3.141790878382219e24*fcs_pow(t,30)*u*v*w*(7326.*fcs_pow(u,6)*(11.*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6)) + 2.52109e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.260545e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 58630.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 715.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 22.*fcs_pow(u,4)*(-3.26481155e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 5.3424189e7*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 1.952379e7*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 693.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 666666.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 55.*u2*(-693.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 5.936021e6*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.084655e6*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1155.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 95238.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 33.*(21.*(1. - 33.*w2 + 231.*fcs_pow(w,4) - 407.*fcs_pow(w,6)) + 1155.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 19721.*fcs_pow(v,10)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 462.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2442.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 5291.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 19721.*fcs_pow(u,10)*(16555.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 59598.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 33.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 400158.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 721497.*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 155617.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 5291.*fcs_pow(u,8)*(-1.35751e6*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.74537e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 11275.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.557355e6*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 33.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 580027.*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return 3.141790878382219e24*fcs_pow(t,31)*u*v*(-26455.*fcs_pow(u,8)*(15785.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 271502.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 33.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 349074.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 911471.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) + 19721.*fcs_pow(u,10)*(-231.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 16555.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 59598.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 400158.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 778085.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 721497.*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 7326.*fcs_pow(u,6)*(-410410.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 2.52109e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 3575.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.260545e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 154.*fcs_pow(u,4)*(1.952379e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 4.6640165e7*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 476190.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 7.632027e6*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 99.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 6105.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 55.*u2*(-693.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 7.592585e6*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 5.936021e6*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 476190.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 1155.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 17094.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 33.*(-21.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 1155.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 138047.*fcs_pow(v,10)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 26455.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 462.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 2442.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -9.425372635146658e24*fcs_pow(t,32)*u*v*w*(-17094.*fcs_pow(u,4)*(2.268981e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 175890.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 99.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1.260545e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 2145.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 165.*u2*(-3.26481155e8*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 6.50793e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 42735.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5.936021e6*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 222222.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 77.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 33.*(1.7808063e7*fcs_pow(v,10)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 231.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 1.084655e6*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 51282.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 95238.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 385.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 5.936021e6*fcs_pow(u,10)*(-99.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 165.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3102.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 43945.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 9306.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1.084655e6*fcs_pow(u,8)*(49665.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 33.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 19866.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 133386.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 240499.*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 155617.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 31746.*fcs_pow(u,6)*(-1.221759e7*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 33825.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1.74537e6*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 4.557355e6*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 99.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 1.740081e6*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return 9.425372635146658e24*fcs_pow(t,33)*u*v*(-85470.*fcs_pow(u,4)*(-738738.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 4.537962e6*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 3.231579e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 1.260545e6*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 2145.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 5775.*u2*(3.904758e6*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 9.328033e6*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 6105.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 31746.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 99.*(-7.592585e6*fcs_pow(v,8)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 5.936021e6*fcs_pow(v,10)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 77.*(-9. + 1575.*w2 - 38850.*fcs_pow(w,4) + 303030.*fcs_pow(w,6) - 887445.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 5775.*v2*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 85470.*fcs_pow(v,4)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 158730.*fcs_pow(v,6)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 476190.*fcs_pow(u,6)*(47355.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 814506.*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4.557355e6*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 581790.*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 33.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 5.936021e6*fcs_pow(u,10)*(-99.*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 46530.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 825.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 381123.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 46530.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 43945.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 1.084655e6*fcs_pow(u,8)*(-693.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 49665.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 2.00079e6*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 99330.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 778085.*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 240499.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
        case 11 : return -5.183954949330662e26*fcs_pow(t,34)*u*v*w*(3.904758e6*fcs_pow(u,6)*(693. - 49665.*w2 + 893970.*fcs_pow(w,4) - 6.00237e6*fcs_pow(w,6) + 1.6339785e7*fcs_pow(w,8) - 1.5151437e7*fcs_pow(w,10) + 4515.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99330.*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 666930.*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 778085.*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 240499.*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10))) - 6.5296231e7*fcs_pow(u,8)*(-9.*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 825.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 15510.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 46530.*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 43945.*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 42347.*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10))) - 222222.*fcs_pow(u,4)*(-222138.*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.74537e6*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4.557355e6*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 1.740081e6*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 9.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 615.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 42735.*u2*(412542.*fcs_pow(v,6)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.260545e6*fcs_pow(v,8)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.077193e6*fcs_pow(v,10)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 9.*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 195.*v2*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 3198.*fcs_pow(v,4)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 5.936021e6*fcs_pow(u,10)*(-99.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 7755.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 65142.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 158202.*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 465817.*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 381123.*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 99.*(-5.936021e6*fcs_pow(v,8)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 5.936021e6*fcs_pow(v,10)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 3885.*v2*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 20202.*fcs_pow(v,4)*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 118326.*fcs_pow(v,6)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10)) - 7.*(-693. + 42735.*w2 - 666666.*fcs_pow(w,4) + 3.904758e6*fcs_pow(w,6) - 9.328033e6*fcs_pow(w,8) + 7.632027e6*fcs_pow(w,10))));
      }
    }
    case 12 : switch(j){
      case 0 : switch(k){
        case 0 : return 467775.*fcs_pow(t,13)*(231. - 18018.*u2 + 225225.*fcs_pow(u,4) - 1.02102e6*fcs_pow(u,6) + 2.078505e6*fcs_pow(u,8) - 1.939938e6*fcs_pow(u,10) + 676039.*fcs_pow(u,12));
      }
      case 1 : switch(k){
        case 0 : return -4.2567525e7*fcs_pow(t,14)*(33. - 2970.*u2 + 42075.*fcs_pow(u,4) - 213180.*fcs_pow(u,6) + 479655.*fcs_pow(u,8) - 490314.*fcs_pow(u,10) + 185725.*fcs_pow(u,12))*v;
        case 1 : return 6.38512875e8*fcs_pow(t,15)*(33. - 3366.*u2 + 53295.*fcs_pow(u,4) - 298452.*fcs_pow(u,6) + 735471.*fcs_pow(u,8) - 817190.*fcs_pow(u,10) + 334305.*fcs_pow(u,12))*v*w;
      }
      case 2 : switch(k){
        case 0 : return 4.2567525e7*fcs_pow(t,15)*(-33. + 495.*v2 - 2970.*u2*(-1. + 17.*v2) + 42075.*fcs_pow(u,4)*(-1. + 19.*v2) - 213180.*fcs_pow(u,6)*(-1. + 21.*v2) + 479655.*fcs_pow(u,8)*(-1. + 23.*v2) - 490314.*fcs_pow(u,10)*(-1. + 25.*v2) + 185725.*fcs_pow(u,12)*(-1. + 27.*v2));
        case 1 : return -6.38512875e8*fcs_pow(t,16)*(-33. + 561.*v2 - 3366.*u2*(-1. + 19.*v2) + 53295.*fcs_pow(u,4)*(-1. + 21.*v2) - 298452.*fcs_pow(u,6)*(-1. + 23.*v2) + 735471.*fcs_pow(u,8)*(-1. + 25.*v2) - 817190.*fcs_pow(u,10)*(-1. + 27.*v2) + 334305.*fcs_pow(u,12)*(-1. + 29.*v2))*w;
        case 2 : return 6.38512875e8*fcs_pow(t,17)*(33.*(1. - 17.*w2 + 17.*v2*(-1. + 19.*w2)) - 3366.*u2*(1. - 19.*w2 + 19.*v2*(-1. + 21.*w2)) + 53295.*fcs_pow(u,4)*(1. - 21.*w2 + 21.*v2*(-1. + 23.*w2)) - 298452.*fcs_pow(u,6)*(1. - 23.*w2 + 23.*v2*(-1. + 25.*w2)) + 735471.*fcs_pow(u,8)*(1. - 25.*w2 + 25.*v2*(-1. + 27.*w2)) - 817190.*fcs_pow(u,10)*(1. - 27.*w2 + 27.*v2*(-1. + 29.*w2)) + 334305.*fcs_pow(u,12)*(1. - 29.*w2 + 29.*v2*(-1. + 31.*w2)));
      }
      case 3 : switch(k){
        case 0 : return -1.915538625e9*fcs_pow(t,16)*v*(-33. + 187.*v2 + 53295.*fcs_pow(u,4)*(-1. + 7.*v2) - 817190.*fcs_pow(u,10)*(-1. + 9.*v2) - 1122.*u2*(-3. + 19.*v2) - 99484.*fcs_pow(u,6)*(-3. + 23.*v2) + 245157.*fcs_pow(u,8)*(-3. + 25.*v2) + 111435.*fcs_pow(u,12)*(-3. + 29.*v2));
        case 1 : return 3.2564156625e10*fcs_pow(t,17)*v*(-33. + 209.*v2 - 3762.*u2*(-1. + 7.*v2) + 1.081575e6*fcs_pow(u,8)*(-1. + 9.*v2) + 21945.*fcs_pow(u,4)*(-3. + 23.*v2) - 134596.*fcs_pow(u,6)*(-3. + 25.*v2) - 432630.*fcs_pow(u,10)*(-3. + 29.*v2) + 190095.*fcs_pow(u,12)*(-3. + 31.*v2))*w;
        case 2 : return -3.2564156625e10*fcs_pow(t,18)*v*(11.*(3. - 57.*w2 + 19.*v2*(-1. + 21.*w2)) - 3762.*u2*(1. - 21.*w2 + 7.*v2*(-1. + 23.*w2)) + 21945.*fcs_pow(u,4)*(3. - 69.*w2 + 23.*v2*(-1. + 25.*w2)) - 134596.*fcs_pow(u,6)*(3. - 75.*w2 + 25.*v2*(-1. + 27.*w2)) + 1.081575e6*fcs_pow(u,8)*(1. - 27.*w2 + 9.*v2*(-1. + 29.*w2)) - 432630.*fcs_pow(u,10)*(3. - 87.*w2 + 29.*v2*(-1. + 31.*w2)) + 190095.*fcs_pow(u,12)*(3. - 93.*w2 + 31.*v2*(-1. + 33.*w2)));
        case 3 : return 1.856156927625e12*fcs_pow(t,19)*v*w*(26565.*fcs_pow(u,4)*(3. - 25.*w2 + 25.*v2*(-1. + 9.*w2)) - 660330.*fcs_pow(u,10)*(3. - 31.*w2 + 31.*v2*(-1. + 11.*w2)) + 11.*(3. - 21.*w2 + 7.*v2*(-3. + 23.*w2)) - 462.*u2*(9. - 69.*w2 + 23.*v2*(-3. + 25.*w2)) + 170775.*fcs_pow(u,8)*(9. - 87.*w2 + 29.*v2*(-3. + 31.*w2)) + 310155.*fcs_pow(u,12)*(3. - 33.*w2 + 11.*v2*(-3. + 35.*w2)) - 531300.*fcs_pow(u,6)*(1. - 9.*w2 + v2*(-9. + 87.*w2)));
      }
      case 4 : switch(k){
        case 0 : return 1.915538625e9*fcs_pow(t,17)*(-3366.*u2*(1. - 38.*v2 + 133.*fcs_pow(v,4)) + 53295.*fcs_pow(u,4)*(1. - 42.*v2 + 161.*fcs_pow(v,4)) + 735471.*fcs_pow(u,8)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) - 817190.*fcs_pow(u,10)*(1. - 54.*v2 + 261.*fcs_pow(v,4)) + 11.*(3. - 102.*v2 + 323.*fcs_pow(v,4)) - 99484.*fcs_pow(u,6)*(3. - 138.*v2 + 575.*fcs_pow(v,4)) + 111435.*fcs_pow(u,12)*(3. - 174.*v2 + 899.*fcs_pow(v,4)));
        case 1 : return -9.7692469875e10*fcs_pow(t,18)*(11.*(1. - 38.*v2 + 133.*fcs_pow(v,4)) - 1254.*u2*(1. - 42.*v2 + 161.*fcs_pow(v,4)) - 134596.*fcs_pow(u,6)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) + 360525.*fcs_pow(u,8)*(1. - 54.*v2 + 261.*fcs_pow(v,4)) + 190095.*fcs_pow(u,12)*(1. - 62.*v2 + 341.*fcs_pow(v,4)) + 7315.*fcs_pow(u,4)*(3. - 138.*v2 + 575.*fcs_pow(v,4)) - 144210.*fcs_pow(u,10)*(3. - 174.*v2 + 899.*fcs_pow(v,4)))*w;
        case 2 : return 9.7692469875e10*fcs_pow(t,19)*(11.*(-1. + 19.*w2 + v2*(38. - 798.*w2) + 133.*fcs_pow(v,4)*(-1. + 23.*w2)) - 1254.*u2*(-1. + 21.*w2 + v2*(42. - 966.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) + 7315.*fcs_pow(u,4)*(-3. + 69.*w2 - 138.*v2*(-1. + 25.*w2) + 575.*fcs_pow(v,4)*(-1. + 27.*w2)) - 134596.*fcs_pow(u,6)*(-1. + 25.*w2 - 50.*v2*(-1. + 27.*w2) + 225.*fcs_pow(v,4)*(-1. + 29.*w2)) + 360525.*fcs_pow(u,8)*(-1. + 27.*w2 - 54.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)) - 144210.*fcs_pow(u,10)*(-3. + 87.*w2 - 174.*v2*(-1. + 31.*w2) + 899.*fcs_pow(v,4)*(-1. + 33.*w2)) + 190095.*fcs_pow(u,12)*(-1. + 31.*w2 - 62.*v2*(-1. + 33.*w2) + 341.*fcs_pow(v,4)*(-1. + 35.*w2)));
        case 3 : return -1.856156927625e12*fcs_pow(t,20)*w*(-1386.*u2*(-3. + 23.*w2 + 575.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) + 11.*(-3. + 21.*w2 - 42.*v2*(-3. + 23.*w2) + 161.*fcs_pow(v,4)*(-3. + 25.*w2)) + 26565.*fcs_pow(u,4)*(-3. + 25.*w2 - 150.*v2*(-1. + 9.*w2) + 225.*fcs_pow(v,4)*(-3. + 29.*w2)) + 512325.*fcs_pow(u,8)*(-3. + 29.*w2 + 899.*fcs_pow(v,4)*(-1. + 11.*w2) - 58.*v2*(-3. + 31.*w2)) - 531300.*fcs_pow(u,6)*(-1. + 9.*w2 + v2*(54. - 522.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)) - 660330.*fcs_pow(u,10)*(-3. + 31.*w2 - 186.*v2*(-1. + 11.*w2) + 341.*fcs_pow(v,4)*(-3. + 35.*w2)) + 310155.*fcs_pow(u,12)*(-3. + 33.*w2 - 66.*v2*(-3. + 35.*w2) + 385.*fcs_pow(v,4)*(-3. + 37.*w2)));
        case 4 : return 5.568470782875e12*fcs_pow(t,21)*(-462.*u2*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 138.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) + 170775.*fcs_pow(u,8)*(3. - 174.*w2 + 899.*fcs_pow(w,4) - 174.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4))) + 11.*(1. - 42.*w2 + 161.*fcs_pow(w,4) + 161.*fcs_pow(v,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 14.*v2*(3. - 138.*w2 + 575.*fcs_pow(w,4))) - 177100.*fcs_pow(u,6)*(1. - 54.*w2 + 261.*fcs_pow(w,4) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 26565.*fcs_pow(u,4)*(1. - 50.*w2 + 225.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 75.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 310155.*fcs_pow(u,12)*(1. - 66.*w2 + 385.*fcs_pow(w,4) + 385.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 22.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 220110.*fcs_pow(u,10)*(3.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 186.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return -3.2564156625e10*fcs_pow(t,18)*v*(109725.*fcs_pow(u,4)*(3. - 46.*v2 + 115.*fcs_pow(v,4)) - 672980.*fcs_pow(u,6)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) - 3762.*u2*(5. - 70.*v2 + 161.*fcs_pow(v,4)) + 1.081575e6*fcs_pow(u,8)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) + 11.*(15. - 190.*v2 + 399.*fcs_pow(v,4)) - 432630.*fcs_pow(u,10)*(15. - 290.*v2 + 899.*fcs_pow(v,4)) + 190095.*fcs_pow(u,12)*(15. - 310.*v2 + 1023.*fcs_pow(v,4)));
        case 1 : return 1.856156927625e12*fcs_pow(t,19)*v*(1.550775e6*fcs_pow(u,12)*(1. - 22.*v2 + 77.*fcs_pow(v,4)) - 2310.*u2*(3. - 46.*v2 + 115.*fcs_pow(v,4)) + 44275.*fcs_pow(u,4)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) + 11.*(5. - 70.*v2 + 161.*fcs_pow(v,4)) - 177100.*fcs_pow(u,6)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) + 170775.*fcs_pow(u,8)*(15. - 290.*v2 + 899.*fcs_pow(v,4)) - 220110.*fcs_pow(u,10)*(15. - 310.*v2 + 1023.*fcs_pow(v,4)))*w;
        case 2 : return -1.856156927625e12*fcs_pow(t,20)*v*(11.*(-5. + 105.*w2 - 70.*v2*(-1. + 23.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) - 2310.*u2*(-3. + 69.*w2 - 46.*v2*(-1. + 25.*w2) + 115.*fcs_pow(v,4)*(-1. + 27.*w2)) + 44275.*fcs_pow(u,4)*(-3. + 75.*w2 - 50.*v2*(-1. + 27.*w2) + 135.*fcs_pow(v,4)*(-1. + 29.*w2)) - 177100.*fcs_pow(u,6)*(-5. + 135.*w2 - 90.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)) + 170775.*fcs_pow(u,8)*(-15. + 435.*w2 - 290.*v2*(-1. + 31.*w2) + 899.*fcs_pow(v,4)*(-1. + 33.*w2)) - 220110.*fcs_pow(u,10)*(-15. + 465.*w2 - 310.*v2*(-1. + 33.*w2) + 1023.*fcs_pow(v,4)*(-1. + 35.*w2)) + 1.550775e6*fcs_pow(u,12)*(-1. + 33.*w2 + v2*(22. - 770.*w2) + 77.*fcs_pow(v,4)*(-1. + 37.*w2)));
        case 3 : return 1.02088631019375e14*fcs_pow(t,21)*v*w*(7.*(-9. + 69.*w2 + 345.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) - 2898.*u2*(-3. + 25.*w2 - 50.*v2*(-1. + 9.*w2) + 45.*fcs_pow(v,4)*(-3. + 29.*w2)) - 28980.*fcs_pow(u,6)*(-45. + 435.*w2 + 2697.*fcs_pow(v,4)*(-1. + 11.*w2) - 290.*v2*(-3. + 31.*w2)) + 36225.*fcs_pow(u,4)*(-5. + 45.*w2 + v2*(90. - 870.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)) + 270135.*fcs_pow(u,8)*(-310.*v2*(-1. + 11.*w2) + 5.*(-3. + 31.*w2) + 341.*fcs_pow(v,4)*(-3. + 35.*w2)) + 310155.*fcs_pow(u,12)*(-9. + 105.*w2 + 777.*fcs_pow(v,4)*(-1. + 13.*w2) - 70.*v2*(-3. + 37.*w2)) - 1.86093e6*fcs_pow(u,10)*(-3. + 33.*w2 + v2*(66. - 770.*w2) + 77.*fcs_pow(v,4)*(-3. + 37.*w2)));
        case 4 : return -3.06265893058125e14*fcs_pow(t,22)*v*(7.*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 46.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 115.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) + 310155.*fcs_pow(u,12)*(3. - 210.*w2 + 1295.*fcs_pow(w,4) - 70.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 259.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4))) - 28980.*fcs_pow(u,6)*(-290.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 12075.*fcs_pow(u,4)*(5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 30.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 966.*u2*(3. - 150.*w2 + 675.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 45.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 620310.*fcs_pow(u,10)*(3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 231.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 22.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 90045.*fcs_pow(u,8)*(15.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 310.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
        case 5 : return 3.522057770168438e16*fcs_pow(t,23)*v*w*(-36540.*fcs_pow(u,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 310.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4))) + 7.*(3. - 50.*w2 + 135.*fcs_pow(w,4) - 10.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 9.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4))) + 1575.*fcs_pow(u,4)*(45. - 870.*w2 + 2697.*fcs_pow(w,4) + 2697.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 58.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 126.*u2*(5.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 30.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 87.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 59334.*fcs_pow(u,10)*(15.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 777.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 70.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 121365.*fcs_pow(u,8)*(15.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 110.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 18879.*fcs_pow(u,12)*(75. - 1850.*w2 + 7215.*fcs_pow(w,4) - 370.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
      }
      case 6 : switch(k){
        case 0 : return 9.7692469875e10*fcs_pow(t,19)*(-6270.*u2*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) + 109725.*fcs_pow(u,4)*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 672980.*fcs_pow(u,6)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 950475.*fcs_pow(u,12)*(-1. + 93.*v2 - 1023.*fcs_pow(v,4) + 2387.*fcs_pow(v,6)) + 11.*(-5. + 285.*v2 - 1995.*fcs_pow(v,4) + 3059.*fcs_pow(v,6)) + 360525.*fcs_pow(u,8)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) - 432630.*fcs_pow(u,10)*(-5. + 435.*v2 - 4495.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)));
        case 1 : return -9.280784638125e12*fcs_pow(t,20)*(11.*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) - 1386.*u2*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 26565.*fcs_pow(u,4)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 660330.*fcs_pow(u,10)*(-1. + 93.*v2 - 1023.*fcs_pow(v,4) + 2387.*fcs_pow(v,6)) + 310155.*fcs_pow(u,12)*(-1. + 99.*v2 - 1155.*fcs_pow(v,4) + 2849.*fcs_pow(v,6)) - 35420.*fcs_pow(u,6)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 102465.*fcs_pow(u,8)*(-5. + 435.*v2 - 4495.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)))*w;
        case 2 : return 9.280784638125e12*fcs_pow(t,21)*(11.*(1. - 21.*w2 + 63.*v2*(-1. + 23.*w2) - 483.*fcs_pow(v,4)*(-1. + 25.*w2) + 805.*fcs_pow(v,6)*(-1. + 27.*w2)) - 1386.*u2*(1. - 23.*w2 + 69.*v2*(-1. + 25.*w2) - 575.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) + 26565.*fcs_pow(u,4)*(1. - 25.*w2 + 75.*v2*(-1. + 27.*w2) - 675.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) - 35420.*fcs_pow(u,6)*(5. - 135.*w2 + 405.*v2*(-1. + 29.*w2) - 3915.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) + 102465.*fcs_pow(u,8)*(5. - 145.*w2 + 435.*v2*(-1. + 31.*w2) - 4495.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)) - 660330.*fcs_pow(u,10)*(1. - 31.*w2 + 93.*v2*(-1. + 33.*w2) - 1023.*fcs_pow(v,4)*(-1. + 35.*w2) + 2387.*fcs_pow(v,6)*(-1. + 37.*w2)) + 310155.*fcs_pow(u,12)*(1. - 33.*w2 + 99.*v2*(-1. + 35.*w2) - 1155.*fcs_pow(v,4)*(-1. + 37.*w2) + 2849.*fcs_pow(v,6)*(-1. + 39.*w2)));
        case 3 : return -3.06265893058125e14*fcs_pow(t,22)*w*(7.*(3. - 23.*w2 - 1725.*fcs_pow(v,4)*(-1. + 9.*w2) + 69.*v2*(-3. + 25.*w2) + 1035.*fcs_pow(v,6)*(-3. + 29.*w2)) + 12075.*fcs_pow(u,4)*(5. - 45.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 135.*v2*(-3. + 29.*w2) - 1305.*fcs_pow(v,4)*(-3. + 31.*w2)) - 966.*u2*(3. - 25.*w2 + 225.*v2*(-1. + 9.*w2) - 675.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)) - 28980.*fcs_pow(u,6)*(15. - 145.*w2 - 13485.*fcs_pow(v,4)*(-1. + 11.*w2) + 435.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) - 1.86093e6*fcs_pow(u,10)*(1. - 11.*w2 + 2849.*fcs_pow(v,6)*(-1. + 13.*w2) + 33.*v2*(-3. + 35.*w2) - 385.*fcs_pow(v,4)*(-3. + 37.*w2)) + 450225.*fcs_pow(u,8)*(3. - 31.*w2 + 279.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 2387.*fcs_pow(v,6)*(-3. + 37.*w2)) + 310155.*fcs_pow(u,12)*(3. - 35.*w2 - 3885.*fcs_pow(v,4)*(-1. + 13.*w2) + 105.*v2*(-3. + 37.*w2) + 3367.*fcs_pow(v,6)*(-3. + 41.*w2)));
        case 4 : return 3.06265893058125e14*fcs_pow(t,23)*(12075.*fcs_pow(u,4)*(-5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 3915.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 135.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 2898.*u2*(-1. + 50.*w2 - 225.*fcs_pow(w,4) + 75.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 225.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 7.*(-3. + 138.*w2 - 575.*fcs_pow(w,4) + 207.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 1725.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1035.*fcs_pow(v,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 1.86093e6*fcs_pow(u,10)*(-1. + 66.*w2 - 385.*fcs_pow(w,4) - 1155.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 2849.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 33.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 1.350675e6*fcs_pow(u,8)*(-1. + 62.*w2 - 341.*fcs_pow(w,4) + 93.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 28980.*fcs_pow(u,6)*(1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 13485.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 310155.*fcs_pow(u,12)*(-3. + 210.*w2 - 1295.*fcs_pow(w,4) + 315.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 3885.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 3367.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -3.522057770168438e16*fcs_pow(t,24)*w*(4725.*fcs_pow(u,4)*(-15. + 290.*w2 - 899.*fcs_pow(w,4) - 13485.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 87.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 630.*u2*(-5. + 90.*w2 - 261.*fcs_pow(w,4) + 8091.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 27.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 261.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 7.*(-3. + 50.*w2 - 135.*fcs_pow(w,4) + 45.*v2*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 135.*fcs_pow(v,4)*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,6)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 364095.*fcs_pow(u,8)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 165.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2849.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 385.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 36540.*fcs_pow(u,6)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 1395.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 5115.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 94395.*fcs_pow(u,12)*(-15. + 370.*w2 - 1443.*fcs_pow(w,4) + 19721.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 333.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 1443.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 178002.*fcs_pow(u,10)*(-5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 3885.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 3367.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return 1.056617331050531e17*fcs_pow(t,25)*(4725.*fcs_pow(u,4)*(5. - 435.*w2 + 4495.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) + 435.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 4495.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6))) - 210.*u2*(5. - 405.*w2 + 3915.*fcs_pow(w,4) - 8091.*fcs_pow(w,6) - 3915.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 81.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 7.*(1. - 75.*w2 + 675.*fcs_pow(w,4) - 1305.*fcs_pow(w,6) + 1305.*fcs_pow(v,6)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 15.*v2*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 135.*fcs_pow(v,4)*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 36540.*fcs_pow(u,6)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 5115.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 2387.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 94395.*fcs_pow(u,12)*(5. - 555.*w2 + 7215.*fcs_pow(w,4) - 19721.*fcs_pow(w,6) - 7215.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 111.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 890010.*fcs_pow(u,10)*(1. - 105.*w2 + 1295.*fcs_pow(w,4) - 3367.*fcs_pow(w,6) + 3367.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 21.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 121365.*fcs_pow(u,8)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 1155.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2849.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
      }
      case 7 : switch(k){
        case 0 : return -1.856156927625e12*fcs_pow(t,20)*v*(1.0855425e7*fcs_pow(u,12)*(-1. + 33.*v2 - 231.*fcs_pow(v,4) + 407.*fcs_pow(v,6)) + 77.*(-5. + 105.*v2 - 483.*fcs_pow(v,4) + 575.*fcs_pow(v,6)) - 6930.*u2*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 132825.*fcs_pow(u,4)*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 4.62231e6*fcs_pow(u,10)*(-5. + 155.*v2 - 1023.*fcs_pow(v,4) + 1705.*fcs_pow(v,6)) - 177100.*fcs_pow(u,6)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 512325.*fcs_pow(u,8)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)));
        case 1 : return 2.143861251406875e15*fcs_pow(t,21)*v*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6) - 620310.*fcs_pow(u,10)*(-1. + 33.*v2 - 231.*fcs_pow(v,4) + 407.*fcs_pow(v,6)) + 310155.*fcs_pow(u,12)*(-1. + 35.*v2 - 259.*fcs_pow(v,4) + 481.*fcs_pow(v,6)) - 138.*u2*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 90045.*fcs_pow(u,8)*(-5. + 155.*v2 - 1023.*fcs_pow(v,4) + 1705.*fcs_pow(v,6)) + 575.*fcs_pow(u,4)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) - 4140.*fcs_pow(u,6)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)))*w;
        case 2 : return -2.143861251406875e15*fcs_pow(t,22)*v*(7. - 161.*w2 + 161.*v2*(-1. + 25.*w2) - 805.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2) - 138.*u2*(7. - 175.*w2 + 175.*v2*(-1. + 27.*w2) - 945.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) + 575.*fcs_pow(u,4)*(35. - 945.*w2 + 945.*v2*(-1. + 29.*w2) - 5481.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) - 4140.*fcs_pow(u,6)*(35. - 1015.*w2 + 1015.*v2*(-1. + 31.*w2) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)) + 90045.*fcs_pow(u,8)*(5. - 155.*w2 + 155.*v2*(-1. + 33.*w2) - 1023.*fcs_pow(v,4)*(-1. + 35.*w2) + 1705.*fcs_pow(v,6)*(-1. + 37.*w2)) - 620310.*fcs_pow(u,10)*(1. - 33.*w2 + 33.*v2*(-1. + 35.*w2) - 231.*fcs_pow(v,4)*(-1. + 37.*w2) + 407.*fcs_pow(v,6)*(-1. + 39.*w2)) + 310155.*fcs_pow(u,12)*(1. - 35.*w2 + 35.*v2*(-1. + 37.*w2) - 259.*fcs_pow(v,4)*(-1. + 39.*w2) + 481.*fcs_pow(v,6)*(-1. + 41.*w2)));
        case 3 : return 4.930880878235812e16*fcs_pow(t,23)*v*w*(21. - 175.*w2 + 525.*v2*(-1. + 9.*w2) - 945.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2) - 90.*u2*(35. - 315.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 1827.*fcs_pow(v,4)*(-3. + 31.*w2)) + 675.*fcs_pow(u,4)*(35.*(3. - 29.*w2) - 18879.*fcs_pow(v,4)*(-1. + 11.*w2) + 1015.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) + 1.820475e6*fcs_pow(u,8)*(1. - 11.*w2 + 407.*fcs_pow(v,6)*(-1. + 13.*w2) + 11.*v2*(-3. + 35.*w2) - 77.*fcs_pow(v,4)*(-3. + 37.*w2)) - 36540.*fcs_pow(u,6)*(15. - 155.*w2 + 465.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 1705.*fcs_pow(v,6)*(-3. + 37.*w2)) - 890010.*fcs_pow(u,10)*(3. - 35.*w2 - 777.*fcs_pow(v,4)*(-1. + 13.*w2) + 35.*v2*(-3. + 37.*w2) + 481.*fcs_pow(v,6)*(-3. + 41.*w2)) + 13485.*fcs_pow(u,12)*(35.*(3. - 37.*w2) + 3885.*v2*(-1. + 13.*w2) - 10101.*fcs_pow(v,4)*(-3. + 41.*w2) + 19721.*fcs_pow(v,6)*(-3. + 43.*w2)));
        case 4 : return -1.479264263470744e17*fcs_pow(t,24)*v*(-7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 175.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 315.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 30.*u2*(-35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 5481.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 606825.*fcs_pow(u,8)*(-1. + 66.*w2 - 385.*fcs_pow(w,4) - 231.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 11.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 36540.*fcs_pow(u,6)*(-5.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 155.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 1705.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 225.*fcs_pow(u,4)*(3045.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18879.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 35.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 13485.*fcs_pow(u,12)*(-35.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 1295.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 3367.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 296670.*fcs_pow(u,10)*(-3. + 210.*w2 - 1295.*fcs_pow(w,4) + 105.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 777.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return 7.396321317353719e17*fcs_pow(t,25)*v*w*(40455.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 35.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 315.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 1827.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) - 270.*u2*(-18879.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 203.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 1.13274e6*fcs_pow(u,6)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 55.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 45675.*fcs_pow(u,4)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 465.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 1023.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 341.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 498945.*fcs_pow(u,12)*(22919.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 3731.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 91.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 59334.*fcs_pow(u,10)*(98605.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3885.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 35.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 10101.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 1.335015e6*fcs_pow(u,8)*(-5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 777.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 35.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 481.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return -3.698160658676859e18*fcs_pow(t,26)*v*(-5481.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) + 189.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) - 162.*u2*(1015.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 6293.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 7.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 27405.*fcs_pow(u,4)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 155.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 1023.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 99789.*fcs_pow(u,12)*(1365.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 11193.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 22919.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 178002.*fcs_pow(u,10)*(-10101.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 259.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 801009.*fcs_pow(u,8)*(5. - 525.*w2 + 6475.*fcs_pow(w,4) - 16835.*fcs_pow(w,6) + 2405.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 35.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 226548.*fcs_pow(u,6)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
        case 7 : return 3.328344592809173e19*fcs_pow(t,27)*v*w*(21.*(35. - 1015.*w2 + 6293.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 203.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6))) + 3.026933e6*fcs_pow(u,12)*(35. - 1435.*w2 + 12341.*fcs_pow(w,4) - 26445.*fcs_pow(w,6) - 12341.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 5289.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 205.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6))) - 3654.*u2*(7.*(5. - 155.*w2 + 1023.*fcs_pow(w,4) - 1705.*fcs_pow(w,6)) + 1085.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 7161.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 94395.*fcs_pow(u,4)*(-35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 1155.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 5.122502e6*fcs_pow(u,10)*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 22919.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1599.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 39.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 89001.*fcs_pow(u,8)*(98605.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 35.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 10101.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 830676.*fcs_pow(u,6)*(-35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 35.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 481.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
      }
      case 8 : switch(k){
        case 0 : return 6.4965492466875e13*fcs_pow(t,21)*(11.*(1. - 84.*v2 + 966.*fcs_pow(v,4) - 3220.*fcs_pow(v,6) + 3105.*fcs_pow(v,8)) - 660330.*fcs_pow(u,10)*(1. - 124.*v2 + 2046.*fcs_pow(v,4) - 9548.*fcs_pow(v,6) + 12617.*fcs_pow(v,8)) + 310155.*fcs_pow(u,12)*(1. - 132.*v2 + 2310.*fcs_pow(v,4) - 11396.*fcs_pow(v,6) + 15873.*fcs_pow(v,8)) - 198.*u2*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) + 3795.*fcs_pow(u,4)*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) + 102465.*fcs_pow(u,8)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) - 5060.*fcs_pow(u,6)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)));
        case 1 : return -2.143861251406875e15*fcs_pow(t,22)*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8) + 450225.*fcs_pow(u,8)*(1. - 124.*v2 + 2046.*fcs_pow(v,4) - 9548.*fcs_pow(v,6) + 12617.*fcs_pow(v,8)) - 620310.*fcs_pow(u,10)*(1. - 132.*v2 + 2310.*fcs_pow(v,4) - 11396.*fcs_pow(v,6) + 15873.*fcs_pow(v,8)) + 310155.*fcs_pow(u,12)*(1. - 140.*v2 + 2590.*fcs_pow(v,4) - 13468.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) - 138.*u2*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) - 28980.*fcs_pow(u,6)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) + 575.*fcs_pow(u,4)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)))*w;
        case 2 : return 2.143861251406875e15*fcs_pow(t,23)*(-7. + 161.*w2 - 644.*v2*(-1. + 25.*w2) + 8050.*fcs_pow(v,4)*(-1. + 27.*w2) - 28980.*fcs_pow(v,6)*(-1. + 29.*w2) + 30015.*fcs_pow(v,8)*(-1. + 31.*w2) - 138.*u2*(-7. + 175.*w2 - 700.*v2*(-1. + 27.*w2) + 9450.*fcs_pow(v,4)*(-1. + 29.*w2) - 36540.*fcs_pow(v,6)*(-1. + 31.*w2) + 40455.*fcs_pow(v,8)*(-1. + 33.*w2)) + 575.*fcs_pow(u,4)*(-35. + 945.*w2 - 3780.*v2*(-1. + 29.*w2) + 54810.*fcs_pow(v,4)*(-1. + 31.*w2) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2)) - 28980.*fcs_pow(u,6)*(-5. + 145.*w2 - 580.*v2*(-1. + 31.*w2) + 8990.*fcs_pow(v,4)*(-1. + 33.*w2) - 39556.*fcs_pow(v,6)*(-1. + 35.*w2) + 49445.*fcs_pow(v,8)*(-1. + 37.*w2)) + 450225.*fcs_pow(u,8)*(-1. + 31.*w2 - 124.*v2*(-1. + 33.*w2) + 2046.*fcs_pow(v,4)*(-1. + 35.*w2) - 9548.*fcs_pow(v,6)*(-1. + 37.*w2) + 12617.*fcs_pow(v,8)*(-1. + 39.*w2)) - 620310.*fcs_pow(u,10)*(-1. + 33.*w2 - 132.*v2*(-1. + 35.*w2) + 2310.*fcs_pow(v,4)*(-1. + 37.*w2) - 11396.*fcs_pow(v,6)*(-1. + 39.*w2) + 15873.*fcs_pow(v,8)*(-1. + 41.*w2)) + 310155.*fcs_pow(u,12)*(-1. + 35.*w2 - 140.*v2*(-1. + 37.*w2) + 2590.*fcs_pow(v,4)*(-1. + 39.*w2) - 13468.*fcs_pow(v,6)*(-1. + 41.*w2) + 19721.*fcs_pow(v,8)*(-1. + 43.*w2)));
        case 3 : return -4.930880878235812e16*fcs_pow(t,24)*w*(-2100.*v2*(-1. + 9.*w2) + 121365.*fcs_pow(v,8)*(-1. + 11.*w2) + 7.*(-3. + 25.*w2) + 9450.*fcs_pow(v,4)*(-3. + 29.*w2) - 36540.*fcs_pow(v,6)*(-3. + 31.*w2) - 90.*u2*(35.*(-1. + 9.*w2) - 226548.*fcs_pow(v,6)*(-1. + 11.*w2) - 1260.*v2*(-3. + 29.*w2) + 18270.*fcs_pow(v,4)*(-3. + 31.*w2) + 89001.*fcs_pow(v,8)*(-3. + 35.*w2)) - 182700.*fcs_pow(u,6)*(-3. + 31.*w2 - 372.*v2*(-1. + 11.*w2) + 37851.*fcs_pow(v,8)*(-1. + 13.*w2) + 2046.*fcs_pow(v,4)*(-3. + 35.*w2) - 9548.*fcs_pow(v,6)*(-3. + 37.*w2)) + 4725.*fcs_pow(u,4)*(26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 5.*(-3. + 29.*w2) - 580.*v2*(-3. + 31.*w2) - 39556.*fcs_pow(v,6)*(-3. + 35.*w2) + 49445.*fcs_pow(v,8)*(-3. + 37.*w2)) + 1.820475e6*fcs_pow(u,8)*(-1. + 11.*w2 - 11396.*fcs_pow(v,6)*(-1. + 13.*w2) - 44.*v2*(-3. + 35.*w2) + 770.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)) + 13485.*fcs_pow(u,12)*(-15540.*v2*(-1. + 13.*w2) + 2.544009e6*fcs_pow(v,8)*(-1. + 15.*w2) + 35.*(-3. + 37.*w2) + 101010.*fcs_pow(v,4)*(-3. + 41.*w2) - 552188.*fcs_pow(v,6)*(-3. + 43.*w2)) - 890010.*fcs_pow(u,10)*(-3. + 35.*w2 + 7770.*fcs_pow(v,4)*(-1. + 13.*w2) - 140.*v2*(-3. + 37.*w2) - 13468.*fcs_pow(v,6)*(-3. + 41.*w2) + 19721.*fcs_pow(v,8)*(-3. + 43.*w2)));
        case 4 : return 1.479264263470744e17*fcs_pow(t,25)*(7.*(1. - 50.*w2 + 225.*fcs_pow(w,4)) - 700.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 36540.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 40455.*fcs_pow(v,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 3150.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 182700.*fcs_pow(u,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4) - 124.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 682.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 1575.*fcs_pow(u,4)*(-1740.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 148335.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 30.*u2*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 54810.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 226548.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 1260.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 13485.*fcs_pow(u,12)*(35.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 5180.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 552188.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 33670.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 296670.*fcs_pow(u,10)*(3. - 210.*w2 + 1295.*fcs_pow(w,4) - 420.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 7770.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59163.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 13468.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 606825.*fcs_pow(u,8)*(1. - 66.*w2 + 385.*fcs_pow(w,4) + 2310.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 44.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -3.698160658676859e18*fcs_pow(t,26)*w*(-226548.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 7.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 252.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 3654.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 9135.*fcs_pow(u,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 10230.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 37851.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 378.*u2*(15. - 290.*w2 + 899.*fcs_pow(w,4) + 26970.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 116.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 59334.*fcs_pow(u,10)*(2.544009e6*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 552188.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 3108.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 7.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 20202.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 267003.*fcs_pow(u,8)*(5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 140.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 13468.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 226548.*fcs_pow(u,6)*(5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 220.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 770.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 99789.*fcs_pow(u,12)*(-641732.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 37310.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 364.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 68757.*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))));
        case 6 : return 3.698160658676859e18*fcs_pow(t,27)*(54810.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 226548.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 267003.*fcs_pow(v,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 7.*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6)) - 756.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) - 1134.*u2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) - 580.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8990.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 39556.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 178002.*fcs_pow(u,10)*(101010.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 552188.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 1036.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 4.005045e6*fcs_pow(u,8)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6) - 13468.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 28.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 518.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 226548.*fcs_pow(u,6)*(5.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 79365.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 27405.*fcs_pow(u,4)*(5.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 620.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 99789.*fcs_pow(u,12)*(-5460.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 111930.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 641732.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 206271.*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))));
        case 7 : return -2.329841214966421e20*fcs_pow(t,28)*w*(432419.*fcs_pow(u,12)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6) + 123410.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 148092.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 820.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 248583.*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6))) + 3.*(-35. + 1015.*w2 - 6293.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) + 62930.*fcs_pow(v,4)*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 276892.*fcs_pow(v,6)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 812.*v2*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 522.*u2*(-4340.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 71610.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 7.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 731786.*fcs_pow(u,10)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6) - 641732.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 206271.*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 15990.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 156.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 445005.*fcs_pow(u,8)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 78884.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 148.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 2886.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 118668.*fcs_pow(u,6)*(35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 140.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2590.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 13468.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 13485.*fcs_pow(u,4)*(35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 15873.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
        case 8 : return 2.329841214966421e20*fcs_pow(t,29)*(-731786.*fcs_pow(u,10)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8) + 111930.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 641732.*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 780.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 206271.*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8))) + 432419.*fcs_pow(u,12)*(-5740.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 123410.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 5.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 148092.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 248583.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) + 3.*(-4060.*v2*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 62930.*fcs_pow(v,4)*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 276892.*fcs_pow(v,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 7.*(5. - 580.*w2 + 8990.*fcs_pow(w,4) - 39556.*fcs_pow(w,6) + 49445.*fcs_pow(w,8)) + 9889.*fcs_pow(v,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 445005.*fcs_pow(u,8)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8) - 552188.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 14430.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 148.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 4.15338e6*fcs_pow(u,6)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8) + 19721.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 1924.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 4.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 74.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 13485.*fcs_pow(u,4)*(35.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 79365.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 11396.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 522.*u2*(35.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 4340.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 9548.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
      }
      case 9 : switch(k){
        case 0 : return -6.431583754220625e15*fcs_pow(t,22)*v*(21. - 644.*v2 + 4830.*fcs_pow(v,4) - 12420.*fcs_pow(v,6) + 10005.*fcs_pow(v,8) - 620310.*fcs_pow(u,10)*(3. - 132.*v2 + 1386.*fcs_pow(v,4) - 4884.*fcs_pow(v,6) + 5291.*fcs_pow(v,8)) - 138.*u2*(21. - 700.*v2 + 5670.*fcs_pow(v,4) - 15660.*fcs_pow(v,6) + 13485.*fcs_pow(v,8)) + 103385.*fcs_pow(u,12)*(9. - 420.*v2 + 4662.*fcs_pow(v,4) - 17316.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) + 1725.*fcs_pow(u,4)*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) + 30015.*fcs_pow(u,8)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) - 1380.*fcs_pow(u,6)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)));
        case 1 : return 1.479264263470744e17*fcs_pow(t,23)*v*(21. - 700.*v2 + 5670.*fcs_pow(v,4) - 15660.*fcs_pow(v,6) + 13485.*fcs_pow(v,8) + 606825.*fcs_pow(u,8)*(3. - 132.*v2 + 1386.*fcs_pow(v,4) - 4884.*fcs_pow(v,6) + 5291.*fcs_pow(v,8)) - 296670.*fcs_pow(u,10)*(9. - 420.*v2 + 4662.*fcs_pow(v,4) - 17316.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) - 90.*u2*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) - 12180.*fcs_pow(u,6)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) + 225.*fcs_pow(u,4)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)) + 4495.*fcs_pow(u,12)*(315. - 15540.*v2 + 181818.*fcs_pow(v,4) - 709956.*fcs_pow(v,6) + 848003.*fcs_pow(v,8)))*w;
        case 2 : return -1.479264263470744e17*fcs_pow(t,24)*v*(-21. + 525.*w2 - 700.*v2*(-1. + 27.*w2) + 5670.*fcs_pow(v,4)*(-1. + 29.*w2) - 15660.*fcs_pow(v,6)*(-1. + 31.*w2) + 13485.*fcs_pow(v,8)*(-1. + 33.*w2) - 90.*u2*(-35. + 945.*w2 - 1260.*v2*(-1. + 29.*w2) + 10962.*fcs_pow(v,4)*(-1. + 31.*w2) - 32364.*fcs_pow(v,6)*(-1. + 33.*w2) + 29667.*fcs_pow(v,8)*(-1. + 35.*w2)) + 225.*fcs_pow(u,4)*(315.*(-1. + 29.*w2) - 12180.*v2*(-1. + 31.*w2) + 113274.*fcs_pow(v,4)*(-1. + 33.*w2) - 356004.*fcs_pow(v,6)*(-1. + 35.*w2) + 346115.*fcs_pow(v,8)*(-1. + 37.*w2)) - 12180.*fcs_pow(u,6)*(45.*(-1. + 31.*w2) - 1860.*v2*(-1. + 33.*w2) + 18414.*fcs_pow(v,4)*(-1. + 35.*w2) - 61380.*fcs_pow(v,6)*(-1. + 37.*w2) + 63085.*fcs_pow(v,8)*(-1. + 39.*w2)) + 606825.*fcs_pow(u,8)*(-3. + 99.*w2 - 132.*v2*(-1. + 35.*w2) + 1386.*fcs_pow(v,4)*(-1. + 37.*w2) - 4884.*fcs_pow(v,6)*(-1. + 39.*w2) + 5291.*fcs_pow(v,8)*(-1. + 41.*w2)) - 296670.*fcs_pow(u,10)*(-9. + 315.*w2 - 420.*v2*(-1. + 37.*w2) + 4662.*fcs_pow(v,4)*(-1. + 39.*w2) - 17316.*fcs_pow(v,6)*(-1. + 41.*w2) + 19721.*fcs_pow(v,8)*(-1. + 43.*w2)) + 4495.*fcs_pow(u,12)*(315.*(-1. + 37.*w2) - 15540.*v2*(-1. + 39.*w2) + 181818.*fcs_pow(v,4)*(-1. + 41.*w2) - 709956.*fcs_pow(v,6)*(-1. + 43.*w2) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2)));
        case 3 : return 6.656689185618347e18*fcs_pow(t,25)*v*w*(35.*(-1. + 9.*w2) - 32364.*fcs_pow(v,6)*(-1. + 11.*w2) - 420.*v2*(-3. + 29.*w2) + 3654.*fcs_pow(v,4)*(-3. + 31.*w2) + 9889.*fcs_pow(v,8)*(-3. + 35.*w2) + 3045.*fcs_pow(u,4)*(-45. + 465.*w2 - 1860.*v2*(-1. + 11.*w2) + 63085.*fcs_pow(v,8)*(-1. + 13.*w2) + 6138.*fcs_pow(v,4)*(-3. + 35.*w2) - 20460.*fcs_pow(v,6)*(-3. + 37.*w2)) - 6.*u2*(339822.*fcs_pow(v,4)*(-1. + 11.*w2) + 315.*(-3. + 29.*w2) - 12180.*v2*(-3. + 31.*w2) - 356004.*fcs_pow(v,6)*(-3. + 35.*w2) + 346115.*fcs_pow(v,8)*(-3. + 37.*w2)) - 125860.*fcs_pow(u,6)*(-9. + 99.*w2 - 14652.*fcs_pow(v,6)*(-1. + 13.*w2) - 132.*v2*(-3. + 35.*w2) + 1386.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)) - 19778.*fcs_pow(u,10)*(-15540.*v2*(-1. + 13.*w2) + 848003.*fcs_pow(v,8)*(-1. + 15.*w2) + 105.*(-3. + 37.*w2) + 60606.*fcs_pow(v,4)*(-3. + 41.*w2) - 236652.*fcs_pow(v,6)*(-3. + 43.*w2)) + 148335.*fcs_pow(u,8)*(13986.*fcs_pow(v,4)*(-1. + 13.*w2) + 9.*(-3. + 35.*w2) - 420.*v2*(-3. + 37.*w2) - 17316.*fcs_pow(v,6)*(-3. + 41.*w2) + 19721.*fcs_pow(v,8)*(-3. + 43.*w2)) + 33263.*fcs_pow(u,12)*(105.*(-1. + 13.*w2) - 275028.*fcs_pow(v,6)*(-1. + 15.*w2) - 1820.*v2*(-3. + 41.*w2) + 22386.*fcs_pow(v,4)*(-3. + 43.*w2) + 114595.*fcs_pow(v,8)*(-3. + 47.*w2)));
        case 4 : return -6.656689185618347e18*fcs_pow(t,26)*v*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 10962.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 32364.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 420.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 3045.*fcs_pow(u,4)*(45.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 1860.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 61380.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 18.*u2*(-12180.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 113274.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 346115.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 105.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 118668.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 19778.*fcs_pow(u,10)*(315.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 15540.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 709956.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 60606.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 445005.*fcs_pow(u,8)*(9. - 630.*w2 + 3885.*fcs_pow(w,4) - 420.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 4662.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 5772.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 125860.*fcs_pow(u,6)*(9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 4158.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 132.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 33263.*fcs_pow(u,12)*(105.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 67158.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 275028.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 1820.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 114595.*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return 3.328344592809173e19*fcs_pow(t,27)*v*w*(339822.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 356004.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 63.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 2436.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 69223.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 3654.*u2*(45. - 930.*w2 + 3069.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 4092.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 89001.*fcs_pow(u,8)*(4.240015e6*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 1.18326e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 15540.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 60606.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 276892.*fcs_pow(u,6)*(45.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 13986.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 420.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 17316.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 94395.*fcs_pow(u,4)*(45.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 660.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1386.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 2.195358e6*fcs_pow(u,10)*(-275028.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 22386.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 21.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 364.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 22919.*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 432419.*fcs_pow(u,12)*(222138.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 5740.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 21.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 63468.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 82861.*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return -9.98503377842752e19*fcs_pow(t,28)*v*(-12180.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 113274.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 356004.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 63.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 69223.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 148335.*fcs_pow(u,8)*(181818.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 709956.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 63.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 3108.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 276892.*fcs_pow(u,6)*(45.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 86580.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 94395.*fcs_pow(u,4)*(15.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 1218.*u2*(45.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 1860.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 18414.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 12276.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 2.195358e6*fcs_pow(u,10)*(-1820.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 22386.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 91676.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 22919.*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 432419.*fcs_pow(u,12)*(105.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 5740.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 74046.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 63468.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 82861.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return 2.026961857020787e22*fcs_pow(t,29)*v*w*(-13020.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 128898.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 63.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 12276.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 984126.*fcs_pow(u,10)*(-105. + 4305.*w2 - 37023.*fcs_pow(w,4) + 79335.*fcs_pow(w,6) + 74046.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 63468.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 820.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 82861.*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6))) + 1.703295e6*fcs_pow(u,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6) - 91676.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 22919.*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 3198.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 52.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 1364.*fcs_pow(u,6)*(2.9680105e7*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 3.54978e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 15540.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 181818.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 5115.*fcs_pow(u,4)*(315.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 17316.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 558.*u2*(105.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 5291.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 611351.*fcs_pow(u,12)*(-6020.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 16254.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 15.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 72756.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 14147.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return -2.026961857020787e22*fcs_pow(t,30)*v*(315.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 13020.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 128898.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 12276.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) + 1.703295e6*fcs_pow(u,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8) + 22386.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 91676.*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 260.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8))) - 984126.*fcs_pow(u,10)*(-5740.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 74046.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 15.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 63468.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 82861.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) + 611351.*fcs_pow(u,12)*(105.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 6020.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 16254.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 72756.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 70735.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) - 47740.*fcs_pow(u,6)*(-709956.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 25974.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 9.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 444.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 5115.*fcs_pow(u,4)*(315.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 690235.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 86580.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 420.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 4662.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 558.*u2*(105.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 26455.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 1386.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 4884.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return 1.885074527029332e24*fcs_pow(t,31)*v*w*(238095.*fcs_pow(u,8)*(666414.*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 2460.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 63468.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 82861.*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 9.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 170940.*fcs_pow(u,6)*(-825084.*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 9594.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 156.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 495.*fcs_pow(u,4)*(2.9680105e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 1.18326e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 15540.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 60606.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 198.*u2*(105.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 98605.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 13986.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 17316.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 140.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 3.*(105.*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) - 1540.*v2*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 14652.*fcs_pow(v,6)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5291.*fcs_pow(v,8)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 462.*fcs_pow(v,4)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 5.936021e6*fcs_pow(u,12)*(45.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 60.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 846.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 3995.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1692.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1.301586e6*fcs_pow(u,10)*(-18060.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 15.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 5418.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 24252.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 14147.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
      }
      case 10 : switch(k){
        case 0 : return 6.431583754220625e15*fcs_pow(t,23)*(-21. + 2415.*v2 - 40250.*fcs_pow(v,4) + 217350.*fcs_pow(v,6) - 450225.*fcs_pow(v,8) + 310155.*fcs_pow(v,10) - 414.*u2*(-7. + 875.*v2 - 15750.*fcs_pow(v,4) + 91350.*fcs_pow(v,6) - 202275.*fcs_pow(v,8) + 148335.*fcs_pow(v,10)) + 450225.*fcs_pow(u,8)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) + 8625.*fcs_pow(u,4)*(-7. + 945.*v2 - 18270.*fcs_pow(v,4) + 113274.*fcs_pow(v,6) - 267003.*fcs_pow(v,8) + 207669.*fcs_pow(v,10)) - 620310.*fcs_pow(u,10)*(-3. + 495.*v2 - 11550.*fcs_pow(v,4) + 85470.*fcs_pow(v,6) - 238095.*fcs_pow(v,8) + 216931.*fcs_pow(v,10)) - 48300.*fcs_pow(u,6)*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)) + 103385.*fcs_pow(u,12)*(-9. + 1575.*v2 - 38850.*fcs_pow(v,4) + 303030.*fcs_pow(v,6) - 887445.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)));
        case 1 : return -4.437792790412231e17*fcs_pow(t,24)*(-7. + 875.*v2 - 15750.*fcs_pow(v,4) + 91350.*fcs_pow(v,6) - 202275.*fcs_pow(v,8) + 148335.*fcs_pow(v,10) - 60900.*fcs_pow(u,6)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) - 150.*u2*(-7. + 945.*v2 - 18270.*fcs_pow(v,4) + 113274.*fcs_pow(v,6) - 267003.*fcs_pow(v,8) + 207669.*fcs_pow(v,10)) + 202275.*fcs_pow(u,8)*(-3. + 495.*v2 - 11550.*fcs_pow(v,4) + 85470.*fcs_pow(v,6) - 238095.*fcs_pow(v,8) + 216931.*fcs_pow(v,10)) + 2625.*fcs_pow(u,4)*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)) - 98890.*fcs_pow(u,10)*(-9. + 1575.*v2 - 38850.*fcs_pow(v,4) + 303030.*fcs_pow(v,6) - 887445.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 67425.*fcs_pow(u,12)*(-7. + 1295.*v2 - 33670.*fcs_pow(v,4) + 276094.*fcs_pow(v,6) - 848003.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)))*w;
        case 2 : return 4.437792790412231e17*fcs_pow(t,25)*(7. - 175.*w2 + 875.*v2*(-1. + 27.*w2) - 15750.*fcs_pow(v,4)*(-1. + 29.*w2) + 91350.*fcs_pow(v,6)*(-1. + 31.*w2) - 202275.*fcs_pow(v,8)*(-1. + 33.*w2) + 148335.*fcs_pow(v,10)*(-1. + 35.*w2) - 150.*u2*(7. - 189.*w2 + 945.*v2*(-1. + 29.*w2) - 18270.*fcs_pow(v,4)*(-1. + 31.*w2) + 113274.*fcs_pow(v,6)*(-1. + 33.*w2) - 267003.*fcs_pow(v,8)*(-1. + 35.*w2) + 207669.*fcs_pow(v,10)*(-1. + 37.*w2)) + 2625.*fcs_pow(u,4)*(9. - 261.*w2 + 1305.*v2*(-1. + 31.*w2) - 26970.*fcs_pow(v,4)*(-1. + 33.*w2) + 178002.*fcs_pow(v,6)*(-1. + 35.*w2) - 445005.*fcs_pow(v,8)*(-1. + 37.*w2) + 365893.*fcs_pow(v,10)*(-1. + 39.*w2)) - 60900.*fcs_pow(u,6)*(3. - 93.*w2 + 465.*v2*(-1. + 33.*w2) - 10230.*fcs_pow(v,4)*(-1. + 35.*w2) + 71610.*fcs_pow(v,6)*(-1. + 37.*w2) - 189255.*fcs_pow(v,8)*(-1. + 39.*w2) + 164021.*fcs_pow(v,10)*(-1. + 41.*w2)) + 202275.*fcs_pow(u,8)*(3. - 99.*w2 + 495.*v2*(-1. + 35.*w2) - 11550.*fcs_pow(v,4)*(-1. + 37.*w2) + 85470.*fcs_pow(v,6)*(-1. + 39.*w2) - 238095.*fcs_pow(v,8)*(-1. + 41.*w2) + 216931.*fcs_pow(v,10)*(-1. + 43.*w2)) - 98890.*fcs_pow(u,10)*(9. - 315.*w2 + 1575.*v2*(-1. + 37.*w2) - 38850.*fcs_pow(v,4)*(-1. + 39.*w2) + 303030.*fcs_pow(v,6)*(-1. + 41.*w2) - 887445.*fcs_pow(v,8)*(-1. + 43.*w2) + 848003.*fcs_pow(v,10)*(-1. + 45.*w2)) + 67425.*fcs_pow(u,12)*(7. - 259.*w2 + 1295.*v2*(-1. + 39.*w2) - 33670.*fcs_pow(v,4)*(-1. + 41.*w2) + 276094.*fcs_pow(v,6)*(-1. + 43.*w2) - 848003.*fcs_pow(v,8)*(-1. + 45.*w2) + 848003.*fcs_pow(v,10)*(-1. + 47.*w2)));
        case 3 : return -3.328344592809173e19*fcs_pow(t,26)*w*(7. - 63.*w2 + 113274.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 6090.*fcs_pow(v,4)*(-3. + 31.*w2) - 89001.*fcs_pow(v,8)*(-3. + 35.*w2) + 69223.*fcs_pow(v,10)*(-3. + 37.*w2) - 126.*u2*(9. - 87.*w2 - 26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 365893.*fcs_pow(v,10)*(-1. + 13.*w2) + 435.*v2*(-3. + 31.*w2) + 59334.*fcs_pow(v,6)*(-3. + 35.*w2) - 148335.*fcs_pow(v,8)*(-3. + 37.*w2)) + 3045.*fcs_pow(u,4)*(9. - 93.*w2 + 1395.*v2*(-1. + 11.*w2) - 567765.*fcs_pow(v,8)*(-1. + 13.*w2) - 10230.*fcs_pow(v,4)*(-3. + 35.*w2) + 71610.*fcs_pow(v,6)*(-3. + 37.*w2) + 164021.*fcs_pow(v,10)*(-3. + 41.*w2)) + 89001.*fcs_pow(u,8)*(9. - 105.*w2 - 38850.*fcs_pow(v,4)*(-1. + 13.*w2) + 848003.*fcs_pow(v,10)*(-1. + 15.*w2) + 525.*v2*(-3. + 37.*w2) + 101010.*fcs_pow(v,6)*(-3. + 41.*w2) - 295815.*fcs_pow(v,8)*(-3. + 43.*w2)) - 25172.*fcs_pow(u,6)*(9. - 99.*w2 + 256410.*fcs_pow(v,6)*(-1. + 13.*w2) + 495.*v2*(-3. + 35.*w2) - 11550.*fcs_pow(v,4)*(-3. + 37.*w2) - 238095.*fcs_pow(v,8)*(-3. + 41.*w2) + 216931.*fcs_pow(v,10)*(-3. + 43.*w2)) - 59334.*fcs_pow(u,10)*(21. - 259.*w2 + 3885.*v2*(-1. + 13.*w2) - 2.544009e6*fcs_pow(v,8)*(-1. + 15.*w2) - 33670.*fcs_pow(v,4)*(-3. + 41.*w2) + 276094.*fcs_pow(v,6)*(-3. + 43.*w2) + 848003.*fcs_pow(v,10)*(-3. + 47.*w2)) + 33263.*fcs_pow(u,12)*(21. - 273.*w2 + 962598.*fcs_pow(v,6)*(-1. + 15.*w2) + 1365.*v2*(-3. + 41.*w2) - 37310.*fcs_pow(v,4)*(-3. + 43.*w2) - 1.031355e6*fcs_pow(v,8)*(-3. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-3. + 49.*w2)));
        case 4 : return 3.328344592809173e19*fcs_pow(t,27)*(-7.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 18270.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 113274.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 207669.*fcs_pow(v,10)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 126.*u2*(-9. + 522.*w2 - 2697.*fcs_pow(w,4) + 1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 445005.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 365893.*fcs_pow(v,10)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59334.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 89001.*fcs_pow(u,8)*(-9. + 630.*w2 - 3885.*fcs_pow(w,4) + 1575.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 38850.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 887445.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 101010.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 75516.*fcs_pow(u,6)*(-3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 11550.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 85470.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 165.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 79365.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 3045.*fcs_pow(u,4)*(-9.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 1395.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 214830.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 567765.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 10230.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 164021.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 99789.*fcs_pow(u,12)*(-7.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 37310.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 320866.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 455.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 343785.*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 59334.*fcs_pow(u,10)*(-21.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 3885.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 828282.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 2.544009e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 33670.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return -9.98503377842752e19*fcs_pow(t,28)*w*(7.*(-45. + 870.*w2 - 2697.*fcs_pow(w,4) - 134850.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 296670.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 365893.*fcs_pow(v,10)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 435.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) - 148335.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 1.38446e6*fcs_pow(u,6)*(-9. + 210.*w2 - 777.*fcs_pow(w,4) + 848003.*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 295815.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 20202.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 157325.*fcs_pow(u,4)*(-9.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 495.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 51282.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 2310.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 47619.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 1218.*u2*(-45. + 930.*w2 - 3069.*fcs_pow(w,4) + 6975.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 51150.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 567765.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 71610.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 164021.*fcs_pow(v,10)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 445005.*fcs_pow(u,8)*(-1.2720045e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1.38047e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3885.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 7.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 33670.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 432419.*fcs_pow(u,12)*(-617050.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 7175.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 7.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 370230.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 1.242915e6*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 731786.*fcs_pow(u,10)*(4.81299e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 186550.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 21.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1365.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.031355e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return 9.98503377842752e19*fcs_pow(t,29)*(-1.38446e6*fcs_pow(u,6)*(9. - 945.*w2 + 11655.*fcs_pow(w,4) - 30303.*fcs_pow(w,6) + 303030.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 315.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 471975.*fcs_pow(u,4)*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 238095.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 216931.*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 18270.*u2*(3. - 279.*w2 + 3069.*fcs_pow(w,4) - 7161.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 164021.*fcs_pow(v,10)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 14322.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 37851.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 7.*(6525.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 134850.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 890010.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) - 445005.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 365893.*fcs_pow(v,10)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 1.335015e6*fcs_pow(u,8)*(-168350.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.38047e6*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.240015e6*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 2.162095e6*fcs_pow(u,12)*(-21.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 4305.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 123410.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 4.060189e6*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 222138.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 745749.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 2.195358e6*fcs_pow(u,10)*(6825.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 186550.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.60433e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.031355e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return -6.08088557106236e22*fcs_pow(t,30)*w*(16275.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 358050.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 21.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) + 71610.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 189255.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 164021.*fcs_pow(v,10)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 567765.*fcs_pow(u,8)*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 1.60433e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.031355e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 26650.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 195.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 143220.*fcs_pow(u,6)*(35. - 1295.*w2 + 10101.*fcs_pow(w,4) - 19721.*fcs_pow(w,6) - 4.240015e6*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 197210.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 185.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 4810.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 8525.*fcs_pow(u,4)*(-63.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 5.936021e6*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 60606.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 930.*u2*(21.*(1. - 33.*w2 + 231.*fcs_pow(w,4) - 407.*fcs_pow(w,6)) + 3465.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 216931.*fcs_pow(v,10)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 47619.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 3.056755e6*fcs_pow(u,12)*(7. - 301.*w2 + 2709.*fcs_pow(w,4) - 6063.*fcs_pow(w,6) + 1505.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 9030.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 84882.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 240499.*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 42441.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 328042.*fcs_pow(u,10)*(105. - 4305.*w2 + 37023.*fcs_pow(w,4) - 79335.*fcs_pow(w,6) - 617050.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.11069e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 3075.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 3.728745e6*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 580027.*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return 3.04044278553118e23*fcs_pow(t,31)*(-21.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) + 3255.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 164021.*fcs_pow(v,10)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 14322.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 37851.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) + 113553.*fcs_pow(u,8)*(-35. + 5460.*w2 - 111930.*fcs_pow(w,4) + 641732.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) - 186550.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.60433e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 975.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) - 328042.*fcs_pow(u,10)*(-21. + 3444.*w2 - 74046.*fcs_pow(w,4) + 444276.*fcs_pow(w,6) - 745749.*fcs_pow(w,8) + 4305.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 123410.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 222138.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 745749.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) + 4.279457e6*fcs_pow(u,12)*(-1. + 172.*w2 - 3870.*fcs_pow(w,4) + 24252.*fcs_pow(w,6) - 42441.*fcs_pow(w,8) + 215.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 1290.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 12126.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 30315.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 34357.*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) - 28644.*fcs_pow(u,6)*(-35. + 5180.*w2 - 101010.*fcs_pow(w,4) + 552188.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 1.38047e6*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 4.240015e6*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 24050.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 185.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 11935.*fcs_pow(u,4)*(-9.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 887445.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 43290.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 45.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 1110.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 186.*u2*(-21.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 3465.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 1.518517e6*fcs_pow(v,10)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 238095.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 17094.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -9.425372635146658e24*fcs_pow(t,32)*w*(-34188.*fcs_pow(u,6)*(1.443897e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 79950.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 585.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 3465.*fcs_pow(u,4)*(-315. + 15540.*w2 - 181818.*fcs_pow(w,4) + 709956.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) - 3.8160135e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 591630.*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1665.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 14430.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 594.*u2*(5.936021e6*fcs_pow(v,10)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 7.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 295815.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 7770.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 20202.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 3.*(-21.*(3. - 132.*w2 + 1386.*fcs_pow(w,4) - 4884.*fcs_pow(w,6) + 5291.*fcs_pow(w,8)) + 1155.*v2*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 216931.*fcs_pow(v,10)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 51282.*fcs_pow(v,6)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 47619.*fcs_pow(v,8)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 770.*fcs_pow(v,4)*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 5.936021e6*fcs_pow(u,12)*(-9.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 45.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1410.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 35955.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 5922.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1.301586e6*fcs_pow(u,10)*(13545.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 3.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 9030.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 84882.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 240499.*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 127323.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 142857.*fcs_pow(u,8)*(-1.85115e6*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 3075.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 370230.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1.242915e6*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 3.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return 2.827611790543997e25*fcs_pow(t,33)*(-170940.*fcs_pow(u,6)*(21. - 4095.*w2 + 111930.*fcs_pow(w,4) - 962598.*fcs_pow(w,6) + 3.094065e6*fcs_pow(w,8) - 3.231579e6*fcs_pow(w,10) - 111930.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 962598.*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.077193e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 343785.*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 195.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 5775.*fcs_pow(u,4)*(2.484846e6*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 7.632027e6*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 9.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 1665.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 14430.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 1386.*u2*(9. - 1575.*w2 + 38850.*fcs_pow(w,4) - 303030.*fcs_pow(w,6) + 887445.*fcs_pow(w,8) - 848003.*fcs_pow(w,10) - 887445.*fcs_pow(v,8)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 225.*v2*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 5550.*fcs_pow(v,4)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 14430.*fcs_pow(v,6)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 3.*(-7.*(-3. + 495.*w2 - 11550.*fcs_pow(w,4) + 85470.*fcs_pow(w,6) - 238095.*fcs_pow(w,8) + 216931.*fcs_pow(w,10)) + 1.518517e6*fcs_pow(v,10)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 385.*v2*(-9. + 1575.*w2 - 38850.*fcs_pow(w,4) + 303030.*fcs_pow(w,6) - 887445.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 11550.*fcs_pow(v,4)*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 85470.*fcs_pow(v,6)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 79365.*fcs_pow(v,8)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 238095.*fcs_pow(u,8)*(3.*(21. - 4305.*w2 + 123410.*fcs_pow(w,4) - 1.11069e6*fcs_pow(w,6) + 3.728745e6*fcs_pow(w,8) - 4.060189e6*fcs_pow(w,10)) + 12915.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 370230.*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 3.728745e6*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 370230.*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 1.7808063e7*fcs_pow(u,12)*(1. - 225.*w2 + 7050.*fcs_pow(w,4) - 69090.*fcs_pow(w,6) + 251685.*fcs_pow(w,8) - 296429.*fcs_pow(w,10) - 2350.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 25.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 42347.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 3290.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 3995.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 3.037034e6*fcs_pow(u,10)*(-9.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 1935.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 181890.*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 6450.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 90945.*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 34357.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
      }
      case 11 : switch(k){
        case 0 : return -1.479264263470744e17*fcs_pow(t,24)*v*(-4950.*u2*(-7. + 315.*v2 - 3654.*fcs_pow(v,4) + 16182.*fcs_pow(v,6) - 29667.*fcs_pow(v,8) + 18879.*fcs_pow(v,10)) + 6.675075e6*fcs_pow(u,8)*(-3. + 165.*v2 - 2310.*fcs_pow(v,4) + 12210.*fcs_pow(v,6) - 26455.*fcs_pow(v,8) + 19721.*fcs_pow(v,10)) + 11.*(-21. + 875.*v2 - 9450.*fcs_pow(v,4) + 39150.*fcs_pow(v,6) - 67425.*fcs_pow(v,8) + 40455.*fcs_pow(v,10)) - 669900.*fcs_pow(u,6)*(-9. + 465.*v2 - 6138.*fcs_pow(v,4) + 30690.*fcs_pow(v,6) - 63085.*fcs_pow(v,8) + 44733.*fcs_pow(v,10)) + 12375.*fcs_pow(u,4)*(-63. + 3045.*v2 - 37758.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 346115.*fcs_pow(v,8) + 232841.*fcs_pow(v,10)) - 296670.*fcs_pow(u,10)*(-99. + 5775.*v2 - 85470.*fcs_pow(v,4) + 476190.*fcs_pow(v,6) - 1.084655e6*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 22475.*fcs_pow(u,12)*(-693. + 42735.*v2 - 666666.*fcs_pow(v,4) + 3.904758e6*fcs_pow(v,6) - 9.328033e6*fcs_pow(v,8) + 7.632027e6*fcs_pow(v,10)));
        case 1 : return 1.109448197603058e19*fcs_pow(t,25)*v*(11.*(-7. + 315.*v2 - 3654.*fcs_pow(v,4) + 16182.*fcs_pow(v,6) - 29667.*fcs_pow(v,8) + 18879.*fcs_pow(v,10)) - 830676.*fcs_pow(u,6)*(-3. + 165.*v2 - 2310.*fcs_pow(v,4) + 12210.*fcs_pow(v,6) - 26455.*fcs_pow(v,8) + 19721.*fcs_pow(v,10)) + 33495.*fcs_pow(u,4)*(-9. + 465.*v2 - 6138.*fcs_pow(v,4) + 30690.*fcs_pow(v,6) - 63085.*fcs_pow(v,8) + 44733.*fcs_pow(v,10)) - 198.*u2*(-63. + 3045.*v2 - 37758.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 346115.*fcs_pow(v,8) + 232841.*fcs_pow(v,10)) + 89001.*fcs_pow(u,8)*(-99. + 5775.*v2 - 85470.*fcs_pow(v,4) + 476190.*fcs_pow(v,6) - 1.084655e6*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 99789.*fcs_pow(u,12)*(-77. + 5005.*v2 - 82082.*fcs_pow(v,4) + 504218.*fcs_pow(v,6) - 1.260545e6*fcs_pow(v,8) + 1.077193e6*fcs_pow(v,10)) - 19778.*fcs_pow(u,10)*(-693. + 42735.*v2 - 666666.*fcs_pow(v,4) + 3.904758e6*fcs_pow(v,6) - 9.328033e6*fcs_pow(v,8) + 7.632027e6*fcs_pow(v,10)))*w;
        case 2 : return -1.109448197603058e19*fcs_pow(t,26)*v*(11.*(7. - 189.*w2 + 315.*v2*(-1. + 29.*w2) - 3654.*fcs_pow(v,4)*(-1. + 31.*w2) + 16182.*fcs_pow(v,6)*(-1. + 33.*w2) - 29667.*fcs_pow(v,8)*(-1. + 35.*w2) + 18879.*fcs_pow(v,10)*(-1. + 37.*w2)) - 198.*u2*(63. - 1827.*w2 + 3045.*v2*(-1. + 31.*w2) - 37758.*fcs_pow(v,4)*(-1. + 33.*w2) + 178002.*fcs_pow(v,6)*(-1. + 35.*w2) - 346115.*fcs_pow(v,8)*(-1. + 37.*w2) + 232841.*fcs_pow(v,10)*(-1. + 39.*w2)) + 33495.*fcs_pow(u,4)*(9. - 279.*w2 + 465.*v2*(-1. + 33.*w2) - 6138.*fcs_pow(v,4)*(-1. + 35.*w2) + 30690.*fcs_pow(v,6)*(-1. + 37.*w2) - 63085.*fcs_pow(v,8)*(-1. + 39.*w2) + 44733.*fcs_pow(v,10)*(-1. + 41.*w2)) - 830676.*fcs_pow(u,6)*(3. - 99.*w2 + 165.*v2*(-1. + 35.*w2) - 2310.*fcs_pow(v,4)*(-1. + 37.*w2) + 12210.*fcs_pow(v,6)*(-1. + 39.*w2) - 26455.*fcs_pow(v,8)*(-1. + 41.*w2) + 19721.*fcs_pow(v,10)*(-1. + 43.*w2)) + 89001.*fcs_pow(u,8)*(99. - 3465.*w2 + 5775.*v2*(-1. + 37.*w2) - 85470.*fcs_pow(v,4)*(-1. + 39.*w2) + 476190.*fcs_pow(v,6)*(-1. + 41.*w2) - 1.084655e6*fcs_pow(v,8)*(-1. + 43.*w2) + 848003.*fcs_pow(v,10)*(-1. + 45.*w2)) - 19778.*fcs_pow(u,10)*(693.*(1. - 37.*w2) + 42735.*v2*(-1. + 39.*w2) - 666666.*fcs_pow(v,4)*(-1. + 41.*w2) + 3.904758e6*fcs_pow(v,6)*(-1. + 43.*w2) - 9.328033e6*fcs_pow(v,8)*(-1. + 45.*w2) + 7.632027e6*fcs_pow(v,10)*(-1. + 47.*w2)) + 99789.*fcs_pow(u,12)*(77. - 3003.*w2 + 5005.*v2*(-1. + 41.*w2) - 82082.*fcs_pow(v,4)*(-1. + 43.*w2) + 504218.*fcs_pow(v,6)*(-1. + 45.*w2) - 1.260545e6*fcs_pow(v,8)*(-1. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-1. + 49.*w2)));
        case 3 : return 3.328344592809173e19*fcs_pow(t,27)*v*w*(11.*(63.*(3. - 29.*w2) - 113274.*fcs_pow(v,4)*(-1. + 11.*w2) + 698523.*fcs_pow(v,10)*(-1. + 13.*w2) + 3045.*v2*(-3. + 31.*w2) + 178002.*fcs_pow(v,6)*(-3. + 35.*w2) - 346115.*fcs_pow(v,8)*(-3. + 37.*w2)) - 40194.*u2*(9. - 93.*w2 + 465.*v2*(-1. + 11.*w2) - 63085.*fcs_pow(v,8)*(-1. + 13.*w2) - 2046.*fcs_pow(v,4)*(-3. + 35.*w2) + 10230.*fcs_pow(v,6)*(-3. + 37.*w2) + 14911.*fcs_pow(v,10)*(-3. + 41.*w2)) - 276892.*fcs_pow(u,6)*(99.*(3. - 35.*w2) - 256410.*fcs_pow(v,4)*(-1. + 13.*w2) + 2.544009e6*fcs_pow(v,10)*(-1. + 15.*w2) + 5775.*v2*(-3. + 37.*w2) + 476190.*fcs_pow(v,6)*(-3. + 41.*w2) - 1.084655e6*fcs_pow(v,8)*(-3. + 43.*w2)) + 1.038345e6*fcs_pow(u,4)*(9. - 99.*w2 + 36630.*fcs_pow(v,6)*(-1. + 13.*w2) + 165.*v2*(-3. + 35.*w2) - 2310.*fcs_pow(v,4)*(-3. + 37.*w2) - 26455.*fcs_pow(v,8)*(-3. + 41.*w2) + 19721.*fcs_pow(v,10)*(-3. + 43.*w2)) + 445005.*fcs_pow(u,8)*(693. - 8547.*w2 + 42735.*v2*(-1. + 13.*w2) - 9.328033e6*fcs_pow(v,8)*(-1. + 15.*w2) - 222222.*fcs_pow(v,4)*(-3. + 41.*w2) + 1.301586e6*fcs_pow(v,6)*(-3. + 43.*w2) + 2.544009e6*fcs_pow(v,10)*(-3. + 47.*w2)) + 432419.*fcs_pow(u,12)*(693. - 9471.*w2 - 814506.*fcs_pow(v,4)*(-1. + 15.*w2) + 1.2180567e7*fcs_pow(v,10)*(-1. + 17.*w2) + 15785.*v2*(-3. + 43.*w2) + 1.74537e6*fcs_pow(v,6)*(-3. + 47.*w2) - 4.557355e6*fcs_pow(v,8)*(-3. + 49.*w2)) - 2.195358e6*fcs_pow(u,10)*(231. - 3003.*w2 + 1.512654e6*fcs_pow(v,6)*(-1. + 15.*w2) + 5005.*v2*(-3. + 41.*w2) - 82082.*fcs_pow(v,4)*(-3. + 43.*w2) - 1.260545e6*fcs_pow(v,8)*(-3. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-3. + 49.*w2)));
        case 4 : return -9.98503377842752e19*fcs_pow(t,28)*v*(11.*(3045.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 37758.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 346115.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 232841.*fcs_pow(v,10)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 21.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 59334.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 276892.*fcs_pow(u,6)*(5775.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 85470.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 1.084655e6*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 33.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 158730.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 346115.*fcs_pow(u,4)*(-9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 6930.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 36630.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59163.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 165.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 26455.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 13398.*u2*(-9.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 465.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 30690.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 2046.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 14911.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 432419.*fcs_pow(u,12)*(15785.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 271502.*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 4.557355e6*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 77.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 581790.*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 731786.*fcs_pow(u,10)*(-231.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 246246.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 1.512654e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 3.231579e6*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 5005.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 1.260545e6*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 148335.*fcs_pow(u,8)*(-693.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 42735.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 3.904758e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 9.328033e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 222222.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 2.544009e6*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return 2.895659795743981e21*fcs_pow(t,29)*v*w*(59675.*fcs_pow(u,4)*(2.544009e6*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 99.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 1.084655e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 51282.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1155.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 95238.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 71610.*u2*(-9.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 165.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 19721.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7326.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 462.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 77.*(-45. + 930.*w2 - 3069.*fcs_pow(w,4) + 2325.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 10230.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 63085.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 10230.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 14911.*fcs_pow(v,10)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 47740.*fcs_pow(u,6)*(-4.6640165e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 6.50793e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 42735.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 231.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 222222.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 2.544009e6*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 3.056755e6*fcs_pow(u,12)*(16555.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 77.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 1.683493e6*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 1.089319e6*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 19866.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 133386.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 328042.*fcs_pow(u,10)*(-4.07253e6*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 78925.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 1.2180567e7*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 231.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 1.74537e6*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 4.557355e6*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 567765.*fcs_pow(u,8)*(7.56327e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 410410.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 231.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 5005.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.260545e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return -1.44782989787199e22*fcs_pow(t,30)*v*(35805.*fcs_pow(u,4)*(-99.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 476190.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 1.084655e6*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 1155.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 42966.*u2*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 462.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2442.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 77.*(9. - 837.*w2 + 9207.*fcs_pow(w,4) - 21483.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 6138.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 44733.*fcs_pow(v,10)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 6138.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 9548.*fcs_pow(u,6)*(-3.33333e6*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.952379e7*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.6640165e7*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 693.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 7.632027e6*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 611351.*fcs_pow(u,12)*(-231.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 16555.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 5.446595e6*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 5.050479e6*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 59598.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 400158.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 984126.*fcs_pow(u,10)*(-231.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 15785.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 271502.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 4.060189e6*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 349074.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 911471.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 340659.*fcs_pow(u,8)*(25025.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 410410.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 2.52109e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 77.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.260545e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return 9.425372635146658e24*fcs_pow(t,31)*v*w*(-34188.*fcs_pow(u,6)*(11.*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6)) + 2.52109e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.260545e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 58630.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 715.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 55.*fcs_pow(u,4)*(-3.26481155e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 5.3424189e7*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 1.952379e7*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 693.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 666666.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 66.*u2*(-693.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 5.936021e6*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.084655e6*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1155.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 95238.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 11.*(21.*(1. - 33.*w2 + 231.*fcs_pow(w,4) - 407.*fcs_pow(w,6)) + 1155.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 19721.*fcs_pow(v,10)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 462.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2442.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 5291.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 5.936021e6*fcs_pow(u,12)*(-11.*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 165.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 3102.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 43945.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 42347.*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 3102.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 433862.*fcs_pow(u,10)*(16555.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 59598.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 33.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 400158.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 721497.*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 155617.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 47619.*fcs_pow(u,8)*(-1.35751e6*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.74537e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 11275.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.557355e6*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 33.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 580027.*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return -9.425372635146658e24*fcs_pow(t,32)*v*(238095.*fcs_pow(u,8)*(15785.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 271502.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 33.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 349074.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 911471.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) - 433862.*fcs_pow(u,10)*(-231.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 16555.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 59598.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 400158.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 778085.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 721497.*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 5.936021e6*fcs_pow(u,12)*(-11.*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 165.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3102.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 15510.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 43945.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 34188.*fcs_pow(u,6)*(-410410.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 2.52109e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 3575.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.260545e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 385.*fcs_pow(u,4)*(1.952379e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 4.6640165e7*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 476190.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 7.632027e6*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 99.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 6105.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 66.*u2*(-693.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 7.592585e6*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 5.936021e6*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 476190.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 1155.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 17094.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 11.*(-21.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 1155.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 138047.*fcs_pow(v,10)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 26455.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 462.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 2442.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return 1.036790989866132e26*fcs_pow(t,33)*v*w*(11655.*fcs_pow(u,4)*(2.268981e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 175890.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 99.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1.260545e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 2145.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 54.*u2*(-3.26481155e8*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 6.50793e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 42735.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5.936021e6*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 222222.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 77.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 3.*(1.7808063e7*fcs_pow(v,10)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 231.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 1.084655e6*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 51282.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 95238.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 385.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 5.936021e6*fcs_pow(u,12)*(-9.*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 705.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 381123.*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 71910.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 211735.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 5922.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 3.5616126e7*fcs_pow(u,10)*(-99.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 165.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3102.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 43945.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 9306.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 2.662335e6*fcs_pow(u,8)*(49665.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 33.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 19866.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 133386.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 240499.*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 155617.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 40404.*fcs_pow(u,6)*(-1.221759e7*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 33825.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1.74537e6*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 4.557355e6*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 99.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 1.740081e6*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return -3.110372969598397e26*fcs_pow(t,34)*v*(19425.*fcs_pow(u,4)*(-738738.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 4.537962e6*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 3.231579e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 1.260545e6*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 2145.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 630.*u2*(3.904758e6*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 9.328033e6*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 6105.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 31746.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 3.*(-7.592585e6*fcs_pow(v,8)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 5.936021e6*fcs_pow(v,10)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 77.*(-9. + 1575.*w2 - 38850.*fcs_pow(w,4) + 303030.*fcs_pow(w,6) - 887445.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 5775.*v2*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 85470.*fcs_pow(v,4)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 158730.*fcs_pow(v,6)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 202020.*fcs_pow(u,6)*(47355.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 814506.*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4.557355e6*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 581790.*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 33.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 8.9040315e7*fcs_pow(u,12)*(9. - 2115.*w2 + 69090.*fcs_pow(w,4) - 704718.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 3.260719e6*fcs_pow(w,10) + 296429.*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 235.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 42347.*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 658.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 1598.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 1.1872042e7*fcs_pow(u,10)*(-99.*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 46530.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 825.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 381123.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 46530.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 43945.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 887445.*fcs_pow(u,8)*(-693.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 49665.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 2.00079e6*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 99330.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 778085.*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 240499.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
        case 11 : return 1.555186484799199e27*fcs_pow(t,35)*v*w*(-1.8222204e7*fcs_pow(u,6)*(693. - 49665.*w2 + 893970.*fcs_pow(w,4) - 6.00237e6*fcs_pow(w,6) + 1.6339785e7*fcs_pow(w,8) - 1.5151437e7*fcs_pow(w,10) + 4515.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99330.*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 666930.*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 778085.*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 240499.*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10))) + 5.87666079e8*fcs_pow(u,8)*(-9.*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 825.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 15510.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 46530.*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 43945.*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 42347.*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10))) + 555555.*fcs_pow(u,4)*(-222138.*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.74537e6*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4.557355e6*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 1.740081e6*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 9.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 615.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) - 51282.*u2*(412542.*fcs_pow(v,6)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.260545e6*fcs_pow(v,8)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.077193e6*fcs_pow(v,10)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 9.*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 195.*v2*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 3198.*fcs_pow(v,4)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 8.36978961e8*fcs_pow(u,12)*(-11.*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 385.*v2*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 119833.*fcs_pow(v,10)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 2618.*fcs_pow(v,4)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 19822.*fcs_pow(v,6)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 49555.*fcs_pow(v,8)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 1.30592462e8*fcs_pow(u,10)*(-99.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 7755.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 65142.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 158202.*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 465817.*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 381123.*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 33.*(-5.936021e6*fcs_pow(v,8)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 5.936021e6*fcs_pow(v,10)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 3885.*v2*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 20202.*fcs_pow(v,4)*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 118326.*fcs_pow(v,6)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10)) - 7.*(-693. + 42735.*w2 - 666666.*fcs_pow(w,4) + 3.904758e6*fcs_pow(w,6) - 9.328033e6*fcs_pow(w,8) + 7.632027e6*fcs_pow(w,10))));
      }
      case 12 : switch(k){
        case 0 : return 4.437792790412231e17*fcs_pow(t,25)*(86625.*fcs_pow(u,4)*(3. - 522.*v2 + 13485.*fcs_pow(v,4) - 118668.*fcs_pow(v,6) + 445005.*fcs_pow(v,8) - 731786.*fcs_pow(v,10) + 432419.*fcs_pow(v,12)) + 11.*(7. - 1050.*v2 + 23625.*fcs_pow(v,4) - 182700.*fcs_pow(v,6) + 606825.*fcs_pow(v,8) - 890010.*fcs_pow(v,10) + 471975.*fcs_pow(v,12)) - 669900.*fcs_pow(u,6)*(3. - 558.*v2 + 15345.*fcs_pow(v,4) - 143220.*fcs_pow(v,6) + 567765.*fcs_pow(v,8) - 984126.*fcs_pow(v,10) + 611351.*fcs_pow(v,12)) - 1650.*u2*(7. - 1134.*v2 + 27405.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 801009.*fcs_pow(v,8) - 1.246014e6*fcs_pow(v,10) + 698523.*fcs_pow(v,12)) + 2.225025e6*fcs_pow(u,8)*(3. - 594.*v2 + 17325.*fcs_pow(v,4) - 170940.*fcs_pow(v,6) + 714285.*fcs_pow(v,8) - 1.301586e6*fcs_pow(v,10) + 848003.*fcs_pow(v,12)) - 296670.*fcs_pow(u,10)*(33. - 6930.*v2 + 213675.*fcs_pow(v,4) - 2.22222e6*fcs_pow(v,6) + 9.761895e6*fcs_pow(v,8) - 1.8656066e7*fcs_pow(v,10) + 1.2720045e7*fcs_pow(v,12)) + 67425.*fcs_pow(u,12)*(77. - 17094.*v2 + 555555.*fcs_pow(v,4) - 6.074068e6*fcs_pow(v,6) + 2.7984099e7*fcs_pow(v,8) - 5.5968198e7*fcs_pow(v,10) + 3.9856141e7*fcs_pow(v,12)));
        case 1 : return -1.109448197603058e19*fcs_pow(t,26)*(-4158.*u2*(3. - 522.*v2 + 13485.*fcs_pow(v,4) - 118668.*fcs_pow(v,6) + 445005.*fcs_pow(v,8) - 731786.*fcs_pow(v,10) + 432419.*fcs_pow(v,12)) + 100485.*fcs_pow(u,4)*(3. - 558.*v2 + 15345.*fcs_pow(v,4) - 143220.*fcs_pow(v,6) + 567765.*fcs_pow(v,8) - 984126.*fcs_pow(v,10) + 611351.*fcs_pow(v,12)) + 11.*(7. - 1134.*v2 + 27405.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 801009.*fcs_pow(v,8) - 1.246014e6*fcs_pow(v,10) + 698523.*fcs_pow(v,12)) - 830676.*fcs_pow(u,6)*(3. - 594.*v2 + 17325.*fcs_pow(v,4) - 170940.*fcs_pow(v,6) + 714285.*fcs_pow(v,8) - 1.301586e6*fcs_pow(v,10) + 848003.*fcs_pow(v,12)) + 267003.*fcs_pow(u,8)*(33. - 6930.*v2 + 213675.*fcs_pow(v,4) - 2.22222e6*fcs_pow(v,6) + 9.761895e6*fcs_pow(v,8) - 1.8656066e7*fcs_pow(v,10) + 1.2720045e7*fcs_pow(v,12)) - 178002.*fcs_pow(u,10)*(77. - 17094.*v2 + 555555.*fcs_pow(v,4) - 6.074068e6*fcs_pow(v,6) + 2.7984099e7*fcs_pow(v,8) - 5.5968198e7*fcs_pow(v,10) + 3.9856141e7*fcs_pow(v,12)) + 99789.*fcs_pow(u,12)*(77. - 18018.*v2 + 615615.*fcs_pow(v,4) - 7.059052e6*fcs_pow(v,6) + 3.4034715e7*fcs_pow(v,8) - 7.1094738e7*fcs_pow(v,10) + 5.2782457e7*fcs_pow(v,12)))*w;
        case 2 : return 1.109448197603058e19*fcs_pow(t,27)*(11.*(-7. + 189.*w2 - 1134.*v2*(-1. + 29.*w2) + 27405.*fcs_pow(v,4)*(-1. + 31.*w2) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2) + 801009.*fcs_pow(v,8)*(-1. + 35.*w2) - 1.246014e6*fcs_pow(v,10)*(-1. + 37.*w2) + 698523.*fcs_pow(v,12)*(-1. + 39.*w2)) - 4158.*u2*(-3. + 87.*w2 - 522.*v2*(-1. + 31.*w2) + 13485.*fcs_pow(v,4)*(-1. + 33.*w2) - 118668.*fcs_pow(v,6)*(-1. + 35.*w2) + 445005.*fcs_pow(v,8)*(-1. + 37.*w2) - 731786.*fcs_pow(v,10)*(-1. + 39.*w2) + 432419.*fcs_pow(v,12)*(-1. + 41.*w2)) + 100485.*fcs_pow(u,4)*(-3. + 93.*w2 - 558.*v2*(-1. + 33.*w2) + 15345.*fcs_pow(v,4)*(-1. + 35.*w2) - 143220.*fcs_pow(v,6)*(-1. + 37.*w2) + 567765.*fcs_pow(v,8)*(-1. + 39.*w2) - 984126.*fcs_pow(v,10)*(-1. + 41.*w2) + 611351.*fcs_pow(v,12)*(-1. + 43.*w2)) - 830676.*fcs_pow(u,6)*(-3. + 99.*w2 - 594.*v2*(-1. + 35.*w2) + 17325.*fcs_pow(v,4)*(-1. + 37.*w2) - 170940.*fcs_pow(v,6)*(-1. + 39.*w2) + 714285.*fcs_pow(v,8)*(-1. + 41.*w2) - 1.301586e6*fcs_pow(v,10)*(-1. + 43.*w2) + 848003.*fcs_pow(v,12)*(-1. + 45.*w2)) + 267003.*fcs_pow(u,8)*(33.*(-1. + 35.*w2) - 6930.*v2*(-1. + 37.*w2) + 213675.*fcs_pow(v,4)*(-1. + 39.*w2) - 2.22222e6*fcs_pow(v,6)*(-1. + 41.*w2) + 9.761895e6*fcs_pow(v,8)*(-1. + 43.*w2) - 1.8656066e7*fcs_pow(v,10)*(-1. + 45.*w2) + 1.2720045e7*fcs_pow(v,12)*(-1. + 47.*w2)) - 178002.*fcs_pow(u,10)*(77.*(-1. + 37.*w2) - 17094.*v2*(-1. + 39.*w2) + 555555.*fcs_pow(v,4)*(-1. + 41.*w2) - 6.074068e6*fcs_pow(v,6)*(-1. + 43.*w2) + 2.7984099e7*fcs_pow(v,8)*(-1. + 45.*w2) - 5.5968198e7*fcs_pow(v,10)*(-1. + 47.*w2) + 3.9856141e7*fcs_pow(v,12)*(-1. + 49.*w2)) + 99789.*fcs_pow(u,12)*(77.*(-1. + 39.*w2) - 18018.*v2*(-1. + 41.*w2) + 615615.*fcs_pow(v,4)*(-1. + 43.*w2) - 7.059052e6*fcs_pow(v,6)*(-1. + 45.*w2) + 3.4034715e7*fcs_pow(v,8)*(-1. + 47.*w2) - 7.1094738e7*fcs_pow(v,10)*(-1. + 49.*w2) + 5.2782457e7*fcs_pow(v,12)*(-1. + 51.*w2)));
        case 3 : return -9.98503377842752e19*fcs_pow(t,28)*w*(77.*(-9. + 87.*w2 + 40455.*fcs_pow(v,4)*(-1. + 11.*w2) - 2.195358e6*fcs_pow(v,10)*(-1. + 13.*w2) - 522.*v2*(-3. + 31.*w2) - 118668.*fcs_pow(v,6)*(-3. + 35.*w2) + 445005.*fcs_pow(v,8)*(-3. + 37.*w2) + 432419.*fcs_pow(v,12)*(-3. + 41.*w2)) + 1.038345e6*fcs_pow(u,4)*(-3. + 33.*w2 - 170940.*fcs_pow(v,6)*(-1. + 13.*w2) + 848003.*fcs_pow(v,12)*(-1. + 15.*w2) - 198.*v2*(-3. + 35.*w2) + 5775.*fcs_pow(v,4)*(-3. + 37.*w2) + 238095.*fcs_pow(v,8)*(-3. + 41.*w2) - 433862.*fcs_pow(v,10)*(-3. + 43.*w2)) - 13398.*u2*(-9. + 93.*w2 - 1674.*v2*(-1. + 11.*w2) + 1.703295e6*fcs_pow(v,8)*(-1. + 13.*w2) + 15345.*fcs_pow(v,4)*(-3. + 35.*w2) - 143220.*fcs_pow(v,6)*(-3. + 37.*w2) - 984126.*fcs_pow(v,10)*(-3. + 41.*w2) + 611351.*fcs_pow(v,12)*(-3. + 43.*w2)) - 830676.*fcs_pow(u,6)*(-33. + 385.*w2 + 213675.*fcs_pow(v,4)*(-1. + 13.*w2) - 1.8656066e7*fcs_pow(v,10)*(-1. + 15.*w2) - 2310.*v2*(-3. + 37.*w2) - 740740.*fcs_pow(v,6)*(-3. + 41.*w2) + 3.253965e6*fcs_pow(v,8)*(-3. + 43.*w2) + 4.240015e6*fcs_pow(v,12)*(-3. + 47.*w2)) - 2.195358e6*fcs_pow(u,10)*(77.*(-1. + 13.*w2) - 7.059052e6*fcs_pow(v,6)*(-1. + 15.*w2) + 5.2782457e7*fcs_pow(v,12)*(-1. + 17.*w2) - 6006.*v2*(-3. + 41.*w2) + 205205.*fcs_pow(v,4)*(-3. + 43.*w2) + 1.1344905e7*fcs_pow(v,8)*(-3. + 47.*w2) - 2.3698246e7*fcs_pow(v,10)*(-3. + 49.*w2)) + 445005.*fcs_pow(u,8)*(-51282.*v2*(-1. + 13.*w2) + 8.3952297e7*fcs_pow(v,8)*(-1. + 15.*w2) + 77.*(-3. + 37.*w2) + 555555.*fcs_pow(v,4)*(-3. + 41.*w2) - 6.074068e6*fcs_pow(v,6)*(-3. + 43.*w2) - 5.5968198e7*fcs_pow(v,10)*(-3. + 47.*w2) + 3.9856141e7*fcs_pow(v,12)*(-3. + 49.*w2)) + 432419.*fcs_pow(u,12)*(2.036265e6*fcs_pow(v,4)*(-1. + 15.*w2) - 2.67972474e8*fcs_pow(v,10)*(-1. + 17.*w2) + 77.*(-3. + 41.*w2) - 18942.*v2*(-3. + 43.*w2) - 8.14506e6*fcs_pow(v,6)*(-3. + 47.*w2) + 4.1016195e7*fcs_pow(v,8)*(-3. + 49.*w2) + 6.9023213e7*fcs_pow(v,12)*(-3. + 53.*w2)));
        case 4 : return 9.98503377842752e19*fcs_pow(t,29)*(1.038345e6*fcs_pow(u,4)*(3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 17325.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 170940.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 1.301586e6*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,12)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 198.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 238095.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 40194.*u2*(3.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 558.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 143220.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 567765.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 611351.*fcs_pow(v,12)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 5115.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 328042.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 77.*(9. - 522.*w2 + 2697.*fcs_pow(w,4) - 1566.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 40455.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 1.335015e6*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 2.195358e6*fcs_pow(v,10)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 118668.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 432419.*fcs_pow(v,12)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 2.195358e6*fcs_pow(u,10)*(77.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 615615.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 7.059052e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 7.1094738e7*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 5.2782457e7*fcs_pow(v,12)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 6006.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 1.1344905e7*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 1.335015e6*fcs_pow(u,8)*(77.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 17094.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 6.074068e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 2.7984099e7*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 3.9856141e7*fcs_pow(v,12)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 185185.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 1.8656066e7*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 830676.*fcs_pow(u,6)*(-6930.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 213675.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 9.761895e6*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 1.8656066e7*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 11.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 740740.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 4.240015e6*fcs_pow(v,12)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 432419.*fcs_pow(u,12)*(-56826.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 2.036265e6*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 1.23048585e8*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 2.67972474e8*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 77.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 8.14506e6*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 6.9023213e7*fcs_pow(v,12)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))));
        case 5 : return -2.895659795743981e21*fcs_pow(t,30)*w*(-214830.*u2*(3. - 66.*w2 + 231.*fcs_pow(w,4) + 848003.*fcs_pow(v,12)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 198.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 433862.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 34188.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1155.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 47619.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 77.*(45. - 930.*w2 + 3069.*fcs_pow(w,4) - 8370.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 76725.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 3.056755e6*fcs_pow(v,12)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 1.703295e6*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 143220.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 984126.*fcs_pow(v,10)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 537075.*fcs_pow(u,4)*(-1.8656066e7*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 11.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 3.253965e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 42735.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 462.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 148148.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,12)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 2.1397285e7*fcs_pow(u,12)*(-8514.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 11.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 1.5872934e7*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 1.2746447e7*fcs_pow(v,12)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 4.201659e6*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 21285.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 266772.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 984126.*fcs_pow(u,10)*(1.0181325e7*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 94710.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3.45116065e8*fcs_pow(v,12)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 2.67972474e8*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 77.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 8.14506e6*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 4.1016195e7*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 1.703295e6*fcs_pow(u,8)*(-3.529526e7*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1.026025e6*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 77.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 5.2782457e7*fcs_pow(v,12)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 6006.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 1.1344905e7*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 2.3698246e7*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 143220.*fcs_pow(u,6)*(4.19761485e8*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 3.037034e7*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 51282.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 77.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 555555.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 5.5968198e7*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 3.9856141e7*fcs_pow(v,12)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return 4.343489693615971e22*fcs_pow(t,31)*(-14322.*u2*(-3. + 297.*w2 - 3465.*fcs_pow(w,4) + 8547.*fcs_pow(w,6) - 594.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 714285.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 1.301586e6*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,12)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 3465.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 34188.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 77.*(-3. + 279.*w2 - 3069.*fcs_pow(w,4) + 7161.*fcs_pow(w,6) - 558.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 15345.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 984126.*fcs_pow(v,10)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 611351.*fcs_pow(v,12)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 28644.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 113553.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 35805.*fcs_pow(u,4)*(33.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 2.22222e6*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 9.761895e6*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.8656066e7*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 1386.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 2.544009e6*fcs_pow(v,12)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 4.279457e6*fcs_pow(u,12)*(11.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2838.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.002765e6*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 1.5872934e7*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.2746447e7*fcs_pow(v,12)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 21285.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 266772.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 984126.*fcs_pow(u,10)*(77.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 18942.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 678755.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 8.9324158e7*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 6.9023213e7*fcs_pow(v,12)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.629012e6*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 8.203239e6*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 113553.*fcs_pow(u,8)*(-90090.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 3.078075e6*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 3.529526e7*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 2.63912285e8*fcs_pow(v,12)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 77.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 3.4034715e7*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 7.1094738e7*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 28644.*fcs_pow(u,6)*(2.777775e6*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 3.037034e7*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.39920495e8*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 77.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 5.5968198e7*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 3.9856141e7*fcs_pow(v,12)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return -1.036790989866132e26*fcs_pow(t,32)*w*(21.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4158.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 5.936021e6*fcs_pow(v,12)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.301586e6*fcs_pow(v,10)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 3465.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 34188.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 142857.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 315.*fcs_pow(u,4)*(1.39920495e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 5.5968198e7*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 4.33862e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 3.9856141e7*fcs_pow(v,12)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 11.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 2442.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 79365.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 18.*u2*(231.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 1.30592462e8*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.7808063e7*fcs_pow(v,12)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 9.761895e6*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1386.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 444444.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 118326.*fcs_pow(u,10)*(-19866.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 148995.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 11.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.867404e6*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 1.5872934e7*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.2746447e7*fcs_pow(v,12)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 1.400553e6*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 12987.*fcs_pow(u,8)*(3.393775e6*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 8.14506e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 13530.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 4.1016195e7*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 4.9302295e7*fcs_pow(v,12)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 11.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 1.2760594e7*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 3108.*fcs_pow(u,6)*(-3.529526e7*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 3.4034715e7*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 439725.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 7.1094738e7*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 11.*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 2574.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 7.540351e6*fcs_pow(v,12)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 5.936021e6*fcs_pow(u,12)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6) - 54.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 2115.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 107865.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 254082.*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 3948.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 42347.*fcs_pow(v,12)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 8 : return 1.036790989866132e26*fcs_pow(t,33)*(21.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4158.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 9.111102e6*fcs_pow(v,10)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 5.936021e6*fcs_pow(v,12)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 714285.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 3465.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 34188.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) + 64935.*fcs_pow(u,8)*(-18942.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 678755.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 11.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.629012e6*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 8.203239e6*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.2760594e7*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 9.860459e6*fcs_pow(v,12)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 5.936021e6*fcs_pow(u,12)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8) - 54.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 2115.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 19740.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 296429.*fcs_pow(v,12)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 107865.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 254082.*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 828282.*fcs_pow(u,10)*(11.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 2838.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 21285.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 266772.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.000395e6*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 2.267562e6*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.820921e6*fcs_pow(v,12)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 3108.*fcs_pow(u,6)*(3.078075e6*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 3.529526e7*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 12870.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 3.4034715e7*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 7.1094738e7*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 3.7701755e7*fcs_pow(v,12)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 315.*fcs_pow(u,4)*(-3.037034e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.39920495e8*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 396825.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 5.5968198e7*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 3.9856141e7*fcs_pow(v,12)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 11.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 2442.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 126.*u2*(33.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 9.761895e6*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 1.8656066e7*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 317460.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2.544009e6*fcs_pow(v,12)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 198.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 6105.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -9.331118908795191e26*fcs_pow(t,34)*w*(-3.91777386e8*fcs_pow(v,10)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 77.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 9.761895e6*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 128205.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5.936021e6*fcs_pow(v,12)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 444444.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 462.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 42.*u2*(1.259284455e9*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 1.301586e7*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 21978.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 5.5968198e7*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 3.9856141e7*fcs_pow(v,12)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 238095.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 11.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 5.936021e6*fcs_pow(u,12)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8) + 5.632151e6*fcs_pow(v,12)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 282.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2.794902e6*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 111860.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 635205.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 4935.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1.30592462e8*fcs_pow(u,10)*(3.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 18.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 705.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 127041.*fcs_pow(v,12)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 35955.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 84694.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 3948.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 295815.*fcs_pow(u,8)*(-178794.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 33.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 148995.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1.867404e6*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.5872934e7*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.2746447e7*fcs_pow(v,12)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 4.201659e6*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 13468.*fcs_pow(u,6)*(3.0543975e7*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 40590.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 8.14506e6*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 4.1016195e7*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 33.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 4.9302295e7*fcs_pow(v,12)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 3.8281782e7*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 3885.*fcs_pow(u,4)*(-1.0588578e8*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 439725.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 33.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 1.1344905e7*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 2.3698246e7*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2574.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 7.540351e6*fcs_pow(v,12)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return 6.531783236156634e27*fcs_pow(t,35)*(2.9285685e7*fcs_pow(v,8)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 5.5968198e7*fcs_pow(v,10)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 11.*(-9. + 1575.*w2 - 38850.*fcs_pow(w,4) + 303030.*fcs_pow(w,6) - 887445.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 2970.*v2*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 91575.*fcs_pow(v,4)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 4.240015e6*fcs_pow(v,12)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 317460.*fcs_pow(v,6)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 90.*u2*(-1.8222204e7*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 8.3952297e7*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 33.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 7326.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 3.9856141e7*fcs_pow(v,12)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 1.8656066e7*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 79365.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 2775.*fcs_pow(u,4)*(1.846845e6*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 2.1177156e7*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 33.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 7.1094738e7*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 1.1344905e7*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 2574.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 7.540351e6*fcs_pow(v,12)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 4.240015e6*fcs_pow(u,12)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10) - 1.9564314e7*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 846.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 5.632151e6*fcs_pow(v,12)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 1.143369e6*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 4935.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 22372.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 5.5968198e7*fcs_pow(u,10)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10) + 1.482145e6*fcs_pow(v,12)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 3525.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 30.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 254082.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 6580.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 11985.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 887445.*fcs_pow(u,8)*(33.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 8514.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.33386e6*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 35475.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 5.462763e6*fcs_pow(v,12)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 1.000395e6*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 755854.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 9620.*fcs_pow(u,6)*(-170478.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 6.108795e6*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.23048585e8*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 8.14506e6*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 33.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 3.8281782e7*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
        case 11 : return -3.265891618078317e28*fcs_pow(t,36)*w*(-867724.*fcs_pow(u,6)*(-16254.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 744975.*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 9.33702e6*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 2.1008295e7*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 1.5872934e7*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 1.2746447e7*fcs_pow(v,12)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 3.*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 79365.*fcs_pow(u,4)*(555345.*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 8.14506e6*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 4.1016195e7*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 3.8281782e7*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 3.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 738.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) - 7326.*u2*(-1.925196e6*fcs_pow(v,6)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.1344905e7*fcs_pow(v,8)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 2.3698246e7*fcs_pow(v,10)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 7.540351e6*fcs_pow(v,12)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 3.*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 234.*v2*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 7995.*fcs_pow(v,4)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 11.*(-693. + 42735.*w2 - 666666.*fcs_pow(w,4) + 3.904758e6*fcs_pow(w,6) - 9.328033e6*fcs_pow(w,8) + 7.632027e6*fcs_pow(w,10) + 2.2896081e7*fcs_pow(v,8)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 5.5968198e7*fcs_pow(v,10)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 3.9856141e7*fcs_pow(v,12)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 1998.*v2*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 21645.*fcs_pow(v,4)*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 236652.*fcs_pow(v,6)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 3.9856141e7*fcs_pow(u,12)*(11.*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 1386.*v2*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 7.908978e6*fcs_pow(v,10)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 19635.*fcs_pow(v,4)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,12)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 277508.*fcs_pow(v,6)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 1.337985e6*fcs_pow(v,8)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 5.5968198e7*fcs_pow(u,10)*(11.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 3102.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 54285.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 5.632151e6*fcs_pow(v,12)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 246092.*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 1.397451e6*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 2.794902e6*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 8.3952297e7*fcs_pow(u,8)*(-33. + 2475.*w2 - 46530.*fcs_pow(w,4) + 325710.*fcs_pow(w,6) - 922845.*fcs_pow(w,8) + 889287.*fcs_pow(w,10) - 990.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 38775.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 217140.*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 395505.*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 931634.*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 635205.*fcs_pow(v,12)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))));
        case 12 : return 9.797674854234951e28*fcs_pow(t,37)*(-2442.*u2*(-2.1177156e7*fcs_pow(v,6)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 3.4034715e7*fcs_pow(v,8)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 7.1094738e7*fcs_pow(v,10)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 167895.*fcs_pow(v,4)*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) + 3.*(77. - 18018.*w2 + 615615.*fcs_pow(w,4) - 7.059052e6*fcs_pow(w,6) + 3.4034715e7*fcs_pow(w,8) - 7.1094738e7*fcs_pow(w,10) + 5.2782457e7*fcs_pow(w,12)) + 7.540351e6*fcs_pow(v,12)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 702.*v2*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12))) + 11.*(8.3952297e7*fcs_pow(v,8)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 5.5968198e7*fcs_pow(v,10)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 3.9856141e7*fcs_pow(v,12)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 1.656564e6*fcs_pow(v,6)*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) + 3.*(77. - 17094.*w2 + 555555.*fcs_pow(w,4) - 6.074068e6*fcs_pow(w,6) + 2.7984099e7*fcs_pow(w,8) - 5.5968198e7*fcs_pow(w,10) + 3.9856141e7*fcs_pow(w,12)) - 666.*v2*(77. - 18018.*w2 + 615615.*fcs_pow(w,4) - 7.059052e6*fcs_pow(w,6) + 3.4034715e7*fcs_pow(w,8) - 7.1094738e7*fcs_pow(w,10) + 5.2782457e7*fcs_pow(w,12)) + 21645.*fcs_pow(v,4)*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12))) - 5.5968198e7*fcs_pow(u,10)*(11.*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 3102.*v2*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 4.192353e6*fcs_pow(v,8)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 6.521438e6*fcs_pow(v,10)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 18095.*fcs_pow(v,4)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 5.632151e6*fcs_pow(v,12)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 246092.*fcs_pow(v,6)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 8.3952297e7*fcs_pow(u,8)*(11.*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 990.*v2*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 38775.*fcs_pow(v,4)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 2.794902e6*fcs_pow(v,10)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 1.482145e6*fcs_pow(v,12)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 72380.*fcs_pow(v,6)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 395505.*fcs_pow(v,8)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 1.8222204e7*fcs_pow(u,6)*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12) - 2838.*v2*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 35475.*fcs_pow(v,4)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 444620.*fcs_pow(v,6)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 1.820921e6*fcs_pow(v,12)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 333465.*fcs_pow(v,8)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 755854.*fcs_pow(v,10)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 79365.*fcs_pow(u,4)*(2.036265e6*fcs_pow(v,4)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 8.14506e6*fcs_pow(v,6)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 4.1016195e7*fcs_pow(v,8)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 5166.*v2*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) - 1.2760594e7*fcs_pow(v,10)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3.*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12)) + 9.860459e6*fcs_pow(v,12)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 3.9856141e7*fcs_pow(u,12)*(11.*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 832524.*fcs_pow(v,6)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 3.121965e6*fcs_pow(v,8)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 462.*v2*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 7.908978e6*fcs_pow(v,10)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 19635.*fcs_pow(v,4)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 7.070147e6*fcs_pow(v,12)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12))));
      }
    }
    case 13 : switch(j){
      case 0 : switch(k){
        case 0 : return -6.081075e6*fcs_pow(t,14)*u*(3003. - 90090.*u2 + 765765.*fcs_pow(u,4) - 2.77134e6*fcs_pow(u,6) + 4.849845e6*fcs_pow(u,8) - 4.056234e6*fcs_pow(u,10) + 1.300075e6*fcs_pow(u,12));
      }
      case 1 : switch(k){
        case 0 : return 6.38512875e8*fcs_pow(t,15)*u*(429. - 14586.*u2 + 138567.*fcs_pow(u,4) - 554268.*fcs_pow(u,6) + 1.062347e6*fcs_pow(u,8) - 965770.*fcs_pow(u,10) + 334305.*fcs_pow(u,12))*v;
        case 1 : return -1.0854718875e10*fcs_pow(t,16)*u*(429. - 16302.*u2 + 171171.*fcs_pow(u,4) - 749892.*fcs_pow(u,6) + 1.562275e6*fcs_pow(u,8) - 1.53387e6*fcs_pow(u,10) + 570285.*fcs_pow(u,12))*v*w;
      }
      case 2 : switch(k){
        case 0 : return -6.38512875e8*fcs_pow(t,16)*u*(429.*(-1. + 17.*v2) - 14586.*u2*(-1. + 19.*v2) + 138567.*fcs_pow(u,4)*(-1. + 21.*v2) - 554268.*fcs_pow(u,6)*(-1. + 23.*v2) + 1.062347e6*fcs_pow(u,8)*(-1. + 25.*v2) - 965770.*fcs_pow(u,10)*(-1. + 27.*v2) + 334305.*fcs_pow(u,12)*(-1. + 29.*v2));
        case 1 : return 1.0854718875e10*fcs_pow(t,17)*u*(429.*(-1. + 19.*v2) - 16302.*u2*(-1. + 21.*v2) + 171171.*fcs_pow(u,4)*(-1. + 23.*v2) - 749892.*fcs_pow(u,6)*(-1. + 25.*v2) + 1.562275e6*fcs_pow(u,8)*(-1. + 27.*v2) - 1.53387e6*fcs_pow(u,10)*(-1. + 29.*v2) + 570285.*fcs_pow(u,12)*(-1. + 31.*v2))*w;
        case 2 : return -1.0854718875e10*fcs_pow(t,18)*u*(429.*(1. - 19.*w2 + 19.*v2*(-1. + 21.*w2)) - 16302.*u2*(1. - 21.*w2 + 21.*v2*(-1. + 23.*w2)) + 171171.*fcs_pow(u,4)*(1. - 23.*w2 + 23.*v2*(-1. + 25.*w2)) - 749892.*fcs_pow(u,6)*(1. - 25.*w2 + 25.*v2*(-1. + 27.*w2)) + 1.562275e6*fcs_pow(u,8)*(1. - 27.*w2 + 27.*v2*(-1. + 29.*w2)) - 1.53387e6*fcs_pow(u,10)*(1. - 29.*w2 + 29.*v2*(-1. + 31.*w2)) + 570285.*fcs_pow(u,12)*(1. - 31.*w2 + 31.*v2*(-1. + 33.*w2)));
      }
      case 3 : switch(k){
        case 0 : return 3.2564156625e10*fcs_pow(t,17)*u*v*(-16302.*u2*(-1. + 7.*v2) + 1.562275e6*fcs_pow(u,8)*(-1. + 9.*v2) + 143.*(-3. + 19.*v2) + 57057.*fcs_pow(u,4)*(-3. + 23.*v2) - 249964.*fcs_pow(u,6)*(-3. + 25.*v2) - 511290.*fcs_pow(u,10)*(-3. + 29.*v2) + 190095.*fcs_pow(u,12)*(-3. + 31.*v2));
        case 1 : return -1.856156927625e12*fcs_pow(t,18)*u*v*(143.*(-1. + 7.*v2) - 328900.*fcs_pow(u,6)*(-1. + 9.*v2) + 310155.*fcs_pow(u,12)*(-1. + 11.*v2) - 2002.*u2*(-3. + 23.*v2) + 23023.*fcs_pow(u,4)*(-3. + 25.*v2) + 246675.*fcs_pow(u,8)*(-3. + 29.*v2) - 260130.*fcs_pow(u,10)*(-3. + 31.*v2))*w;
        case 2 : return 1.856156927625e12*fcs_pow(t,19)*u*v*(143.*(1. - 21.*w2 + 7.*v2*(-1. + 23.*w2)) - 2002.*u2*(3. - 69.*w2 + 23.*v2*(-1. + 25.*w2)) + 23023.*fcs_pow(u,4)*(3. - 75.*w2 + 25.*v2*(-1. + 27.*w2)) - 328900.*fcs_pow(u,6)*(1. - 27.*w2 + 9.*v2*(-1. + 29.*w2)) + 246675.*fcs_pow(u,8)*(3. - 87.*w2 + 29.*v2*(-1. + 31.*w2)) - 260130.*fcs_pow(u,10)*(3. - 93.*w2 + 31.*v2*(-1. + 33.*w2)) + 310155.*fcs_pow(u,12)*(1. - 33.*w2 + 11.*v2*(-1. + 35.*w2)));
        case 3 : return -1.856156927625e12*fcs_pow(t,20)*u*v*w*(-138138.*u2*(3. - 25.*w2 + 25.*v2*(-1. + 9.*w2)) + 2.1460725e7*fcs_pow(u,8)*(3. - 31.*w2 + 31.*v2*(-1. + 11.*w2)) + 1001.*(9. - 69.*w2 + 23.*v2*(-3. + 25.*w2)) - 2.9601e6*fcs_pow(u,6)*(9. - 87.*w2 + 29.*v2*(-3. + 31.*w2)) - 2.419209e7*fcs_pow(u,10)*(3. - 33.*w2 + 11.*v2*(-3. + 35.*w2)) + 3.411705e6*fcs_pow(u,12)*(9. - 105.*w2 + 35.*v2*(-3. + 37.*w2)) + 5.180175e6*fcs_pow(u,4)*(1. - 9.*w2 + v2*(-9. + 87.*w2)));
      }
      case 4 : switch(k){
        case 0 : return -3.2564156625e10*fcs_pow(t,18)*u*(429.*(1. - 38.*v2 + 133.*fcs_pow(v,4)) - 16302.*u2*(1. - 42.*v2 + 161.*fcs_pow(v,4)) - 749892.*fcs_pow(u,6)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) + 1.562275e6*fcs_pow(u,8)*(1. - 54.*v2 + 261.*fcs_pow(v,4)) + 570285.*fcs_pow(u,12)*(1. - 62.*v2 + 341.*fcs_pow(v,4)) + 57057.*fcs_pow(u,4)*(3. - 138.*v2 + 575.*fcs_pow(v,4)) - 511290.*fcs_pow(u,10)*(3. - 174.*v2 + 899.*fcs_pow(v,4)));
        case 1 : return 1.856156927625e12*fcs_pow(t,19)*u*(143.*(1. - 42.*v2 + 161.*fcs_pow(v,4)) + 69069.*fcs_pow(u,4)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) - 328900.*fcs_pow(u,6)*(1. - 54.*v2 + 261.*fcs_pow(v,4)) - 780390.*fcs_pow(u,10)*(1. - 62.*v2 + 341.*fcs_pow(v,4)) + 310155.*fcs_pow(u,12)*(1. - 66.*v2 + 385.*fcs_pow(v,4)) - 2002.*u2*(3. - 138.*v2 + 575.*fcs_pow(v,4)) + 246675.*fcs_pow(u,8)*(3. - 174.*v2 + 899.*fcs_pow(v,4)))*w;
        case 2 : return -1.856156927625e12*fcs_pow(t,20)*u*(143.*(-1. + 21.*w2 + v2*(42. - 966.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) - 2002.*u2*(-3. + 69.*w2 - 138.*v2*(-1. + 25.*w2) + 575.*fcs_pow(v,4)*(-1. + 27.*w2)) + 69069.*fcs_pow(u,4)*(-1. + 25.*w2 - 50.*v2*(-1. + 27.*w2) + 225.*fcs_pow(v,4)*(-1. + 29.*w2)) - 328900.*fcs_pow(u,6)*(-1. + 27.*w2 - 54.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)) + 246675.*fcs_pow(u,8)*(-3. + 87.*w2 - 174.*v2*(-1. + 31.*w2) + 899.*fcs_pow(v,4)*(-1. + 33.*w2)) - 780390.*fcs_pow(u,10)*(-1. + 31.*w2 - 62.*v2*(-1. + 33.*w2) + 341.*fcs_pow(v,4)*(-1. + 35.*w2)) + 310155.*fcs_pow(u,12)*(-1. + 33.*w2 - 66.*v2*(-1. + 35.*w2) + 385.*fcs_pow(v,4)*(-1. + 37.*w2)));
        case 3 : return 5.568470782875e12*fcs_pow(t,21)*u*w*(1001.*(-3. + 23.*w2 + 575.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) - 46046.*u2*(-3. + 25.*w2 - 150.*v2*(-1. + 9.*w2) + 225.*fcs_pow(v,4)*(-3. + 29.*w2)) - 2.9601e6*fcs_pow(u,6)*(-3. + 29.*w2 + 899.*fcs_pow(v,4)*(-1. + 11.*w2) - 58.*v2*(-3. + 31.*w2)) + 1.726725e6*fcs_pow(u,4)*(-1. + 9.*w2 + v2*(54. - 522.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)) + 7.153575e6*fcs_pow(u,8)*(-3. + 31.*w2 - 186.*v2*(-1. + 11.*w2) + 341.*fcs_pow(v,4)*(-3. + 35.*w2)) + 3.411705e6*fcs_pow(u,12)*(-3. + 35.*w2 + 1295.*fcs_pow(v,4)*(-1. + 13.*w2) - 70.*v2*(-3. + 37.*w2)) - 8.06403e6*fcs_pow(u,10)*(-3. + 33.*w2 - 66.*v2*(-3. + 35.*w2) + 385.*fcs_pow(v,4)*(-3. + 37.*w2)));
        case 4 : return -5.568470782875e12*fcs_pow(t,22)*u*(1001.*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 138.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) - 2.9601e6*fcs_pow(u,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4) - 174.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4))) + 3.411705e6*fcs_pow(u,12)*(3. - 210.*w2 + 1295.*fcs_pow(w,4) - 210.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 1295.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4))) + 1.726725e6*fcs_pow(u,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 138138.*u2*(1. - 50.*w2 + 225.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 75.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 2.419209e7*fcs_pow(u,10)*(1. - 66.*w2 + 385.*fcs_pow(w,4) + 385.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 22.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 7.153575e6*fcs_pow(u,8)*(3.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 186.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return 1.856156927625e12*fcs_pow(t,19)*u*v*(1.550775e6*fcs_pow(u,12)*(1. - 22.*v2 + 77.*fcs_pow(v,4)) - 10010.*u2*(3. - 46.*v2 + 115.*fcs_pow(v,4)) + 115115.*fcs_pow(u,4)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) + 143.*(5. - 70.*v2 + 161.*fcs_pow(v,4)) - 328900.*fcs_pow(u,6)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) + 246675.*fcs_pow(u,8)*(15. - 290.*v2 + 899.*fcs_pow(v,4)) - 260130.*fcs_pow(u,10)*(15. - 310.*v2 + 1023.*fcs_pow(v,4)));
        case 1 : return -9.280784638125e12*fcs_pow(t,20)*u*v*(-2.419209e7*fcs_pow(u,10)*(1. - 22.*v2 + 77.*fcs_pow(v,4)) + 1001.*(3. - 46.*v2 + 115.*fcs_pow(v,4)) - 46046.*u2*(3. - 50.*v2 + 135.*fcs_pow(v,4)) + 3.411705e6*fcs_pow(u,12)*(3. - 70.*v2 + 259.*fcs_pow(v,4)) + 345345.*fcs_pow(u,4)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) - 592020.*fcs_pow(u,6)*(15. - 290.*v2 + 899.*fcs_pow(v,4)) + 1.430715e6*fcs_pow(u,8)*(15. - 310.*v2 + 1023.*fcs_pow(v,4)))*w;
        case 2 : return 9.280784638125e12*fcs_pow(t,21)*u*v*(1001.*(-3. + 69.*w2 - 46.*v2*(-1. + 25.*w2) + 115.*fcs_pow(v,4)*(-1. + 27.*w2)) - 46046.*u2*(-3. + 75.*w2 - 50.*v2*(-1. + 27.*w2) + 135.*fcs_pow(v,4)*(-1. + 29.*w2)) + 345345.*fcs_pow(u,4)*(-5. + 135.*w2 - 90.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)) - 592020.*fcs_pow(u,6)*(-15. + 435.*w2 - 290.*v2*(-1. + 31.*w2) + 899.*fcs_pow(v,4)*(-1. + 33.*w2)) + 1.430715e6*fcs_pow(u,8)*(-15. + 465.*w2 - 310.*v2*(-1. + 33.*w2) + 1023.*fcs_pow(v,4)*(-1. + 35.*w2)) - 2.419209e7*fcs_pow(u,10)*(-1. + 33.*w2 + v2*(22. - 770.*w2) + 77.*fcs_pow(v,4)*(-1. + 37.*w2)) + 3.411705e6*fcs_pow(u,12)*(-3. + 105.*w2 - 70.*v2*(-1. + 37.*w2) + 259.*fcs_pow(v,4)*(-1. + 39.*w2)));
        case 3 : return -7.044115540336875e15*fcs_pow(t,22)*u*v*w*(91.*(-3. + 25.*w2 - 50.*v2*(-1. + 9.*w2) + 45.*fcs_pow(v,4)*(-3. + 29.*w2)) + 4095.*fcs_pow(u,4)*(-45. + 435.*w2 + 2697.*fcs_pow(v,4)*(-1. + 11.*w2) - 290.*v2*(-3. + 31.*w2)) - 2730.*u2*(-5. + 45.*w2 + v2*(90. - 870.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)) - 67860.*fcs_pow(u,6)*(-310.*v2*(-1. + 11.*w2) + 5.*(-3. + 31.*w2) + 341.*fcs_pow(v,4)*(-3. + 35.*w2)) - 350610.*fcs_pow(u,10)*(-9. + 105.*w2 + 777.*fcs_pow(v,4)*(-1. + 13.*w2) - 70.*v2*(-3. + 37.*w2)) + 876525.*fcs_pow(u,8)*(-3. + 33.*w2 + v2*(66. - 770.*w2) + 77.*fcs_pow(v,4)*(-3. + 37.*w2)) + 94395.*fcs_pow(u,12)*(-370.*v2*(-1. + 13.*w2) + 5.*(-3. + 37.*w2) + 481.*fcs_pow(v,4)*(-3. + 41.*w2)));
        case 4 : return 7.044115540336875e15*fcs_pow(t,23)*u*v*(-1.05183e6*fcs_pow(u,10)*(3. - 210.*w2 + 1295.*fcs_pow(w,4) - 70.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 259.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4))) + 12285.*fcs_pow(u,4)*(-290.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 2730.*u2*(5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 30.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 91.*(3. - 150.*w2 + 675.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 45.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 876525.*fcs_pow(u,8)*(3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 231.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 22.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 67860.*fcs_pow(u,6)*(15.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 310.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 94395.*fcs_pow(u,12)*(15.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 370.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -1.056617331050531e17*fcs_pow(t,24)*u*v*w*(118755.*fcs_pow(u,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 310.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4))) - 2730.*u2*(45. - 870.*w2 + 2697.*fcs_pow(w,4) + 2697.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 58.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 91.*(5.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) - 30.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 87.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 642785.*fcs_pow(u,8)*(15.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 777.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 70.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 701220.*fcs_pow(u,6)*(15.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 110.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 1.164205e6*fcs_pow(u,12)*(533.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 26.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 490854.*fcs_pow(u,10)*(75. - 1850.*w2 + 7215.*fcs_pow(w,4) - 370.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
      }
      case 6 : switch(k){
        case 0 : return -9.280784638125e12*fcs_pow(t,20)*u*(143.*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) - 6006.*u2*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 69069.*fcs_pow(u,4)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 780390.*fcs_pow(u,10)*(-1. + 93.*v2 - 1023.*fcs_pow(v,4) + 2387.*fcs_pow(v,6)) + 310155.*fcs_pow(u,12)*(-1. + 99.*v2 - 1155.*fcs_pow(v,4) + 2849.*fcs_pow(v,6)) - 65780.*fcs_pow(u,6)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 148005.*fcs_pow(u,8)*(-5. + 435.*v2 - 4495.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)));
        case 1 : return 2.7842353914375e13*fcs_pow(t,21)*u*(1001.*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 46046.*u2*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 7.153575e6*fcs_pow(u,8)*(-1. + 93.*v2 - 1023.*fcs_pow(v,4) + 2387.*fcs_pow(v,6)) - 8.06403e6*fcs_pow(u,10)*(-1. + 99.*v2 - 1155.*fcs_pow(v,4) + 2849.*fcs_pow(v,6)) + 3.411705e6*fcs_pow(u,12)*(-1. + 105.*v2 - 1295.*fcs_pow(v,4) + 3367.*fcs_pow(v,6)) + 115115.*fcs_pow(u,4)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) - 592020.*fcs_pow(u,6)*(-5. + 435.*v2 - 4495.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)))*w;
        case 2 : return -2.7842353914375e13*fcs_pow(t,22)*u*(1001.*(1. - 23.*w2 + 69.*v2*(-1. + 25.*w2) - 575.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) - 46046.*u2*(1. - 25.*w2 + 75.*v2*(-1. + 27.*w2) - 675.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) + 115115.*fcs_pow(u,4)*(5. - 135.*w2 + 405.*v2*(-1. + 29.*w2) - 3915.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) - 592020.*fcs_pow(u,6)*(5. - 145.*w2 + 435.*v2*(-1. + 31.*w2) - 4495.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)) + 7.153575e6*fcs_pow(u,8)*(1. - 31.*w2 + 93.*v2*(-1. + 33.*w2) - 1023.*fcs_pow(v,4)*(-1. + 35.*w2) + 2387.*fcs_pow(v,6)*(-1. + 37.*w2)) - 8.06403e6*fcs_pow(u,10)*(1. - 33.*w2 + 99.*v2*(-1. + 35.*w2) - 1155.*fcs_pow(v,4)*(-1. + 37.*w2) + 2849.*fcs_pow(v,6)*(-1. + 39.*w2)) + 3.411705e6*fcs_pow(u,12)*(1. - 35.*w2 + 105.*v2*(-1. + 37.*w2) - 1295.*fcs_pow(v,4)*(-1. + 39.*w2) + 3367.*fcs_pow(v,6)*(-1. + 41.*w2)));
        case 3 : return 7.044115540336875e15*fcs_pow(t,23)*u*w*(-2730.*u2*(5. - 45.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 135.*v2*(-3. + 29.*w2) - 1305.*fcs_pow(v,4)*(-3. + 31.*w2)) + 91.*(3. - 25.*w2 + 225.*v2*(-1. + 9.*w2) - 675.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)) + 12285.*fcs_pow(u,4)*(15. - 145.*w2 - 13485.*fcs_pow(v,4)*(-1. + 11.*w2) + 435.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) + 2.629575e6*fcs_pow(u,8)*(1. - 11.*w2 + 2849.*fcs_pow(v,6)*(-1. + 13.*w2) + 33.*v2*(-3. + 35.*w2) - 385.*fcs_pow(v,4)*(-3. + 37.*w2)) - 339300.*fcs_pow(u,6)*(3. - 31.*w2 + 279.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 2387.*fcs_pow(v,6)*(-3. + 37.*w2)) - 1.05183e6*fcs_pow(u,10)*(3. - 35.*w2 - 3885.*fcs_pow(v,4)*(-1. + 13.*w2) + 105.*v2*(-3. + 37.*w2) + 3367.*fcs_pow(v,6)*(-3. + 41.*w2)) + 94395.*fcs_pow(u,12)*(15. - 185.*w2 + 1665.*v2*(-1. + 13.*w2) - 7215.*fcs_pow(v,4)*(-3. + 41.*w2) + 19721.*fcs_pow(v,6)*(-3. + 43.*w2)));
        case 4 : return -2.113234662101062e16*fcs_pow(t,24)*u*(-910.*u2*(-5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 3915.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 135.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 91.*(-1. + 50.*w2 - 225.*fcs_pow(w,4) + 75.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 225.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 876525.*fcs_pow(u,8)*(-1. + 66.*w2 - 385.*fcs_pow(w,4) - 1155.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 2849.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 33.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 339300.*fcs_pow(u,6)*(-1. + 62.*w2 - 341.*fcs_pow(w,4) + 93.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 4095.*fcs_pow(u,4)*(1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 13485.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 94395.*fcs_pow(u,12)*(-5.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 555.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 2405.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 350610.*fcs_pow(u,10)*(-3. + 210.*w2 - 1295.*fcs_pow(w,4) + 315.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 3885.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 3367.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return 5.283086655252656e17*fcs_pow(t,25)*u*w*(-1638.*u2*(-15. + 290.*w2 - 899.*fcs_pow(w,4) - 13485.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 87.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 91.*(-5. + 90.*w2 - 261.*fcs_pow(w,4) + 8091.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 27.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 261.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 420732.*fcs_pow(u,6)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 165.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2849.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 385.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 23751.*fcs_pow(u,4)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 1395.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 5115.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 698523.*fcs_pow(u,12)*(-5. + 130.*w2 - 533.*fcs_pow(w,4) + 22919.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 2665.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 39.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 490854.*fcs_pow(u,10)*(-15. + 370.*w2 - 1443.*fcs_pow(w,4) + 19721.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 333.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 1443.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 385671.*fcs_pow(u,8)*(-5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 3885.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 3367.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
        case 6 : return -5.283086655252656e17*fcs_pow(t,26)*u*(-4914.*u2*(5. - 435.*w2 + 4495.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) + 435.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 4495.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6))) + 698523.*fcs_pow(u,12)*(5. - 585.*w2 + 7995.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 585.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 7995.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 22919.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6))) + 91.*(5. - 405.*w2 + 3915.*fcs_pow(w,4) - 8091.*fcs_pow(w,6) - 3915.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 81.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 71253.*fcs_pow(u,4)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 5115.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 2387.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 1.472562e6*fcs_pow(u,10)*(5. - 555.*w2 + 7215.*fcs_pow(w,4) - 19721.*fcs_pow(w,6) - 7215.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 111.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 5.785065e6*fcs_pow(u,8)*(1. - 105.*w2 + 1295.*fcs_pow(w,4) - 3367.*fcs_pow(w,6) + 3367.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 21.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 420732.*fcs_pow(u,6)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 1155.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2849.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))));
      }
      case 7 : switch(k){
        case 0 : return 2.7842353914375e13*fcs_pow(t,21)*u*v*(-5.644821e7*fcs_pow(u,10)*(-1. + 33.*v2 - 231.*fcs_pow(v,4) + 407.*fcs_pow(v,6)) + 2.3881935e7*fcs_pow(u,12)*(-1. + 35.*v2 - 259.*fcs_pow(v,4) + 481.*fcs_pow(v,6)) + 1001.*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 46046.*u2*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 1.0015005e7*fcs_pow(u,8)*(-5. + 155.*v2 - 1023.*fcs_pow(v,4) + 1705.*fcs_pow(v,6)) + 115115.*fcs_pow(u,4)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) - 592020.*fcs_pow(u,6)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)));
        case 1 : return -4.930880878235812e16*fcs_pow(t,22)*u*v*(876525.*fcs_pow(u,8)*(-1. + 33.*v2 - 231.*fcs_pow(v,4) + 407.*fcs_pow(v,6)) - 1.05183e6*fcs_pow(u,10)*(-1. + 35.*v2 - 259.*fcs_pow(v,4) + 481.*fcs_pow(v,6)) + 13.*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 67860.*fcs_pow(u,6)*(-5. + 155.*v2 - 1023.*fcs_pow(v,4) + 1705.*fcs_pow(v,6)) - 130.*u2*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 1755.*fcs_pow(u,4)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)) + 13485.*fcs_pow(u,12)*(-35. + 1295.*v2 - 10101.*fcs_pow(v,4) + 19721.*fcs_pow(v,6)))*w;
        case 2 : return 4.930880878235812e16*fcs_pow(t,23)*u*v*(13.*(7. - 175.*w2 + 175.*v2*(-1. + 27.*w2) - 945.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) - 130.*u2*(35. - 945.*w2 + 945.*v2*(-1. + 29.*w2) - 5481.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) + 1755.*fcs_pow(u,4)*(35. - 1015.*w2 + 1015.*v2*(-1. + 31.*w2) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)) - 67860.*fcs_pow(u,6)*(5. - 155.*w2 + 155.*v2*(-1. + 33.*w2) - 1023.*fcs_pow(v,4)*(-1. + 35.*w2) + 1705.*fcs_pow(v,6)*(-1. + 37.*w2)) + 876525.*fcs_pow(u,8)*(1. - 33.*w2 + 33.*v2*(-1. + 35.*w2) - 231.*fcs_pow(v,4)*(-1. + 37.*w2) + 407.*fcs_pow(v,6)*(-1. + 39.*w2)) - 1.05183e6*fcs_pow(u,10)*(1. - 35.*w2 + 35.*v2*(-1. + 37.*w2) - 259.*fcs_pow(v,4)*(-1. + 39.*w2) + 481.*fcs_pow(v,6)*(-1. + 41.*w2)) + 13485.*fcs_pow(u,12)*(35. - 1295.*w2 + 1295.*v2*(-1. + 39.*w2) - 10101.*fcs_pow(v,4)*(-1. + 41.*w2) + 19721.*fcs_pow(v,6)*(-1. + 43.*w2)));
        case 3 : return -7.396321317353719e17*fcs_pow(t,24)*u*v*w*(13.*(35. - 315.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 315.*v2*(-3. + 29.*w2) - 1827.*fcs_pow(v,4)*(-3. + 31.*w2)) - 234.*u2*(35.*(3. - 29.*w2) - 18879.*fcs_pow(v,4)*(-1. + 11.*w2) + 1015.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) - 2.10366e6*fcs_pow(u,6)*(1. - 11.*w2 + 407.*fcs_pow(v,6)*(-1. + 13.*w2) + 11.*v2*(-3. + 35.*w2) - 77.*fcs_pow(v,4)*(-3. + 37.*w2)) + 23751.*fcs_pow(u,4)*(15. - 155.*w2 + 465.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 1705.*fcs_pow(v,6)*(-3. + 37.*w2)) + 1.928355e6*fcs_pow(u,8)*(3. - 35.*w2 - 777.*fcs_pow(v,4)*(-1. + 13.*w2) + 35.*v2*(-3. + 37.*w2) + 481.*fcs_pow(v,6)*(-3. + 41.*w2)) + 99789.*fcs_pow(u,12)*(35. - 455.*w2 + 22919.*fcs_pow(v,6)*(-1. + 15.*w2) + 455.*v2*(-3. + 41.*w2) - 3731.*fcs_pow(v,4)*(-3. + 43.*w2)) - 70122.*fcs_pow(u,10)*(35.*(3. - 37.*w2) + 3885.*v2*(-1. + 13.*w2) - 10101.*fcs_pow(v,4)*(-3. + 41.*w2) + 19721.*fcs_pow(v,6)*(-3. + 43.*w2)));
        case 4 : return 7.396321317353719e17*fcs_pow(t,25)*u*v*(13.*(-35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 5481.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 2.10366e6*fcs_pow(u,6)*(-1. + 66.*w2 - 385.*fcs_pow(w,4) - 231.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 11.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 71253.*fcs_pow(u,4)*(-5.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 155.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 1705.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 234.*u2*(3045.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18879.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 35.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 99789.*fcs_pow(u,12)*(-35.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 11193.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 22919.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 455.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 210366.*fcs_pow(u,10)*(-35.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 1295.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 3367.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 1.928355e6*fcs_pow(u,8)*(-3. + 210.*w2 - 1295.*fcs_pow(w,4) + 105.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 777.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -1.442282656883975e20*fcs_pow(t,26)*u*v*w*(3.*(-18879.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 203.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 56637.*fcs_pow(u,4)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 55.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 1218.*u2*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 465.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 1023.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 341.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 199578.*fcs_pow(u,10)*(22919.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 3731.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 91.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 9889.*fcs_pow(u,8)*(98605.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3885.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 35.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 10101.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 118668.*fcs_pow(u,6)*(-5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 777.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 35.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 481.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 33263.*fcs_pow(u,12)*(-37023.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1435.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 5289.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))));
        case 6 : return 4.326847970651925e20*fcs_pow(t,27)*u*v*(3.*(1015.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 6293.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 7.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) - 1218.*u2*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 155.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 1023.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 66526.*fcs_pow(u,10)*(1365.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 11193.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 22919.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 49445.*fcs_pow(u,8)*(-10101.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 259.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 118668.*fcs_pow(u,6)*(5. - 525.*w2 + 6475.*fcs_pow(w,4) - 16835.*fcs_pow(w,6) + 2405.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 35.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 18879.*fcs_pow(u,4)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 33263.*fcs_pow(u,12)*(-35.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1435.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 12341.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 5289.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))));
        case 7 : return -1.254785911489058e22*fcs_pow(t,28)*u*v*w*(-626262.*fcs_pow(u,10)*(35. - 1435.*w2 + 12341.*fcs_pow(w,4) - 26445.*fcs_pow(w,6) - 12341.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 5289.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 205.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6))) + 329189.*fcs_pow(u,12)*(1505.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 2709.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 5.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 6063.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6))) + 21.*(7.*(5. - 155.*w2 + 1023.*fcs_pow(w,4) - 1705.*fcs_pow(w,6)) + 1085.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 7161.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 1302.*u2*(-35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 1155.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 441595.*fcs_pow(u,8)*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 22919.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1599.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 39.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 4092.*fcs_pow(u,6)*(98605.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 35.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 10101.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 21483.*fcs_pow(u,4)*(-35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 35.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 481.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))));
      }
      case 8 : switch(k){
        case 0 : return -1.94896477400625e14*fcs_pow(t,22)*u*(7.153575e6*fcs_pow(u,8)*(1. - 124.*v2 + 2046.*fcs_pow(v,4) - 9548.*fcs_pow(v,6) + 12617.*fcs_pow(v,8)) - 8.06403e6*fcs_pow(u,10)*(1. - 132.*v2 + 2310.*fcs_pow(v,4) - 11396.*fcs_pow(v,6) + 15873.*fcs_pow(v,8)) + 3.411705e6*fcs_pow(u,12)*(1. - 140.*v2 + 2590.*fcs_pow(v,4) - 13468.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) + 143.*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) - 6578.*u2*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) - 592020.*fcs_pow(u,6)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) + 16445.*fcs_pow(u,4)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)));
        case 1 : return 4.930880878235812e16*fcs_pow(t,23)*u*(-339300.*fcs_pow(u,6)*(1. - 124.*v2 + 2046.*fcs_pow(v,4) - 9548.*fcs_pow(v,6) + 12617.*fcs_pow(v,8)) + 876525.*fcs_pow(u,8)*(1. - 132.*v2 + 2310.*fcs_pow(v,4) - 11396.*fcs_pow(v,6) + 15873.*fcs_pow(v,8)) - 1.05183e6*fcs_pow(u,10)*(1. - 140.*v2 + 2590.*fcs_pow(v,4) - 13468.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) + 13.*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) + 12285.*fcs_pow(u,4)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) - 130.*u2*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)) + 13485.*fcs_pow(u,12)*(35. - 5180.*v2 + 101010.*fcs_pow(v,4) - 552188.*fcs_pow(v,6) + 848003.*fcs_pow(v,8)))*w;
        case 2 : return -4.930880878235812e16*fcs_pow(t,24)*u*(13.*(-7. + 175.*w2 - 700.*v2*(-1. + 27.*w2) + 9450.*fcs_pow(v,4)*(-1. + 29.*w2) - 36540.*fcs_pow(v,6)*(-1. + 31.*w2) + 40455.*fcs_pow(v,8)*(-1. + 33.*w2)) - 130.*u2*(-35. + 945.*w2 - 3780.*v2*(-1. + 29.*w2) + 54810.*fcs_pow(v,4)*(-1. + 31.*w2) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2)) + 12285.*fcs_pow(u,4)*(-5. + 145.*w2 - 580.*v2*(-1. + 31.*w2) + 8990.*fcs_pow(v,4)*(-1. + 33.*w2) - 39556.*fcs_pow(v,6)*(-1. + 35.*w2) + 49445.*fcs_pow(v,8)*(-1. + 37.*w2)) - 339300.*fcs_pow(u,6)*(-1. + 31.*w2 - 124.*v2*(-1. + 33.*w2) + 2046.*fcs_pow(v,4)*(-1. + 35.*w2) - 9548.*fcs_pow(v,6)*(-1. + 37.*w2) + 12617.*fcs_pow(v,8)*(-1. + 39.*w2)) + 876525.*fcs_pow(u,8)*(-1. + 33.*w2 - 132.*v2*(-1. + 35.*w2) + 2310.*fcs_pow(v,4)*(-1. + 37.*w2) - 11396.*fcs_pow(v,6)*(-1. + 39.*w2) + 15873.*fcs_pow(v,8)*(-1. + 41.*w2)) - 1.05183e6*fcs_pow(u,10)*(-1. + 35.*w2 - 140.*v2*(-1. + 37.*w2) + 2590.*fcs_pow(v,4)*(-1. + 39.*w2) - 13468.*fcs_pow(v,6)*(-1. + 41.*w2) + 19721.*fcs_pow(v,8)*(-1. + 43.*w2)) + 13485.*fcs_pow(u,12)*(35.*(-1. + 37.*w2) - 5180.*v2*(-1. + 39.*w2) + 101010.*fcs_pow(v,4)*(-1. + 41.*w2) - 552188.*fcs_pow(v,6)*(-1. + 43.*w2) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2)));
        case 3 : return 7.396321317353719e17*fcs_pow(t,25)*u*w*(13.*(35.*(-1. + 9.*w2) - 226548.*fcs_pow(v,6)*(-1. + 11.*w2) - 1260.*v2*(-3. + 29.*w2) + 18270.*fcs_pow(v,4)*(-3. + 31.*w2) + 89001.*fcs_pow(v,8)*(-3. + 35.*w2)) + 118755.*fcs_pow(u,4)*(-3. + 31.*w2 - 372.*v2*(-1. + 11.*w2) + 37851.*fcs_pow(v,8)*(-1. + 13.*w2) + 2046.*fcs_pow(v,4)*(-3. + 35.*w2) - 9548.*fcs_pow(v,6)*(-3. + 37.*w2)) - 1638.*u2*(26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 5.*(-3. + 29.*w2) - 580.*v2*(-3. + 31.*w2) - 39556.*fcs_pow(v,6)*(-3. + 35.*w2) + 49445.*fcs_pow(v,8)*(-3. + 37.*w2)) - 2.10366e6*fcs_pow(u,6)*(-1. + 11.*w2 - 11396.*fcs_pow(v,6)*(-1. + 13.*w2) - 44.*v2*(-3. + 35.*w2) + 770.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)) - 70122.*fcs_pow(u,10)*(-15540.*v2*(-1. + 13.*w2) + 2.544009e6*fcs_pow(v,8)*(-1. + 15.*w2) + 35.*(-3. + 37.*w2) + 101010.*fcs_pow(v,4)*(-3. + 41.*w2) - 552188.*fcs_pow(v,6)*(-3. + 43.*w2)) + 1.928355e6*fcs_pow(u,8)*(-3. + 35.*w2 + 7770.*fcs_pow(v,4)*(-1. + 13.*w2) - 140.*v2*(-3. + 37.*w2) - 13468.*fcs_pow(v,6)*(-3. + 41.*w2) + 19721.*fcs_pow(v,8)*(-3. + 43.*w2)) + 99789.*fcs_pow(u,12)*(-35. + 455.*w2 - 641732.*fcs_pow(v,6)*(-1. + 15.*w2) - 1820.*v2*(-3. + 41.*w2) + 37310.*fcs_pow(v,4)*(-3. + 43.*w2) + 343785.*fcs_pow(v,8)*(-3. + 47.*w2)));
        case 4 : return -7.396321317353719e17*fcs_pow(t,26)*u*(356265.*fcs_pow(u,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4) - 124.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 682.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 1638.*u2*(-1740.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 148335.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 13.*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 54810.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 226548.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 1260.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 210366.*fcs_pow(u,10)*(35.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 5180.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 552188.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 33670.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 1.928355e6*fcs_pow(u,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4) - 420.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 7770.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59163.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 13468.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 2.10366e6*fcs_pow(u,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4) + 2310.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 44.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 99789.*fcs_pow(u,12)*(35.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 111930.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 641732.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 1820.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 343785.*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return 1.442282656883975e20*fcs_pow(t,27)*u*w*(-1218.*u2*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 10230.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 37851.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 21.*(15. - 290.*w2 + 899.*fcs_pow(w,4) + 26970.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 116.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 49445.*fcs_pow(u,8)*(2.544009e6*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 552188.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 3108.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 7.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 20202.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 118668.*fcs_pow(u,6)*(5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 140.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 13468.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 56637.*fcs_pow(u,4)*(5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 220.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 770.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 199578.*fcs_pow(u,10)*(-641732.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 37310.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 364.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 68757.*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 33263.*fcs_pow(u,12)*(370230.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 5740.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 148092.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 248583.*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return -4.326847970651925e20*fcs_pow(t,28)*u*(21.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) - 580.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8990.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 39556.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 49445.*fcs_pow(u,8)*(101010.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 552188.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 1036.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 593340.*fcs_pow(u,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6) - 13468.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 28.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 518.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 18879.*fcs_pow(u,4)*(5.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 79365.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 1218.*u2*(5.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 620.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 66526.*fcs_pow(u,10)*(-5460.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 111930.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 641732.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 206271.*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 33263.*fcs_pow(u,12)*(35.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 5740.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 123410.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 148092.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 248583.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return 8.783501380423409e22*fcs_pow(t,29)*u*w*(-89466.*fcs_pow(u,10)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6) + 123410.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 148092.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 820.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 248583.*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6))) + 3.*(-4340.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 71610.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 7.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 63085.*fcs_pow(u,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6) - 641732.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 206271.*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 15990.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 156.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 20460.*fcs_pow(u,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 78884.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 148.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 2886.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 3069.*fcs_pow(u,4)*(35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 140.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2590.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 13468.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 186.*u2*(35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 15873.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 47027.*fcs_pow(u,12)*(-6020.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 27090.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 5.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 169764.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 42441.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return -8.783501380423409e22*fcs_pow(t,30)*u*(63085.*fcs_pow(u,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8) + 111930.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 641732.*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 780.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 206271.*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8))) - 89466.*fcs_pow(u,10)*(-5740.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 123410.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 5.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 148092.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 248583.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) + 329189.*fcs_pow(u,12)*(5.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 860.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 3870.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 24252.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 30315.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) - 20460.*fcs_pow(u,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8) - 552188.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 14430.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 148.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 107415.*fcs_pow(u,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8) + 19721.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 1924.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 4.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 74.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 186.*u2*(35.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 79365.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 11396.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 3.*(35.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 4340.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 9548.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
      }
      case 9 : switch(k){
        case 0 : return 1.479264263470744e17*fcs_pow(t,23)*u*v*(876525.*fcs_pow(u,8)*(3. - 132.*v2 + 1386.*fcs_pow(v,4) - 4884.*fcs_pow(v,6) + 5291.*fcs_pow(v,8)) + 13.*(21. - 700.*v2 + 5670.*fcs_pow(v,4) - 15660.*fcs_pow(v,6) + 13485.*fcs_pow(v,8)) - 350610.*fcs_pow(u,10)*(9. - 420.*v2 + 4662.*fcs_pow(v,4) - 17316.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) - 390.*u2*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) - 22620.*fcs_pow(u,6)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) + 585.*fcs_pow(u,4)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)) + 4495.*fcs_pow(u,12)*(315. - 15540.*v2 + 181818.*fcs_pow(v,4) - 709956.*fcs_pow(v,6) + 848003.*fcs_pow(v,8)));
        case 1 : return -2.218896395206116e18*fcs_pow(t,24)*u*v*(-701220.*fcs_pow(u,6)*(3. - 132.*v2 + 1386.*fcs_pow(v,4) - 4884.*fcs_pow(v,6) + 5291.*fcs_pow(v,8)) + 642785.*fcs_pow(u,8)*(9. - 420.*v2 + 4662.*fcs_pow(v,4) - 17316.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) + 13.*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) + 7917.*fcs_pow(u,4)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) + 99789.*fcs_pow(u,12)*(35. - 1820.*v2 + 22386.*fcs_pow(v,4) - 91676.*fcs_pow(v,6) + 114595.*fcs_pow(v,8)) - 78.*u2*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)) - 23374.*fcs_pow(u,10)*(315. - 15540.*v2 + 181818.*fcs_pow(v,4) - 709956.*fcs_pow(v,6) + 848003.*fcs_pow(v,8)))*w;
        case 2 : return 2.218896395206116e18*fcs_pow(t,25)*u*v*(13.*(-35. + 945.*w2 - 1260.*v2*(-1. + 29.*w2) + 10962.*fcs_pow(v,4)*(-1. + 31.*w2) - 32364.*fcs_pow(v,6)*(-1. + 33.*w2) + 29667.*fcs_pow(v,8)*(-1. + 35.*w2)) - 78.*u2*(315.*(-1. + 29.*w2) - 12180.*v2*(-1. + 31.*w2) + 113274.*fcs_pow(v,4)*(-1. + 33.*w2) - 356004.*fcs_pow(v,6)*(-1. + 35.*w2) + 346115.*fcs_pow(v,8)*(-1. + 37.*w2)) + 7917.*fcs_pow(u,4)*(45.*(-1. + 31.*w2) - 1860.*v2*(-1. + 33.*w2) + 18414.*fcs_pow(v,4)*(-1. + 35.*w2) - 61380.*fcs_pow(v,6)*(-1. + 37.*w2) + 63085.*fcs_pow(v,8)*(-1. + 39.*w2)) - 701220.*fcs_pow(u,6)*(-3. + 99.*w2 - 132.*v2*(-1. + 35.*w2) + 1386.*fcs_pow(v,4)*(-1. + 37.*w2) - 4884.*fcs_pow(v,6)*(-1. + 39.*w2) + 5291.*fcs_pow(v,8)*(-1. + 41.*w2)) + 642785.*fcs_pow(u,8)*(-9. + 315.*w2 - 420.*v2*(-1. + 37.*w2) + 4662.*fcs_pow(v,4)*(-1. + 39.*w2) - 17316.*fcs_pow(v,6)*(-1. + 41.*w2) + 19721.*fcs_pow(v,8)*(-1. + 43.*w2)) - 23374.*fcs_pow(u,10)*(315.*(-1. + 37.*w2) - 15540.*v2*(-1. + 39.*w2) + 181818.*fcs_pow(v,4)*(-1. + 41.*w2) - 709956.*fcs_pow(v,6)*(-1. + 43.*w2) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2)) + 99789.*fcs_pow(u,12)*(35.*(-1. + 39.*w2) - 1820.*v2*(-1. + 41.*w2) + 22386.*fcs_pow(v,4)*(-1. + 43.*w2) - 91676.*fcs_pow(v,6)*(-1. + 45.*w2) + 114595.*fcs_pow(v,8)*(-1. + 47.*w2)));
        case 3 : return -8.653695941303851e19*fcs_pow(t,26)*u*v*w*(339822.*fcs_pow(v,4)*(-1. + 11.*w2) + 315.*(-3. + 29.*w2) - 12180.*v2*(-3. + 31.*w2) - 356004.*fcs_pow(v,6)*(-3. + 35.*w2) + 346115.*fcs_pow(v,8)*(-3. + 37.*w2) - 1218.*u2*(-45. + 465.*w2 - 1860.*v2*(-1. + 11.*w2) + 63085.*fcs_pow(v,8)*(-1. + 13.*w2) + 6138.*fcs_pow(v,4)*(-3. + 35.*w2) - 20460.*fcs_pow(v,6)*(-3. + 37.*w2)) + 94395.*fcs_pow(u,4)*(-9. + 99.*w2 - 14652.*fcs_pow(v,6)*(-1. + 13.*w2) - 132.*v2*(-3. + 35.*w2) + 1386.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)) + 49445.*fcs_pow(u,8)*(-15540.*v2*(-1. + 13.*w2) + 848003.*fcs_pow(v,8)*(-1. + 15.*w2) + 105.*(-3. + 37.*w2) + 60606.*fcs_pow(v,4)*(-3. + 41.*w2) - 236652.*fcs_pow(v,6)*(-3. + 43.*w2)) - 197780.*fcs_pow(u,6)*(13986.*fcs_pow(v,4)*(-1. + 13.*w2) + 9.*(-3. + 35.*w2) - 420.*v2*(-3. + 37.*w2) - 17316.*fcs_pow(v,6)*(-3. + 41.*w2) + 19721.*fcs_pow(v,8)*(-3. + 43.*w2)) - 199578.*fcs_pow(u,10)*(105.*(-1. + 13.*w2) - 275028.*fcs_pow(v,6)*(-1. + 15.*w2) - 1820.*v2*(-3. + 41.*w2) + 22386.*fcs_pow(v,4)*(-3. + 43.*w2) + 114595.*fcs_pow(v,8)*(-3. + 47.*w2)) + 33263.*fcs_pow(u,12)*(222138.*fcs_pow(v,4)*(-1. + 15.*w2) + 105.*(-3. + 41.*w2) - 5740.*v2*(-3. + 43.*w2) - 317340.*fcs_pow(v,6)*(-3. + 47.*w2) + 414305.*fcs_pow(v,8)*(-3. + 49.*w2)));
        case 4 : return 8.653695941303851e19*fcs_pow(t,27)*u*v*(-1218.*u2*(45.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 1860.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 61380.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 3.*(-12180.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 113274.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 346115.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 105.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 118668.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 49445.*fcs_pow(u,8)*(315.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 15540.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 709956.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 60606.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 593340.*fcs_pow(u,6)*(9. - 630.*w2 + 3885.*fcs_pow(w,4) - 420.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 4662.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 5772.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 94395.*fcs_pow(u,4)*(9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 4158.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 132.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 99789.*fcs_pow(u,12)*(-5740.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 74046.*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 414305.*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 35.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 105780.*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 199578.*fcs_pow(u,10)*(105.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 67158.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 275028.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 1820.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 114595.*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return -3.764357734467175e22*fcs_pow(t,28)*u*v*w*(7.*(45. - 930.*w2 + 3069.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 4092.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 1364.*fcs_pow(u,6)*(4.240015e6*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 1.18326e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 15540.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 60606.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 2387.*fcs_pow(u,4)*(45.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 13986.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 420.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 17316.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 434.*u2*(45.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 660.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1386.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 63085.*fcs_pow(u,8)*(-275028.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 22386.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 21.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 364.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 22919.*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 47027.*fcs_pow(u,12)*(-6020.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 35.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 99029.*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 5418.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 24252.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 29822.*fcs_pow(u,10)*(222138.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 5740.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 21.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 63468.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 82861.*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return 3.764357734467175e22*fcs_pow(t,29)*u*v*(-6820.*fcs_pow(u,6)*(181818.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 709956.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 63.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 3108.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 7161.*fcs_pow(u,4)*(45.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 86580.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 1302.*u2*(15.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 7.*(45.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 1860.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 18414.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 12276.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 189255.*fcs_pow(u,8)*(-1820.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 22386.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 91676.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 22919.*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 47027.*fcs_pow(u,12)*(105.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 6020.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 495145.*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 16254.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 72756.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 89466.*fcs_pow(u,10)*(105.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 5740.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 74046.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 63468.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 82861.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return -2.450596885138131e25*fcs_pow(t,30)*u*v*w*(105.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 5291.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 26455.*fcs_pow(u,8)*(-105. + 4305.*w2 - 37023.*fcs_pow(w,4) + 79335.*fcs_pow(w,6) + 74046.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 63468.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 820.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 82861.*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6))) - 24420.*fcs_pow(u,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6) - 91676.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 22919.*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 3198.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 52.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 11.*fcs_pow(u,4)*(2.9680105e7*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 3.54978e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 15540.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 181818.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 22.*u2*(315.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 17316.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 456617.*fcs_pow(u,12)*(5.*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 60.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 846.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 3995.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 564.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 39442.*fcs_pow(u,10)*(-6020.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 16254.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 15.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 72756.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 14147.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return 2.450596885138131e25*fcs_pow(t,31)*u*v*(105.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 26455.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 1386.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 4884.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) - 24420.*fcs_pow(u,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8) + 22386.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 91676.*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 260.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8))) + 26455.*fcs_pow(u,8)*(-5740.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 74046.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 15.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 63468.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 82861.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) - 39442.*fcs_pow(u,10)*(105.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 6020.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 16254.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 72756.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 70735.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) + 456617.*fcs_pow(u,12)*(5.*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 60.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 846.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2820.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 3995.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 385.*fcs_pow(u,4)*(-709956.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 25974.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 9.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 444.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 22.*u2*(315.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 690235.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 86580.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 420.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 4662.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -2.450596885138131e25*fcs_pow(t,32)*u*v*w*(-317460.*fcs_pow(u,6)*(666414.*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 2460.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 63468.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 82861.*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 9.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 128205.*fcs_pow(u,4)*(-825084.*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 9594.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 156.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 198.*u2*(2.9680105e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 1.18326e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 15540.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 60606.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 33.*(105.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 98605.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 13986.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 17316.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 140.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 456617.*fcs_pow(u,12)*(45.*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 2820.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 143820.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 211735.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 17766.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 3.5616126e7*fcs_pow(u,10)*(45.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 60.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 846.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 3995.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1692.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 3.253965e6*fcs_pow(u,8)*(-18060.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 15.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 5418.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 24252.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 14147.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
      }
      case 10 : switch(k){
        case 0 : return -4.437792790412231e17*fcs_pow(t,24)*u*(13.*(-7. + 875.*v2 - 15750.*fcs_pow(v,4) + 91350.*fcs_pow(v,6) - 202275.*fcs_pow(v,8) + 148335.*fcs_pow(v,10)) - 113100.*fcs_pow(u,6)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) - 650.*u2*(-7. + 945.*v2 - 18270.*fcs_pow(v,4) + 113274.*fcs_pow(v,6) - 267003.*fcs_pow(v,8) + 207669.*fcs_pow(v,10)) + 292175.*fcs_pow(u,8)*(-3. + 495.*v2 - 11550.*fcs_pow(v,4) + 85470.*fcs_pow(v,6) - 238095.*fcs_pow(v,8) + 216931.*fcs_pow(v,10)) + 6825.*fcs_pow(u,4)*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)) - 116870.*fcs_pow(u,10)*(-9. + 1575.*v2 - 38850.*fcs_pow(v,4) + 303030.*fcs_pow(v,6) - 887445.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 67425.*fcs_pow(u,12)*(-7. + 1295.*v2 - 33670.*fcs_pow(v,4) + 276094.*fcs_pow(v,6) - 848003.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)));
        case 1 : return 1.109448197603058e19*fcs_pow(t,25)*u*(23751.*fcs_pow(u,4)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) + 13.*(-7. + 945.*v2 - 18270.*fcs_pow(v,4) + 113274.*fcs_pow(v,6) - 267003.*fcs_pow(v,8) + 207669.*fcs_pow(v,10)) - 140244.*fcs_pow(u,6)*(-3. + 495.*v2 - 11550.*fcs_pow(v,4) + 85470.*fcs_pow(v,6) - 238095.*fcs_pow(v,8) + 216931.*fcs_pow(v,10)) - 546.*u2*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)) + 128557.*fcs_pow(u,8)*(-9. + 1575.*v2 - 38850.*fcs_pow(v,4) + 303030.*fcs_pow(v,6) - 887445.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) - 210366.*fcs_pow(u,10)*(-7. + 1295.*v2 - 33670.*fcs_pow(v,4) + 276094.*fcs_pow(v,6) - 848003.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 99789.*fcs_pow(u,12)*(-7. + 1365.*v2 - 37310.*fcs_pow(v,4) + 320866.*fcs_pow(v,6) - 1.031355e6*fcs_pow(v,8) + 1.077193e6*fcs_pow(v,10)))*w;
        case 2 : return -1.109448197603058e19*fcs_pow(t,26)*u*(13.*(7. - 189.*w2 + 945.*v2*(-1. + 29.*w2) - 18270.*fcs_pow(v,4)*(-1. + 31.*w2) + 113274.*fcs_pow(v,6)*(-1. + 33.*w2) - 267003.*fcs_pow(v,8)*(-1. + 35.*w2) + 207669.*fcs_pow(v,10)*(-1. + 37.*w2)) - 546.*u2*(9. - 261.*w2 + 1305.*v2*(-1. + 31.*w2) - 26970.*fcs_pow(v,4)*(-1. + 33.*w2) + 178002.*fcs_pow(v,6)*(-1. + 35.*w2) - 445005.*fcs_pow(v,8)*(-1. + 37.*w2) + 365893.*fcs_pow(v,10)*(-1. + 39.*w2)) + 23751.*fcs_pow(u,4)*(3. - 93.*w2 + 465.*v2*(-1. + 33.*w2) - 10230.*fcs_pow(v,4)*(-1. + 35.*w2) + 71610.*fcs_pow(v,6)*(-1. + 37.*w2) - 189255.*fcs_pow(v,8)*(-1. + 39.*w2) + 164021.*fcs_pow(v,10)*(-1. + 41.*w2)) - 140244.*fcs_pow(u,6)*(3. - 99.*w2 + 495.*v2*(-1. + 35.*w2) - 11550.*fcs_pow(v,4)*(-1. + 37.*w2) + 85470.*fcs_pow(v,6)*(-1. + 39.*w2) - 238095.*fcs_pow(v,8)*(-1. + 41.*w2) + 216931.*fcs_pow(v,10)*(-1. + 43.*w2)) + 128557.*fcs_pow(u,8)*(9. - 315.*w2 + 1575.*v2*(-1. + 37.*w2) - 38850.*fcs_pow(v,4)*(-1. + 39.*w2) + 303030.*fcs_pow(v,6)*(-1. + 41.*w2) - 887445.*fcs_pow(v,8)*(-1. + 43.*w2) + 848003.*fcs_pow(v,10)*(-1. + 45.*w2)) - 210366.*fcs_pow(u,10)*(7. - 259.*w2 + 1295.*v2*(-1. + 39.*w2) - 33670.*fcs_pow(v,4)*(-1. + 41.*w2) + 276094.*fcs_pow(v,6)*(-1. + 43.*w2) - 848003.*fcs_pow(v,8)*(-1. + 45.*w2) + 848003.*fcs_pow(v,10)*(-1. + 47.*w2)) + 99789.*fcs_pow(u,12)*(7. - 273.*w2 + 1365.*v2*(-1. + 41.*w2) - 37310.*fcs_pow(v,4)*(-1. + 43.*w2) + 320866.*fcs_pow(v,6)*(-1. + 45.*w2) - 1.031355e6*fcs_pow(v,8)*(-1. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-1. + 49.*w2)));
        case 3 : return 1.298054391195578e21*fcs_pow(t,27)*u*w*(7.*(9. - 87.*w2 - 26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 365893.*fcs_pow(v,10)*(-1. + 13.*w2) + 435.*v2*(-3. + 31.*w2) + 59334.*fcs_pow(v,6)*(-3. + 35.*w2) - 148335.*fcs_pow(v,8)*(-3. + 37.*w2)) - 406.*u2*(9. - 93.*w2 + 1395.*v2*(-1. + 11.*w2) - 567765.*fcs_pow(v,8)*(-1. + 13.*w2) - 10230.*fcs_pow(v,4)*(-3. + 35.*w2) + 71610.*fcs_pow(v,6)*(-3. + 37.*w2) + 164021.*fcs_pow(v,10)*(-3. + 41.*w2)) - 39556.*fcs_pow(u,6)*(9. - 105.*w2 - 38850.*fcs_pow(v,4)*(-1. + 13.*w2) + 848003.*fcs_pow(v,10)*(-1. + 15.*w2) + 525.*v2*(-3. + 37.*w2) + 101010.*fcs_pow(v,6)*(-3. + 41.*w2) - 295815.*fcs_pow(v,8)*(-3. + 43.*w2)) + 6293.*fcs_pow(u,4)*(9. - 99.*w2 + 256410.*fcs_pow(v,6)*(-1. + 13.*w2) + 495.*v2*(-3. + 35.*w2) - 11550.*fcs_pow(v,4)*(-3. + 37.*w2) - 238095.*fcs_pow(v,8)*(-3. + 41.*w2) + 216931.*fcs_pow(v,10)*(-3. + 43.*w2)) + 49445.*fcs_pow(u,8)*(21. - 259.*w2 + 3885.*v2*(-1. + 13.*w2) - 2.544009e6*fcs_pow(v,8)*(-1. + 15.*w2) - 33670.*fcs_pow(v,4)*(-3. + 41.*w2) + 276094.*fcs_pow(v,6)*(-3. + 43.*w2) + 848003.*fcs_pow(v,10)*(-3. + 47.*w2)) + 33263.*fcs_pow(u,12)*(21. - 287.*w2 - 123410.*fcs_pow(v,4)*(-1. + 15.*w2) + 4.060189e6*fcs_pow(v,10)*(-1. + 17.*w2) + 1435.*v2*(-3. + 43.*w2) + 370230.*fcs_pow(v,6)*(-3. + 47.*w2) - 1.242915e6*fcs_pow(v,8)*(-3. + 49.*w2)) - 66526.*fcs_pow(u,10)*(21. - 273.*w2 + 962598.*fcs_pow(v,6)*(-1. + 15.*w2) + 1365.*v2*(-3. + 41.*w2) - 37310.*fcs_pow(v,4)*(-3. + 43.*w2) - 1.031355e6*fcs_pow(v,8)*(-3. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-3. + 49.*w2)));
        case 4 : return -1.298054391195578e21*fcs_pow(t,28)*u*(7.*(-9. + 522.*w2 - 2697.*fcs_pow(w,4) + 1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 445005.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 365893.*fcs_pow(v,10)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59334.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 39556.*fcs_pow(u,6)*(-9. + 630.*w2 - 3885.*fcs_pow(w,4) + 1575.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 38850.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 887445.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 101010.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 18879.*fcs_pow(u,4)*(-3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 11550.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 85470.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 165.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 79365.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 406.*u2*(-9.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 1395.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 214830.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 567765.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 10230.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 164021.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 33263.*fcs_pow(u,12)*(4305.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 123410.*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 3.728745e6*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 7.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 370230.*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 199578.*fcs_pow(u,10)*(-7.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 37310.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 320866.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 455.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 343785.*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 49445.*fcs_pow(u,8)*(-21.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 3885.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 828282.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 2.544009e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 33670.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return 3.764357734467175e22*fcs_pow(t,29)*u*w*(35805.*fcs_pow(u,4)*(-9. + 210.*w2 - 777.*fcs_pow(w,4) + 848003.*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 295815.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 20202.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 2170.*u2*(-9.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 495.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 51282.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 2310.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 47619.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 7.*(-45. + 930.*w2 - 3069.*fcs_pow(w,4) + 6975.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 51150.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 567765.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 71610.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 164021.*fcs_pow(v,10)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 20460.*fcs_pow(u,6)*(-1.2720045e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1.38047e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3885.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 7.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 33670.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 1.645945e6*fcs_pow(u,12)*(-3. + 86.*w2 - 387.*fcs_pow(w,4) + 645.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 240499.*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 127323.*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 1290.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 12126.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 89466.*fcs_pow(u,10)*(-617050.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 7175.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 7.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 370230.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 1.242915e6*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 63085.*fcs_pow(u,8)*(4.81299e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 186550.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 21.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1365.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.031355e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return -5.646536601700763e23*fcs_pow(t,30)*u*(2387.*fcs_pow(u,4)*(9. - 945.*w2 + 11655.*fcs_pow(w,4) - 30303.*fcs_pow(w,6) + 303030.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 315.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 434.*u2*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 238095.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 216931.*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 7.*(3. - 279.*w2 + 3069.*fcs_pow(w,4) - 7161.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 164021.*fcs_pow(v,10)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 14322.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 37851.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 4092.*fcs_pow(u,6)*(-168350.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.38047e6*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.240015e6*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 329189.*fcs_pow(u,12)*(1. - 129.*w2 + 1935.*fcs_pow(w,4) - 6063.*fcs_pow(w,6) + 215.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 212205.*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 240499.*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1290.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 12126.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 29822.*fcs_pow(u,10)*(-21.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 4305.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 123410.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 4.060189e6*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 222138.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 745749.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 12617.*fcs_pow(u,8)*(6825.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 186550.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.60433e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.031355e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return 1.225298442569066e26*fcs_pow(t,31)*u*w*(21.*(1. - 33.*w2 + 231.*fcs_pow(w,4) - 407.*fcs_pow(w,6)) + 3465.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 216931.*fcs_pow(v,10)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 47619.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 4884.*fcs_pow(u,6)*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 1.60433e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.031355e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 26650.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 195.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 693.*fcs_pow(u,4)*(35. - 1295.*w2 + 10101.*fcs_pow(w,4) - 19721.*fcs_pow(w,6) - 4.240015e6*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 197210.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 185.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 4810.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 22.*u2*(-63.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 5.936021e6*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 60606.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 456617.*fcs_pow(u,12)*(1. - 45.*w2 + 423.*fcs_pow(w,4) - 987.*fcs_pow(w,6) + 45.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 1410.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 35955.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 42347.*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 1974.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 118326.*fcs_pow(u,10)*(7. - 301.*w2 + 2709.*fcs_pow(w,4) - 6063.*fcs_pow(w,6) + 1505.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 9030.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 84882.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 240499.*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 42441.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 5291.*fcs_pow(u,8)*(105. - 4305.*w2 + 37023.*fcs_pow(w,4) - 79335.*fcs_pow(w,6) - 617050.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.11069e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 3075.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 3.728745e6*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 580027.*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return -1.225298442569066e26*fcs_pow(t,32)*u*(-21.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 3465.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 1.518517e6*fcs_pow(v,10)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 238095.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 17094.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) - 4884.*fcs_pow(u,6)*(-35. + 5460.*w2 - 111930.*fcs_pow(w,4) + 641732.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) - 186550.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.60433e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 975.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) + 26455.*fcs_pow(u,8)*(-21. + 3444.*w2 - 74046.*fcs_pow(w,4) + 444276.*fcs_pow(w,6) - 745749.*fcs_pow(w,8) + 4305.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 123410.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 222138.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 745749.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) - 828282.*fcs_pow(u,10)*(-1. + 172.*w2 - 3870.*fcs_pow(w,4) + 24252.*fcs_pow(w,6) - 42441.*fcs_pow(w,8) + 215.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 1290.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 12126.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 30315.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 34357.*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 456617.*fcs_pow(u,12)*(-1. + 180.*w2 - 4230.*fcs_pow(w,4) + 27636.*fcs_pow(w,6) - 50337.*fcs_pow(w,8) + 45.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1410.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 9870.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 35955.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 693.*fcs_pow(u,4)*(-35. + 5180.*w2 - 101010.*fcs_pow(w,4) + 552188.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 1.38047e6*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 4.240015e6*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 24050.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 185.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 154.*u2*(-9.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 887445.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 43290.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 45.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 1110.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return 3.675895327707197e26*fcs_pow(t,33)*u*w*(8547.*fcs_pow(u,4)*(1.443897e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 79950.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 585.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 462.*u2*(-315. + 15540.*w2 - 181818.*fcs_pow(w,4) + 709956.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) - 3.8160135e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 591630.*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1665.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 14430.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 33.*(5.936021e6*fcs_pow(v,10)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 7.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 295815.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 7770.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 20202.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 1.369851e6*fcs_pow(u,12)*(-45. + 2820.*w2 - 41454.*fcs_pow(w,4) + 201348.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 235.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 465817.*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 55930.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 211735.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 3290.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1.1872042e7*fcs_pow(u,10)*(-9.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 45.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1410.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 35955.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 5922.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 1.084655e6*fcs_pow(u,8)*(13545.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 3.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 9030.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 84882.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 240499.*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 127323.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 63492.*fcs_pow(u,6)*(-1.85115e6*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 3075.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 370230.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1.242915e6*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 3.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return -3.675895327707197e26*fcs_pow(t,34)*u*(128205.*fcs_pow(u,4)*(21. - 4095.*w2 + 111930.*fcs_pow(w,4) - 962598.*fcs_pow(w,6) + 3.094065e6*fcs_pow(w,8) - 3.231579e6*fcs_pow(w,10) - 111930.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 962598.*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.077193e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 343785.*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 195.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 2310.*u2*(2.484846e6*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 7.632027e6*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 9.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 1665.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 14430.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 231.*(9. - 1575.*w2 + 38850.*fcs_pow(w,4) - 303030.*fcs_pow(w,6) + 887445.*fcs_pow(w,8) - 848003.*fcs_pow(w,10) - 887445.*fcs_pow(v,8)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 225.*v2*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 5550.*fcs_pow(v,4)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 14430.*fcs_pow(v,6)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 317460.*fcs_pow(u,6)*(3.*(21. - 4305.*w2 + 123410.*fcs_pow(w,4) - 1.11069e6*fcs_pow(w,6) + 3.728745e6*fcs_pow(w,8) - 4.060189e6*fcs_pow(w,10)) + 12915.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 370230.*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 3.728745e6*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 370230.*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 6.849255e6*fcs_pow(u,12)*(9. - 2115.*w2 + 69090.*fcs_pow(w,4) - 704718.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 3.260719e6*fcs_pow(w,10) + 3.260719e6*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 705.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 381123.*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 3290.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 11186.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 1.06848378e8*fcs_pow(u,10)*(1. - 225.*w2 + 7050.*fcs_pow(w,4) - 69090.*fcs_pow(w,6) + 251685.*fcs_pow(w,8) - 296429.*fcs_pow(w,10) - 2350.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 25.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 42347.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 3290.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 3995.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 7.592585e6*fcs_pow(u,8)*(-9.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 1935.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 181890.*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 6450.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 90945.*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 34357.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
      }
      case 11 : switch(k){
        case 0 : return 1.109448197603058e19*fcs_pow(t,25)*u*v*(143.*(-7. + 315.*v2 - 3654.*fcs_pow(v,4) + 16182.*fcs_pow(v,6) - 29667.*fcs_pow(v,8) + 18879.*fcs_pow(v,10)) - 1.542684e6*fcs_pow(u,6)*(-3. + 165.*v2 - 2310.*fcs_pow(v,4) + 12210.*fcs_pow(v,6) - 26455.*fcs_pow(v,8) + 19721.*fcs_pow(v,10)) + 87087.*fcs_pow(u,4)*(-9. + 465.*v2 - 6138.*fcs_pow(v,4) + 30690.*fcs_pow(v,6) - 63085.*fcs_pow(v,8) + 44733.*fcs_pow(v,10)) - 858.*u2*(-63. + 3045.*v2 - 37758.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 346115.*fcs_pow(v,8) + 232841.*fcs_pow(v,10)) + 128557.*fcs_pow(u,8)*(-99. + 5775.*v2 - 85470.*fcs_pow(v,4) + 476190.*fcs_pow(v,6) - 1.084655e6*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 99789.*fcs_pow(u,12)*(-77. + 5005.*v2 - 82082.*fcs_pow(v,4) + 504218.*fcs_pow(v,6) - 1.260545e6*fcs_pow(v,8) + 1.077193e6*fcs_pow(v,10)) - 23374.*fcs_pow(u,10)*(-693. + 42735.*v2 - 666666.*fcs_pow(v,4) + 3.904758e6*fcs_pow(v,6) - 9.328033e6*fcs_pow(v,8) + 7.632027e6*fcs_pow(v,10)));
        case 1 : return -1.442282656883975e20*fcs_pow(t,26)*u*v*(623007.*fcs_pow(u,4)*(-3. + 165.*v2 - 2310.*fcs_pow(v,4) + 12210.*fcs_pow(v,6) - 26455.*fcs_pow(v,8) + 19721.*fcs_pow(v,10)) - 13398.*u2*(-9. + 465.*v2 - 6138.*fcs_pow(v,4) + 30690.*fcs_pow(v,6) - 63085.*fcs_pow(v,8) + 44733.*fcs_pow(v,10)) + 33.*(-63. + 3045.*v2 - 37758.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 346115.*fcs_pow(v,8) + 232841.*fcs_pow(v,10)) - 118668.*fcs_pow(u,6)*(-99. + 5775.*v2 - 85470.*fcs_pow(v,4) + 476190.*fcs_pow(v,6) - 1.084655e6*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) - 598734.*fcs_pow(u,10)*(-77. + 5005.*v2 - 82082.*fcs_pow(v,4) + 504218.*fcs_pow(v,6) - 1.260545e6*fcs_pow(v,8) + 1.077193e6*fcs_pow(v,10)) + 99789.*fcs_pow(u,12)*(-231. + 15785.*v2 - 271502.*fcs_pow(v,4) + 1.74537e6*fcs_pow(v,6) - 4.557355e6*fcs_pow(v,8) + 4.060189e6*fcs_pow(v,10)) + 49445.*fcs_pow(u,8)*(-693. + 42735.*v2 - 666666.*fcs_pow(v,4) + 3.904758e6*fcs_pow(v,6) - 9.328033e6*fcs_pow(v,8) + 7.632027e6*fcs_pow(v,10)))*w;
        case 2 : return 1.442282656883975e20*fcs_pow(t,27)*u*v*(33.*(63. - 1827.*w2 + 3045.*v2*(-1. + 31.*w2) - 37758.*fcs_pow(v,4)*(-1. + 33.*w2) + 178002.*fcs_pow(v,6)*(-1. + 35.*w2) - 346115.*fcs_pow(v,8)*(-1. + 37.*w2) + 232841.*fcs_pow(v,10)*(-1. + 39.*w2)) - 13398.*u2*(9. - 279.*w2 + 465.*v2*(-1. + 33.*w2) - 6138.*fcs_pow(v,4)*(-1. + 35.*w2) + 30690.*fcs_pow(v,6)*(-1. + 37.*w2) - 63085.*fcs_pow(v,8)*(-1. + 39.*w2) + 44733.*fcs_pow(v,10)*(-1. + 41.*w2)) + 623007.*fcs_pow(u,4)*(3. - 99.*w2 + 165.*v2*(-1. + 35.*w2) - 2310.*fcs_pow(v,4)*(-1. + 37.*w2) + 12210.*fcs_pow(v,6)*(-1. + 39.*w2) - 26455.*fcs_pow(v,8)*(-1. + 41.*w2) + 19721.*fcs_pow(v,10)*(-1. + 43.*w2)) - 118668.*fcs_pow(u,6)*(99. - 3465.*w2 + 5775.*v2*(-1. + 37.*w2) - 85470.*fcs_pow(v,4)*(-1. + 39.*w2) + 476190.*fcs_pow(v,6)*(-1. + 41.*w2) - 1.084655e6*fcs_pow(v,8)*(-1. + 43.*w2) + 848003.*fcs_pow(v,10)*(-1. + 45.*w2)) + 49445.*fcs_pow(u,8)*(693.*(1. - 37.*w2) + 42735.*v2*(-1. + 39.*w2) - 666666.*fcs_pow(v,4)*(-1. + 41.*w2) + 3.904758e6*fcs_pow(v,6)*(-1. + 43.*w2) - 9.328033e6*fcs_pow(v,8)*(-1. + 45.*w2) + 7.632027e6*fcs_pow(v,10)*(-1. + 47.*w2)) - 598734.*fcs_pow(u,10)*(77. - 3003.*w2 + 5005.*v2*(-1. + 41.*w2) - 82082.*fcs_pow(v,4)*(-1. + 43.*w2) + 504218.*fcs_pow(v,6)*(-1. + 45.*w2) - 1.260545e6*fcs_pow(v,8)*(-1. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-1. + 49.*w2)) + 99789.*fcs_pow(u,12)*(231. - 9471.*w2 + 15785.*v2*(-1. + 43.*w2) - 271502.*fcs_pow(v,4)*(-1. + 45.*w2) + 1.74537e6*fcs_pow(v,6)*(-1. + 47.*w2) - 4.557355e6*fcs_pow(v,8)*(-1. + 49.*w2) + 4.060189e6*fcs_pow(v,10)*(-1. + 51.*w2)));
        case 3 : return -3.764357734467175e22*fcs_pow(t,28)*u*v*w*(77.*(9. - 93.*w2 + 465.*v2*(-1. + 11.*w2) - 63085.*fcs_pow(v,8)*(-1. + 13.*w2) - 2046.*fcs_pow(v,4)*(-3. + 35.*w2) + 10230.*fcs_pow(v,6)*(-3. + 37.*w2) + 14911.*fcs_pow(v,10)*(-3. + 41.*w2)) + 2387.*fcs_pow(u,4)*(99.*(3. - 35.*w2) - 256410.*fcs_pow(v,4)*(-1. + 13.*w2) + 2.544009e6*fcs_pow(v,10)*(-1. + 15.*w2) + 5775.*v2*(-3. + 37.*w2) + 476190.*fcs_pow(v,6)*(-3. + 41.*w2) - 1.084655e6*fcs_pow(v,8)*(-3. + 43.*w2)) - 4774.*u2*(9. - 99.*w2 + 36630.*fcs_pow(v,6)*(-1. + 13.*w2) + 165.*v2*(-3. + 35.*w2) - 2310.*fcs_pow(v,4)*(-3. + 37.*w2) - 26455.*fcs_pow(v,8)*(-3. + 41.*w2) + 19721.*fcs_pow(v,10)*(-3. + 43.*w2)) - 6820.*fcs_pow(u,6)*(693. - 8547.*w2 + 42735.*v2*(-1. + 13.*w2) - 9.328033e6*fcs_pow(v,8)*(-1. + 15.*w2) - 222222.*fcs_pow(v,4)*(-3. + 41.*w2) + 1.301586e6*fcs_pow(v,6)*(-3. + 43.*w2) + 2.544009e6*fcs_pow(v,10)*(-3. + 47.*w2)) - 29822.*fcs_pow(u,10)*(693. - 9471.*w2 - 814506.*fcs_pow(v,4)*(-1. + 15.*w2) + 1.2180567e7*fcs_pow(v,10)*(-1. + 17.*w2) + 15785.*v2*(-3. + 43.*w2) + 1.74537e6*fcs_pow(v,6)*(-3. + 47.*w2) - 4.557355e6*fcs_pow(v,8)*(-3. + 49.*w2)) + 63085.*fcs_pow(u,8)*(231. - 3003.*w2 + 1.512654e6*fcs_pow(v,6)*(-1. + 15.*w2) + 5005.*v2*(-3. + 41.*w2) - 82082.*fcs_pow(v,4)*(-3. + 43.*w2) - 1.260545e6*fcs_pow(v,8)*(-3. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-3. + 49.*w2)) + 47027.*fcs_pow(u,12)*(77.*(3. - 43.*w2) + 16555.*v2*(-1. + 15.*w2) - 5.446595e6*fcs_pow(v,8)*(-1. + 17.*w2) - 99330.*fcs_pow(v,4)*(-3. + 47.*w2) + 666930.*fcs_pow(v,6)*(-3. + 49.*w2) + 1.683493e6*fcs_pow(v,10)*(-3. + 53.*w2)));
        case 4 : return 3.764357734467175e22*fcs_pow(t,29)*u*v*(7161.*fcs_pow(u,4)*(5775.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 85470.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 1.084655e6*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 33.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 158730.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 4774.*u2*(-9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 6930.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 36630.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59163.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 165.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 26455.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 77.*(-9.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 465.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 30690.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 2046.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 14911.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 89466.*fcs_pow(u,10)*(15785.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 271502.*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 4.557355e6*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 77.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 581790.*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 63085.*fcs_pow(u,8)*(-231.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 246246.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 1.512654e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 3.231579e6*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 5005.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 1.260545e6*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 6820.*fcs_pow(u,6)*(-693.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 42735.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 3.904758e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 9.328033e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 222222.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 2.544009e6*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 47027.*fcs_pow(u,12)*(-231.*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 16555.*v2*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 2.00079e6*fcs_pow(v,6)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 5.446595e6*fcs_pow(v,8)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 99330.*fcs_pow(v,4)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 1.683493e6*fcs_pow(v,10)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))));
        case 5 : return -5.834754488424121e24*fcs_pow(t,30)*u*v*w*(-154.*u2*(2.544009e6*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 99.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 1.084655e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 51282.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1155.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 95238.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 77.*(-9.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 165.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 19721.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7326.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 462.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 231.*fcs_pow(u,4)*(-4.6640165e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 6.50793e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 42735.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 231.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 222222.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 2.544009e6*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 456617.*fcs_pow(u,12)*(-33.*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 307615.*fcs_pow(v,8)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 296429.*fcs_pow(v,10)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 65142.*fcs_pow(v,6)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 165.*v2*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 3102.*fcs_pow(v,4)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 118326.*fcs_pow(u,10)*(16555.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 77.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 1.683493e6*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 1.089319e6*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 19866.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 133386.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 5291.*fcs_pow(u,8)*(-4.07253e6*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 78925.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 1.2180567e7*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 231.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 1.74537e6*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 4.557355e6*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 4884.*fcs_pow(u,6)*(7.56327e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 410410.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 231.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 5005.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.260545e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return 1.750426346527236e25*fcs_pow(t,31)*u*v*(-154.*u2*(-99.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 476190.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 1.084655e6*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 1155.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 77.*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 462.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2442.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 77.*fcs_pow(u,4)*(-3.33333e6*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.952379e7*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.6640165e7*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 693.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 7.632027e6*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 456617.*fcs_pow(u,12)*(11.*(1. - 135.*w2 + 2115.*fcs_pow(w,4) - 6909.*fcs_pow(w,6)) + 108570.*fcs_pow(v,6)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 307615.*fcs_pow(v,8)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 296429.*fcs_pow(v,10)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 165.*v2*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 3102.*fcs_pow(v,4)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 39442.*fcs_pow(u,10)*(-231.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 16555.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 5.446595e6*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 5.050479e6*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 59598.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 400158.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 26455.*fcs_pow(u,8)*(-231.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 15785.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 271502.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 4.060189e6*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 349074.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 911471.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 4884.*fcs_pow(u,6)*(25025.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 410410.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 2.52109e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 77.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.260545e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return -1.225298442569066e26*fcs_pow(t,32)*u*v*w*(25641.*fcs_pow(u,4)*(11.*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6)) + 2.52109e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.260545e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 58630.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 715.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 22.*u2*(-3.26481155e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 5.3424189e7*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 1.952379e7*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 693.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 666666.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 11.*(-693.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 5.936021e6*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.084655e6*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1155.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 95238.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 3.5616126e7*fcs_pow(u,10)*(-11.*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 165.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 3102.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 43945.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 42347.*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 3102.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 1.084655e6*fcs_pow(u,8)*(16555.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 59598.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 33.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 400158.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 721497.*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 155617.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 63492.*fcs_pow(u,6)*(-1.35751e6*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.74537e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 11275.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.557355e6*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 33.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 580027.*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 5.022787e6*fcs_pow(u,12)*(-9.*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 705.*v2*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 71910.*fcs_pow(v,6)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 211735.*fcs_pow(v,8)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 1974.*fcs_pow(v,4)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 42347.*fcs_pow(v,10)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 8 : return 1.225298442569066e26*fcs_pow(t,33)*u*v*(-317460.*fcs_pow(u,6)*(15785.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 271502.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 33.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 349074.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 911471.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) + 1.084655e6*fcs_pow(u,8)*(-231.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 16555.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 59598.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 400158.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 778085.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 721497.*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 5.022787e6*fcs_pow(u,12)*(-9.*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 705.*v2*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 9870.*fcs_pow(v,4)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 296429.*fcs_pow(v,10)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 71910.*fcs_pow(v,6)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 211735.*fcs_pow(v,8)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 3.5616126e7*fcs_pow(u,10)*(-11.*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 165.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3102.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 15510.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 43945.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 25641.*fcs_pow(u,4)*(-410410.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 2.52109e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 3575.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.260545e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 154.*u2*(1.952379e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 4.6640165e7*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 476190.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 7.632027e6*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 99.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 6105.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 11.*(-693.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 7.592585e6*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 5.936021e6*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 476190.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 1155.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 17094.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -4.043484860477916e27*fcs_pow(t,34)*u*v*w*(-1554.*u2*(2.268981e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 175890.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 99.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1.260545e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 2145.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 3.*(-3.26481155e8*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 6.50793e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 42735.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5.936021e6*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 222222.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 77.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 6.4382997e7*fcs_pow(u,12)*(-45. + 2940.*w2 - 44982.*fcs_pow(w,4) + 227052.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 119833.*fcs_pow(v,10)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 49555.*fcs_pow(v,8)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 1190.*fcs_pow(v,4)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 9010.*fcs_pow(v,6)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 35.*v2*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1.1872042e7*fcs_pow(u,10)*(-9.*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 705.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 381123.*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 71910.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 211735.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 5922.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 2.9680105e7*fcs_pow(u,8)*(-99.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 165.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3102.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 43945.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 9306.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1.18326e6*fcs_pow(u,6)*(49665.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 33.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 19866.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 133386.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 240499.*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 155617.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 10101.*fcs_pow(u,4)*(-1.221759e7*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 33825.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1.74537e6*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 4.557355e6*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 99.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 1.740081e6*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return 2.021742430238958e28*fcs_pow(t,35)*u*v*(-1554.*u2*(-738738.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 4.537962e6*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 3.231579e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 1.260545e6*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 2145.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 21.*(3.904758e6*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 9.328033e6*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 6105.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 31746.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 30303.*fcs_pow(u,4)*(47355.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 814506.*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4.557355e6*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 581790.*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 33.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 6.4382997e7*fcs_pow(u,12)*(9. - 2205.*w2 + 74970.*fcs_pow(w,4) - 794682.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 3.954489e6*fcs_pow(w,10) - 346885.*fcs_pow(v,8)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 119833.*fcs_pow(v,10)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 16218.*fcs_pow(v,6)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 35.*v2*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 238.*fcs_pow(v,4)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 1.06848378e8*fcs_pow(u,10)*(9. - 2115.*w2 + 69090.*fcs_pow(w,4) - 704718.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 3.260719e6*fcs_pow(w,10) + 296429.*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 235.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 42347.*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 658.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 1598.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 5.936021e6*fcs_pow(u,8)*(-99.*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 46530.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 825.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 381123.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 46530.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 43945.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 236652.*fcs_pow(u,6)*(-693.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 49665.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 2.00079e6*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 99330.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 778085.*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 240499.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
        case 11 : return -5.236312894318901e30*fcs_pow(t,36)*u*v*w*(52767.*fcs_pow(u,4)*(693. - 49665.*w2 + 893970.*fcs_pow(w,4) - 6.00237e6*fcs_pow(w,6) + 1.6339785e7*fcs_pow(w,8) - 1.5151437e7*fcs_pow(w,10) + 4515.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99330.*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 666930.*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 778085.*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 240499.*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10))) - 3.025308e6*fcs_pow(u,6)*(-9.*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 825.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 15510.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 46530.*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 43945.*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 42347.*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10))) - 858.*u2*(-222138.*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.74537e6*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4.557355e6*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 1.740081e6*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 9.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 615.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 33.*(412542.*fcs_pow(v,6)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.260545e6*fcs_pow(v,8)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.077193e6*fcs_pow(v,10)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 9.*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 195.*v2*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 3198.*fcs_pow(v,4)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 248583.*fcs_pow(u,12)*(-231.*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 6.590815e6*fcs_pow(v,8)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 6545.*v2*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,10)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 138754.*fcs_pow(v,4)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 891990.*fcs_pow(v,6)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 1.9389474e7*fcs_pow(u,10)*(-11.*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 385.*v2*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 119833.*fcs_pow(v,10)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 2618.*fcs_pow(v,4)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 19822.*fcs_pow(v,6)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 49555.*fcs_pow(v,8)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 1.260545e6*fcs_pow(u,8)*(-99.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 7755.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 65142.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 158202.*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 465817.*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 381123.*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))));
      }
      case 12 : switch(k){
        case 0 : return -1.109448197603058e19*fcs_pow(t,26)*u*(-18018.*u2*(3. - 522.*v2 + 13485.*fcs_pow(v,4) - 118668.*fcs_pow(v,6) + 445005.*fcs_pow(v,8) - 731786.*fcs_pow(v,10) + 432419.*fcs_pow(v,12)) + 261261.*fcs_pow(u,4)*(3. - 558.*v2 + 15345.*fcs_pow(v,4) - 143220.*fcs_pow(v,6) + 567765.*fcs_pow(v,8) - 984126.*fcs_pow(v,10) + 611351.*fcs_pow(v,12)) + 143.*(7. - 1134.*v2 + 27405.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 801009.*fcs_pow(v,8) - 1.246014e6*fcs_pow(v,10) + 698523.*fcs_pow(v,12)) - 1.542684e6*fcs_pow(u,6)*(3. - 594.*v2 + 17325.*fcs_pow(v,4) - 170940.*fcs_pow(v,6) + 714285.*fcs_pow(v,8) - 1.301586e6*fcs_pow(v,10) + 848003.*fcs_pow(v,12)) + 385671.*fcs_pow(u,8)*(33. - 6930.*v2 + 213675.*fcs_pow(v,4) - 2.22222e6*fcs_pow(v,6) + 9.761895e6*fcs_pow(v,8) - 1.8656066e7*fcs_pow(v,10) + 1.2720045e7*fcs_pow(v,12)) - 210366.*fcs_pow(u,10)*(77. - 17094.*v2 + 555555.*fcs_pow(v,4) - 6.074068e6*fcs_pow(v,6) + 2.7984099e7*fcs_pow(v,8) - 5.5968198e7*fcs_pow(v,10) + 3.9856141e7*fcs_pow(v,12)) + 99789.*fcs_pow(u,12)*(77. - 18018.*v2 + 615615.*fcs_pow(v,4) - 7.059052e6*fcs_pow(v,6) + 3.4034715e7*fcs_pow(v,8) - 7.1094738e7*fcs_pow(v,10) + 5.2782457e7*fcs_pow(v,12)));
        case 1 : return 1.298054391195578e21*fcs_pow(t,27)*u*(77.*(3. - 522.*v2 + 13485.*fcs_pow(v,4) - 118668.*fcs_pow(v,6) + 445005.*fcs_pow(v,8) - 731786.*fcs_pow(v,10) + 432419.*fcs_pow(v,12)) - 4466.*u2*(3. - 558.*v2 + 15345.*fcs_pow(v,4) - 143220.*fcs_pow(v,6) + 567765.*fcs_pow(v,8) - 984126.*fcs_pow(v,10) + 611351.*fcs_pow(v,12)) + 69223.*fcs_pow(u,4)*(3. - 594.*v2 + 17325.*fcs_pow(v,4) - 170940.*fcs_pow(v,6) + 714285.*fcs_pow(v,8) - 1.301586e6*fcs_pow(v,10) + 848003.*fcs_pow(v,12)) - 39556.*fcs_pow(u,6)*(33. - 6930.*v2 + 213675.*fcs_pow(v,4) - 2.22222e6*fcs_pow(v,6) + 9.761895e6*fcs_pow(v,8) - 1.8656066e7*fcs_pow(v,10) + 1.2720045e7*fcs_pow(v,12)) + 49445.*fcs_pow(u,8)*(77. - 17094.*v2 + 555555.*fcs_pow(v,4) - 6.074068e6*fcs_pow(v,6) + 2.7984099e7*fcs_pow(v,8) - 5.5968198e7*fcs_pow(v,10) + 3.9856141e7*fcs_pow(v,12)) - 66526.*fcs_pow(u,10)*(77. - 18018.*v2 + 615615.*fcs_pow(v,4) - 7.059052e6*fcs_pow(v,6) + 3.4034715e7*fcs_pow(v,8) - 7.1094738e7*fcs_pow(v,10) + 5.2782457e7*fcs_pow(v,12)) + 33263.*fcs_pow(u,12)*(77. - 18942.*v2 + 678755.*fcs_pow(v,4) - 8.14506e6*fcs_pow(v,6) + 4.1016195e7*fcs_pow(v,8) - 8.9324158e7*fcs_pow(v,10) + 6.9023213e7*fcs_pow(v,12)))*w;
        case 2 : return -1.298054391195578e21*fcs_pow(t,28)*u*(77.*(-3. + 87.*w2 - 522.*v2*(-1. + 31.*w2) + 13485.*fcs_pow(v,4)*(-1. + 33.*w2) - 118668.*fcs_pow(v,6)*(-1. + 35.*w2) + 445005.*fcs_pow(v,8)*(-1. + 37.*w2) - 731786.*fcs_pow(v,10)*(-1. + 39.*w2) + 432419.*fcs_pow(v,12)*(-1. + 41.*w2)) - 4466.*u2*(-3. + 93.*w2 - 558.*v2*(-1. + 33.*w2) + 15345.*fcs_pow(v,4)*(-1. + 35.*w2) - 143220.*fcs_pow(v,6)*(-1. + 37.*w2) + 567765.*fcs_pow(v,8)*(-1. + 39.*w2) - 984126.*fcs_pow(v,10)*(-1. + 41.*w2) + 611351.*fcs_pow(v,12)*(-1. + 43.*w2)) + 69223.*fcs_pow(u,4)*(-3. + 99.*w2 - 594.*v2*(-1. + 35.*w2) + 17325.*fcs_pow(v,4)*(-1. + 37.*w2) - 170940.*fcs_pow(v,6)*(-1. + 39.*w2) + 714285.*fcs_pow(v,8)*(-1. + 41.*w2) - 1.301586e6*fcs_pow(v,10)*(-1. + 43.*w2) + 848003.*fcs_pow(v,12)*(-1. + 45.*w2)) - 39556.*fcs_pow(u,6)*(33.*(-1. + 35.*w2) - 6930.*v2*(-1. + 37.*w2) + 213675.*fcs_pow(v,4)*(-1. + 39.*w2) - 2.22222e6*fcs_pow(v,6)*(-1. + 41.*w2) + 9.761895e6*fcs_pow(v,8)*(-1. + 43.*w2) - 1.8656066e7*fcs_pow(v,10)*(-1. + 45.*w2) + 1.2720045e7*fcs_pow(v,12)*(-1. + 47.*w2)) + 49445.*fcs_pow(u,8)*(77.*(-1. + 37.*w2) - 17094.*v2*(-1. + 39.*w2) + 555555.*fcs_pow(v,4)*(-1. + 41.*w2) - 6.074068e6*fcs_pow(v,6)*(-1. + 43.*w2) + 2.7984099e7*fcs_pow(v,8)*(-1. + 45.*w2) - 5.5968198e7*fcs_pow(v,10)*(-1. + 47.*w2) + 3.9856141e7*fcs_pow(v,12)*(-1. + 49.*w2)) - 66526.*fcs_pow(u,10)*(77.*(-1. + 39.*w2) - 18018.*v2*(-1. + 41.*w2) + 615615.*fcs_pow(v,4)*(-1. + 43.*w2) - 7.059052e6*fcs_pow(v,6)*(-1. + 45.*w2) + 3.4034715e7*fcs_pow(v,8)*(-1. + 47.*w2) - 7.1094738e7*fcs_pow(v,10)*(-1. + 49.*w2) + 5.2782457e7*fcs_pow(v,12)*(-1. + 51.*w2)) + 33263.*fcs_pow(u,12)*(77.*(-1. + 41.*w2) - 18942.*v2*(-1. + 43.*w2) + 678755.*fcs_pow(v,4)*(-1. + 45.*w2) - 8.14506e6*fcs_pow(v,6)*(-1. + 47.*w2) + 4.1016195e7*fcs_pow(v,8)*(-1. + 49.*w2) - 8.9324158e7*fcs_pow(v,10)*(-1. + 51.*w2) + 6.9023213e7*fcs_pow(v,12)*(-1. + 53.*w2)));
        case 3 : return 3.764357734467175e22*fcs_pow(t,29)*u*w*(-14322.*u2*(-3. + 33.*w2 - 170940.*fcs_pow(v,6)*(-1. + 13.*w2) + 848003.*fcs_pow(v,12)*(-1. + 15.*w2) - 198.*v2*(-3. + 35.*w2) + 5775.*fcs_pow(v,4)*(-3. + 37.*w2) + 238095.*fcs_pow(v,8)*(-3. + 41.*w2) - 433862.*fcs_pow(v,10)*(-3. + 43.*w2)) + 77.*(-9. + 93.*w2 - 1674.*v2*(-1. + 11.*w2) + 1.703295e6*fcs_pow(v,8)*(-1. + 13.*w2) + 15345.*fcs_pow(v,4)*(-3. + 35.*w2) - 143220.*fcs_pow(v,6)*(-3. + 37.*w2) - 984126.*fcs_pow(v,10)*(-3. + 41.*w2) + 611351.*fcs_pow(v,12)*(-3. + 43.*w2)) + 21483.*fcs_pow(u,4)*(-33. + 385.*w2 + 213675.*fcs_pow(v,4)*(-1. + 13.*w2) - 1.8656066e7*fcs_pow(v,10)*(-1. + 15.*w2) - 2310.*v2*(-3. + 37.*w2) - 740740.*fcs_pow(v,6)*(-3. + 41.*w2) + 3.253965e6*fcs_pow(v,8)*(-3. + 43.*w2) + 4.240015e6*fcs_pow(v,12)*(-3. + 47.*w2)) + 189255.*fcs_pow(u,8)*(77.*(-1. + 13.*w2) - 7.059052e6*fcs_pow(v,6)*(-1. + 15.*w2) + 5.2782457e7*fcs_pow(v,12)*(-1. + 17.*w2) - 6006.*v2*(-3. + 41.*w2) + 205205.*fcs_pow(v,4)*(-3. + 43.*w2) + 1.1344905e7*fcs_pow(v,8)*(-3. + 47.*w2) - 2.3698246e7*fcs_pow(v,10)*(-3. + 49.*w2)) - 20460.*fcs_pow(u,6)*(-51282.*v2*(-1. + 13.*w2) + 8.3952297e7*fcs_pow(v,8)*(-1. + 15.*w2) + 77.*(-3. + 37.*w2) + 555555.*fcs_pow(v,4)*(-3. + 41.*w2) - 6.074068e6*fcs_pow(v,6)*(-3. + 43.*w2) - 5.5968198e7*fcs_pow(v,10)*(-3. + 47.*w2) + 3.9856141e7*fcs_pow(v,12)*(-3. + 49.*w2)) - 89466.*fcs_pow(u,10)*(2.036265e6*fcs_pow(v,4)*(-1. + 15.*w2) - 2.67972474e8*fcs_pow(v,10)*(-1. + 17.*w2) + 77.*(-3. + 41.*w2) - 18942.*v2*(-3. + 43.*w2) - 8.14506e6*fcs_pow(v,6)*(-3. + 47.*w2) + 4.1016195e7*fcs_pow(v,8)*(-3. + 49.*w2) + 6.9023213e7*fcs_pow(v,12)*(-3. + 53.*w2)) + 329189.*fcs_pow(u,12)*(-33. + 473.*w2 - 8514.*v2*(-1. + 15.*w2) + 2.1008295e7*fcs_pow(v,8)*(-1. + 17.*w2) + 106425.*fcs_pow(v,4)*(-3. + 47.*w2) - 1.33386e6*fcs_pow(v,6)*(-3. + 49.*w2) - 1.5872934e7*fcs_pow(v,10)*(-3. + 53.*w2) + 1.2746447e7*fcs_pow(v,12)*(-3. + 55.*w2)));
        case 4 : return -1.129307320340153e23*fcs_pow(t,30)*u*(-4774.*u2*(3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 17325.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 170940.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 1.301586e6*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,12)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 198.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 238095.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 77.*(3.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 558.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 143220.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 567765.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 611351.*fcs_pow(v,12)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 5115.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 328042.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 63085.*fcs_pow(u,8)*(77.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 615615.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 7.059052e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 7.1094738e7*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 5.2782457e7*fcs_pow(v,12)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 6006.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 1.1344905e7*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 20460.*fcs_pow(u,6)*(77.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 17094.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 6.074068e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 2.7984099e7*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 3.9856141e7*fcs_pow(v,12)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 185185.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 1.8656066e7*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 7161.*fcs_pow(u,4)*(-6930.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 213675.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 9.761895e6*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 1.8656066e7*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 11.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 740740.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 4.240015e6*fcs_pow(v,12)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 329189.*fcs_pow(u,12)*(11.*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 2838.*v2*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 1.33386e6*fcs_pow(v,6)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 7.002765e6*fcs_pow(v,8)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 1.2746447e7*fcs_pow(v,12)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 35475.*fcs_pow(v,4)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) - 5.290978e6*fcs_pow(v,10)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) - 29822.*fcs_pow(u,10)*(-56826.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 2.036265e6*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 1.23048585e8*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 2.67972474e8*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 77.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 8.14506e6*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 6.9023213e7*fcs_pow(v,12)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))));
        case 5 : return 1.750426346527236e25*fcs_pow(t,31)*u*w*(77.*(3. - 66.*w2 + 231.*fcs_pow(w,4) + 848003.*fcs_pow(v,12)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 198.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 433862.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 34188.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1155.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 47619.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 462.*u2*(-1.8656066e7*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 11.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 3.253965e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 42735.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 462.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 148148.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,12)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 5.022787e6*fcs_pow(u,12)*(1. - 30.*w2 + 141.*fcs_pow(w,4) + 251685.*fcs_pow(v,8)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 592858.*fcs_pow(v,10)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) - 27636.*fcs_pow(v,6)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 296429.*fcs_pow(v,12)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) - 18.*v2*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 705.*fcs_pow(v,4)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 276094.*fcs_pow(u,10)*(-8514.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 11.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 1.5872934e7*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 1.2746447e7*fcs_pow(v,12)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 4.201659e6*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 21285.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 266772.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 5291.*fcs_pow(u,8)*(1.0181325e7*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 94710.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3.45116065e8*fcs_pow(v,12)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 2.67972474e8*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 77.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 8.14506e6*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 4.1016195e7*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 4884.*fcs_pow(u,6)*(-3.529526e7*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1.026025e6*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 77.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 5.2782457e7*fcs_pow(v,12)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 6006.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 1.1344905e7*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 2.3698246e7*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 231.*fcs_pow(u,4)*(4.19761485e8*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 3.037034e7*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 51282.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 77.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 555555.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 5.5968198e7*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 3.9856141e7*fcs_pow(v,12)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return -1.750426346527236e25*fcs_pow(t,32)*u*(77.*(-3. + 297.*w2 - 3465.*fcs_pow(w,4) + 8547.*fcs_pow(w,6) - 594.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 714285.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 1.301586e6*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,12)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 3465.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 34188.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 462.*u2*(33.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 2.22222e6*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 9.761895e6*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.8656066e7*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 1386.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 2.544009e6*fcs_pow(v,12)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) - 828282.*fcs_pow(u,10)*(11.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2838.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.002765e6*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 1.5872934e7*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.2746447e7*fcs_pow(v,12)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 21285.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 266772.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 79365.*fcs_pow(u,8)*(77.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 18942.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 678755.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 8.9324158e7*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 6.9023213e7*fcs_pow(v,12)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.629012e6*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 8.203239e6*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 4884.*fcs_pow(u,6)*(-90090.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 3.078075e6*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 3.529526e7*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 2.63912285e8*fcs_pow(v,12)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 77.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 3.4034715e7*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 7.1094738e7*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 693.*fcs_pow(u,4)*(2.777775e6*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 3.037034e7*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.39920495e8*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 77.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 5.5968198e7*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 3.9856141e7*fcs_pow(v,12)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 5.022787e6*fcs_pow(u,12)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6) - 138180.*fcs_pow(v,6)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 755055.*fcs_pow(v,8)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.778574e6*fcs_pow(v,10)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 54.*v2*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 2115.*fcs_pow(v,4)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) + 296429.*fcs_pow(v,12)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 7 : return 4.043484860477916e27*fcs_pow(t,33)*u*w*(231.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 1.30592462e8*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.7808063e7*fcs_pow(v,12)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 9.761895e6*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1386.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 444444.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 42.*u2*(1.39920495e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 5.5968198e7*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 4.33862e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 3.9856141e7*fcs_pow(v,12)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 11.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 2442.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 79365.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 98605.*fcs_pow(u,8)*(-19866.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 148995.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 11.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.867404e6*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 1.5872934e7*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.2746447e7*fcs_pow(v,12)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 1.400553e6*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 5772.*fcs_pow(u,6)*(3.393775e6*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 8.14506e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 13530.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 4.1016195e7*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 4.9302295e7*fcs_pow(v,12)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 11.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 1.2760594e7*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 777.*fcs_pow(u,4)*(-3.529526e7*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 3.4034715e7*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 439725.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 7.1094738e7*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 11.*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 2574.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 7.540351e6*fcs_pow(v,12)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 456617.*fcs_pow(u,12)*(3.*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 846.*v2*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 5.632151e6*fcs_pow(v,12)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) - 335580.*fcs_pow(v,6)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.905615e6*fcs_pow(v,8)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 4935.*fcs_pow(v,4)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) - 931634.*fcs_pow(v,10)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) - 1.1872042e7*fcs_pow(u,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6) - 54.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 2115.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 107865.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 254082.*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 3948.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 42347.*fcs_pow(v,12)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 8 : return -4.043484860477916e27*fcs_pow(t,34)*u*(-28860.*fcs_pow(u,6)*(-18942.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 678755.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 11.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.629012e6*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 8.203239e6*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.2760594e7*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 9.860459e6*fcs_pow(v,12)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 456617.*fcs_pow(u,12)*(3.*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 846.*v2*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 24675.*fcs_pow(v,4)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 6.521438e6*fcs_pow(v,10)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) - 335580.*fcs_pow(v,6)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 5.632151e6*fcs_pow(v,12)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) + 1.905615e6*fcs_pow(v,8)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 1.1872042e7*fcs_pow(u,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8) - 54.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 2115.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 19740.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 296429.*fcs_pow(v,12)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 107865.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 254082.*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 690235.*fcs_pow(u,8)*(11.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 2838.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 21285.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 266772.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.000395e6*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 2.267562e6*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.820921e6*fcs_pow(v,12)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 777.*fcs_pow(u,4)*(3.078075e6*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 3.529526e7*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 12870.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 3.4034715e7*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 7.1094738e7*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 3.7701755e7*fcs_pow(v,12)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 42.*u2*(-3.037034e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.39920495e8*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 396825.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 5.5968198e7*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 3.9856141e7*fcs_pow(v,12)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 11.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 2442.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 7.*(33.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 9.761895e6*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 1.8656066e7*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 317460.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2.544009e6*fcs_pow(v,12)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 198.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 6105.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return 1.213045458143375e28*fcs_pow(t,35)*u*w*(7.*(1.259284455e9*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 1.301586e7*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 21978.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 5.5968198e7*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 3.9856141e7*fcs_pow(v,12)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 238095.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 11.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 2.1460999e7*fcs_pow(u,12)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8) - 7.908978e6*fcs_pow(v,10)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 7.070147e6*fcs_pow(v,12)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) + 1.337985e6*fcs_pow(v,8)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 8925.*fcs_pow(v,4)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 126140.*fcs_pow(v,6)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 126.*v2*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 3.5616126e7*fcs_pow(u,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8) + 5.632151e6*fcs_pow(v,12)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 282.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2.794902e6*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 111860.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 635205.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 4935.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 3.26481155e8*fcs_pow(u,8)*(3.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 18.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 705.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 127041.*fcs_pow(v,12)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 35955.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 84694.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 3948.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 394420.*fcs_pow(u,6)*(-178794.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 33.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 148995.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1.867404e6*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.5872934e7*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.2746447e7*fcs_pow(v,12)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 4.201659e6*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 10101.*fcs_pow(u,4)*(3.0543975e7*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 40590.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 8.14506e6*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 4.1016195e7*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 33.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 4.9302295e7*fcs_pow(v,12)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 3.8281782e7*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1554.*u2*(-1.0588578e8*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 439725.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 33.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 1.1344905e7*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 2.3698246e7*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2574.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 7.540351e6*fcs_pow(v,12)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return -1.273697731050544e30*fcs_pow(t,36)*u*(-1.8222204e7*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 8.3952297e7*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 33.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 7326.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 3.9856141e7*fcs_pow(v,12)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 1.8656066e7*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 79365.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 74.*u2*(1.846845e6*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 2.1177156e7*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 33.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 7.1094738e7*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 1.1344905e7*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 2574.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 7.540351e6*fcs_pow(v,12)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 3.065857e6*fcs_pow(u,12)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10) + 3.121965e6*fcs_pow(v,8)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,12)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) - 2.636326e6*fcs_pow(v,10)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 75684.*fcs_pow(v,6)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 42.*v2*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 595.*fcs_pow(v,4)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 1.696006e6*fcs_pow(u,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10) - 1.9564314e7*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 846.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 5.632151e6*fcs_pow(v,12)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 1.143369e6*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 4935.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 22372.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 9.328033e6*fcs_pow(u,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10) + 1.482145e6*fcs_pow(v,12)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 3525.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 30.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 254082.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 6580.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 11985.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 78884.*fcs_pow(u,6)*(33.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 8514.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.33386e6*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 35475.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 5.462763e6*fcs_pow(v,12)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 1.000395e6*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 755854.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 481.*fcs_pow(u,4)*(-170478.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 6.108795e6*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.23048585e8*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 8.14506e6*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 33.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 3.8281782e7*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
        case 11 : return 4.712681604887011e31*fcs_pow(t,37)*u*w*(5863.*fcs_pow(u,4)*(-16254.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 744975.*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 9.33702e6*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 2.1008295e7*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 1.5872934e7*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 1.2746447e7*fcs_pow(v,12)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 3.*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) - 286.*u2*(555345.*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 8.14506e6*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 4.1016195e7*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 3.8281782e7*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 3.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 738.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 11.*(-1.925196e6*fcs_pow(v,6)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.1344905e7*fcs_pow(v,8)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 2.3698246e7*fcs_pow(v,10)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 7.540351e6*fcs_pow(v,12)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 3.*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 234.*v2*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 7995.*fcs_pow(v,4)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 580027.*fcs_pow(u,12)*(11.*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 6.1611281e7*fcs_pow(v,12)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 8.473905e6*fcs_pow(v,8)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 1122.*v2*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 2.2220462e7*fcs_pow(v,10)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) + 49555.*fcs_pow(v,4)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 594660.*fcs_pow(v,6)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 2.154386e6*fcs_pow(u,10)*(11.*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 1386.*v2*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 7.908978e6*fcs_pow(v,10)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 19635.*fcs_pow(v,4)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,12)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 277508.*fcs_pow(v,6)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 1.337985e6*fcs_pow(v,8)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 1.260545e6*fcs_pow(u,8)*(11.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 3102.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 54285.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 5.632151e6*fcs_pow(v,12)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 246092.*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 1.397451e6*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 2.794902e6*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 1.008436e6*fcs_pow(u,6)*(-33. + 2475.*w2 - 46530.*fcs_pow(w,4) + 325710.*fcs_pow(w,6) - 922845.*fcs_pow(w,8) + 889287.*fcs_pow(w,10) - 990.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 38775.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 217140.*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 395505.*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 931634.*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 635205.*fcs_pow(v,12)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))));
        case 12 : return -4.712681604887011e31*fcs_pow(t,38)*u*(11.*(-2.1177156e7*fcs_pow(v,6)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 3.4034715e7*fcs_pow(v,8)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 7.1094738e7*fcs_pow(v,10)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 167895.*fcs_pow(v,4)*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) + 3.*(77. - 18018.*w2 + 615615.*fcs_pow(w,4) - 7.059052e6*fcs_pow(w,6) + 3.4034715e7*fcs_pow(w,8) - 7.1094738e7*fcs_pow(w,10) + 5.2782457e7*fcs_pow(w,12)) + 7.540351e6*fcs_pow(v,12)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 702.*v2*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12))) + 3.781635e6*fcs_pow(u,8)*(11.*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 3102.*v2*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 4.192353e6*fcs_pow(v,8)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 6.521438e6*fcs_pow(v,10)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 18095.*fcs_pow(v,4)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 5.632151e6*fcs_pow(v,12)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 246092.*fcs_pow(v,6)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 3.025308e6*fcs_pow(u,6)*(11.*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 990.*v2*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 38775.*fcs_pow(v,4)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 2.794902e6*fcs_pow(v,10)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 1.482145e6*fcs_pow(v,12)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 72380.*fcs_pow(v,6)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 395505.*fcs_pow(v,8)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 369369.*fcs_pow(u,4)*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12) - 2838.*v2*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 35475.*fcs_pow(v,4)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 444620.*fcs_pow(v,6)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 1.820921e6*fcs_pow(v,12)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 333465.*fcs_pow(v,8)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 755854.*fcs_pow(v,10)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 858.*u2*(2.036265e6*fcs_pow(v,4)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 8.14506e6*fcs_pow(v,6)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 4.1016195e7*fcs_pow(v,8)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 5166.*v2*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) - 1.2760594e7*fcs_pow(v,10)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3.*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12)) + 9.860459e6*fcs_pow(v,12)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 6.463158e6*fcs_pow(u,10)*(11.*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 832524.*fcs_pow(v,6)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 3.121965e6*fcs_pow(v,8)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 462.*v2*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 7.908978e6*fcs_pow(v,10)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 19635.*fcs_pow(v,4)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 7.070147e6*fcs_pow(v,12)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12))) + 580027.*fcs_pow(u,12)*(445995.*fcs_pow(v,4)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 4.16262e6*fcs_pow(v,6)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 11.*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 2.5421715e7*fcs_pow(v,8)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 3366.*v2*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) - 6.6661386e7*fcs_pow(v,10)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) + 6.1611281e7*fcs_pow(v,12)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12))));
      }
      case 13 : switch(k){
        case 0 : return 4.326847970651925e20*fcs_pow(t,27)*u*v*(-174174.*u2*(3. - 186.*v2 + 3069.*fcs_pow(v,4) - 20460.*fcs_pow(v,6) + 63085.*fcs_pow(v,8) - 89466.*fcs_pow(v,10) + 47027.*fcs_pow(v,12)) + 2.699697e6*fcs_pow(u,4)*(3. - 198.*v2 + 3465.*fcs_pow(v,4) - 24420.*fcs_pow(v,6) + 79365.*fcs_pow(v,8) - 118326.*fcs_pow(v,10) + 65231.*fcs_pow(v,12)) + 429.*(21. - 1218.*v2 + 18879.*fcs_pow(v,4) - 118668.*fcs_pow(v,6) + 346115.*fcs_pow(v,8) - 465682.*fcs_pow(v,10) + 232841.*fcs_pow(v,12)) - 1.542684e6*fcs_pow(u,6)*(33. - 2310.*v2 + 42735.*fcs_pow(v,4) - 317460.*fcs_pow(v,6) + 1.084655e6*fcs_pow(v,8) - 1.696006e6*fcs_pow(v,10) + 978465.*fcs_pow(v,12)) - 2.594514e6*fcs_pow(u,10)*(77. - 6006.*v2 + 123123.*fcs_pow(v,4) - 1.008436e6*fcs_pow(v,6) + 3.781635e6*fcs_pow(v,8) - 6.463158e6*fcs_pow(v,10) + 4.060189e6*fcs_pow(v,12)) + 642785.*fcs_pow(u,8)*(231. - 17094.*v2 + 333333.*fcs_pow(v,4) - 2.603172e6*fcs_pow(v,6) + 9.328033e6*fcs_pow(v,8) - 1.5264054e7*fcs_pow(v,10) + 9.197571e6*fcs_pow(v,12)) + 99789.*fcs_pow(u,12)*(1001. - 82082.*v2 + 1.764763e6*fcs_pow(v,4) - 1.512654e7*fcs_pow(v,6) + 5.9245615e7*fcs_pow(v,8) - 1.05564914e8*fcs_pow(v,10) + 6.9023213e7*fcs_pow(v,12)));
        case 1 : return -3.764357734467175e22*fcs_pow(t,28)*u*v*(1001.*(3. - 186.*v2 + 3069.*fcs_pow(v,4) - 20460.*fcs_pow(v,6) + 63085.*fcs_pow(v,8) - 89466.*fcs_pow(v,10) + 47027.*fcs_pow(v,12)) - 62062.*u2*(3. - 198.*v2 + 3465.*fcs_pow(v,4) - 24420.*fcs_pow(v,6) + 79365.*fcs_pow(v,8) - 118326.*fcs_pow(v,10) + 65231.*fcs_pow(v,12)) + 93093.*fcs_pow(u,4)*(33. - 2310.*v2 + 42735.*fcs_pow(v,4) - 317460.*fcs_pow(v,6) + 1.084655e6*fcs_pow(v,8) - 1.696006e6*fcs_pow(v,10) + 978465.*fcs_pow(v,12)) + 820105.*fcs_pow(u,8)*(77. - 6006.*v2 + 123123.*fcs_pow(v,4) - 1.008436e6*fcs_pow(v,6) + 3.781635e6*fcs_pow(v,8) - 6.463158e6*fcs_pow(v,10) + 4.060189e6*fcs_pow(v,12)) - 88660.*fcs_pow(u,6)*(231. - 17094.*v2 + 333333.*fcs_pow(v,4) - 2.603172e6*fcs_pow(v,6) + 9.328033e6*fcs_pow(v,8) - 1.5264054e7*fcs_pow(v,10) + 9.197571e6*fcs_pow(v,12)) - 89466.*fcs_pow(u,10)*(1001. - 82082.*v2 + 1.764763e6*fcs_pow(v,4) - 1.512654e7*fcs_pow(v,6) + 5.9245615e7*fcs_pow(v,8) - 1.05564914e8*fcs_pow(v,10) + 6.9023213e7*fcs_pow(v,12)) + 47027.*fcs_pow(u,12)*(1001. - 86086.*v2 + 1.936935e6*fcs_pow(v,4) - 1.734018e7*fcs_pow(v,6) + 7.0805735e7*fcs_pow(v,8) - 1.31312454e8*fcs_pow(v,10) + 8.9225129e7*fcs_pow(v,12)))*w;
        case 2 : return 3.764357734467175e22*fcs_pow(t,29)*u*v*(1001.*(-3. + 93.*w2 - 186.*v2*(-1. + 33.*w2) + 3069.*fcs_pow(v,4)*(-1. + 35.*w2) - 20460.*fcs_pow(v,6)*(-1. + 37.*w2) + 63085.*fcs_pow(v,8)*(-1. + 39.*w2) - 89466.*fcs_pow(v,10)*(-1. + 41.*w2) + 47027.*fcs_pow(v,12)*(-1. + 43.*w2)) - 62062.*u2*(-3. + 99.*w2 - 198.*v2*(-1. + 35.*w2) + 3465.*fcs_pow(v,4)*(-1. + 37.*w2) - 24420.*fcs_pow(v,6)*(-1. + 39.*w2) + 79365.*fcs_pow(v,8)*(-1. + 41.*w2) - 118326.*fcs_pow(v,10)*(-1. + 43.*w2) + 65231.*fcs_pow(v,12)*(-1. + 45.*w2)) + 93093.*fcs_pow(u,4)*(33.*(-1. + 35.*w2) - 2310.*v2*(-1. + 37.*w2) + 42735.*fcs_pow(v,4)*(-1. + 39.*w2) - 317460.*fcs_pow(v,6)*(-1. + 41.*w2) + 1.084655e6*fcs_pow(v,8)*(-1. + 43.*w2) - 1.696006e6*fcs_pow(v,10)*(-1. + 45.*w2) + 978465.*fcs_pow(v,12)*(-1. + 47.*w2)) - 88660.*fcs_pow(u,6)*(231.*(-1. + 37.*w2) - 17094.*v2*(-1. + 39.*w2) + 333333.*fcs_pow(v,4)*(-1. + 41.*w2) - 2.603172e6*fcs_pow(v,6)*(-1. + 43.*w2) + 9.328033e6*fcs_pow(v,8)*(-1. + 45.*w2) - 1.5264054e7*fcs_pow(v,10)*(-1. + 47.*w2) + 9.197571e6*fcs_pow(v,12)*(-1. + 49.*w2)) + 820105.*fcs_pow(u,8)*(77.*(-1. + 39.*w2) - 6006.*v2*(-1. + 41.*w2) + 123123.*fcs_pow(v,4)*(-1. + 43.*w2) - 1.008436e6*fcs_pow(v,6)*(-1. + 45.*w2) + 3.781635e6*fcs_pow(v,8)*(-1. + 47.*w2) - 6.463158e6*fcs_pow(v,10)*(-1. + 49.*w2) + 4.060189e6*fcs_pow(v,12)*(-1. + 51.*w2)) - 89466.*fcs_pow(u,10)*(1001.*(-1. + 41.*w2) - 82082.*v2*(-1. + 43.*w2) + 1.764763e6*fcs_pow(v,4)*(-1. + 45.*w2) - 1.512654e7*fcs_pow(v,6)*(-1. + 47.*w2) + 5.9245615e7*fcs_pow(v,8)*(-1. + 49.*w2) - 1.05564914e8*fcs_pow(v,10)*(-1. + 51.*w2) + 6.9023213e7*fcs_pow(v,12)*(-1. + 53.*w2)) + 47027.*fcs_pow(u,12)*(1001.*(-1. + 43.*w2) - 86086.*v2*(-1. + 45.*w2) + 1.936935e6*fcs_pow(v,4)*(-1. + 47.*w2) - 1.734018e7*fcs_pow(v,6)*(-1. + 49.*w2) + 7.0805735e7*fcs_pow(v,8)*(-1. + 51.*w2) - 1.31312454e8*fcs_pow(v,10)*(-1. + 53.*w2) + 8.9225129e7*fcs_pow(v,12)*(-1. + 55.*w2)));
        case 3 : return -3.500852693054473e24*fcs_pow(t,30)*u*v*w*(1001.*(-3. + 33.*w2 - 24420.*fcs_pow(v,6)*(-1. + 13.*w2) + 65231.*fcs_pow(v,12)*(-1. + 15.*w2) - 66.*v2*(-3. + 35.*w2) + 1155.*fcs_pow(v,4)*(-3. + 37.*w2) + 26455.*fcs_pow(v,8)*(-3. + 41.*w2) - 39442.*fcs_pow(v,10)*(-3. + 43.*w2)) - 2002.*u2*(128205.*fcs_pow(v,4)*(-1. + 13.*w2) - 5.088018e6*fcs_pow(v,10)*(-1. + 15.*w2) + 33.*(-3. + 35.*w2) - 2310.*v2*(-3. + 37.*w2) - 317460.*fcs_pow(v,6)*(-3. + 41.*w2) + 1.084655e6*fcs_pow(v,8)*(-3. + 43.*w2) + 978465.*fcs_pow(v,12)*(-3. + 47.*w2)) - 317460.*fcs_pow(u,6)*(77.*(-1. + 13.*w2) - 1.008436e6*fcs_pow(v,6)*(-1. + 15.*w2) + 4.060189e6*fcs_pow(v,12)*(-1. + 17.*w2) - 2002.*v2*(-3. + 41.*w2) + 41041.*fcs_pow(v,4)*(-3. + 43.*w2) + 1.260545e6*fcs_pow(v,8)*(-3. + 47.*w2) - 2.154386e6*fcs_pow(v,10)*(-3. + 49.*w2)) + 15015.*fcs_pow(u,4)*(-17094.*v2*(-1. + 13.*w2) + 9.328033e6*fcs_pow(v,8)*(-1. + 15.*w2) + 77.*(-3. + 37.*w2) + 111111.*fcs_pow(v,4)*(-3. + 41.*w2) - 867724.*fcs_pow(v,6)*(-3. + 43.*w2) - 5.088018e6*fcs_pow(v,10)*(-3. + 47.*w2) + 3.065857e6*fcs_pow(v,12)*(-3. + 49.*w2)) + 26455.*fcs_pow(u,8)*(5.294289e6*fcs_pow(v,4)*(-1. + 15.*w2) - 3.16694742e8*fcs_pow(v,10)*(-1. + 17.*w2) + 1001.*(-3. + 41.*w2) - 82082.*v2*(-3. + 43.*w2) - 1.512654e7*fcs_pow(v,6)*(-3. + 47.*w2) + 5.9245615e7*fcs_pow(v,8)*(-3. + 49.*w2) + 6.9023213e7*fcs_pow(v,12)*(-3. + 53.*w2)) + 456617.*fcs_pow(u,12)*(143.*(-1. + 15.*w2) - 2.82282e6*fcs_pow(v,6)*(-1. + 17.*w2) + 1.6303595e7*fcs_pow(v,12)*(-1. + 19.*w2) - 4290.*v2*(-3. + 47.*w2) + 100815.*fcs_pow(v,4)*(-3. + 49.*w2) + 3.998995e6*fcs_pow(v,8)*(-3. + 53.*w2) - 7.707154e6*fcs_pow(v,10)*(-3. + 55.*w2)) - 39442.*fcs_pow(u,10)*(-258258.*v2*(-1. + 15.*w2) + 2.12417205e8*fcs_pow(v,8)*(-1. + 17.*w2) + 1001.*(-3. + 43.*w2) + 1.936935e6*fcs_pow(v,4)*(-3. + 47.*w2) - 1.734018e7*fcs_pow(v,6)*(-3. + 49.*w2) - 1.31312454e8*fcs_pow(v,10)*(-3. + 53.*w2) + 8.9225129e7*fcs_pow(v,12)*(-3. + 55.*w2)));
        case 4 : return 3.500852693054473e24*fcs_pow(t,31)*u*v*(1001.*(3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 3465.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 24420.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 118326.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 65231.*fcs_pow(v,12)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 66.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 26455.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 317460.*fcs_pow(u,6)*(77.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 123123.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 1.008436e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 6.463158e6*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,12)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 2002.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 1.260545e6*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 15015.*fcs_pow(u,4)*(231.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 17094.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 2.603172e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 9.328033e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 9.197571e6*fcs_pow(v,12)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 111111.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 5.088018e6*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 6006.*u2*(-2310.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 42735.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 1.084655e6*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 1.696006e6*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 11.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 105820.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 326155.*fcs_pow(v,12)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 456617.*fcs_pow(u,12)*(143.*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 302445.*fcs_pow(v,4)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 2.82282e6*fcs_pow(v,6)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 2.3121462e7*fcs_pow(v,10)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 1.6303595e7*fcs_pow(v,12)*(1. - 114.*w2 + 1121.*fcs_pow(w,4)) - 4290.*v2*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 3.998995e6*fcs_pow(v,8)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) - 118326.*fcs_pow(u,10)*(1001.*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 86086.*v2*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 1.734018e7*fcs_pow(v,6)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 7.0805735e7*fcs_pow(v,8)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 8.9225129e7*fcs_pow(v,12)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 645645.*fcs_pow(v,4)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) - 4.3770818e7*fcs_pow(v,10)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) + 26455.*fcs_pow(u,8)*(-246246.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 5.294289e6*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 1.77736845e8*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 3.16694742e8*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 1001.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 1.512654e7*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 6.9023213e7*fcs_pow(v,12)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))));
        case 5 : return -1.750426346527236e25*fcs_pow(t,32)*u*v*w*(1001.*(-5.088018e6*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 33.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 1.084655e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 25641.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 462.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 63492.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 195693.*fcs_pow(v,12)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) - 3.5616126e7*fcs_pow(u,10)*(143.*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 3.998995e6*fcs_pow(v,8)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 7.707154e6*fcs_pow(v,10)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) - 564564.*fcs_pow(v,6)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 3.260719e6*fcs_pow(v,12)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) - 858.*v2*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 20163.*fcs_pow(v,4)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 1.084655e6*fcs_pow(u,8)*(-258258.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1001.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 1.31312454e8*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 8.9225129e7*fcs_pow(v,12)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 4.2483441e7*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 387387.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 3.468036e6*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 63492.*fcs_pow(u,6)*(2.6471445e7*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 410410.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3.45116065e8*fcs_pow(v,12)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 3.16694742e8*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 1001.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.512654e7*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 5.9245615e7*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 333333.*fcs_pow(u,4)*(-5.04218e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 205205.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 77.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,12)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 2002.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 1.260545e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 2.154386e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 6006.*u2*(4.6640165e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 4.33862e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 17094.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 77.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 111111.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 5.088018e6*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 3.065857e6*fcs_pow(v,12)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 5.022787e6*fcs_pow(u,12)*(-4.36254e6*fcs_pow(v,6)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 1.9267885e7*fcs_pow(v,8)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 269451.*fcs_pow(v,4)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 2.3121462e7*fcs_pow(v,10)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) + 39.*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 3666.*v2*(15. - 490.*w2 + 2499.*fcs_pow(w,4)) + 5.632151e6*fcs_pow(v,12)*(15. - 590.*w2 + 3599.*fcs_pow(w,4))));
        case 6 : return 5.251279039581709e25*fcs_pow(t,33)*u*v*(1001.*(33.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 317460.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.084655e6*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.696006e6*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 462.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 8547.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 195693.*fcs_pow(v,12)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 1.084655e6*fcs_pow(u,8)*(1001.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 86086.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.0805735e7*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 1.31312454e8*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 8.9225129e7*fcs_pow(v,12)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 387387.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 3.468036e6*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 317460.*fcs_pow(u,6)*(1001.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 82082.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.764763e6*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 1.05564914e8*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 6.9023213e7*fcs_pow(v,12)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 3.025308e6*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.1849123e7*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 111111.*fcs_pow(u,4)*(-30030.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 615615.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 5.04218e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 2.0300945e7*fcs_pow(v,12)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 77.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 3.781635e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 6.463158e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 2002.*u2*(1.666665e6*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 1.301586e7*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 4.6640165e7*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 231.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.5264054e7*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 9.197571e6*fcs_pow(v,12)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 1.1872042e7*fcs_pow(u,10)*(143.*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 2.82282e6*fcs_pow(v,6)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 1.1996985e7*fcs_pow(v,8)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 2.3121462e7*fcs_pow(v,10)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 2574.*v2*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 60489.*fcs_pow(v,4)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) + 3.260719e6*fcs_pow(v,12)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) + 5.022787e6*fcs_pow(u,12)*(449085.*fcs_pow(v,4)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 4.36254e6*fcs_pow(v,6)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.9267885e7*fcs_pow(v,8)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 39.*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 3666.*v2*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) - 7.707154e6*fcs_pow(v,10)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6)) + 5.632151e6*fcs_pow(v,12)*(-5. + 885.*w2 - 17995.*fcs_pow(w,4) + 75579.*fcs_pow(w,6))));
        case 7 : return -4.043484860477916e27*fcs_pow(t,34)*u*v*w*(13.*(3.26481155e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.06848378e8*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 1.301586e7*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 6.4382997e7*fcs_pow(v,12)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 231.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 333333.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 1.18326e6*fcs_pow(u,6)*(-86086.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 387387.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 143.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 3.468036e6*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 1.8758922e7*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.2746447e7*fcs_pow(v,12)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 2.023021e6*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 30303.*fcs_pow(u,4)*(8.823815e6*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.512654e7*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 58630.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 5.9245615e7*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 4.9302295e7*fcs_pow(v,12)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 143.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 1.5080702e7*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 20202.*u2*(-5.04218e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 3.781635e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 87945.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 6.463158e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 11.*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 858.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 580027.*fcs_pow(v,12)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 2.1460999e7*fcs_pow(u,12)*(39.*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 9.346974e6*fcs_pow(v,10)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) + 69615.*fcs_pow(v,4)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 7.070147e6*fcs_pow(v,12)*(-5. + 305.*w2 - 3843.*fcs_pow(w,4) + 11895.*fcs_pow(w,6)) - 702780.*fcs_pow(v,6)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 546.*v2*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 644215.*fcs_pow(v,8)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) - 3.5616126e7*fcs_pow(u,10)*(39.*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 3666.*v2*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 5.632151e6*fcs_pow(v,12)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) - 623220.*fcs_pow(v,6)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 2.752555e6*fcs_pow(v,8)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 12831.*fcs_pow(v,4)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) - 1.101022e6*fcs_pow(v,10)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) + 2.9680105e7*fcs_pow(u,8)*(143.*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 2574.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 60489.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 1.713855e6*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 3.303066e6*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 80652.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 465817.*fcs_pow(v,12)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 8 : return 4.043484860477916e27*fcs_pow(t,35)*u*v*(151515.*fcs_pow(u,4)*(-82082.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.764763e6*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 143.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 3.025308e6*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.1849123e7*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.5080702e7*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 9.860459e6*fcs_pow(v,12)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 2.1460999e7*fcs_pow(u,12)*(3.5350735e7*fcs_pow(v,12)*(1. - 244.*w2 + 7686.*fcs_pow(w,4) - 66612.*fcs_pow(w,6) + 159393.*fcs_pow(w,8)) + 39.*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2730.*v2*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 4.509505e6*fcs_pow(v,8)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 69615.*fcs_pow(v,4)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 9.346974e6*fcs_pow(v,10)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) - 702780.*fcs_pow(v,6)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 3.5616126e7*fcs_pow(u,10)*(39.*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3666.*v2*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 64155.*fcs_pow(v,4)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 7.707154e6*fcs_pow(v,10)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) - 623220.*fcs_pow(v,6)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 5.632151e6*fcs_pow(v,12)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) + 2.752555e6*fcs_pow(v,8)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 2.9680105e7*fcs_pow(u,8)*(143.*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 2574.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 60489.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 403260.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 3.260719e6*fcs_pow(v,12)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 1.713855e6*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 3.303066e6*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 1.18326e6*fcs_pow(u,6)*(1001.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 86086.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 387387.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3.468036e6*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.0115105e7*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 1.8758922e7*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.2746447e7*fcs_pow(v,12)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 20202.*u2*(615615.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 5.04218e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 4290.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 3.781635e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 6.463158e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 2.900135e6*fcs_pow(v,12)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 91.*(-1.301586e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 4.6640165e7*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 238095.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.5264054e7*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 9.197571e6*fcs_pow(v,12)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 33.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 2442.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -3.141787736591341e30*fcs_pow(t,36)*u*v*w*(248583.*fcs_pow(u,12)*(2.156394835e9*fcs_pow(v,12)*(1. - 84.*w2 + 1638.*fcs_pow(w,4) - 10452.*fcs_pow(w,6) + 20033.*fcs_pow(w,8)) + 8.5680595e7*fcs_pow(v,8)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 1.83823822e8*fcs_pow(v,10)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) - 7.73058e6*fcs_pow(v,6)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 15470.*v2*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 409955.*fcs_pow(v,4)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 91.*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 6.463158e6*fcs_pow(u,10)*(-9.346974e6*fcs_pow(v,10)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 7.070147e6*fcs_pow(v,12)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) + 13.*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.932645e6*fcs_pow(v,8)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 23205.*fcs_pow(v,4)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 234260.*fcs_pow(v,6)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 546.*v2*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 1.260545e6*fcs_pow(u,8)*(1.6896453e7*fcs_pow(v,12)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 39.*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3666.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 9.909198e6*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 623220.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 2.752555e6*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 38493.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1.37514e6*fcs_pow(u,6)*(429.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 858.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 20163.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.397451e6*fcs_pow(v,12)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 571285.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1.101022e6*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 80652.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 7995.*fcs_pow(u,4)*(-774774.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 429.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 387387.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3.468036e6*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.8758922e7*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.2746447e7*fcs_pow(v,12)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 6.069063e6*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 78.*u2*(7.9414335e7*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 175890.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 1.512654e7*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 5.9245615e7*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 429.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 4.9302295e7*fcs_pow(v,12)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 4.5242106e7*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 39.*(-1.512654e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 87945.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 33.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 1.260545e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 2.154386e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 858.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 580027.*fcs_pow(v,12)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return 4.712681604887011e31*fcs_pow(t,37)*u*v*(13.*(369369.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 3.025308e6*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 33.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 6.463158e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 1.260545e6*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 858.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,12)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 580027.*fcs_pow(u,12)*(-7.73058e6*fcs_pow(v,6)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 7.8781638e7*fcs_pow(v,10)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) + 6.1611281e7*fcs_pow(v,12)*(-1. + 315.*w2 - 13650.*fcs_pow(w,4) + 182910.*fcs_pow(w,6) - 901485.*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 1.2240085e7*fcs_pow(v,8)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 105417.*fcs_pow(v,4)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 13.*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 442.*v2*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 6.463158e6*fcs_pow(u,10)*(4.509505e6*fcs_pow(v,8)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,12)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) + 13.*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 3.115658e6*fcs_pow(v,10)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 140556.*fcs_pow(v,6)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 182.*v2*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 1547.*fcs_pow(v,4)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 1.260545e6*fcs_pow(u,8)*(-2.3121462e7*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 3666.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 5.632151e6*fcs_pow(v,12)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 13.*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 1.651533e6*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 12831.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 41548.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 91676.*fcs_pow(u,6)*(429.*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 4.8910785e7*fcs_pow(v,12)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 302445.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 4290.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 9.909198e6*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 403260.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 571285.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 1599.*fcs_pow(u,4)*(3003.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 258258.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.734018e7*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 645645.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 3.8239341e7*fcs_pow(v,12)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 1.0115105e7*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 6.252974e6*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 26.*u2*(-738738.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 1.5882867e7*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.77736845e8*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 1.512654e7*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 429.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 4.5242106e7*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
        case 11 : return -4.712681604887011e31*fcs_pow(t,38)*u*v*w*(-11726.*u2*(-70434.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.936935e6*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.734018e7*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 3.0345315e7*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 1.8758922e7*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 1.2746447e7*fcs_pow(v,12)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 39.*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 143.*(1.443897e6*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.512654e7*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 5.9245615e7*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 4.5242106e7*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 39.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 3198.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) - 6.463158e6*fcs_pow(u,10)*(1001.*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 4.31278967e8*fcs_pow(v,12)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 8.5680595e7*fcs_pow(v,8)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 34034.*v2*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 1.83823822e8*fcs_pow(v,10)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) + 901901.*fcs_pow(v,4)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 7.73058e6*fcs_pow(v,6)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 5.9245615e7*fcs_pow(u,8)*(143.*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 6006.*v2*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 9.346974e6*fcs_pow(v,10)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 51051.*fcs_pow(v,4)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,12)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 515372.*fcs_pow(v,6)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 1.932645e6*fcs_pow(v,8)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 5.04218e6*fcs_pow(u,6)*(429.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 40326.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 423423.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 1.6896453e7*fcs_pow(v,12)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 1.371084e6*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 6.055621e6*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 9.909198e6*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 5.294289e6*fcs_pow(u,4)*(39.*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4290.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 100815.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 403260.*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 571285.*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 1.101022e6*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 635205.*fcs_pow(v,12)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 9.860459e6*fcs_pow(u,12)*(-2.82687054e8*fcs_pow(v,10)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) - 8.64006e6*fcs_pow(v,6)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 143.*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 4.2480295e7*fcs_pow(v,8)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 15158.*v2*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 341055.*fcs_pow(v,4)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10)) + 3.624193e6*fcs_pow(v,12)*(-693. + 75075.*w2 - 2.01201e6*fcs_pow(w,4) + 1.983267e7*fcs_pow(w,6) - 7.8228865e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10))));
        case 12 : return 1.413804481466103e32*fcs_pow(t,39)*u*v*(-1.512654e7*fcs_pow(u,6)*(143.*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 13442.*v2*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 6.055621e6*fcs_pow(v,8)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 7.707154e6*fcs_pow(v,10)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 47047.*fcs_pow(v,4)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 5.632151e6*fcs_pow(v,12)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 457028.*fcs_pow(v,6)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 5.294289e6*fcs_pow(u,4)*(143.*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 4290.*v2*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 100815.*fcs_pow(v,4)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 3.303066e6*fcs_pow(v,10)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 1.482145e6*fcs_pow(v,12)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 134420.*fcs_pow(v,6)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 571285.*fcs_pow(v,8)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 11726.*u2*(-258258.*v2*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 1.936935e6*fcs_pow(v,4)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 1.734018e7*fcs_pow(v,6)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 273.*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) + 3.8239341e7*fcs_pow(v,12)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 1.0115105e7*fcs_pow(v,8)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 1.8758922e7*fcs_pow(v,10)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 143.*(5.294289e6*fcs_pow(v,4)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 1.512654e7*fcs_pow(v,6)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 5.9245615e7*fcs_pow(v,8)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 22386.*v2*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) - 1.5080702e7*fcs_pow(v,10)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 39.*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12)) + 9.860459e6*fcs_pow(v,12)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 5.9245615e7*fcs_pow(u,8)*(143.*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 1.546116e6*fcs_pow(v,6)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 4.509505e6*fcs_pow(v,8)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 2002.*v2*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 9.346974e6*fcs_pow(v,10)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 51051.*fcs_pow(v,4)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 7.070147e6*fcs_pow(v,12)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12))) - 1.5080702e7*fcs_pow(u,10)*(1.159587e6*fcs_pow(v,4)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 7.73058e6*fcs_pow(v,6)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 143.*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3.6720255e7*fcs_pow(v,8)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 14586.*v2*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) - 7.8781638e7*fcs_pow(v,10)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) + 6.1611281e7*fcs_pow(v,12)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12))) + 9.860459e6*fcs_pow(u,12)*(-45474.*v2*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 795795.*fcs_pow(v,4)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 8.64006e6*fcs_pow(v,6)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 143.*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 4.2480295e7*fcs_pow(v,8)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) - 9.4229018e7*fcs_pow(v,10)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12)) + 3.624193e6*fcs_pow(v,12)*(231. - 90090.*w2 + 5.030025e6*fcs_pow(w,4) - 9.255246e7*fcs_pow(w,6) + 7.04059785e8*fcs_pow(w,8) - 2.284282858e9*fcs_pow(w,10) + 2.595775975e9*fcs_pow(w,12))));
        case 13 : return -7.535577886214331e34*fcs_pow(t,40)*u*v*w*(1.445015e6*fcs_pow(u,8)*(2.57041785e8*fcs_pow(v,8)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) - 5.51471466e8*fcs_pow(v,10)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) + 2.705703e6*fcs_pow(v,4)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 4.31278967e8*fcs_pow(v,12)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) - 7.73058e6*fcs_pow(v,6)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 77.*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 7854.*v2*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 1.734018e7*fcs_pow(u,6)*(-9.346974e6*fcs_pow(v,10)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 7.070147e6*fcs_pow(v,12)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) + 11.*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 515372.*fcs_pow(v,6)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 644215.*fcs_pow(v,8)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 154.*v2*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3927.*fcs_pow(v,4)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 645645.*fcs_pow(u,4)*(1.6896453e7*fcs_pow(v,12)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 33.*(39. - 3666.*w2 + 89817.*fcs_pow(w,4) - 872508.*fcs_pow(w,6) + 3.853577e6*fcs_pow(w,8) - 7.707154e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 3102.*v2*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 6.055621e6*fcs_pow(v,8)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) - 3.303066e6*fcs_pow(v,10)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 10857.*fcs_pow(v,4)*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 105468.*fcs_pow(v,6)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 258258.*u2*(143. - 12870.*w2 + 302445.*fcs_pow(w,4) - 2.82282e6*fcs_pow(w,6) + 1.1996985e7*fcs_pow(w,8) - 2.3121462e7*fcs_pow(w,10) + 1.6303595e7*fcs_pow(w,12) - 330.*v2*(39. - 3666.*w2 + 89817.*fcs_pow(w,4) - 872508.*fcs_pow(w,6) + 3.853577e6*fcs_pow(w,8) - 7.707154e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 7755.*fcs_pow(v,4)*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 1.101022e6*fcs_pow(v,10)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 211735.*fcs_pow(v,12)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 10340.*fcs_pow(v,6)*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 43945.*fcs_pow(v,8)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 143.*(148995.*fcs_pow(v,4)*(39. - 3666.*w2 + 89817.*fcs_pow(w,4) - 872508.*fcs_pow(w,6) + 3.853577e6*fcs_pow(w,8) - 7.707154e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 1.33386e6*fcs_pow(v,6)*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 1.2746447e7*fcs_pow(v,12)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) - 1806.*v2*(143. - 12870.*w2 + 302445.*fcs_pow(w,4) - 2.82282e6*fcs_pow(w,6) + 1.1996985e7*fcs_pow(w,8) - 2.3121462e7*fcs_pow(w,10) + 1.6303595e7*fcs_pow(w,12)) + 778085.*fcs_pow(v,8)*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 1.442994e6*fcs_pow(v,10)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 3.*(1001. - 86086.*w2 + 1.936935e6*fcs_pow(w,4) - 1.734018e7*fcs_pow(w,6) + 7.0805735e7*fcs_pow(w,8) - 1.31312454e8*fcs_pow(w,10) + 8.9225129e7*fcs_pow(w,12))) - 1.8758922e7*fcs_pow(u,10)*(-8.64006e6*fcs_pow(v,6)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 4.2480295e7*fcs_pow(v,8)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) - 15158.*v2*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) - 9.4229018e7*fcs_pow(v,10)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) + 113685.*fcs_pow(v,4)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 11.*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 3.624193e6*fcs_pow(v,12)*(231. - 30030.*w2 + 1.006005e6*fcs_pow(w,4) - 1.322178e7*fcs_pow(w,6) + 7.8228865e7*fcs_pow(w,8) - 2.07662078e8*fcs_pow(w,10) + 1.99675075e8*fcs_pow(w,12))) + 1.2746447e7*fcs_pow(u,12)*(855855.*fcs_pow(v,4)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) - 9.61818e6*fcs_pow(v,6)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) + 143.*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 4.8892415e7*fcs_pow(v,8)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) - 4290.*v2*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 5.333718e6*fcs_pow(v,10)*(231. - 30030.*w2 + 1.006005e6*fcs_pow(w,4) - 1.322178e7*fcs_pow(w,6) + 7.8228865e7*fcs_pow(w,8) - 2.07662078e8*fcs_pow(w,10) + 1.99675075e8*fcs_pow(w,12)) + 341905.*fcs_pow(v,12)*(3003. - 402402.*w2 + 1.3882869e7*fcs_pow(w,4) - 1.87749276e8*fcs_pow(w,6) + 1.142141429e9*fcs_pow(w,8) - 3.11493117e9*fcs_pow(w,10) + 3.074996155e9*fcs_pow(w,12))));
      }
    }
    case 14 : switch(j){
      case 0 : switch(k){
        case 0 : return 4.2567525e7*fcs_pow(t,15)*(-429. + 45045.*u2 - 765765.*fcs_pow(u,4) + 4.849845e6*fcs_pow(u,6) - 1.4549535e7*fcs_pow(u,8) + 2.2309287e7*fcs_pow(u,10) - 1.6900975e7*fcs_pow(u,12) + 5.014575e6*fcs_pow(u,14));
      }
      case 1 : switch(k){
        case 0 : return -6.38512875e8*fcs_pow(t,16)*(-429. + 51051.*u2 - 969969.*fcs_pow(u,4) + 6.789783e6*fcs_pow(u,6) - 2.2309287e7*fcs_pow(u,8) + 3.7182145e7*fcs_pow(u,10) - 3.0421755e7*fcs_pow(u,12) + 9.694845e6*fcs_pow(u,14))*v;
        case 1 : return 3.2564156625e10*fcs_pow(t,17)*(-143. + 19019.*u2 - 399399.*fcs_pow(u,4) + 3.062059e6*fcs_pow(u,6) - 1.0935925e7*fcs_pow(u,8) + 1.9684665e7*fcs_pow(u,10) - 1.7298645e7*fcs_pow(u,12) + 5.892945e6*fcs_pow(u,14))*v*w;
      }
      case 2 : switch(k){
        case 0 : return 6.38512875e8*fcs_pow(t,17)*(429. - 7293.*v2 + 51051.*u2*(-1. + 19.*v2) - 969969.*fcs_pow(u,4)*(-1. + 21.*v2) + 6.789783e6*fcs_pow(u,6)*(-1. + 23.*v2) - 2.2309287e7*fcs_pow(u,8)*(-1. + 25.*v2) + 3.7182145e7*fcs_pow(u,10)*(-1. + 27.*v2) - 3.0421755e7*fcs_pow(u,12)*(-1. + 29.*v2) + 9.694845e6*fcs_pow(u,14)*(-1. + 31.*v2));
        case 1 : return -3.2564156625e10*fcs_pow(t,18)*(143. - 2717.*v2 + 19019.*u2*(-1. + 21.*v2) - 399399.*fcs_pow(u,4)*(-1. + 23.*v2) + 3.062059e6*fcs_pow(u,6)*(-1. + 25.*v2) - 1.0935925e7*fcs_pow(u,8)*(-1. + 27.*v2) + 1.9684665e7*fcs_pow(u,10)*(-1. + 29.*v2) - 1.7298645e7*fcs_pow(u,12)*(-1. + 31.*v2) + 5.892945e6*fcs_pow(u,14)*(-1. + 33.*v2))*w;
        case 2 : return 3.2564156625e10*fcs_pow(t,19)*(-143.*(1. - 19.*w2 + 19.*v2*(-1. + 21.*w2)) + 19019.*u2*(1. - 21.*w2 + 21.*v2*(-1. + 23.*w2)) - 399399.*fcs_pow(u,4)*(1. - 23.*w2 + 23.*v2*(-1. + 25.*w2)) + 3.062059e6*fcs_pow(u,6)*(1. - 25.*w2 + 25.*v2*(-1. + 27.*w2)) - 1.0935925e7*fcs_pow(u,8)*(1. - 27.*w2 + 27.*v2*(-1. + 29.*w2)) + 1.9684665e7*fcs_pow(u,10)*(1. - 29.*w2 + 29.*v2*(-1. + 31.*w2)) - 1.7298645e7*fcs_pow(u,12)*(1. - 31.*w2 + 31.*v2*(-1. + 33.*w2)) + 5.892945e6*fcs_pow(u,14)*(1. - 33.*w2 + 33.*v2*(-1. + 35.*w2)));
      }
      case 3 : switch(k){
        case 0 : return -3.2564156625e10*fcs_pow(t,18)*v*(429. - 2717.*v2 + 57057.*u2*(-1. + 7.*v2) - 3.2807775e7*fcs_pow(u,8)*(-1. + 9.*v2) + 1.7678835e7*fcs_pow(u,14)*(-1. + 11.*v2) - 399399.*fcs_pow(u,4)*(-3. + 23.*v2) + 3.062059e6*fcs_pow(u,6)*(-3. + 25.*v2) + 1.9684665e7*fcs_pow(u,10)*(-3. + 29.*v2) - 1.7298645e7*fcs_pow(u,12)*(-3. + 31.*v2));
        case 1 : return 1.856156927625e12*fcs_pow(t,19)*v*(143.*(1. - 7.*v2) + 4.029025e6*fcs_pow(u,6)*(-1. + 9.*v2) - 2.8224105e7*fcs_pow(u,12)*(-1. + 11.*v2) + 7007.*u2*(-3. + 23.*v2) - 161161.*fcs_pow(u,4)*(-3. + 25.*v2) - 5.180175e6*fcs_pow(u,8)*(-3. + 29.*v2) + 1.0015005e7*fcs_pow(u,10)*(-3. + 31.*v2) + 3.411705e6*fcs_pow(u,14)*(-3. + 35.*v2))*w;
        case 2 : return -1.856156927625e12*fcs_pow(t,20)*v*(-143.*(1. - 21.*w2 + 7.*v2*(-1. + 23.*w2)) + 7007.*u2*(3. - 69.*w2 + 23.*v2*(-1. + 25.*w2)) - 161161.*fcs_pow(u,4)*(3. - 75.*w2 + 25.*v2*(-1. + 27.*w2)) + 4.029025e6*fcs_pow(u,6)*(1. - 27.*w2 + 9.*v2*(-1. + 29.*w2)) - 5.180175e6*fcs_pow(u,8)*(3. - 87.*w2 + 29.*v2*(-1. + 31.*w2)) + 1.0015005e7*fcs_pow(u,10)*(3. - 93.*w2 + 31.*v2*(-1. + 33.*w2)) - 2.8224105e7*fcs_pow(u,12)*(1. - 33.*w2 + 11.*v2*(-1. + 35.*w2)) + 3.411705e6*fcs_pow(u,14)*(3. - 105.*w2 + 35.*v2*(-1. + 37.*w2)));
        case 3 : return 1.42924083427125e14*fcs_pow(t,21)*v*w*(6279.*u2*(3. - 25.*w2 + 25.*v2*(-1. + 9.*w2)) - 5.852925e6*fcs_pow(u,8)*(3. - 31.*w2 + 31.*v2*(-1. + 11.*w2)) + 4.652325e6*fcs_pow(u,14)*(3. - 37.*w2 + 37.*v2*(-1. + 13.*w2)) - 13.*(9. - 69.*w2 + 23.*v2*(-3. + 25.*w2)) + 470925.*fcs_pow(u,6)*(9. - 87.*w2 + 29.*v2*(-3. + 31.*w2)) + 1.2096045e7*fcs_pow(u,10)*(3. - 33.*w2 + 11.*v2*(-3. + 35.*w2)) - 4.032015e6*fcs_pow(u,12)*(9. - 105.*w2 + 35.*v2*(-3. + 37.*w2)) - 470925.*fcs_pow(u,4)*(1. - 9.*w2 + v2*(-9. + 87.*w2)));
      }
      case 4 : switch(k){
        case 0 : return 9.7692469875e10*fcs_pow(t,19)*(-143.*(1. - 38.*v2 + 133.*fcs_pow(v,4)) + 19019.*u2*(1. - 42.*v2 + 161.*fcs_pow(v,4)) + 3.062059e6*fcs_pow(u,6)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) - 1.0935925e7*fcs_pow(u,8)*(1. - 54.*v2 + 261.*fcs_pow(v,4)) - 1.7298645e7*fcs_pow(u,12)*(1. - 62.*v2 + 341.*fcs_pow(v,4)) + 5.892945e6*fcs_pow(u,14)*(1. - 66.*v2 + 385.*fcs_pow(v,4)) - 133133.*fcs_pow(u,4)*(3. - 138.*v2 + 575.*fcs_pow(v,4)) + 6.561555e6*fcs_pow(u,10)*(3. - 174.*v2 + 899.*fcs_pow(v,4)));
        case 1 : return -1.856156927625e12*fcs_pow(t,20)*(-143.*(1. - 42.*v2 + 161.*fcs_pow(v,4)) - 483483.*fcs_pow(u,4)*(1. - 50.*v2 + 225.*fcs_pow(v,4)) + 4.029025e6*fcs_pow(u,6)*(1. - 54.*v2 + 261.*fcs_pow(v,4)) + 3.0045015e7*fcs_pow(u,10)*(1. - 62.*v2 + 341.*fcs_pow(v,4)) - 2.8224105e7*fcs_pow(u,12)*(1. - 66.*v2 + 385.*fcs_pow(v,4)) + 7007.*u2*(3. - 138.*v2 + 575.*fcs_pow(v,4)) - 5.180175e6*fcs_pow(u,8)*(3. - 174.*v2 + 899.*fcs_pow(v,4)) + 3.411705e6*fcs_pow(u,14)*(3. - 210.*v2 + 1295.*fcs_pow(v,4)))*w;
        case 2 : return 1.856156927625e12*fcs_pow(t,21)*(-143.*(-1. + 21.*w2 + v2*(42. - 966.*w2) + 161.*fcs_pow(v,4)*(-1. + 25.*w2)) + 7007.*u2*(-3. + 69.*w2 - 138.*v2*(-1. + 25.*w2) + 575.*fcs_pow(v,4)*(-1. + 27.*w2)) - 483483.*fcs_pow(u,4)*(-1. + 25.*w2 - 50.*v2*(-1. + 27.*w2) + 225.*fcs_pow(v,4)*(-1. + 29.*w2)) + 4.029025e6*fcs_pow(u,6)*(-1. + 27.*w2 - 54.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)) - 5.180175e6*fcs_pow(u,8)*(-3. + 87.*w2 - 174.*v2*(-1. + 31.*w2) + 899.*fcs_pow(v,4)*(-1. + 33.*w2)) + 3.0045015e7*fcs_pow(u,10)*(-1. + 31.*w2 - 62.*v2*(-1. + 33.*w2) + 341.*fcs_pow(v,4)*(-1. + 35.*w2)) - 2.8224105e7*fcs_pow(u,12)*(-1. + 33.*w2 - 66.*v2*(-1. + 35.*w2) + 385.*fcs_pow(v,4)*(-1. + 37.*w2)) + 3.411705e6*fcs_pow(u,14)*(-3. + 105.*w2 - 210.*v2*(-1. + 37.*w2) + 1295.*fcs_pow(v,4)*(-1. + 39.*w2)));
        case 3 : return -4.28772250281375e14*fcs_pow(t,22)*w*(-13.*(-3. + 23.*w2 + 575.*fcs_pow(v,4)*(-1. + 9.*w2) - 46.*v2*(-3. + 25.*w2)) + 2093.*u2*(-3. + 25.*w2 - 150.*v2*(-1. + 9.*w2) + 225.*fcs_pow(v,4)*(-3. + 29.*w2)) + 470925.*fcs_pow(u,6)*(-3. + 29.*w2 + 899.*fcs_pow(v,4)*(-1. + 11.*w2) - 58.*v2*(-3. + 31.*w2)) - 156975.*fcs_pow(u,4)*(-1. + 9.*w2 + v2*(54. - 522.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)) - 1.950975e6*fcs_pow(u,8)*(-3. + 31.*w2 - 186.*v2*(-1. + 11.*w2) + 341.*fcs_pow(v,4)*(-3. + 35.*w2)) - 4.032015e6*fcs_pow(u,12)*(-3. + 35.*w2 + 1295.*fcs_pow(v,4)*(-1. + 13.*w2) - 70.*v2*(-3. + 37.*w2)) + 4.032015e6*fcs_pow(u,10)*(-3. + 33.*w2 - 66.*v2*(-3. + 35.*w2) + 385.*fcs_pow(v,4)*(-3. + 37.*w2)) + 1.550775e6*fcs_pow(u,14)*(-3. + 37.*w2 - 222.*v2*(-1. + 13.*w2) + 481.*fcs_pow(v,4)*(-3. + 41.*w2)));
        case 4 : return 4.28772250281375e14*fcs_pow(t,23)*(-13.*(3. - 138.*w2 + 575.*fcs_pow(w,4) - 138.*v2*(1. - 50.*w2 + 225.*fcs_pow(w,4)) + 575.*fcs_pow(v,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4))) + 470925.*fcs_pow(u,6)*(3. - 174.*w2 + 899.*fcs_pow(w,4) - 174.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4))) - 4.032015e6*fcs_pow(u,12)*(3. - 210.*w2 + 1295.*fcs_pow(w,4) - 210.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 1295.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4))) - 156975.*fcs_pow(u,4)*(1. - 54.*w2 + 261.*fcs_pow(w,4) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 6279.*u2*(1. - 50.*w2 + 225.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 75.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 1.2096045e7*fcs_pow(u,10)*(1. - 66.*w2 + 385.*fcs_pow(w,4) + 385.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 22.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 1.950975e6*fcs_pow(u,8)*(3.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 186.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 1.550775e6*fcs_pow(u,14)*(3.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 222.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
      }
      case 5 : switch(k){
        case 0 : return -1.856156927625e12*fcs_pow(t,20)*v*(-1.41120525e8*fcs_pow(u,12)*(1. - 22.*v2 + 77.*fcs_pow(v,4)) + 35035.*u2*(3. - 46.*v2 + 115.*fcs_pow(v,4)) - 805805.*fcs_pow(u,4)*(3. - 50.*v2 + 135.*fcs_pow(v,4)) - 143.*(5. - 70.*v2 + 161.*fcs_pow(v,4)) + 1.7058525e7*fcs_pow(u,14)*(3. - 70.*v2 + 259.*fcs_pow(v,4)) + 4.029025e6*fcs_pow(u,6)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) - 5.180175e6*fcs_pow(u,8)*(15. - 290.*v2 + 899.*fcs_pow(v,4)) + 1.0015005e7*fcs_pow(u,10)*(15. - 310.*v2 + 1023.*fcs_pow(v,4)));
        case 1 : return 7.14620417135625e14*fcs_pow(t,21)*v*(1.2096045e7*fcs_pow(u,10)*(1. - 22.*v2 + 77.*fcs_pow(v,4)) - 13.*(3. - 46.*v2 + 115.*fcs_pow(v,4)) + 2093.*u2*(3. - 50.*v2 + 135.*fcs_pow(v,4)) - 4.032015e6*fcs_pow(u,12)*(3. - 70.*v2 + 259.*fcs_pow(v,4)) - 31395.*fcs_pow(u,4)*(5. - 90.*v2 + 261.*fcs_pow(v,4)) + 94185.*fcs_pow(u,6)*(15. - 290.*v2 + 899.*fcs_pow(v,4)) - 390195.*fcs_pow(u,8)*(15. - 310.*v2 + 1023.*fcs_pow(v,4)) + 310155.*fcs_pow(u,14)*(15. - 370.*v2 + 1443.*fcs_pow(v,4)))*w;
        case 2 : return -7.14620417135625e14*fcs_pow(t,22)*v*(-13.*(-3. + 69.*w2 - 46.*v2*(-1. + 25.*w2) + 115.*fcs_pow(v,4)*(-1. + 27.*w2)) + 2093.*u2*(-3. + 75.*w2 - 50.*v2*(-1. + 27.*w2) + 135.*fcs_pow(v,4)*(-1. + 29.*w2)) - 31395.*fcs_pow(u,4)*(-5. + 135.*w2 - 90.*v2*(-1. + 29.*w2) + 261.*fcs_pow(v,4)*(-1. + 31.*w2)) + 94185.*fcs_pow(u,6)*(-15. + 435.*w2 - 290.*v2*(-1. + 31.*w2) + 899.*fcs_pow(v,4)*(-1. + 33.*w2)) - 390195.*fcs_pow(u,8)*(-15. + 465.*w2 - 310.*v2*(-1. + 33.*w2) + 1023.*fcs_pow(v,4)*(-1. + 35.*w2)) + 1.2096045e7*fcs_pow(u,10)*(-1. + 33.*w2 + v2*(22. - 770.*w2) + 77.*fcs_pow(v,4)*(-1. + 37.*w2)) - 4.032015e6*fcs_pow(u,12)*(-3. + 105.*w2 - 70.*v2*(-1. + 37.*w2) + 259.*fcs_pow(v,4)*(-1. + 39.*w2)) + 310155.*fcs_pow(u,14)*(-15. + 555.*w2 - 370.*v2*(-1. + 39.*w2) + 1443.*fcs_pow(v,4)*(-1. + 41.*w2)));
        case 3 : return 4.930880878235812e16*fcs_pow(t,23)*v*w*(-13.*(-3. + 25.*w2 - 50.*v2*(-1. + 9.*w2) + 45.*fcs_pow(v,4)*(-3. + 29.*w2)) - 4095.*fcs_pow(u,4)*(-45. + 435.*w2 + 2697.*fcs_pow(v,4)*(-1. + 11.*w2) - 290.*v2*(-3. + 31.*w2)) + 1365.*u2*(-5. + 45.*w2 + v2*(90. - 870.*w2) + 87.*fcs_pow(v,4)*(-3. + 31.*w2)) + 118755.*fcs_pow(u,6)*(-310.*v2*(-1. + 11.*w2) + 5.*(-3. + 31.*w2) + 341.*fcs_pow(v,4)*(-3. + 35.*w2)) + 1.928355e6*fcs_pow(u,10)*(-9. + 105.*w2 + 777.*fcs_pow(v,4)*(-1. + 13.*w2) - 70.*v2*(-3. + 37.*w2)) - 2.629575e6*fcs_pow(u,8)*(-3. + 33.*w2 + v2*(66. - 770.*w2) + 77.*fcs_pow(v,4)*(-3. + 37.*w2)) - 1.227135e6*fcs_pow(u,12)*(-370.*v2*(-1. + 13.*w2) + 5.*(-3. + 37.*w2) + 481.*fcs_pow(v,4)*(-3. + 41.*w2)) + 498945.*fcs_pow(u,14)*(-15. + 195.*w2 - 130.*v2*(-3. + 41.*w2) + 533.*fcs_pow(v,4)*(-3. + 43.*w2)));
        case 4 : return -4.930880878235812e16*fcs_pow(t,24)*v*(5.785065e6*fcs_pow(u,10)*(3. - 210.*w2 + 1295.*fcs_pow(w,4) - 70.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 259.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4))) - 12285.*fcs_pow(u,4)*(-290.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 899.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4))) + 1365.*u2*(5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 261.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 30.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 13.*(3. - 150.*w2 + 675.*fcs_pow(w,4) - 50.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 45.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 2.629575e6*fcs_pow(u,8)*(3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 231.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 22.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 118755.*fcs_pow(u,6)*(15.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 310.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 498945.*fcs_pow(u,14)*(15.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 1599.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 130.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 1.227135e6*fcs_pow(u,12)*(15.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 370.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return 9.615217712559834e18*fcs_pow(t,25)*v*w*(-5.*(5. - 90.*w2 + 261.*fcs_pow(w,4)) + 30.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) - 87.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) - 9135.*fcs_pow(u,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 310.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 341.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4))) + 166315.*fcs_pow(u,14)*(45. - 1230.*w2 + 5289.*fcs_pow(w,4) + 5289.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 410.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4))) + 105.*u2*(45. - 870.*w2 + 2697.*fcs_pow(w,4) + 2697.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 58.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 148335.*fcs_pow(u,8)*(15.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 777.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 70.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 94395.*fcs_pow(u,6)*(15.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 110.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 1.164205e6*fcs_pow(u,12)*(533.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 26.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 207669.*fcs_pow(u,10)*(75. - 1850.*w2 + 7215.*fcs_pow(w,4) - 370.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))));
      }
      case 6 : switch(k){
        case 0 : return 9.280784638125e12*fcs_pow(t,21)*(-143.*(-1. + 63.*v2 - 483.*fcs_pow(v,4) + 805.*fcs_pow(v,6)) + 21021.*u2*(-1. + 69.*v2 - 575.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) - 483483.*fcs_pow(u,4)*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 3.0045015e7*fcs_pow(u,10)*(-1. + 93.*v2 - 1023.*fcs_pow(v,4) + 2387.*fcs_pow(v,6)) - 2.8224105e7*fcs_pow(u,12)*(-1. + 99.*v2 - 1155.*fcs_pow(v,4) + 2849.*fcs_pow(v,6)) + 1.0235115e7*fcs_pow(u,14)*(-1. + 105.*v2 - 1295.*fcs_pow(v,4) + 3367.*fcs_pow(v,6)) + 805805.*fcs_pow(u,6)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) - 3.108105e6*fcs_pow(u,8)*(-5. + 435.*v2 - 4495.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)));
        case 1 : return -2.143861251406875e15*fcs_pow(t,22)*(13.*(1. - 69.*v2 + 575.*fcs_pow(v,4) - 1035.*fcs_pow(v,6)) + 2093.*u2*(-1. + 75.*v2 - 675.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 1.950975e6*fcs_pow(u,8)*(-1. + 93.*v2 - 1023.*fcs_pow(v,4) + 2387.*fcs_pow(v,6)) + 4.032015e6*fcs_pow(u,10)*(-1. + 99.*v2 - 1155.*fcs_pow(v,4) + 2849.*fcs_pow(v,6)) - 4.032015e6*fcs_pow(u,12)*(-1. + 105.*v2 - 1295.*fcs_pow(v,4) + 3367.*fcs_pow(v,6)) - 10465.*fcs_pow(u,4)*(-5. + 405.*v2 - 3915.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 94185.*fcs_pow(u,6)*(-5. + 435.*v2 - 4495.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)) + 310155.*fcs_pow(u,14)*(-5. + 555.*v2 - 7215.*fcs_pow(v,4) + 19721.*fcs_pow(v,6)))*w;
        case 2 : return 2.143861251406875e15*fcs_pow(t,23)*(-13.*(1. - 23.*w2 + 69.*v2*(-1. + 25.*w2) - 575.*fcs_pow(v,4)*(-1. + 27.*w2) + 1035.*fcs_pow(v,6)*(-1. + 29.*w2)) + 2093.*u2*(1. - 25.*w2 + 75.*v2*(-1. + 27.*w2) - 675.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) - 10465.*fcs_pow(u,4)*(5. - 135.*w2 + 405.*v2*(-1. + 29.*w2) - 3915.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) + 94185.*fcs_pow(u,6)*(5. - 145.*w2 + 435.*v2*(-1. + 31.*w2) - 4495.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)) - 1.950975e6*fcs_pow(u,8)*(1. - 31.*w2 + 93.*v2*(-1. + 33.*w2) - 1023.*fcs_pow(v,4)*(-1. + 35.*w2) + 2387.*fcs_pow(v,6)*(-1. + 37.*w2)) + 4.032015e6*fcs_pow(u,10)*(1. - 33.*w2 + 99.*v2*(-1. + 35.*w2) - 1155.*fcs_pow(v,4)*(-1. + 37.*w2) + 2849.*fcs_pow(v,6)*(-1. + 39.*w2)) - 4.032015e6*fcs_pow(u,12)*(1. - 35.*w2 + 105.*v2*(-1. + 37.*w2) - 1295.*fcs_pow(v,4)*(-1. + 39.*w2) + 3367.*fcs_pow(v,6)*(-1. + 41.*w2)) + 310155.*fcs_pow(u,14)*(5. - 185.*w2 + 555.*v2*(-1. + 39.*w2) - 7215.*fcs_pow(v,4)*(-1. + 41.*w2) + 19721.*fcs_pow(v,6)*(-1. + 43.*w2)));
        case 3 : return -4.930880878235812e16*fcs_pow(t,24)*w*(1365.*u2*(5. - 45.*w2 + 8091.*fcs_pow(v,6)*(-1. + 11.*w2) + 135.*v2*(-3. + 29.*w2) - 1305.*fcs_pow(v,4)*(-3. + 31.*w2)) - 13.*(3. - 25.*w2 + 225.*v2*(-1. + 9.*w2) - 675.*fcs_pow(v,4)*(-3. + 29.*w2) + 1305.*fcs_pow(v,6)*(-3. + 31.*w2)) - 12285.*fcs_pow(u,4)*(15. - 145.*w2 - 13485.*fcs_pow(v,4)*(-1. + 11.*w2) + 435.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) - 7.888725e6*fcs_pow(u,8)*(1. - 11.*w2 + 2849.*fcs_pow(v,6)*(-1. + 13.*w2) + 33.*v2*(-3. + 35.*w2) - 385.*fcs_pow(v,4)*(-3. + 37.*w2)) + 593775.*fcs_pow(u,6)*(3. - 31.*w2 + 279.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 2387.*fcs_pow(v,6)*(-3. + 37.*w2)) + 5.785065e6*fcs_pow(u,10)*(3. - 35.*w2 - 3885.*fcs_pow(v,4)*(-1. + 13.*w2) + 105.*v2*(-3. + 37.*w2) + 3367.*fcs_pow(v,6)*(-3. + 41.*w2)) + 1.496835e6*fcs_pow(u,14)*(5. - 65.*w2 + 22919.*fcs_pow(v,6)*(-1. + 15.*w2) + 195.*v2*(-3. + 41.*w2) - 2665.*fcs_pow(v,4)*(-3. + 43.*w2)) - 1.227135e6*fcs_pow(u,12)*(15. - 185.*w2 + 1665.*v2*(-1. + 13.*w2) - 7215.*fcs_pow(v,4)*(-3. + 41.*w2) + 19721.*fcs_pow(v,6)*(-3. + 43.*w2)));
        case 4 : return 1.479264263470744e17*fcs_pow(t,25)*(455.*u2*(-5.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 3915.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 135.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 13.*(-1. + 50.*w2 - 225.*fcs_pow(w,4) + 75.*v2*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 1305.*fcs_pow(v,6)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 225.*fcs_pow(v,4)*(3. - 174.*w2 + 899.*fcs_pow(w,4))) - 2.629575e6*fcs_pow(u,8)*(-1. + 66.*w2 - 385.*fcs_pow(w,4) - 1155.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 2849.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 33.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 593775.*fcs_pow(u,6)*(-1. + 62.*w2 - 341.*fcs_pow(w,4) + 93.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 4095.*fcs_pow(u,4)*(1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 13485.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 498945.*fcs_pow(u,14)*(-5.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 7995.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 22919.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 195.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 1.227135e6*fcs_pow(u,12)*(-5.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 555.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 2405.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 1.928355e6*fcs_pow(u,10)*(-3. + 210.*w2 - 1295.*fcs_pow(w,4) + 315.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 3885.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 3367.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))));
        case 5 : return -4.807608856279917e19*fcs_pow(t,26)*w*(5. - 90.*w2 + 261.*fcs_pow(w,4) - 8091.*fcs_pow(v,6)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 27.*v2*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 261.*fcs_pow(v,4)*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 63.*u2*(-15. + 290.*w2 - 899.*fcs_pow(w,4) - 13485.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 87.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) + 56637.*fcs_pow(u,6)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 165.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2849.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 385.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 1827.*fcs_pow(u,4)*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 1395.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 5115.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 2387.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 698523.*fcs_pow(u,12)*(-5. + 130.*w2 - 533.*fcs_pow(w,4) + 22919.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 2665.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 39.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 207669.*fcs_pow(u,10)*(-15. + 370.*w2 - 1443.*fcs_pow(w,4) + 19721.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 333.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 1443.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 89001.*fcs_pow(u,8)*(-5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 3885.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 3367.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 99789.*fcs_pow(u,14)*(-15. + 410.*w2 - 1763.*fcs_pow(w,4) - 26445.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 615.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 5289.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))));
        case 6 : return 4.807608856279917e19*fcs_pow(t,27)*(-5. + 405.*w2 - 3915.*fcs_pow(w,4) + 8091.*fcs_pow(w,6) + 3915.*fcs_pow(v,4)*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 8091.*fcs_pow(v,6)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 81.*v2*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6)) + 189.*u2*(5. - 435.*w2 + 4495.*fcs_pow(w,4) - 9889.*fcs_pow(w,6) + 435.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 4495.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6))) - 698523.*fcs_pow(u,12)*(5. - 585.*w2 + 7995.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 585.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 7995.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 22919.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6))) - 5481.*fcs_pow(u,4)*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 5115.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 2387.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 623007.*fcs_pow(u,10)*(5. - 555.*w2 + 7215.*fcs_pow(w,4) - 19721.*fcs_pow(w,6) - 7215.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 111.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 1.335015e6*fcs_pow(u,8)*(1. - 105.*w2 + 1295.*fcs_pow(w,4) - 3367.*fcs_pow(w,6) + 3367.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 21.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 56637.*fcs_pow(u,6)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 1155.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2849.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 299367.*fcs_pow(u,14)*(5. - 615.*w2 + 8815.*fcs_pow(w,4) - 26445.*fcs_pow(w,6) + 615.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 8815.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 5289.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))));
      }
      case 7 : switch(k){
        case 0 : return -2.143861251406875e15*fcs_pow(t,22)*v*(2.8224105e7*fcs_pow(u,10)*(-1. + 33.*v2 - 231.*fcs_pow(v,4) + 407.*fcs_pow(v,6)) - 2.8224105e7*fcs_pow(u,12)*(-1. + 35.*v2 - 259.*fcs_pow(v,4) + 481.*fcs_pow(v,6)) - 13.*(-7. + 161.*v2 - 805.*fcs_pow(v,4) + 1035.*fcs_pow(v,6)) + 2093.*u2*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) - 2.731365e6*fcs_pow(u,8)*(-5. + 155.*v2 - 1023.*fcs_pow(v,4) + 1705.*fcs_pow(v,6)) - 10465.*fcs_pow(u,4)*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) + 94185.*fcs_pow(u,6)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)) + 310155.*fcs_pow(u,14)*(-35. + 1295.*v2 - 10101.*fcs_pow(v,4) + 19721.*fcs_pow(v,6)));
        case 1 : return 4.930880878235812e16*fcs_pow(t,23)*v*(-1.8407025e7*fcs_pow(u,8)*(-1. + 33.*v2 - 231.*fcs_pow(v,4) + 407.*fcs_pow(v,6)) + 4.0495455e7*fcs_pow(u,10)*(-1. + 35.*v2 - 259.*fcs_pow(v,4) + 481.*fcs_pow(v,6)) - 13.*(-7. + 175.*v2 - 945.*fcs_pow(v,4) + 1305.*fcs_pow(v,6)) + 831285.*fcs_pow(u,6)*(-5. + 155.*v2 - 1023.*fcs_pow(v,4) + 1705.*fcs_pow(v,6)) + 455.*u2*(-35. + 945.*v2 - 5481.*fcs_pow(v,4) + 8091.*fcs_pow(v,6)) - 12285.*fcs_pow(u,4)*(-35. + 1015.*v2 - 6293.*fcs_pow(v,4) + 9889.*fcs_pow(v,6)) - 1.227135e6*fcs_pow(u,12)*(-35. + 1295.*v2 - 10101.*fcs_pow(v,4) + 19721.*fcs_pow(v,6)) + 498945.*fcs_pow(u,14)*(-35. + 1365.*v2 - 11193.*fcs_pow(v,4) + 22919.*fcs_pow(v,6)))*w;
        case 2 : return -4.930880878235812e16*fcs_pow(t,24)*v*(-13.*(7. - 175.*w2 + 175.*v2*(-1. + 27.*w2) - 945.*fcs_pow(v,4)*(-1. + 29.*w2) + 1305.*fcs_pow(v,6)*(-1. + 31.*w2)) + 455.*u2*(35. - 945.*w2 + 945.*v2*(-1. + 29.*w2) - 5481.*fcs_pow(v,4)*(-1. + 31.*w2) + 8091.*fcs_pow(v,6)*(-1. + 33.*w2)) - 12285.*fcs_pow(u,4)*(35. - 1015.*w2 + 1015.*v2*(-1. + 31.*w2) - 6293.*fcs_pow(v,4)*(-1. + 33.*w2) + 9889.*fcs_pow(v,6)*(-1. + 35.*w2)) + 831285.*fcs_pow(u,6)*(5. - 155.*w2 + 155.*v2*(-1. + 33.*w2) - 1023.*fcs_pow(v,4)*(-1. + 35.*w2) + 1705.*fcs_pow(v,6)*(-1. + 37.*w2)) - 1.8407025e7*fcs_pow(u,8)*(1. - 33.*w2 + 33.*v2*(-1. + 35.*w2) - 231.*fcs_pow(v,4)*(-1. + 37.*w2) + 407.*fcs_pow(v,6)*(-1. + 39.*w2)) + 4.0495455e7*fcs_pow(u,10)*(1. - 35.*w2 + 35.*v2*(-1. + 37.*w2) - 259.*fcs_pow(v,4)*(-1. + 39.*w2) + 481.*fcs_pow(v,6)*(-1. + 41.*w2)) - 1.227135e6*fcs_pow(u,12)*(35. - 1295.*w2 + 1295.*v2*(-1. + 39.*w2) - 10101.*fcs_pow(v,4)*(-1. + 41.*w2) + 19721.*fcs_pow(v,6)*(-1. + 43.*w2)) + 498945.*fcs_pow(u,14)*(35. - 1365.*w2 + 1365.*v2*(-1. + 41.*w2) - 11193.*fcs_pow(v,4)*(-1. + 43.*w2) + 22919.*fcs_pow(v,6)*(-1. + 45.*w2)));
        case 3 : return 9.615217712559834e18*fcs_pow(t,25)*v*w*(35.*(-1. + 9.*w2) - 8091.*fcs_pow(v,6)*(-1. + 11.*w2) - 315.*v2*(-3. + 29.*w2) + 1827.*fcs_pow(v,4)*(-3. + 31.*w2) + 63.*u2*(35.*(3. - 29.*w2) - 18879.*fcs_pow(v,4)*(-1. + 11.*w2) + 1015.*v2*(-3. + 31.*w2) + 9889.*fcs_pow(v,6)*(-3. + 35.*w2)) + 1.982295e6*fcs_pow(u,6)*(1. - 11.*w2 + 407.*fcs_pow(v,6)*(-1. + 13.*w2) + 11.*v2*(-3. + 35.*w2) - 77.*fcs_pow(v,4)*(-3. + 37.*w2)) - 12789.*fcs_pow(u,4)*(15. - 155.*w2 + 465.*v2*(-1. + 11.*w2) - 1023.*fcs_pow(v,4)*(-3. + 35.*w2) + 1705.*fcs_pow(v,6)*(-3. + 37.*w2)) - 3.115035e6*fcs_pow(u,8)*(3. - 35.*w2 - 777.*fcs_pow(v,4)*(-1. + 13.*w2) + 35.*v2*(-3. + 37.*w2) + 481.*fcs_pow(v,6)*(-3. + 41.*w2)) - 698523.*fcs_pow(u,12)*(35. - 455.*w2 + 22919.*fcs_pow(v,6)*(-1. + 15.*w2) + 455.*v2*(-3. + 41.*w2) - 3731.*fcs_pow(v,4)*(-3. + 43.*w2)) + 207669.*fcs_pow(u,10)*(35.*(3. - 37.*w2) + 3885.*v2*(-1. + 13.*w2) - 10101.*fcs_pow(v,4)*(-3. + 41.*w2) + 19721.*fcs_pow(v,6)*(-3. + 43.*w2)) + 99789.*fcs_pow(u,14)*(35.*(3. - 41.*w2) - 37023.*fcs_pow(v,4)*(-1. + 15.*w2) + 1435.*v2*(-3. + 43.*w2) + 26445.*fcs_pow(v,6)*(-3. + 47.*w2)));
        case 4 : return -9.615217712559834e18*fcs_pow(t,26)*v*(35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) + 5481.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 8091.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 315.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 1.982295e6*fcs_pow(u,6)*(-1. + 66.*w2 - 385.*fcs_pow(w,4) - 231.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 11.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 38367.*fcs_pow(u,4)*(-5.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 155.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 1705.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 341.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 63.*u2*(3045.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 18879.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 35.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 698523.*fcs_pow(u,12)*(-35.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 11193.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 22919.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 455.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 623007.*fcs_pow(u,10)*(-35.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 1295.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 3367.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 3.115035e6*fcs_pow(u,8)*(-3. + 210.*w2 - 1295.*fcs_pow(w,4) + 105.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 777.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 481.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 99789.*fcs_pow(u,14)*(4305.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 37023.*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 35.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 26445.*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return 1.442282656883975e20*fcs_pow(t,27)*v*w*(-3.*(-18879.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 9889.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.*(15. - 290.*w2 + 899.*fcs_pow(w,4)) + 203.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4))) - 396459.*fcs_pow(u,4)*(-5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 55.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 407.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 77.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 4263.*u2*(-15. + 310.*w2 - 1023.*fcs_pow(w,4) + 465.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 1023.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 341.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) + 7.683753e6*fcs_pow(u,10)*(22919.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 3731.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 91.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 207669.*fcs_pow(u,8)*(98605.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3885.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 35.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 10101.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 1.453683e6*fcs_pow(u,6)*(-5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 777.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 35.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 481.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 3.026933e6*fcs_pow(u,12)*(-37023.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1435.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 5289.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 1.363783e6*fcs_pow(u,14)*(4515.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 35.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 2709.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 6063.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return -4.326847970651925e20*fcs_pow(t,28)*v*(-3.*(1015.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 6293.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 9889.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 7.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6))) + 4263.*u2*(5. - 465.*w2 + 5115.*fcs_pow(w,4) - 11935.*fcs_pow(w,6) + 155.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 1023.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 2.561251e6*fcs_pow(u,10)*(1365.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 11193.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 22919.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 1.038345e6*fcs_pow(u,8)*(-10101.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 259.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 1.453683e6*fcs_pow(u,6)*(5. - 525.*w2 + 6475.*fcs_pow(w,4) - 16835.*fcs_pow(w,6) + 2405.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 35.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 132153.*fcs_pow(u,4)*(5. - 495.*w2 + 5775.*fcs_pow(w,4) - 14245.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 3.026933e6*fcs_pow(u,12)*(-35.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1435.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 12341.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 5289.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 1.363783e6*fcs_pow(u,14)*(-35.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1505.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 2709.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 6063.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return 8.783501380423409e22*fcs_pow(t,29)*v*w*(3.444441e6*fcs_pow(u,10)*(35. - 1435.*w2 + 12341.*fcs_pow(w,4) - 26445.*fcs_pow(w,6) - 12341.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 5289.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 205.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6))) - 4.279457e6*fcs_pow(u,12)*(1505.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 2709.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 5.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 6063.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6))) - 3.*(7.*(5. - 155.*w2 + 1023.*fcs_pow(w,4) - 1705.*fcs_pow(w,6)) + 1085.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 7161.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 341.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) + 651.*u2*(-35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 1155.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 231.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 407.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 1.324785e6*fcs_pow(u,8)*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 22919.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1599.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 39.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 7161.*fcs_pow(u,6)*(98605.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 35.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 10101.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 21483.*fcs_pow(u,4)*(-35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 35.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 259.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 481.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 1.4155127e7*fcs_pow(u,14)*(5.*(1. - 45.*w2 + 423.*fcs_pow(w,4) - 987.*fcs_pow(w,6)) + 45.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 423.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 141.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
      }
      case 8 : switch(k){
        case 0 : return 2.143861251406875e15*fcs_pow(t,23)*(-1.3656825e7*fcs_pow(u,8)*(1. - 124.*v2 + 2046.*fcs_pow(v,4) - 9548.*fcs_pow(v,6) + 12617.*fcs_pow(v,8)) + 2.8224105e7*fcs_pow(u,10)*(1. - 132.*v2 + 2310.*fcs_pow(v,4) - 11396.*fcs_pow(v,6) + 15873.*fcs_pow(v,8)) - 2.8224105e7*fcs_pow(u,12)*(1. - 140.*v2 + 2590.*fcs_pow(v,4) - 13468.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) - 13.*(7. - 644.*v2 + 8050.*fcs_pow(v,4) - 28980.*fcs_pow(v,6) + 30015.*fcs_pow(v,8)) + 2093.*u2*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) + 659295.*fcs_pow(u,6)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) - 10465.*fcs_pow(u,4)*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)) + 310155.*fcs_pow(u,14)*(35. - 5180.*v2 + 101010.*fcs_pow(v,4) - 552188.*fcs_pow(v,6) + 848003.*fcs_pow(v,8)));
        case 1 : return -4.930880878235812e16*fcs_pow(t,24)*(4.156425e6*fcs_pow(u,6)*(1. - 124.*v2 + 2046.*fcs_pow(v,4) - 9548.*fcs_pow(v,6) + 12617.*fcs_pow(v,8)) - 1.8407025e7*fcs_pow(u,8)*(1. - 132.*v2 + 2310.*fcs_pow(v,4) - 11396.*fcs_pow(v,6) + 15873.*fcs_pow(v,8)) + 4.0495455e7*fcs_pow(u,10)*(1. - 140.*v2 + 2590.*fcs_pow(v,4) - 13468.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) - 13.*(7. - 700.*v2 + 9450.*fcs_pow(v,4) - 36540.*fcs_pow(v,6) + 40455.*fcs_pow(v,8)) - 85995.*fcs_pow(u,4)*(5. - 580.*v2 + 8990.*fcs_pow(v,4) - 39556.*fcs_pow(v,6) + 49445.*fcs_pow(v,8)) + 455.*u2*(35. - 3780.*v2 + 54810.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 267003.*fcs_pow(v,8)) - 1.227135e6*fcs_pow(u,12)*(35. - 5180.*v2 + 101010.*fcs_pow(v,4) - 552188.*fcs_pow(v,6) + 848003.*fcs_pow(v,8)) + 498945.*fcs_pow(u,14)*(35. - 5460.*v2 + 111930.*fcs_pow(v,4) - 641732.*fcs_pow(v,6) + 1.031355e6*fcs_pow(v,8)))*w;
        case 2 : return 4.930880878235812e16*fcs_pow(t,25)*(-13.*(-7. + 175.*w2 - 700.*v2*(-1. + 27.*w2) + 9450.*fcs_pow(v,4)*(-1. + 29.*w2) - 36540.*fcs_pow(v,6)*(-1. + 31.*w2) + 40455.*fcs_pow(v,8)*(-1. + 33.*w2)) + 455.*u2*(-35. + 945.*w2 - 3780.*v2*(-1. + 29.*w2) + 54810.*fcs_pow(v,4)*(-1. + 31.*w2) - 226548.*fcs_pow(v,6)*(-1. + 33.*w2) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2)) - 85995.*fcs_pow(u,4)*(-5. + 145.*w2 - 580.*v2*(-1. + 31.*w2) + 8990.*fcs_pow(v,4)*(-1. + 33.*w2) - 39556.*fcs_pow(v,6)*(-1. + 35.*w2) + 49445.*fcs_pow(v,8)*(-1. + 37.*w2)) + 4.156425e6*fcs_pow(u,6)*(-1. + 31.*w2 - 124.*v2*(-1. + 33.*w2) + 2046.*fcs_pow(v,4)*(-1. + 35.*w2) - 9548.*fcs_pow(v,6)*(-1. + 37.*w2) + 12617.*fcs_pow(v,8)*(-1. + 39.*w2)) - 1.8407025e7*fcs_pow(u,8)*(-1. + 33.*w2 - 132.*v2*(-1. + 35.*w2) + 2310.*fcs_pow(v,4)*(-1. + 37.*w2) - 11396.*fcs_pow(v,6)*(-1. + 39.*w2) + 15873.*fcs_pow(v,8)*(-1. + 41.*w2)) + 4.0495455e7*fcs_pow(u,10)*(-1. + 35.*w2 - 140.*v2*(-1. + 37.*w2) + 2590.*fcs_pow(v,4)*(-1. + 39.*w2) - 13468.*fcs_pow(v,6)*(-1. + 41.*w2) + 19721.*fcs_pow(v,8)*(-1. + 43.*w2)) - 1.227135e6*fcs_pow(u,12)*(35.*(-1. + 37.*w2) - 5180.*v2*(-1. + 39.*w2) + 101010.*fcs_pow(v,4)*(-1. + 41.*w2) - 552188.*fcs_pow(v,6)*(-1. + 43.*w2) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2)) + 498945.*fcs_pow(u,14)*(35.*(-1. + 39.*w2) - 5460.*v2*(-1. + 41.*w2) + 111930.*fcs_pow(v,4)*(-1. + 43.*w2) - 641732.*fcs_pow(v,6)*(-1. + 45.*w2) + 1.031355e6*fcs_pow(v,8)*(-1. + 47.*w2)));
        case 3 : return -9.615217712559834e18*fcs_pow(t,26)*w*(35. - 315.*w2 + 226548.*fcs_pow(v,6)*(-1. + 11.*w2) + 1260.*v2*(-3. + 29.*w2) - 18270.*fcs_pow(v,4)*(-3. + 31.*w2) - 89001.*fcs_pow(v,8)*(-3. + 35.*w2) - 63945.*fcs_pow(u,4)*(-3. + 31.*w2 - 372.*v2*(-1. + 11.*w2) + 37851.*fcs_pow(v,8)*(-1. + 13.*w2) + 2046.*fcs_pow(v,4)*(-3. + 35.*w2) - 9548.*fcs_pow(v,6)*(-3. + 37.*w2)) + 441.*u2*(26970.*fcs_pow(v,4)*(-1. + 11.*w2) + 5.*(-3. + 29.*w2) - 580.*v2*(-3. + 31.*w2) - 39556.*fcs_pow(v,6)*(-3. + 35.*w2) + 49445.*fcs_pow(v,8)*(-3. + 37.*w2)) + 1.982295e6*fcs_pow(u,6)*(-1. + 11.*w2 - 11396.*fcs_pow(v,6)*(-1. + 13.*w2) - 44.*v2*(-3. + 35.*w2) + 770.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)) + 207669.*fcs_pow(u,10)*(-15540.*v2*(-1. + 13.*w2) + 2.544009e6*fcs_pow(v,8)*(-1. + 15.*w2) + 35.*(-3. + 37.*w2) + 101010.*fcs_pow(v,4)*(-3. + 41.*w2) - 552188.*fcs_pow(v,6)*(-3. + 43.*w2)) - 3.115035e6*fcs_pow(u,8)*(-3. + 35.*w2 + 7770.*fcs_pow(v,4)*(-1. + 13.*w2) - 140.*v2*(-3. + 37.*w2) - 13468.*fcs_pow(v,6)*(-3. + 41.*w2) + 19721.*fcs_pow(v,8)*(-3. + 43.*w2)) - 698523.*fcs_pow(u,12)*(-35. + 455.*w2 - 641732.*fcs_pow(v,6)*(-1. + 15.*w2) - 1820.*v2*(-3. + 41.*w2) + 37310.*fcs_pow(v,4)*(-3. + 43.*w2) + 343785.*fcs_pow(v,8)*(-3. + 47.*w2)) + 498945.*fcs_pow(u,14)*(74046.*fcs_pow(v,4)*(-1. + 15.*w2) + 7.*(-3. + 41.*w2) - 1148.*v2*(-3. + 43.*w2) - 148092.*fcs_pow(v,6)*(-3. + 47.*w2) + 248583.*fcs_pow(v,8)*(-3. + 49.*w2)));
        case 4 : return 9.615217712559834e18*fcs_pow(t,27)*(-35.*(1. - 54.*w2 + 261.*fcs_pow(w,4)) - 54810.*fcs_pow(v,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 226548.*fcs_pow(v,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 1260.*v2*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 89001.*fcs_pow(v,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 191835.*fcs_pow(u,4)*(1. - 62.*w2 + 341.*fcs_pow(w,4) - 124.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 12617.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 682.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 441.*u2*(-1740.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 148335.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 5.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) + 623007.*fcs_pow(u,10)*(35.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 5180.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 552188.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 33670.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 3.115035e6*fcs_pow(u,8)*(3. - 210.*w2 + 1295.*fcs_pow(w,4) - 420.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 7770.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59163.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 13468.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 1.982295e6*fcs_pow(u,6)*(1. - 66.*w2 + 385.*fcs_pow(w,4) + 2310.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 44.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 498945.*fcs_pow(u,14)*(-3444.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 74046.*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 745749.*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 7.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 148092.*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 698523.*fcs_pow(u,12)*(35.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 111930.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 641732.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 1820.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 343785.*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return -1.009597859818783e21*fcs_pow(t,28)*w*(609.*u2*(15. - 310.*w2 + 1023.*fcs_pow(w,4) - 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 10230.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 37851.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 9548.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 3.*(15. - 290.*w2 + 899.*fcs_pow(w,4) + 26970.*fcs_pow(v,4)*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 39556.*fcs_pow(v,6)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 116.*v2*(15. - 310.*w2 + 1023.*fcs_pow(w,4)) + 9889.*fcs_pow(v,8)*(15. - 370.*w2 + 1443.*fcs_pow(w,4))) - 148335.*fcs_pow(u,8)*(2.544009e6*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 552188.*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 3108.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 7.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 20202.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 207669.*fcs_pow(u,6)*(5.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 140.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 13468.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 56637.*fcs_pow(u,4)*(5.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 220.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 11396.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 770.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 1.097679e6*fcs_pow(u,10)*(-641732.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 37310.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 364.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 68757.*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 1.363783e6*fcs_pow(u,14)*(-2580.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 5.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 127323.*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 3870.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 24252.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 432419.*fcs_pow(u,12)*(370230.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 5740.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 148092.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 248583.*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return 3.028793579456348e21*fcs_pow(t,29)*(-3.*(-5. + 435.*w2 - 4495.*fcs_pow(w,4) + 9889.*fcs_pow(w,6) - 580.*v2*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 8990.*fcs_pow(v,4)*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 39556.*fcs_pow(v,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 9889.*fcs_pow(v,8)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6))) - 148335.*fcs_pow(u,8)*(101010.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 552188.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 1036.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 1.038345e6*fcs_pow(u,6)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6) - 13468.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 28.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 518.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 18879.*fcs_pow(u,4)*(5.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 79365.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 609.*u2*(5.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) - 620.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 365893.*fcs_pow(u,10)*(-5460.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 111930.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 641732.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 206271.*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 1.363783e6*fcs_pow(u,14)*(-5. + 645.*w2 - 9675.*fcs_pow(w,4) + 30315.*fcs_pow(w,6) - 860.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 212205.*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 3870.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 24252.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 432419.*fcs_pow(u,12)*(35.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 5740.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 123410.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 148092.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 248583.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return -8.783501380423409e22*fcs_pow(t,30)*w*(3.444441e6*fcs_pow(u,10)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6) + 123410.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 148092.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 820.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 248583.*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6))) - 3.*(-4340.*v2*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 71610.*fcs_pow(v,4)*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 7.*(-5. + 155.*w2 - 1023.*fcs_pow(w,4) + 1705.*fcs_pow(w,6)) - 9548.*fcs_pow(v,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 12617.*fcs_pow(v,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 1.324785e6*fcs_pow(u,8)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6) - 641732.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 206271.*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 15990.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 156.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 250635.*fcs_pow(u,6)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 78884.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 148.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 2886.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 21483.*fcs_pow(u,4)*(35.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 140.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2590.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 13468.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 651.*u2*(35.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 11396.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 15873.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 1.4155127e7*fcs_pow(u,14)*(5.*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 180.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 4230.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 35955.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 3948.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 4.279457e6*fcs_pow(u,12)*(-6020.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 27090.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 5.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 169764.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 42441.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return 8.783501380423409e22*fcs_pow(t,31)*(-1.324785e6*fcs_pow(u,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8) + 111930.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 641732.*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 780.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 206271.*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8))) + 3.444441e6*fcs_pow(u,10)*(-5740.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 123410.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 5.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 148092.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 248583.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) - 2.9956199e7*fcs_pow(u,12)*(5.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 860.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 3870.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 24252.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 30315.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) + 7.0775635e7*fcs_pow(u,14)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8) - 36.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 846.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 3948.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 7191.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 250635.*fcs_pow(u,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8) - 552188.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 14430.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 148.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 751905.*fcs_pow(u,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8) + 19721.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 1924.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 4.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 74.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 651.*u2*(35.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 79365.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 11396.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 3.*(35.*(1. - 124.*w2 + 2046.*fcs_pow(w,4) - 9548.*fcs_pow(w,6) + 12617.*fcs_pow(w,8)) - 4340.*v2*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 71610.*fcs_pow(v,4)*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 9548.*fcs_pow(v,6)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 12617.*fcs_pow(v,8)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
      }
      case 9 : switch(k){
        case 0 : return -1.479264263470744e17*fcs_pow(t,24)*v*(-1.8407025e7*fcs_pow(u,8)*(3. - 132.*v2 + 1386.*fcs_pow(v,4) - 4884.*fcs_pow(v,6) + 5291.*fcs_pow(v,8)) - 13.*(21. - 700.*v2 + 5670.*fcs_pow(v,4) - 15660.*fcs_pow(v,6) + 13485.*fcs_pow(v,8)) + 1.3498485e7*fcs_pow(u,10)*(9. - 420.*v2 + 4662.*fcs_pow(v,4) - 17316.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) + 1365.*u2*(35. - 1260.*v2 + 10962.*fcs_pow(v,4) - 32364.*fcs_pow(v,6) + 29667.*fcs_pow(v,8)) + 277095.*fcs_pow(u,6)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) + 1.496835e6*fcs_pow(u,14)*(35. - 1820.*v2 + 22386.*fcs_pow(v,4) - 91676.*fcs_pow(v,6) + 114595.*fcs_pow(v,8)) - 4095.*fcs_pow(u,4)*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)) - 409045.*fcs_pow(u,12)*(315. - 15540.*v2 + 181818.*fcs_pow(v,4) - 709956.*fcs_pow(v,6) + 848003.*fcs_pow(v,8)));
        case 1 : return 2.88456531376795e19*fcs_pow(t,25)*v*(-35. + 1260.*v2 - 10962.*fcs_pow(v,4) + 32364.*fcs_pow(v,6) - 29667.*fcs_pow(v,8) + 660765.*fcs_pow(u,6)*(3. - 132.*v2 + 1386.*fcs_pow(v,4) - 4884.*fcs_pow(v,6) + 5291.*fcs_pow(v,8)) - 1.038345e6*fcs_pow(u,8)*(9. - 420.*v2 + 4662.*fcs_pow(v,4) - 17316.*fcs_pow(v,6) + 19721.*fcs_pow(v,8)) - 4263.*fcs_pow(u,4)*(45. - 1860.*v2 + 18414.*fcs_pow(v,4) - 61380.*fcs_pow(v,6) + 63085.*fcs_pow(v,8)) - 698523.*fcs_pow(u,12)*(35. - 1820.*v2 + 22386.*fcs_pow(v,4) - 91676.*fcs_pow(v,6) + 114595.*fcs_pow(v,8)) + 21.*u2*(315. - 12180.*v2 + 113274.*fcs_pow(v,4) - 356004.*fcs_pow(v,6) + 346115.*fcs_pow(v,8)) + 99789.*fcs_pow(u,14)*(105. - 5740.*v2 + 74046.*fcs_pow(v,4) - 317340.*fcs_pow(v,6) + 414305.*fcs_pow(v,8)) + 69223.*fcs_pow(u,10)*(315. - 15540.*v2 + 181818.*fcs_pow(v,4) - 709956.*fcs_pow(v,6) + 848003.*fcs_pow(v,8)))*w;
        case 2 : return -2.88456531376795e19*fcs_pow(t,26)*v*(35. - 945.*w2 + 1260.*v2*(-1. + 29.*w2) - 10962.*fcs_pow(v,4)*(-1. + 31.*w2) + 32364.*fcs_pow(v,6)*(-1. + 33.*w2) - 29667.*fcs_pow(v,8)*(-1. + 35.*w2) + 21.*u2*(315.*(-1. + 29.*w2) - 12180.*v2*(-1. + 31.*w2) + 113274.*fcs_pow(v,4)*(-1. + 33.*w2) - 356004.*fcs_pow(v,6)*(-1. + 35.*w2) + 346115.*fcs_pow(v,8)*(-1. + 37.*w2)) - 4263.*fcs_pow(u,4)*(45.*(-1. + 31.*w2) - 1860.*v2*(-1. + 33.*w2) + 18414.*fcs_pow(v,4)*(-1. + 35.*w2) - 61380.*fcs_pow(v,6)*(-1. + 37.*w2) + 63085.*fcs_pow(v,8)*(-1. + 39.*w2)) + 660765.*fcs_pow(u,6)*(-3. + 99.*w2 - 132.*v2*(-1. + 35.*w2) + 1386.*fcs_pow(v,4)*(-1. + 37.*w2) - 4884.*fcs_pow(v,6)*(-1. + 39.*w2) + 5291.*fcs_pow(v,8)*(-1. + 41.*w2)) - 1.038345e6*fcs_pow(u,8)*(-9. + 315.*w2 - 420.*v2*(-1. + 37.*w2) + 4662.*fcs_pow(v,4)*(-1. + 39.*w2) - 17316.*fcs_pow(v,6)*(-1. + 41.*w2) + 19721.*fcs_pow(v,8)*(-1. + 43.*w2)) + 69223.*fcs_pow(u,10)*(315.*(-1. + 37.*w2) - 15540.*v2*(-1. + 39.*w2) + 181818.*fcs_pow(v,4)*(-1. + 41.*w2) - 709956.*fcs_pow(v,6)*(-1. + 43.*w2) + 848003.*fcs_pow(v,8)*(-1. + 45.*w2)) - 698523.*fcs_pow(u,12)*(35.*(-1. + 39.*w2) - 1820.*v2*(-1. + 41.*w2) + 22386.*fcs_pow(v,4)*(-1. + 43.*w2) - 91676.*fcs_pow(v,6)*(-1. + 45.*w2) + 114595.*fcs_pow(v,8)*(-1. + 47.*w2)) + 99789.*fcs_pow(u,14)*(105.*(-1. + 41.*w2) - 5740.*v2*(-1. + 43.*w2) + 74046.*fcs_pow(v,4)*(-1. + 45.*w2) - 317340.*fcs_pow(v,6)*(-1. + 47.*w2) + 414305.*fcs_pow(v,8)*(-1. + 49.*w2)));
        case 3 : return 8.653695941303851e19*fcs_pow(t,27)*v*w*(945. - 9135.*w2 - 339822.*fcs_pow(v,4)*(-1. + 11.*w2) + 12180.*v2*(-3. + 31.*w2) + 356004.*fcs_pow(v,6)*(-3. + 35.*w2) - 346115.*fcs_pow(v,8)*(-3. + 37.*w2) + 4263.*u2*(-45. + 465.*w2 - 1860.*v2*(-1. + 11.*w2) + 63085.*fcs_pow(v,8)*(-1. + 13.*w2) + 6138.*fcs_pow(v,4)*(-3. + 35.*w2) - 20460.*fcs_pow(v,6)*(-3. + 37.*w2)) - 660765.*fcs_pow(u,4)*(-9. + 99.*w2 - 14652.*fcs_pow(v,6)*(-1. + 13.*w2) - 132.*v2*(-3. + 35.*w2) + 1386.*fcs_pow(v,4)*(-3. + 37.*w2) + 5291.*fcs_pow(v,8)*(-3. + 41.*w2)) - 1.038345e6*fcs_pow(u,8)*(-15540.*v2*(-1. + 13.*w2) + 848003.*fcs_pow(v,8)*(-1. + 15.*w2) + 105.*(-3. + 37.*w2) + 60606.*fcs_pow(v,4)*(-3. + 41.*w2) - 236652.*fcs_pow(v,6)*(-3. + 43.*w2)) + 2.422805e6*fcs_pow(u,6)*(13986.*fcs_pow(v,4)*(-1. + 13.*w2) + 9.*(-3. + 35.*w2) - 420.*v2*(-3. + 37.*w2) - 17316.*fcs_pow(v,6)*(-3. + 41.*w2) + 19721.*fcs_pow(v,8)*(-3. + 43.*w2)) + 7.683753e6*fcs_pow(u,10)*(105.*(-1. + 13.*w2) - 275028.*fcs_pow(v,6)*(-1. + 15.*w2) - 1820.*v2*(-3. + 41.*w2) + 22386.*fcs_pow(v,4)*(-3. + 43.*w2) + 114595.*fcs_pow(v,8)*(-3. + 47.*w2)) + 2.0456745e7*fcs_pow(u,14)*(-1204.*v2*(-1. + 15.*w2) + 99029.*fcs_pow(v,8)*(-1. + 17.*w2) + 7.*(-3. + 43.*w2) + 5418.*fcs_pow(v,4)*(-3. + 47.*w2) - 24252.*fcs_pow(v,6)*(-3. + 49.*w2)) - 3.026933e6*fcs_pow(u,12)*(222138.*fcs_pow(v,4)*(-1. + 15.*w2) + 105.*(-3. + 41.*w2) - 5740.*v2*(-3. + 43.*w2) - 317340.*fcs_pow(v,6)*(-3. + 47.*w2) + 414305.*fcs_pow(v,8)*(-3. + 49.*w2)));
        case 4 : return -2.596108782391155e20*fcs_pow(t,28)*v*(12180.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 113274.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 346115.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 105.*(3. - 174.*w2 + 899.*fcs_pow(w,4)) + 118668.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 1421.*u2*(45.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 1860.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 61380.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 6138.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4))) - 346115.*fcs_pow(u,8)*(315.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 15540.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 709956.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 60606.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 2.422805e6*fcs_pow(u,6)*(9. - 630.*w2 + 3885.*fcs_pow(w,4) - 420.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 4662.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 19721.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 5772.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 220255.*fcs_pow(u,4)*(9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 4158.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 132.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 6.818915e6*fcs_pow(u,14)*(21.*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 1204.*v2*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 72756.*fcs_pow(v,6)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 99029.*fcs_pow(v,8)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 5418.*fcs_pow(v,4)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 3.026933e6*fcs_pow(u,12)*(-5740.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 74046.*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 414305.*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 35.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 105780.*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 2.561251e6*fcs_pow(u,10)*(105.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 67158.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 275028.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 1820.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 114595.*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))));
        case 5 : return 2.635050414127023e23*fcs_pow(t,29)*v*w*(-45. + 930.*w2 - 3069.*fcs_pow(w,4) + 1860.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 6138.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 12617.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 4092.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 2387.*fcs_pow(u,6)*(4.240015e6*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 1.18326e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 15540.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 60606.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 2387.*fcs_pow(u,4)*(45.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 98605.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 13986.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 420.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 17316.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 217.*u2*(45.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) - 660.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 14652.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1386.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 189255.*fcs_pow(u,8)*(-275028.*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 22386.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 21.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 364.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 22919.*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 2.022161e6*fcs_pow(u,14)*(15.*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 27965.*fcs_pow(v,8)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 11844.*fcs_pow(v,6)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 60.*v2*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 846.*fcs_pow(v,4)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 611351.*fcs_pow(u,12)*(-6020.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 35.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 99029.*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 5418.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 24252.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 164021.*fcs_pow(u,10)*(222138.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 5740.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 21.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 63468.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 82861.*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return -2.635050414127023e23*fcs_pow(t,30)*v*(-45.*(-1. + 93.*w2 - 1023.*fcs_pow(w,4) + 2387.*fcs_pow(w,6)) + 1860.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 18414.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 12276.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 12617.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 11935.*fcs_pow(u,6)*(181818.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 709956.*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 63.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 3108.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 7161.*fcs_pow(u,4)*(45.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 86580.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 651.*u2*(15.*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) - 660.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1386.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 4884.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 567765.*fcs_pow(u,8)*(-1820.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 22386.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 91676.*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 22919.*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 6.066483e6*fcs_pow(u,14)*(5.*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 19740.*fcs_pow(v,6)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 27965.*fcs_pow(v,8)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 60.*v2*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 846.*fcs_pow(v,4)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 611351.*fcs_pow(u,12)*(105.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 6020.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 495145.*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 16254.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 72756.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 492063.*fcs_pow(u,10)*(105.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 5740.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 74046.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 63468.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 82861.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return 2.450596885138131e25*fcs_pow(t,31)*v*w*(-105.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) + 4620.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 1386.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4884.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 5291.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 555555.*fcs_pow(u,8)*(-105. + 4305.*w2 - 37023.*fcs_pow(w,4) + 79335.*fcs_pow(w,6) + 74046.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 63468.*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 820.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 82861.*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6))) + 299145.*fcs_pow(u,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6) - 91676.*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 22919.*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 3198.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 52.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 77.*fcs_pow(u,4)*(2.9680105e7*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 3.54978e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 15540.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 181818.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 77.*u2*(315.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 98605.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 420.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 4662.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 17316.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 456617.*fcs_pow(u,14)*(45.*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 2820.*v2*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 143820.*fcs_pow(v,6)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 211735.*fcs_pow(v,8)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 5922.*fcs_pow(v,4)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 4.1552147e7*fcs_pow(u,12)*(5.*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 60.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 846.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 3995.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 564.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 1.518517e6*fcs_pow(u,10)*(-6020.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 16254.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 15.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 72756.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 14147.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))));
        case 8 : return -2.450596885138131e25*fcs_pow(t,32)*v*(-105.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) + 4620.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 26455.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1386.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 4884.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) + 299145.*fcs_pow(u,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8) + 22386.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 91676.*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 260.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8))) - 555555.*fcs_pow(u,8)*(-5740.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 74046.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 15.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 63468.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 82861.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) + 1.518517e6*fcs_pow(u,10)*(105.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 6020.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 16254.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 72756.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 70735.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) - 4.1552147e7*fcs_pow(u,12)*(5.*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 60.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 846.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2820.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 3995.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 2.283085e6*fcs_pow(u,14)*(9.*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 564.*v2*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 5922.*fcs_pow(v,4)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 28764.*fcs_pow(v,6)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,8)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 2695.*fcs_pow(u,4)*(-709956.*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 25974.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 9.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 444.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 77.*u2*(315.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 690235.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 86580.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 420.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 4662.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return 2.450596885138131e25*fcs_pow(t,33)*v*w*(3.888885e6*fcs_pow(u,6)*(666414.*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 2460.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 63468.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 82861.*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 9.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 897435.*fcs_pow(u,4)*(-825084.*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 9594.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 22919.*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 156.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 693.*u2*(2.9680105e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 1.18326e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 15540.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 60606.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) - 33.*(105.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 98605.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 13986.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 17316.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 140.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 9.65744955e8*fcs_pow(u,14)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8) + 9911.*fcs_pow(v,8)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 714.*fcs_pow(v,4)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 3604.*fcs_pow(v,6)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 28.*v2*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 4.1552147e7*fcs_pow(u,12)*(45.*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 2820.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 143820.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 211735.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 17766.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 1.371220851e9*fcs_pow(u,10)*(45.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 60.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 846.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 3995.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1692.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 6.8333265e7*fcs_pow(u,8)*(-18060.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 15.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 5418.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 24252.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 14147.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
      }
      case 10 : switch(k){
        case 0 : return 4.437792790412231e17*fcs_pow(t,25)*(-13.*(-7. + 875.*v2 - 15750.*fcs_pow(v,4) + 91350.*fcs_pow(v,6) - 202275.*fcs_pow(v,8) + 148335.*fcs_pow(v,10)) + 1.385475e6*fcs_pow(u,6)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) + 2275.*u2*(-7. + 945.*v2 - 18270.*fcs_pow(v,4) + 113274.*fcs_pow(v,6) - 267003.*fcs_pow(v,8) + 207669.*fcs_pow(v,10)) - 6.135675e6*fcs_pow(u,8)*(-3. + 495.*v2 - 11550.*fcs_pow(v,4) + 85470.*fcs_pow(v,6) - 238095.*fcs_pow(v,8) + 216931.*fcs_pow(v,10)) - 47775.*fcs_pow(u,4)*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)) + 4.499495e6*fcs_pow(u,10)*(-9. + 1575.*v2 - 38850.*fcs_pow(v,4) + 303030.*fcs_pow(v,6) - 887445.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) - 6.135675e6*fcs_pow(u,12)*(-7. + 1295.*v2 - 33670.*fcs_pow(v,4) + 276094.*fcs_pow(v,6) - 848003.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 2.494725e6*fcs_pow(u,14)*(-7. + 1365.*v2 - 37310.*fcs_pow(v,4) + 320866.*fcs_pow(v,6) - 1.031355e6*fcs_pow(v,8) + 1.077193e6*fcs_pow(v,10)));
        case 1 : return -1.442282656883975e20*fcs_pow(t,26)*(7. - 945.*v2 + 18270.*fcs_pow(v,4) - 113274.*fcs_pow(v,6) + 267003.*fcs_pow(v,8) - 207669.*fcs_pow(v,10) - 12789.*fcs_pow(u,4)*(-3. + 465.*v2 - 10230.*fcs_pow(v,4) + 71610.*fcs_pow(v,6) - 189255.*fcs_pow(v,8) + 164021.*fcs_pow(v,10)) + 132153.*fcs_pow(u,6)*(-3. + 495.*v2 - 11550.*fcs_pow(v,4) + 85470.*fcs_pow(v,6) - 238095.*fcs_pow(v,8) + 216931.*fcs_pow(v,10)) + 147.*u2*(-9. + 1305.*v2 - 26970.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 445005.*fcs_pow(v,8) + 365893.*fcs_pow(v,10)) - 207669.*fcs_pow(u,8)*(-9. + 1575.*v2 - 38850.*fcs_pow(v,4) + 303030.*fcs_pow(v,6) - 887445.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 623007.*fcs_pow(u,10)*(-7. + 1295.*v2 - 33670.*fcs_pow(v,4) + 276094.*fcs_pow(v,6) - 848003.*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) - 698523.*fcs_pow(u,12)*(-7. + 1365.*v2 - 37310.*fcs_pow(v,4) + 320866.*fcs_pow(v,6) - 1.031355e6*fcs_pow(v,8) + 1.077193e6*fcs_pow(v,10)) + 99789.*fcs_pow(u,14)*(-21. + 4305.*v2 - 123410.*fcs_pow(v,4) + 1.11069e6*fcs_pow(v,6) - 3.728745e6*fcs_pow(v,8) + 4.060189e6*fcs_pow(v,10)))*w;
        case 2 : return 1.442282656883975e20*fcs_pow(t,27)*(-7. + 189.*w2 - 945.*v2*(-1. + 29.*w2) + 18270.*fcs_pow(v,4)*(-1. + 31.*w2) - 113274.*fcs_pow(v,6)*(-1. + 33.*w2) + 267003.*fcs_pow(v,8)*(-1. + 35.*w2) - 207669.*fcs_pow(v,10)*(-1. + 37.*w2) + 147.*u2*(9. - 261.*w2 + 1305.*v2*(-1. + 31.*w2) - 26970.*fcs_pow(v,4)*(-1. + 33.*w2) + 178002.*fcs_pow(v,6)*(-1. + 35.*w2) - 445005.*fcs_pow(v,8)*(-1. + 37.*w2) + 365893.*fcs_pow(v,10)*(-1. + 39.*w2)) - 12789.*fcs_pow(u,4)*(3. - 93.*w2 + 465.*v2*(-1. + 33.*w2) - 10230.*fcs_pow(v,4)*(-1. + 35.*w2) + 71610.*fcs_pow(v,6)*(-1. + 37.*w2) - 189255.*fcs_pow(v,8)*(-1. + 39.*w2) + 164021.*fcs_pow(v,10)*(-1. + 41.*w2)) + 132153.*fcs_pow(u,6)*(3. - 99.*w2 + 495.*v2*(-1. + 35.*w2) - 11550.*fcs_pow(v,4)*(-1. + 37.*w2) + 85470.*fcs_pow(v,6)*(-1. + 39.*w2) - 238095.*fcs_pow(v,8)*(-1. + 41.*w2) + 216931.*fcs_pow(v,10)*(-1. + 43.*w2)) - 207669.*fcs_pow(u,8)*(9. - 315.*w2 + 1575.*v2*(-1. + 37.*w2) - 38850.*fcs_pow(v,4)*(-1. + 39.*w2) + 303030.*fcs_pow(v,6)*(-1. + 41.*w2) - 887445.*fcs_pow(v,8)*(-1. + 43.*w2) + 848003.*fcs_pow(v,10)*(-1. + 45.*w2)) + 623007.*fcs_pow(u,10)*(7. - 259.*w2 + 1295.*v2*(-1. + 39.*w2) - 33670.*fcs_pow(v,4)*(-1. + 41.*w2) + 276094.*fcs_pow(v,6)*(-1. + 43.*w2) - 848003.*fcs_pow(v,8)*(-1. + 45.*w2) + 848003.*fcs_pow(v,10)*(-1. + 47.*w2)) - 698523.*fcs_pow(u,12)*(7. - 273.*w2 + 1365.*v2*(-1. + 41.*w2) - 37310.*fcs_pow(v,4)*(-1. + 43.*w2) + 320866.*fcs_pow(v,6)*(-1. + 45.*w2) - 1.031355e6*fcs_pow(v,8)*(-1. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-1. + 49.*w2)) + 99789.*fcs_pow(u,14)*(21. - 861.*w2 + 4305.*v2*(-1. + 43.*w2) - 123410.*fcs_pow(v,4)*(-1. + 45.*w2) + 1.11069e6*fcs_pow(v,6)*(-1. + 47.*w2) - 3.728745e6*fcs_pow(v,8)*(-1. + 49.*w2) + 4.060189e6*fcs_pow(v,10)*(-1. + 51.*w2)));
        case 3 : return -9.086380738369043e21*fcs_pow(t,28)*w*(-9. + 87.*w2 + 26970.*fcs_pow(v,4)*(-1. + 11.*w2) - 365893.*fcs_pow(v,10)*(-1. + 13.*w2) - 435.*v2*(-3. + 31.*w2) - 59334.*fcs_pow(v,6)*(-3. + 35.*w2) + 148335.*fcs_pow(v,8)*(-3. + 37.*w2) + 203.*u2*(9. - 93.*w2 + 1395.*v2*(-1. + 11.*w2) - 567765.*fcs_pow(v,8)*(-1. + 13.*w2) - 10230.*fcs_pow(v,4)*(-3. + 35.*w2) + 71610.*fcs_pow(v,6)*(-3. + 37.*w2) + 164021.*fcs_pow(v,10)*(-3. + 41.*w2)) + 69223.*fcs_pow(u,6)*(9. - 105.*w2 - 38850.*fcs_pow(v,4)*(-1. + 13.*w2) + 848003.*fcs_pow(v,10)*(-1. + 15.*w2) + 525.*v2*(-3. + 37.*w2) + 101010.*fcs_pow(v,6)*(-3. + 41.*w2) - 295815.*fcs_pow(v,8)*(-3. + 43.*w2)) - 6293.*fcs_pow(u,4)*(9. - 99.*w2 + 256410.*fcs_pow(v,6)*(-1. + 13.*w2) + 495.*v2*(-3. + 35.*w2) - 11550.*fcs_pow(v,4)*(-3. + 37.*w2) - 238095.*fcs_pow(v,8)*(-3. + 41.*w2) + 216931.*fcs_pow(v,10)*(-3. + 43.*w2)) - 148335.*fcs_pow(u,8)*(21. - 259.*w2 + 3885.*v2*(-1. + 13.*w2) - 2.544009e6*fcs_pow(v,8)*(-1. + 15.*w2) - 33670.*fcs_pow(v,4)*(-3. + 41.*w2) + 276094.*fcs_pow(v,6)*(-3. + 43.*w2) + 848003.*fcs_pow(v,10)*(-3. + 47.*w2)) - 432419.*fcs_pow(u,12)*(21. - 287.*w2 - 123410.*fcs_pow(v,4)*(-1. + 15.*w2) + 4.060189e6*fcs_pow(v,10)*(-1. + 17.*w2) + 1435.*v2*(-3. + 43.*w2) + 370230.*fcs_pow(v,6)*(-3. + 47.*w2) - 1.242915e6*fcs_pow(v,8)*(-3. + 49.*w2)) + 365893.*fcs_pow(u,10)*(21. - 273.*w2 + 962598.*fcs_pow(v,6)*(-1. + 15.*w2) + 1365.*v2*(-3. + 41.*w2) - 37310.*fcs_pow(v,4)*(-3. + 43.*w2) - 1.031355e6*fcs_pow(v,8)*(-3. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-3. + 49.*w2)) + 1.363783e6*fcs_pow(u,14)*(3. - 43.*w2 + 645.*v2*(-1. + 15.*w2) - 636615.*fcs_pow(v,8)*(-1. + 17.*w2) - 6450.*fcs_pow(v,4)*(-3. + 47.*w2) + 60630.*fcs_pow(v,6)*(-3. + 49.*w2) + 240499.*fcs_pow(v,10)*(-3. + 53.*w2)));
        case 4 : return 9.086380738369043e21*fcs_pow(t,29)*(9. - 522.*w2 + 2697.*fcs_pow(w,4) - 1305.*v2*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 26970.*fcs_pow(v,4)*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 445005.*fcs_pow(v,8)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 365893.*fcs_pow(v,10)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 59334.*fcs_pow(v,6)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 69223.*fcs_pow(u,6)*(-9. + 630.*w2 - 3885.*fcs_pow(w,4) + 1575.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 38850.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 887445.*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 101010.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 18879.*fcs_pow(u,4)*(-3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 11550.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 85470.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 165.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 79365.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 203.*u2*(-9.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 1395.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 214830.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 567765.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 10230.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 164021.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 432419.*fcs_pow(u,12)*(4305.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 123410.*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 3.728745e6*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 7.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 370230.*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 1.097679e6*fcs_pow(u,10)*(-7.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 37310.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 320866.*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 455.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 343785.*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 148335.*fcs_pow(u,8)*(-21.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 3885.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 828282.*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 2.544009e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 33670.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 1.363783e6*fcs_pow(u,14)*(-3.*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 645.*v2*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 181890.*fcs_pow(v,6)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 636615.*fcs_pow(v,8)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 6450.*fcs_pow(v,4)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 240499.*fcs_pow(v,10)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))));
        case 5 : return -2.635050414127023e23*fcs_pow(t,30)*w*(45. - 930.*w2 + 3069.*fcs_pow(w,4) - 6975.*v2*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 51150.*fcs_pow(v,4)*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 567765.*fcs_pow(v,8)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 71610.*fcs_pow(v,6)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 164021.*fcs_pow(v,10)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 35805.*fcs_pow(u,4)*(-9. + 210.*w2 - 777.*fcs_pow(w,4) + 848003.*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 295815.*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 7770.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 105.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 20202.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 1085.*u2*(-9.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 495.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 216931.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 51282.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 2310.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 47619.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) + 35805.*fcs_pow(u,6)*(-1.2720045e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1.38047e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3885.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 7.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 33670.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 1.0110805e7*fcs_pow(u,14)*(-3. + 90.*w2 - 423.*fcs_pow(w,4) - 251685.*fcs_pow(v,8)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 296429.*fcs_pow(v,10)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 41454.*fcs_pow(v,6)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 45.*v2*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 1410.*fcs_pow(v,4)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 2.1397285e7*fcs_pow(u,12)*(-3. + 86.*w2 - 387.*fcs_pow(w,4) + 645.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 240499.*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 127323.*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 1290.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 12126.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 492063.*fcs_pow(u,10)*(-617050.*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 7175.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 7.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 370230.*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 1.242915e6*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 189255.*fcs_pow(u,8)*(4.81299e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 186550.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 21.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1365.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.031355e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return 3.952575621190534e24*fcs_pow(t,31)*(-3. + 279.*w2 - 3069.*fcs_pow(w,4) + 7161.*fcs_pow(w,6) - 465.*v2*(-1. + 99.*w2 - 1155.*fcs_pow(w,4) + 2849.*fcs_pow(w,6)) + 10230.*fcs_pow(v,4)*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 164021.*fcs_pow(v,10)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 14322.*fcs_pow(v,6)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 37851.*fcs_pow(v,8)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 2387.*fcs_pow(u,4)*(9. - 945.*w2 + 11655.*fcs_pow(w,4) - 30303.*fcs_pow(w,6) + 303030.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 315.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 217.*u2*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 495.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 238095.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 216931.*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2310.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) + 7161.*fcs_pow(u,6)*(-168350.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.38047e6*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.240015e6*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 1295.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 2.022161e6*fcs_pow(u,14)*(1. - 135.*w2 + 2115.*fcs_pow(w,4) - 6909.*fcs_pow(w,6) + 69090.*fcs_pow(v,6)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 251685.*fcs_pow(v,8)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 296429.*fcs_pow(v,10)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 45.*v2*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 1410.*fcs_pow(v,4)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 4.279457e6*fcs_pow(u,12)*(1. - 129.*w2 + 1935.*fcs_pow(w,4) - 6063.*fcs_pow(w,6) + 215.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 212205.*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 240499.*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1290.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 12126.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 164021.*fcs_pow(u,10)*(-21.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 4305.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 123410.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 4.060189e6*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 222138.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 745749.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 37851.*fcs_pow(u,8)*(6825.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 186550.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.60433e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 7.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.031355e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))));
        case 7 : return -1.225298442569066e26*fcs_pow(t,32)*w*(21.*(-1. + 33.*w2 - 231.*fcs_pow(w,4) + 407.*fcs_pow(w,6)) - 3465.*v2*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 216931.*fcs_pow(v,10)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 2310.*fcs_pow(v,4)*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*fcs_pow(v,6)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 47619.*fcs_pow(v,8)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 59829.*fcs_pow(u,6)*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6) + 1.60433e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.031355e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 26650.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 195.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 4851.*fcs_pow(u,4)*(35. - 1295.*w2 + 10101.*fcs_pow(w,4) - 19721.*fcs_pow(w,6) - 4.240015e6*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 197210.*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 185.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 4810.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 77.*u2*(-63.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 5.936021e6*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 887445.*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 315.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 7770.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 60606.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 4.1552147e7*fcs_pow(u,12)*(1. - 45.*w2 + 423.*fcs_pow(w,4) - 987.*fcs_pow(w,6) + 45.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 1410.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 35955.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 42347.*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 1974.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 4.555551e6*fcs_pow(u,10)*(7. - 301.*w2 + 2709.*fcs_pow(w,4) - 6063.*fcs_pow(w,6) + 1505.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 9030.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 84882.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 240499.*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 42441.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 111111.*fcs_pow(u,8)*(105. - 4305.*w2 + 37023.*fcs_pow(w,4) - 79335.*fcs_pow(w,6) - 617050.*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.11069e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 3075.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 3.728745e6*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 580027.*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 456617.*fcs_pow(u,14)*(-9.*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 2115.*v2*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 503370.*fcs_pow(v,6)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.905615e6*fcs_pow(v,8)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 9870.*fcs_pow(v,4)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 465817.*fcs_pow(v,10)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 8 : return 1.225298442569066e26*fcs_pow(t,33)*(21.*(1. - 132.*w2 + 2310.*fcs_pow(w,4) - 11396.*fcs_pow(w,6) + 15873.*fcs_pow(w,8)) - 3465.*v2*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 1.518517e6*fcs_pow(v,10)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 238095.*fcs_pow(v,8)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 2310.*fcs_pow(v,4)*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 17094.*fcs_pow(v,6)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) + 59829.*fcs_pow(u,6)*(-35. + 5460.*w2 - 111930.*fcs_pow(w,4) + 641732.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) - 186550.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.60433e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 975.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8))) - 555555.*fcs_pow(u,8)*(-21. + 3444.*w2 - 74046.*fcs_pow(w,4) + 444276.*fcs_pow(w,6) - 745749.*fcs_pow(w,8) + 4305.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 123410.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 222138.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 745749.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) + 3.1888857e7*fcs_pow(u,10)*(-1. + 172.*w2 - 3870.*fcs_pow(w,4) + 24252.*fcs_pow(w,6) - 42441.*fcs_pow(w,8) + 215.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 1290.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 12126.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 30315.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 34357.*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 456617.*fcs_pow(u,14)*(-9.*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 2115.*v2*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 49350.*fcs_pow(v,4)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 3.260719e6*fcs_pow(v,10)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 503370.*fcs_pow(v,6)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1.905615e6*fcs_pow(v,8)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 4.1552147e7*fcs_pow(u,12)*(-1. + 180.*w2 - 4230.*fcs_pow(w,4) + 27636.*fcs_pow(w,6) - 50337.*fcs_pow(w,8) + 45.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1410.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 9870.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 35955.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 4851.*fcs_pow(u,4)*(-35. + 5180.*w2 - 101010.*fcs_pow(w,4) + 552188.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 1.38047e6*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 4.240015e6*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 24050.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 185.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 539.*u2*(-9.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 887445.*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 43290.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 45.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 1110.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -3.675895327707197e26*fcs_pow(t,34)*w*(-59829.*fcs_pow(u,4)*(1.443897e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 79950.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 9.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1.031355e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 585.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) + 1617.*u2*(-315. + 15540.*w2 - 181818.*fcs_pow(w,4) + 709956.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) - 3.8160135e7*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 591630.*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1665.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 848003.*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 14430.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 33.*(5.936021e6*fcs_pow(v,10)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 7.*(9. - 420.*w2 + 4662.*fcs_pow(w,4) - 17316.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 295815.*fcs_pow(v,8)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 7770.*fcs_pow(v,4)*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 20202.*fcs_pow(v,6)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 35.*v2*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 6.4382997e7*fcs_pow(u,14)*(-45. + 2940.*w2 - 44982.*fcs_pow(w,4) + 227052.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 1.318163e6*fcs_pow(v,10)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 445995.*fcs_pow(v,8)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 5950.*fcs_pow(v,4)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 63070.*fcs_pow(v,6)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 105.*v2*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1.24656441e8*fcs_pow(u,12)*(-45. + 2820.*w2 - 41454.*fcs_pow(w,4) + 201348.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 235.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 465817.*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 55930.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 211735.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 3290.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 4.57073617e8*fcs_pow(u,10)*(-9.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 45.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1410.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 35955.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 5922.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 2.2777755e7*fcs_pow(u,8)*(13545.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 3.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 9030.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 84882.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 240499.*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 127323.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 777777.*fcs_pow(u,6)*(-1.85115e6*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 3075.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 370230.*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1.242915e6*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 3.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return 7.719380188185113e27*fcs_pow(t,35)*(-42735.*fcs_pow(u,4)*(21. - 4095.*w2 + 111930.*fcs_pow(w,4) - 962598.*fcs_pow(w,6) + 3.094065e6*fcs_pow(w,8) - 3.231579e6*fcs_pow(w,10) - 111930.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 962598.*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.077193e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 343785.*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 195.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 385.*u2*(2.484846e6*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 7.632027e6*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 9.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 1665.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 14430.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 11.*(9. - 1575.*w2 + 38850.*fcs_pow(w,4) - 303030.*fcs_pow(w,6) + 887445.*fcs_pow(w,8) - 848003.*fcs_pow(w,10) - 887445.*fcs_pow(v,8)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 848003.*fcs_pow(v,10)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 225.*v2*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 5550.*fcs_pow(v,4)*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 14430.*fcs_pow(v,6)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) + 185185.*fcs_pow(u,6)*(3.*(21. - 4305.*w2 + 123410.*fcs_pow(w,4) - 1.11069e6*fcs_pow(w,6) + 3.728745e6*fcs_pow(w,8) - 4.060189e6*fcs_pow(w,10)) + 12915.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 370230.*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 3.728745e6*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 370230.*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 1.5329285e7*fcs_pow(u,14)*(9. - 2205.*w2 + 74970.*fcs_pow(w,4) - 794682.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 3.954489e6*fcs_pow(w,10) - 3.121965e6*fcs_pow(v,8)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 1.318163e6*fcs_pow(v,10)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 113526.*fcs_pow(v,6)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 105.*v2*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 1190.*fcs_pow(v,4)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 2.9680105e7*fcs_pow(u,12)*(9. - 2115.*w2 + 69090.*fcs_pow(w,4) - 704718.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 3.260719e6*fcs_pow(w,10) + 3.260719e6*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 705.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 381123.*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 3290.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 11186.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 1.95888693e8*fcs_pow(u,10)*(1. - 225.*w2 + 7050.*fcs_pow(w,4) - 69090.*fcs_pow(w,6) + 251685.*fcs_pow(w,8) - 296429.*fcs_pow(w,10) - 2350.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 25.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 42347.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 3290.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 3995.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 7.592585e6*fcs_pow(u,8)*(-9.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 1935.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 181890.*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 6450.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 90945.*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 34357.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
      }
      case 11 : switch(k){
        case 0 : return -1.442282656883975e20*fcs_pow(t,26)*v*(-11.*(-7. + 315.*v2 - 3654.*fcs_pow(v,4) + 16182.*fcs_pow(v,6) - 29667.*fcs_pow(v,8) + 18879.*fcs_pow(v,10)) + 1.453683e6*fcs_pow(u,6)*(-3. + 165.*v2 - 2310.*fcs_pow(v,4) + 12210.*fcs_pow(v,6) - 26455.*fcs_pow(v,8) + 19721.*fcs_pow(v,10)) - 46893.*fcs_pow(u,4)*(-9. + 465.*v2 - 6138.*fcs_pow(v,4) + 30690.*fcs_pow(v,6) - 63085.*fcs_pow(v,8) + 44733.*fcs_pow(v,10)) + 231.*u2*(-63. + 3045.*v2 - 37758.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 346115.*fcs_pow(v,8) + 232841.*fcs_pow(v,10)) - 207669.*fcs_pow(u,8)*(-99. + 5775.*v2 - 85470.*fcs_pow(v,4) + 476190.*fcs_pow(v,6) - 1.084655e6*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) - 698523.*fcs_pow(u,12)*(-77. + 5005.*v2 - 82082.*fcs_pow(v,4) + 504218.*fcs_pow(v,6) - 1.260545e6*fcs_pow(v,8) + 1.077193e6*fcs_pow(v,10)) + 99789.*fcs_pow(u,14)*(-231. + 15785.*v2 - 271502.*fcs_pow(v,4) + 1.74537e6*fcs_pow(v,6) - 4.557355e6*fcs_pow(v,8) + 4.060189e6*fcs_pow(v,10)) + 69223.*fcs_pow(u,10)*(-693. + 42735.*v2 - 666666.*fcs_pow(v,4) + 3.904758e6*fcs_pow(v,6) - 9.328033e6*fcs_pow(v,8) + 7.632027e6*fcs_pow(v,10)));
        case 1 : return 4.326847970651925e20*fcs_pow(t,27)*v*(-1.453683e6*fcs_pow(u,4)*(-3. + 165.*v2 - 2310.*fcs_pow(v,4) + 12210.*fcs_pow(v,6) - 26455.*fcs_pow(v,8) + 19721.*fcs_pow(v,10)) + 15631.*u2*(-9. + 465.*v2 - 6138.*fcs_pow(v,4) + 30690.*fcs_pow(v,6) - 63085.*fcs_pow(v,8) + 44733.*fcs_pow(v,10)) - 11.*(-63. + 3045.*v2 - 37758.*fcs_pow(v,4) + 178002.*fcs_pow(v,6) - 346115.*fcs_pow(v,8) + 232841.*fcs_pow(v,10)) + 484561.*fcs_pow(u,6)*(-99. + 5775.*v2 - 85470.*fcs_pow(v,4) + 476190.*fcs_pow(v,6) - 1.084655e6*fcs_pow(v,8) + 848003.*fcs_pow(v,10)) + 7.683753e6*fcs_pow(u,10)*(-77. + 5005.*v2 - 82082.*fcs_pow(v,4) + 504218.*fcs_pow(v,6) - 1.260545e6*fcs_pow(v,8) + 1.077193e6*fcs_pow(v,10)) - 3.026933e6*fcs_pow(u,12)*(-231. + 15785.*v2 - 271502.*fcs_pow(v,4) + 1.74537e6*fcs_pow(v,6) - 4.557355e6*fcs_pow(v,8) + 4.060189e6*fcs_pow(v,10)) + 1.363783e6*fcs_pow(u,14)*(-231. + 16555.*v2 - 297990.*fcs_pow(v,4) + 2.00079e6*fcs_pow(v,6) - 5.446595e6*fcs_pow(v,8) + 5.050479e6*fcs_pow(v,10)) - 346115.*fcs_pow(u,8)*(-693. + 42735.*v2 - 666666.*fcs_pow(v,4) + 3.904758e6*fcs_pow(v,6) - 9.328033e6*fcs_pow(v,8) + 7.632027e6*fcs_pow(v,10)))*w;
        case 2 : return -4.326847970651925e20*fcs_pow(t,28)*v*(-11.*(63. - 1827.*w2 + 3045.*v2*(-1. + 31.*w2) - 37758.*fcs_pow(v,4)*(-1. + 33.*w2) + 178002.*fcs_pow(v,6)*(-1. + 35.*w2) - 346115.*fcs_pow(v,8)*(-1. + 37.*w2) + 232841.*fcs_pow(v,10)*(-1. + 39.*w2)) + 15631.*u2*(9. - 279.*w2 + 465.*v2*(-1. + 33.*w2) - 6138.*fcs_pow(v,4)*(-1. + 35.*w2) + 30690.*fcs_pow(v,6)*(-1. + 37.*w2) - 63085.*fcs_pow(v,8)*(-1. + 39.*w2) + 44733.*fcs_pow(v,10)*(-1. + 41.*w2)) - 1.453683e6*fcs_pow(u,4)*(3. - 99.*w2 + 165.*v2*(-1. + 35.*w2) - 2310.*fcs_pow(v,4)*(-1. + 37.*w2) + 12210.*fcs_pow(v,6)*(-1. + 39.*w2) - 26455.*fcs_pow(v,8)*(-1. + 41.*w2) + 19721.*fcs_pow(v,10)*(-1. + 43.*w2)) + 484561.*fcs_pow(u,6)*(99. - 3465.*w2 + 5775.*v2*(-1. + 37.*w2) - 85470.*fcs_pow(v,4)*(-1. + 39.*w2) + 476190.*fcs_pow(v,6)*(-1. + 41.*w2) - 1.084655e6*fcs_pow(v,8)*(-1. + 43.*w2) + 848003.*fcs_pow(v,10)*(-1. + 45.*w2)) - 346115.*fcs_pow(u,8)*(693.*(1. - 37.*w2) + 42735.*v2*(-1. + 39.*w2) - 666666.*fcs_pow(v,4)*(-1. + 41.*w2) + 3.904758e6*fcs_pow(v,6)*(-1. + 43.*w2) - 9.328033e6*fcs_pow(v,8)*(-1. + 45.*w2) + 7.632027e6*fcs_pow(v,10)*(-1. + 47.*w2)) + 7.683753e6*fcs_pow(u,10)*(77. - 3003.*w2 + 5005.*v2*(-1. + 41.*w2) - 82082.*fcs_pow(v,4)*(-1. + 43.*w2) + 504218.*fcs_pow(v,6)*(-1. + 45.*w2) - 1.260545e6*fcs_pow(v,8)*(-1. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-1. + 49.*w2)) - 3.026933e6*fcs_pow(u,12)*(231. - 9471.*w2 + 15785.*v2*(-1. + 43.*w2) - 271502.*fcs_pow(v,4)*(-1. + 45.*w2) + 1.74537e6*fcs_pow(v,6)*(-1. + 47.*w2) - 4.557355e6*fcs_pow(v,8)*(-1. + 49.*w2) + 4.060189e6*fcs_pow(v,10)*(-1. + 51.*w2)) + 1.363783e6*fcs_pow(u,14)*(231. - 9933.*w2 + 16555.*v2*(-1. + 45.*w2) - 297990.*fcs_pow(v,4)*(-1. + 47.*w2) + 2.00079e6*fcs_pow(v,6)*(-1. + 49.*w2) - 5.446595e6*fcs_pow(v,8)*(-1. + 51.*w2) + 5.050479e6*fcs_pow(v,10)*(-1. + 53.*w2)));
        case 3 : return 2.635050414127023e23*fcs_pow(t,29)*v*w*(-11.*(9. - 93.*w2 + 465.*v2*(-1. + 11.*w2) - 63085.*fcs_pow(v,8)*(-1. + 13.*w2) - 2046.*fcs_pow(v,4)*(-3. + 35.*w2) + 10230.*fcs_pow(v,6)*(-3. + 37.*w2) + 14911.*fcs_pow(v,10)*(-3. + 41.*w2)) - 2387.*fcs_pow(u,4)*(99.*(3. - 35.*w2) - 256410.*fcs_pow(v,4)*(-1. + 13.*w2) + 2.544009e6*fcs_pow(v,10)*(-1. + 15.*w2) + 5775.*v2*(-3. + 37.*w2) + 476190.*fcs_pow(v,6)*(-3. + 41.*w2) - 1.084655e6*fcs_pow(v,8)*(-3. + 43.*w2)) + 2387.*u2*(9. - 99.*w2 + 36630.*fcs_pow(v,6)*(-1. + 13.*w2) + 165.*v2*(-3. + 35.*w2) - 2310.*fcs_pow(v,4)*(-3. + 37.*w2) - 26455.*fcs_pow(v,8)*(-3. + 41.*w2) + 19721.*fcs_pow(v,10)*(-3. + 43.*w2)) + 11935.*fcs_pow(u,6)*(693. - 8547.*w2 + 42735.*v2*(-1. + 13.*w2) - 9.328033e6*fcs_pow(v,8)*(-1. + 15.*w2) - 222222.*fcs_pow(v,4)*(-3. + 41.*w2) + 1.301586e6*fcs_pow(v,6)*(-3. + 43.*w2) + 2.544009e6*fcs_pow(v,10)*(-3. + 47.*w2)) + 164021.*fcs_pow(u,10)*(693. - 9471.*w2 - 814506.*fcs_pow(v,4)*(-1. + 15.*w2) + 1.2180567e7*fcs_pow(v,10)*(-1. + 17.*w2) + 15785.*v2*(-3. + 43.*w2) + 1.74537e6*fcs_pow(v,6)*(-3. + 47.*w2) - 4.557355e6*fcs_pow(v,8)*(-3. + 49.*w2)) - 189255.*fcs_pow(u,8)*(231. - 3003.*w2 + 1.512654e6*fcs_pow(v,6)*(-1. + 15.*w2) + 5005.*v2*(-3. + 41.*w2) - 82082.*fcs_pow(v,4)*(-3. + 43.*w2) - 1.260545e6*fcs_pow(v,8)*(-3. + 47.*w2) + 1.077193e6*fcs_pow(v,10)*(-3. + 49.*w2)) - 611351.*fcs_pow(u,12)*(77.*(3. - 43.*w2) + 16555.*v2*(-1. + 15.*w2) - 5.446595e6*fcs_pow(v,8)*(-1. + 17.*w2) - 99330.*fcs_pow(v,4)*(-3. + 47.*w2) + 666930.*fcs_pow(v,6)*(-3. + 49.*w2) + 1.683493e6*fcs_pow(v,10)*(-3. + 53.*w2)) + 2.022161e6*fcs_pow(u,14)*(33. - 495.*w2 + 325710.*fcs_pow(v,6)*(-1. + 17.*w2) + 825.*v2*(-3. + 47.*w2) - 15510.*fcs_pow(v,4)*(-3. + 49.*w2) - 307615.*fcs_pow(v,8)*(-3. + 53.*w2) + 296429.*fcs_pow(v,10)*(-3. + 55.*w2)));
        case 4 : return -2.635050414127023e23*fcs_pow(t,30)*v*(-7161.*fcs_pow(u,4)*(5775.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 85470.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 1.084655e6*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 33.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 158730.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 2387.*u2*(-9.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 6930.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 36630.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 59163.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 165.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 26455.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 11.*(-9.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) + 465.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 30690.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 63085.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 2046.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 14911.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 492063.*fcs_pow(u,10)*(15785.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 271502.*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 4.557355e6*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 77.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 581790.*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 189255.*fcs_pow(u,8)*(-231.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 246246.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 1.512654e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 3.231579e6*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 5005.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 1.260545e6*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 11935.*fcs_pow(u,6)*(-693.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 42735.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 3.904758e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 9.328033e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 222222.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 2.544009e6*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 2.022161e6*fcs_pow(u,14)*(-33.*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 46530.*fcs_pow(v,4)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 325710.*fcs_pow(v,6)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 889287.*fcs_pow(v,10)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 825.*v2*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) - 307615.*fcs_pow(v,8)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) - 611351.*fcs_pow(u,12)*(-231.*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 16555.*v2*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 2.00079e6*fcs_pow(v,6)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 5.446595e6*fcs_pow(v,8)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 99330.*fcs_pow(v,4)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 1.683493e6*fcs_pow(v,10)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))));
        case 5 : return 4.084328141896885e25*fcs_pow(t,31)*v*w*(77.*u2*(2.544009e6*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 99.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 1.084655e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 51282.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 1155.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 95238.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 11.*(-9.*(1. - 22.*w2 + 77.*fcs_pow(w,4)) + 165.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 19721.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7326.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 462.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 5291.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4))) - 231.*fcs_pow(u,4)*(-4.6640165e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 6.50793e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 42735.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 231.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 222222.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 2.544009e6*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 717541.*fcs_pow(u,14)*(503370.*fcs_pow(v,6)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 1.482145e6*fcs_pow(v,8)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) - 41454.*fcs_pow(v,4)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 889287.*fcs_pow(v,10)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) - 9.*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 705.*v2*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 5.936021e6*fcs_pow(u,12)*(-33.*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 307615.*fcs_pow(v,8)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 296429.*fcs_pow(v,10)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 65142.*fcs_pow(v,6)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 165.*v2*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 3102.*fcs_pow(v,4)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 650793.*fcs_pow(u,10)*(16555.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 77.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 1.683493e6*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 1.089319e6*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 19866.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 133386.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 15873.*fcs_pow(u,8)*(-4.07253e6*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 78925.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 1.2180567e7*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 231.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 1.74537e6*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 4.557355e6*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 8547.*fcs_pow(u,6)*(7.56327e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 410410.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 231.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 5005.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.260545e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 1.077193e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))));
        case 6 : return -1.225298442569066e26*fcs_pow(t,32)*v*(77.*u2*(-99.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 476190.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 1.084655e6*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 848003.*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 1155.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 11.*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 165.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 26455.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 19721.*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 462.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 2442.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6))) - 77.*fcs_pow(u,4)*(-3.33333e6*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.952379e7*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.6640165e7*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 693.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 7.632027e6*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) - 5.936021e6*fcs_pow(u,12)*(11.*(1. - 135.*w2 + 2115.*fcs_pow(w,4) - 6909.*fcs_pow(w,6)) + 108570.*fcs_pow(v,6)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 307615.*fcs_pow(v,8)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 296429.*fcs_pow(v,10)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 165.*v2*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 3102.*fcs_pow(v,4)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 216931.*fcs_pow(u,10)*(-231.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 16555.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 5.446595e6*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 5.050479e6*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 59598.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 400158.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 79365.*fcs_pow(u,8)*(-231.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 15785.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 271502.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 4.060189e6*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 349074.*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 911471.*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 8547.*fcs_pow(u,6)*(25025.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 410410.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 2.52109e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 77.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.260545e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 717541.*fcs_pow(u,14)*(-69090.*fcs_pow(v,4)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 503370.*fcs_pow(v,6)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.482145e6*fcs_pow(v,8)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 9.*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 705.*v2*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) + 296429.*fcs_pow(v,10)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 7 : return 1.347828286825972e27*fcs_pow(t,33)*v*w*(693.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) - 5.936021e6*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.084655e6*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1155.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 17094.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 95238.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 16317.*fcs_pow(u,4)*(11.*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6)) + 2.52109e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.260545e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 58630.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.077193e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 715.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 7.*u2*(-3.26481155e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 5.3424189e7*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 1.952379e7*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 693.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 666666.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 1.24656441e8*fcs_pow(u,10)*(-11.*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 165.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 3102.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 43945.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 42347.*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 3102.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 2.070705e6*fcs_pow(u,8)*(16555.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 59598.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 33.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 400158.*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 721497.*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 155617.*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 70707.*fcs_pow(u,6)*(-1.35751e6*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.74537e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 11275.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.557355e6*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 33.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 580027.*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 2.1460999e7*fcs_pow(u,14)*(-9.*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 359499.*fcs_pow(v,10)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) - 10710.*fcs_pow(v,4)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 81090.*fcs_pow(v,6)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 105.*v2*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) - 49555.*fcs_pow(v,8)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) - 4.1552147e7*fcs_pow(u,12)*(-9.*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 705.*v2*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 71910.*fcs_pow(v,6)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 211735.*fcs_pow(v,8)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 1974.*fcs_pow(v,4)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 42347.*fcs_pow(v,10)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 8 : return -1.347828286825972e27*fcs_pow(t,34)*v*(693.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) + 7.592585e6*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 5.936021e6*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 476190.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1155.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 17094.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) + 353535.*fcs_pow(u,6)*(15785.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 271502.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 33.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 349074.*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 911471.*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 580027.*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8))) - 2.070705e6*fcs_pow(u,8)*(-231.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 16555.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 59598.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 400158.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 778085.*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 721497.*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 2.1460999e7*fcs_pow(u,14)*(-9.*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 525.*v2*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 346885.*fcs_pow(v,8)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) - 10710.*fcs_pow(v,4)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 359499.*fcs_pow(v,10)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) + 81090.*fcs_pow(v,6)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 4.1552147e7*fcs_pow(u,12)*(-9.*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 705.*v2*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 9870.*fcs_pow(v,4)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 296429.*fcs_pow(v,10)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 71910.*fcs_pow(v,6)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 211735.*fcs_pow(v,8)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 1.24656441e8*fcs_pow(u,10)*(-11.*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 165.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3102.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 15510.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 43945.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 16317.*fcs_pow(u,4)*(-410410.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 2.52109e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 3575.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.260545e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 49.*u2*(1.952379e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 4.6640165e7*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 476190.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 7.632027e6*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 99.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 6105.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return 4.043484860477916e27*fcs_pow(t,35)*v*w*(5439.*u2*(2.268981e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 175890.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 99.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) - 1.260545e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.077193e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 2145.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8))) - 3.*(-3.26481155e8*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 6.50793e6*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 42735.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5.936021e6*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 222222.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 77.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8))) + 6.4382997e7*fcs_pow(u,14)*(-6.590815e6*fcs_pow(v,8)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 7.070147e6*fcs_pow(v,10)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) + 891990.*fcs_pow(v,6)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 2975.*v2*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 63070.*fcs_pow(v,4)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 21.*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 5.858852727e9*fcs_pow(u,12)*(-45. + 2940.*w2 - 44982.*fcs_pow(w,4) + 227052.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 119833.*fcs_pow(v,10)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 49555.*fcs_pow(v,8)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 1190.*fcs_pow(v,4)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 9010.*fcs_pow(v,6)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 35.*v2*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 4.57073617e8*fcs_pow(u,10)*(-9.*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 705.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 381123.*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 71910.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 211735.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 5922.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 6.23282205e8*fcs_pow(u,8)*(-99.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 165.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3102.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 43945.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 42347.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 9306.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 1.4494935e7*fcs_pow(u,6)*(49665.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 33.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 19866.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 133386.*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 240499.*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 155617.*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 70707.*fcs_pow(u,4)*(-1.221759e7*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 33825.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1.74537e6*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 4.557355e6*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 99.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 1.740081e6*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return -4.245659103501812e29*fcs_pow(t,36)*v*(-3.904758e6*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 9.328033e6*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 99.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) - 6105.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 848003.*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 31746.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 259.*u2*(-738738.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 4.537962e6*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 3.231579e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 1.260545e6*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 2145.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10))) - 10101.*fcs_pow(u,4)*(47355.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 814506.*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4.557355e6*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 581790.*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 33.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 2.1460999e7*fcs_pow(u,14)*(63. - 16065.*w2 + 567630.*fcs_pow(w,4) - 6.24393e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 3.3330693e7*fcs_pow(w,10) + 891990.*fcs_pow(v,6)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 3.030063e6*fcs_pow(v,10)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) - 941545.*fcs_pow(v,8)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 16218.*fcs_pow(v,4)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 85.*v2*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 2.78992987e8*fcs_pow(u,12)*(9. - 2205.*w2 + 74970.*fcs_pow(w,4) - 794682.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 3.954489e6*fcs_pow(w,10) - 346885.*fcs_pow(v,8)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 119833.*fcs_pow(v,10)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 16218.*fcs_pow(v,6)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 35.*v2*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 238.*fcs_pow(v,4)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 1.95888693e8*fcs_pow(u,10)*(9. - 2115.*w2 + 69090.*fcs_pow(w,4) - 704718.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 3.260719e6*fcs_pow(w,10) + 296429.*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 235.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 42347.*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 658.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 1598.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 5.936021e6*fcs_pow(u,8)*(-99.*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 46530.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 825.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 381123.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 46530.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 43945.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 138047.*fcs_pow(u,6)*(-693.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 49665.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 2.00079e6*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 99330.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 778085.*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 240499.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
        case 11 : return 1.57089386829567e31*fcs_pow(t,37)*v*w*(-123123.*fcs_pow(u,4)*(693. - 49665.*w2 + 893970.*fcs_pow(w,4) - 6.00237e6*fcs_pow(w,6) + 1.6339785e7*fcs_pow(w,8) - 1.5151437e7*fcs_pow(w,10) + 4515.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 99330.*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 666930.*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 778085.*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 240499.*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10))) + 1.2353341e7*fcs_pow(u,6)*(-9.*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 825.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 15510.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 46530.*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 43945.*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 42347.*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10))) + 1001.*u2*(-222138.*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.74537e6*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4.557355e6*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 1.740081e6*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 9.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 615.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) - 11.*(412542.*fcs_pow(v,6)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.260545e6*fcs_pow(v,8)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.077193e6*fcs_pow(v,10)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 9.*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) + 195.*v2*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 3198.*fcs_pow(v,4)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 9.860459e6*fcs_pow(u,14)*(1.0872579e7*fcs_pow(v,10)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 996930.*fcs_pow(v,6)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 33.*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 3.267715e6*fcs_pow(v,8)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) + 2915.*v2*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 52470.*fcs_pow(v,4)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 7.540351e6*fcs_pow(u,12)*(-231.*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 6.590815e6*fcs_pow(v,8)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 6545.*v2*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,10)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 138754.*fcs_pow(v,4)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 891990.*fcs_pow(v,6)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 2.48831583e8*fcs_pow(u,10)*(-11.*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 385.*v2*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 119833.*fcs_pow(v,10)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 2618.*fcs_pow(v,4)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 19822.*fcs_pow(v,6)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 49555.*fcs_pow(v,8)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 8.823815e6*fcs_pow(u,8)*(-99.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 7755.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 65142.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 158202.*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 465817.*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 381123.*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))));
      }
      case 12 : switch(k){
        case 0 : return 1.442282656883975e20*fcs_pow(t,27)*(4851.*u2*(3. - 522.*v2 + 13485.*fcs_pow(v,4) - 118668.*fcs_pow(v,6) + 445005.*fcs_pow(v,8) - 731786.*fcs_pow(v,10) + 432419.*fcs_pow(v,12)) - 140679.*fcs_pow(u,4)*(3. - 558.*v2 + 15345.*fcs_pow(v,4) - 143220.*fcs_pow(v,6) + 567765.*fcs_pow(v,8) - 984126.*fcs_pow(v,10) + 611351.*fcs_pow(v,12)) - 11.*(7. - 1134.*v2 + 27405.*fcs_pow(v,4) - 226548.*fcs_pow(v,6) + 801009.*fcs_pow(v,8) - 1.246014e6*fcs_pow(v,10) + 698523.*fcs_pow(v,12)) + 1.453683e6*fcs_pow(u,6)*(3. - 594.*v2 + 17325.*fcs_pow(v,4) - 170940.*fcs_pow(v,6) + 714285.*fcs_pow(v,8) - 1.301586e6*fcs_pow(v,10) + 848003.*fcs_pow(v,12)) - 623007.*fcs_pow(u,8)*(33. - 6930.*v2 + 213675.*fcs_pow(v,4) - 2.22222e6*fcs_pow(v,6) + 9.761895e6*fcs_pow(v,8) - 1.8656066e7*fcs_pow(v,10) + 1.2720045e7*fcs_pow(v,12)) + 623007.*fcs_pow(u,10)*(77. - 17094.*v2 + 555555.*fcs_pow(v,4) - 6.074068e6*fcs_pow(v,6) + 2.7984099e7*fcs_pow(v,8) - 5.5968198e7*fcs_pow(v,10) + 3.9856141e7*fcs_pow(v,12)) - 698523.*fcs_pow(u,12)*(77. - 18018.*v2 + 615615.*fcs_pow(v,4) - 7.059052e6*fcs_pow(v,6) + 3.4034715e7*fcs_pow(v,8) - 7.1094738e7*fcs_pow(v,10) + 5.2782457e7*fcs_pow(v,12)) + 299367.*fcs_pow(u,14)*(77. - 18942.*v2 + 678755.*fcs_pow(v,4) - 8.14506e6*fcs_pow(v,6) + 4.1016195e7*fcs_pow(v,8) - 8.9324158e7*fcs_pow(v,10) + 6.9023213e7*fcs_pow(v,12)));
        case 1 : return -9.086380738369043e21*fcs_pow(t,28)*(-11.*(3. - 522.*v2 + 13485.*fcs_pow(v,4) - 118668.*fcs_pow(v,6) + 445005.*fcs_pow(v,8) - 731786.*fcs_pow(v,10) + 432419.*fcs_pow(v,12)) + 2233.*u2*(3. - 558.*v2 + 15345.*fcs_pow(v,4) - 143220.*fcs_pow(v,6) + 567765.*fcs_pow(v,8) - 984126.*fcs_pow(v,10) + 611351.*fcs_pow(v,12)) - 69223.*fcs_pow(u,4)*(3. - 594.*v2 + 17325.*fcs_pow(v,4) - 170940.*fcs_pow(v,6) + 714285.*fcs_pow(v,8) - 1.301586e6*fcs_pow(v,10) + 848003.*fcs_pow(v,12)) + 69223.*fcs_pow(u,6)*(33. - 6930.*v2 + 213675.*fcs_pow(v,4) - 2.22222e6*fcs_pow(v,6) + 9.761895e6*fcs_pow(v,8) - 1.8656066e7*fcs_pow(v,10) + 1.2720045e7*fcs_pow(v,12)) + 1.363783e6*fcs_pow(u,14)*(11. - 2838.*v2 + 106425.*fcs_pow(v,4) - 1.33386e6*fcs_pow(v,6) + 7.002765e6*fcs_pow(v,8) - 1.5872934e7*fcs_pow(v,10) + 1.2746447e7*fcs_pow(v,12)) - 148335.*fcs_pow(u,8)*(77. - 17094.*v2 + 555555.*fcs_pow(v,4) - 6.074068e6*fcs_pow(v,6) + 2.7984099e7*fcs_pow(v,8) - 5.5968198e7*fcs_pow(v,10) + 3.9856141e7*fcs_pow(v,12)) + 365893.*fcs_pow(u,10)*(77. - 18018.*v2 + 615615.*fcs_pow(v,4) - 7.059052e6*fcs_pow(v,6) + 3.4034715e7*fcs_pow(v,8) - 7.1094738e7*fcs_pow(v,10) + 5.2782457e7*fcs_pow(v,12)) - 432419.*fcs_pow(u,12)*(77. - 18942.*v2 + 678755.*fcs_pow(v,4) - 8.14506e6*fcs_pow(v,6) + 4.1016195e7*fcs_pow(v,8) - 8.9324158e7*fcs_pow(v,10) + 6.9023213e7*fcs_pow(v,12)))*w;
        case 2 : return 9.086380738369043e21*fcs_pow(t,29)*(-11.*(-3. + 87.*w2 - 522.*v2*(-1. + 31.*w2) + 13485.*fcs_pow(v,4)*(-1. + 33.*w2) - 118668.*fcs_pow(v,6)*(-1. + 35.*w2) + 445005.*fcs_pow(v,8)*(-1. + 37.*w2) - 731786.*fcs_pow(v,10)*(-1. + 39.*w2) + 432419.*fcs_pow(v,12)*(-1. + 41.*w2)) + 2233.*u2*(-3. + 93.*w2 - 558.*v2*(-1. + 33.*w2) + 15345.*fcs_pow(v,4)*(-1. + 35.*w2) - 143220.*fcs_pow(v,6)*(-1. + 37.*w2) + 567765.*fcs_pow(v,8)*(-1. + 39.*w2) - 984126.*fcs_pow(v,10)*(-1. + 41.*w2) + 611351.*fcs_pow(v,12)*(-1. + 43.*w2)) - 69223.*fcs_pow(u,4)*(-3. + 99.*w2 - 594.*v2*(-1. + 35.*w2) + 17325.*fcs_pow(v,4)*(-1. + 37.*w2) - 170940.*fcs_pow(v,6)*(-1. + 39.*w2) + 714285.*fcs_pow(v,8)*(-1. + 41.*w2) - 1.301586e6*fcs_pow(v,10)*(-1. + 43.*w2) + 848003.*fcs_pow(v,12)*(-1. + 45.*w2)) + 69223.*fcs_pow(u,6)*(33.*(-1. + 35.*w2) - 6930.*v2*(-1. + 37.*w2) + 213675.*fcs_pow(v,4)*(-1. + 39.*w2) - 2.22222e6*fcs_pow(v,6)*(-1. + 41.*w2) + 9.761895e6*fcs_pow(v,8)*(-1. + 43.*w2) - 1.8656066e7*fcs_pow(v,10)*(-1. + 45.*w2) + 1.2720045e7*fcs_pow(v,12)*(-1. + 47.*w2)) - 148335.*fcs_pow(u,8)*(77.*(-1. + 37.*w2) - 17094.*v2*(-1. + 39.*w2) + 555555.*fcs_pow(v,4)*(-1. + 41.*w2) - 6.074068e6*fcs_pow(v,6)*(-1. + 43.*w2) + 2.7984099e7*fcs_pow(v,8)*(-1. + 45.*w2) - 5.5968198e7*fcs_pow(v,10)*(-1. + 47.*w2) + 3.9856141e7*fcs_pow(v,12)*(-1. + 49.*w2)) + 365893.*fcs_pow(u,10)*(77.*(-1. + 39.*w2) - 18018.*v2*(-1. + 41.*w2) + 615615.*fcs_pow(v,4)*(-1. + 43.*w2) - 7.059052e6*fcs_pow(v,6)*(-1. + 45.*w2) + 3.4034715e7*fcs_pow(v,8)*(-1. + 47.*w2) - 7.1094738e7*fcs_pow(v,10)*(-1. + 49.*w2) + 5.2782457e7*fcs_pow(v,12)*(-1. + 51.*w2)) - 432419.*fcs_pow(u,12)*(77.*(-1. + 41.*w2) - 18942.*v2*(-1. + 43.*w2) + 678755.*fcs_pow(v,4)*(-1. + 45.*w2) - 8.14506e6*fcs_pow(v,6)*(-1. + 47.*w2) + 4.1016195e7*fcs_pow(v,8)*(-1. + 49.*w2) - 8.9324158e7*fcs_pow(v,10)*(-1. + 51.*w2) + 6.9023213e7*fcs_pow(v,12)*(-1. + 53.*w2)) + 1.363783e6*fcs_pow(u,14)*(-11. + 473.*w2 - 2838.*v2*(-1. + 45.*w2) + 106425.*fcs_pow(v,4)*(-1. + 47.*w2) - 1.33386e6*fcs_pow(v,6)*(-1. + 49.*w2) + 7.002765e6*fcs_pow(v,8)*(-1. + 51.*w2) - 1.5872934e7*fcs_pow(v,10)*(-1. + 53.*w2) + 1.2746447e7*fcs_pow(v,12)*(-1. + 55.*w2)));
        case 3 : return -2.635050414127023e23*fcs_pow(t,30)*w*(7161.*u2*(-3. + 33.*w2 - 170940.*fcs_pow(v,6)*(-1. + 13.*w2) + 848003.*fcs_pow(v,12)*(-1. + 15.*w2) - 198.*v2*(-3. + 35.*w2) + 5775.*fcs_pow(v,4)*(-3. + 37.*w2) + 238095.*fcs_pow(v,8)*(-3. + 41.*w2) - 433862.*fcs_pow(v,10)*(-3. + 43.*w2)) - 11.*(-9. + 93.*w2 - 1674.*v2*(-1. + 11.*w2) + 1.703295e6*fcs_pow(v,8)*(-1. + 13.*w2) + 15345.*fcs_pow(v,4)*(-3. + 35.*w2) - 143220.*fcs_pow(v,6)*(-3. + 37.*w2) - 984126.*fcs_pow(v,10)*(-3. + 41.*w2) + 611351.*fcs_pow(v,12)*(-3. + 43.*w2)) - 21483.*fcs_pow(u,4)*(-33. + 385.*w2 + 213675.*fcs_pow(v,4)*(-1. + 13.*w2) - 1.8656066e7*fcs_pow(v,10)*(-1. + 15.*w2) - 2310.*v2*(-3. + 37.*w2) - 740740.*fcs_pow(v,6)*(-3. + 41.*w2) + 3.253965e6*fcs_pow(v,8)*(-3. + 43.*w2) + 4.240015e6*fcs_pow(v,12)*(-3. + 47.*w2)) - 567765.*fcs_pow(u,8)*(77.*(-1. + 13.*w2) - 7.059052e6*fcs_pow(v,6)*(-1. + 15.*w2) + 5.2782457e7*fcs_pow(v,12)*(-1. + 17.*w2) - 6006.*v2*(-3. + 41.*w2) + 205205.*fcs_pow(v,4)*(-3. + 43.*w2) + 1.1344905e7*fcs_pow(v,8)*(-3. + 47.*w2) - 2.3698246e7*fcs_pow(v,10)*(-3. + 49.*w2)) + 35805.*fcs_pow(u,6)*(-51282.*v2*(-1. + 13.*w2) + 8.3952297e7*fcs_pow(v,8)*(-1. + 15.*w2) + 77.*(-3. + 37.*w2) + 555555.*fcs_pow(v,4)*(-3. + 41.*w2) - 6.074068e6*fcs_pow(v,6)*(-3. + 43.*w2) - 5.5968198e7*fcs_pow(v,10)*(-3. + 47.*w2) + 3.9856141e7*fcs_pow(v,12)*(-3. + 49.*w2)) + 492063.*fcs_pow(u,10)*(2.036265e6*fcs_pow(v,4)*(-1. + 15.*w2) - 2.67972474e8*fcs_pow(v,10)*(-1. + 17.*w2) + 77.*(-3. + 41.*w2) - 18942.*v2*(-3. + 43.*w2) - 8.14506e6*fcs_pow(v,6)*(-3. + 47.*w2) + 4.1016195e7*fcs_pow(v,8)*(-3. + 49.*w2) + 6.9023213e7*fcs_pow(v,12)*(-3. + 53.*w2)) + 6.6731313e7*fcs_pow(u,14)*(-1. + 15.*w2 - 138180.*fcs_pow(v,6)*(-1. + 17.*w2) + 1.482145e6*fcs_pow(v,12)*(-1. + 19.*w2) - 90.*v2*(-3. + 47.*w2) + 3525.*fcs_pow(v,4)*(-3. + 49.*w2) + 251685.*fcs_pow(v,8)*(-3. + 53.*w2) - 592858.*fcs_pow(v,10)*(-3. + 55.*w2)) - 4.279457e6*fcs_pow(u,12)*(-33. + 473.*w2 - 8514.*v2*(-1. + 15.*w2) + 2.1008295e7*fcs_pow(v,8)*(-1. + 17.*w2) + 106425.*fcs_pow(v,4)*(-3. + 47.*w2) - 1.33386e6*fcs_pow(v,6)*(-3. + 49.*w2) - 1.5872934e7*fcs_pow(v,10)*(-3. + 53.*w2) + 1.2746447e7*fcs_pow(v,12)*(-3. + 55.*w2)));
        case 4 : return 7.905151242381068e23*fcs_pow(t,31)*(2387.*u2*(3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 17325.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 170940.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 1.301586e6*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 848003.*fcs_pow(v,12)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 198.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 238095.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 11.*(3.*(1. - 62.*w2 + 341.*fcs_pow(w,4)) - 558.*v2*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 143220.*fcs_pow(v,6)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 567765.*fcs_pow(v,8)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 611351.*fcs_pow(v,12)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 5115.*fcs_pow(v,4)*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 328042.*fcs_pow(v,10)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) - 189255.*fcs_pow(u,8)*(77.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 615615.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 7.059052e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 7.1094738e7*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 5.2782457e7*fcs_pow(v,12)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 6006.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 1.1344905e7*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 35805.*fcs_pow(u,6)*(77.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 17094.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 6.074068e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 2.7984099e7*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 3.9856141e7*fcs_pow(v,12)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 185185.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 1.8656066e7*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 7161.*fcs_pow(u,4)*(-6930.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 213675.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 9.761895e6*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 1.8656066e7*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 11.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 740740.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 4.240015e6*fcs_pow(v,12)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 2.2243771e7*fcs_pow(u,14)*(1. - 90.*w2 + 705.*fcs_pow(w,4) + 10575.*fcs_pow(v,4)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 138180.*fcs_pow(v,6)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 1.778574e6*fcs_pow(v,10)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 1.482145e6*fcs_pow(v,12)*(1. - 114.*w2 + 1121.*fcs_pow(w,4)) - 90.*v2*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 251685.*fcs_pow(v,8)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) - 4.279457e6*fcs_pow(u,12)*(11.*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 2838.*v2*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 1.33386e6*fcs_pow(v,6)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 7.002765e6*fcs_pow(v,8)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 1.2746447e7*fcs_pow(v,12)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 35475.*fcs_pow(v,4)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) - 5.290978e6*fcs_pow(v,10)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) + 164021.*fcs_pow(u,10)*(-56826.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 2.036265e6*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 1.23048585e8*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 2.67972474e8*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 77.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 8.14506e6*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 6.9023213e7*fcs_pow(v,12)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))));
        case 5 : return -1.347828286825972e27*fcs_pow(t,32)*w*(-3. + 66.*w2 - 231.*fcs_pow(w,4) - 848003.*fcs_pow(v,12)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 198.*v2*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 433862.*fcs_pow(v,10)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 34188.*fcs_pow(v,6)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 1155.*fcs_pow(v,4)*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 47619.*fcs_pow(v,8)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 21.*u2*(-1.8656066e7*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 11.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 3.253965e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 42735.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 462.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 148148.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 848003.*fcs_pow(v,12)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) - 5.936021e6*fcs_pow(u,12)*(1. - 30.*w2 + 141.*fcs_pow(w,4) + 251685.*fcs_pow(v,8)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 592858.*fcs_pow(v,10)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) - 27636.*fcs_pow(v,6)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 296429.*fcs_pow(v,12)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) - 18.*v2*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 705.*fcs_pow(v,4)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 138047.*fcs_pow(u,10)*(-8514.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 11.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 1.5872934e7*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 1.2746447e7*fcs_pow(v,12)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 4.201659e6*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 21285.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 266772.*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 1443.*fcs_pow(u,8)*(1.0181325e7*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 94710.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3.45116065e8*fcs_pow(v,12)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 2.67972474e8*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 77.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 8.14506e6*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 4.1016195e7*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 777.*fcs_pow(u,6)*(-3.529526e7*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1.026025e6*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 77.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 5.2782457e7*fcs_pow(v,12)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 6006.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 1.1344905e7*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 2.3698246e7*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 21.*fcs_pow(u,4)*(4.19761485e8*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 3.037034e7*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 51282.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 77.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 555555.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 5.5968198e7*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 3.9856141e7*fcs_pow(v,12)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 65231.*fcs_pow(u,14)*(45. - 1410.*w2 + 6909.*fcs_pow(w,4) - 2.34906e6*fcs_pow(v,6)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 1.3339305e7*fcs_pow(v,8)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 103635.*fcs_pow(v,4)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 1.9564314e7*fcs_pow(v,10)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) - 846.*v2*(15. - 490.*w2 + 2499.*fcs_pow(w,4)) + 5.632151e6*fcs_pow(v,12)*(15. - 590.*w2 + 3599.*fcs_pow(w,4))));
        case 6 : return 1.347828286825972e27*fcs_pow(t,33)*(3. - 297.*w2 + 3465.*fcs_pow(w,4) - 8547.*fcs_pow(w,6) + 594.*v2*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 714285.*fcs_pow(v,8)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.301586e6*fcs_pow(v,10)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 848003.*fcs_pow(v,12)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 3465.*fcs_pow(v,4)*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 34188.*fcs_pow(v,6)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 21.*u2*(33.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 2.22222e6*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 9.761895e6*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.8656066e7*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 1386.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 42735.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 2.544009e6*fcs_pow(v,12)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) + 414141.*fcs_pow(u,10)*(11.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2838.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.002765e6*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 1.5872934e7*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.2746447e7*fcs_pow(v,12)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 21285.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 266772.*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 21645.*fcs_pow(u,8)*(77.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 18942.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 678755.*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 8.9324158e7*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 6.9023213e7*fcs_pow(v,12)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.629012e6*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 8.203239e6*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 777.*fcs_pow(u,6)*(-90090.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 3.078075e6*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 3.529526e7*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 2.63912285e8*fcs_pow(v,12)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 77.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 3.4034715e7*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 7.1094738e7*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 63.*fcs_pow(u,4)*(2.777775e6*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 3.037034e7*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.39920495e8*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 77.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 5.5968198e7*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 3.9856141e7*fcs_pow(v,12)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 5.936021e6*fcs_pow(u,12)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6) - 138180.*fcs_pow(v,6)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 755055.*fcs_pow(v,8)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.778574e6*fcs_pow(v,10)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 54.*v2*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 2115.*fcs_pow(v,4)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) + 296429.*fcs_pow(v,12)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) + 195693.*fcs_pow(u,14)*(172725.*fcs_pow(v,4)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 2.34906e6*fcs_pow(v,6)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.3339305e7*fcs_pow(v,8)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 3.*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 846.*v2*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) - 6.521438e6*fcs_pow(v,10)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6)) + 5.632151e6*fcs_pow(v,12)*(-5. + 885.*w2 - 17995.*fcs_pow(w,4) + 75579.*fcs_pow(w,6))));
        case 7 : return -4.043484860477916e27*fcs_pow(t,34)*w*(-231.*(-1. + 35.*w2 - 259.*fcs_pow(w,4) + 481.*fcs_pow(w,6)) + 1.30592462e8*fcs_pow(v,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.7808063e7*fcs_pow(v,12)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 9.761895e6*fcs_pow(v,8)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1386.*v2*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 42735.*fcs_pow(v,4)*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 444444.*fcs_pow(v,6)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 147.*u2*(1.39920495e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 5.5968198e7*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 4.33862e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 3.9856141e7*fcs_pow(v,12)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 11.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 2442.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 79365.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) - 2.070705e6*fcs_pow(u,8)*(-19866.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 148995.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 11.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.867404e6*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 1.5872934e7*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.2746447e7*fcs_pow(v,12)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 1.400553e6*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 70707.*fcs_pow(u,6)*(3.393775e6*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 8.14506e6*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 13530.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 4.1016195e7*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 4.9302295e7*fcs_pow(v,12)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 11.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 1.2760594e7*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 5439.*fcs_pow(u,4)*(-3.529526e7*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 3.4034715e7*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 439725.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 7.1094738e7*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 11.*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 2574.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 7.540351e6*fcs_pow(v,12)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 2.1460999e7*fcs_pow(u,14)*(3.*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 7.908978e6*fcs_pow(v,10)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) + 26775.*fcs_pow(v,4)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 7.070147e6*fcs_pow(v,12)*(-5. + 305.*w2 - 3843.*fcs_pow(w,4) + 11895.*fcs_pow(w,6)) - 378420.*fcs_pow(v,6)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 126.*v2*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 445995.*fcs_pow(v,8)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) - 4.1552147e7*fcs_pow(u,12)*(3.*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 846.*v2*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 5.632151e6*fcs_pow(v,12)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) - 335580.*fcs_pow(v,6)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.905615e6*fcs_pow(v,8)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 4935.*fcs_pow(v,4)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) - 931634.*fcs_pow(v,10)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) + 4.57073617e8*fcs_pow(u,10)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6) - 54.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 2115.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 107865.*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 254082.*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 3948.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 42347.*fcs_pow(v,12)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 8 : return 2.830439402334541e28*fcs_pow(t,35)*(-33.*(1. - 140.*w2 + 2590.*fcs_pow(w,4) - 13468.*fcs_pow(w,6) + 19721.*fcs_pow(w,8)) - 9.761895e6*fcs_pow(v,8)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.8656066e7*fcs_pow(v,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 317460.*fcs_pow(v,6)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 2.544009e6*fcs_pow(v,12)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 198.*v2*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 6105.*fcs_pow(v,4)*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8)) + 50505.*fcs_pow(u,6)*(-18942.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 678755.*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 11.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.629012e6*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 8.203239e6*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.2760594e7*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 9.860459e6*fcs_pow(v,12)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 3.065857e6*fcs_pow(u,14)*(3.5350735e7*fcs_pow(v,12)*(1. - 244.*w2 + 7686.*fcs_pow(w,4) - 66612.*fcs_pow(w,6) + 159393.*fcs_pow(w,8)) + 3.*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 630.*v2*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 3.121965e6*fcs_pow(v,8)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 26775.*fcs_pow(v,4)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 7.908978e6*fcs_pow(v,10)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) - 378420.*fcs_pow(v,6)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 5.936021e6*fcs_pow(u,12)*(3.*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 846.*v2*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 24675.*fcs_pow(v,4)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 6.521438e6*fcs_pow(v,10)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) - 335580.*fcs_pow(v,6)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 5.632151e6*fcs_pow(v,12)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) + 1.905615e6*fcs_pow(v,8)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 6.5296231e7*fcs_pow(u,10)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8) - 54.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 2115.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 19740.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 296429.*fcs_pow(v,12)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 107865.*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 254082.*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 2.070705e6*fcs_pow(u,8)*(11.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 2838.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 21285.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 266772.*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.000395e6*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 2.267562e6*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.820921e6*fcs_pow(v,12)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 777.*fcs_pow(u,4)*(3.078075e6*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 3.529526e7*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 12870.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 3.4034715e7*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 7.1094738e7*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 3.7701755e7*fcs_pow(v,12)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 21.*u2*(-3.037034e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.39920495e8*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 396825.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 5.5968198e7*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 3.9856141e7*fcs_pow(v,12)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 11.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 2442.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return -8.491318207003624e28*fcs_pow(t,36)*w*(-1.259284455e9*fcs_pow(v,8)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 1.301586e7*fcs_pow(v,6)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 21978.*v2*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 5.5968198e7*fcs_pow(v,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3.9856141e7*fcs_pow(v,12)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 238095.*fcs_pow(v,4)*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 11.*(315. - 15540.*w2 + 181818.*fcs_pow(w,4) - 709956.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 6.4382997e7*fcs_pow(u,14)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8) + 3.08056405e8*fcs_pow(v,12)*(1. - 84.*w2 + 1638.*fcs_pow(w,4) - 10452.*fcs_pow(w,6) + 20033.*fcs_pow(w,8)) + 8.473905e6*fcs_pow(v,8)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 2.2220462e7*fcs_pow(v,10)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) - 594660.*fcs_pow(v,6)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 510.*v2*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 22525.*fcs_pow(v,4)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 2.78992987e8*fcs_pow(u,12)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8) - 7.908978e6*fcs_pow(v,10)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 7.070147e6*fcs_pow(v,12)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) + 1.337985e6*fcs_pow(v,8)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 8925.*fcs_pow(v,4)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 126140.*fcs_pow(v,6)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 126.*v2*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 1.95888693e8*fcs_pow(u,10)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8) + 5.632151e6*fcs_pow(v,12)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 282.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2.794902e6*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 111860.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 635205.*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 4935.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 9.79443465e8*fcs_pow(u,8)*(3.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 18.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 705.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 127041.*fcs_pow(v,12)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 35955.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 84694.*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 3948.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 690235.*fcs_pow(u,6)*(-178794.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 33.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 148995.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1.867404e6*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.5872934e7*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.2746447e7*fcs_pow(v,12)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 4.201659e6*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 10101.*fcs_pow(u,4)*(3.0543975e7*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 40590.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 8.14506e6*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 4.1016195e7*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 33.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 4.9302295e7*fcs_pow(v,12)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 3.8281782e7*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 777.*u2*(-1.0588578e8*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 439725.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 33.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 1.1344905e7*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 2.3698246e7*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2574.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 7.540351e6*fcs_pow(v,12)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))));
        case 10 : return 1.273697731050544e30*fcs_pow(t,37)*(1.8222204e7*fcs_pow(v,6)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 8.3952297e7*fcs_pow(v,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 33.*(-7. + 1295.*w2 - 33670.*fcs_pow(w,4) + 276094.*fcs_pow(w,6) - 848003.*fcs_pow(w,8) + 848003.*fcs_pow(w,10)) + 7326.*v2*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 3.9856141e7*fcs_pow(v,12)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 1.8656066e7*fcs_pow(v,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 79365.*fcs_pow(v,4)*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 259.*u2*(1.846845e6*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 2.1177156e7*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 33.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 7.1094738e7*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 1.1344905e7*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 2574.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 7.540351e6*fcs_pow(v,12)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) + 2.1460999e7*fcs_pow(u,14)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10) - 4.16262e6*fcs_pow(v,6)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 6.6661386e7*fcs_pow(v,10)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) + 6.1611281e7*fcs_pow(v,12)*(-1. + 315.*w2 - 13650.*fcs_pow(w,4) + 182910.*fcs_pow(w,6) - 901485.*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 8.473905e6*fcs_pow(v,8)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 40545.*fcs_pow(v,4)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 102.*v2*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 2.78992987e8*fcs_pow(u,12)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10) + 3.121965e6*fcs_pow(v,8)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,12)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) - 2.636326e6*fcs_pow(v,10)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 75684.*fcs_pow(v,6)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 42.*v2*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 595.*fcs_pow(v,4)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 6.5296231e7*fcs_pow(u,10)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10) - 1.9564314e7*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 846.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 5.632151e6*fcs_pow(v,12)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 1.143369e6*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 4935.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 22372.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 1.95888693e8*fcs_pow(u,8)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10) + 1.482145e6*fcs_pow(v,12)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 3525.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 30.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 254082.*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 6580.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 11985.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 966329.*fcs_pow(u,6)*(33.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 8514.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.33386e6*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 35475.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 5.462763e6*fcs_pow(v,12)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 1.000395e6*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 755854.*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 3367.*fcs_pow(u,4)*(-170478.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 6.108795e6*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.23048585e8*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 8.14506e6*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 33.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 3.8281782e7*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))));
        case 11 : return -4.712681604887011e31*fcs_pow(t,38)*w*(-41041.*fcs_pow(u,4)*(-16254.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 744975.*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 9.33702e6*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 2.1008295e7*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 1.5872934e7*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 1.2746447e7*fcs_pow(v,12)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 3.*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 1001.*u2*(555345.*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 8.14506e6*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 4.1016195e7*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 3.8281782e7*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 3.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 738.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) - 11.*(-1.925196e6*fcs_pow(v,6)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.1344905e7*fcs_pow(v,8)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 2.3698246e7*fcs_pow(v,10)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 7.540351e6*fcs_pow(v,12)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 3.*(-77. + 5005.*w2 - 82082.*fcs_pow(w,4) + 504218.*fcs_pow(w,6) - 1.260545e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 234.*v2*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 7995.*fcs_pow(v,4)*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) - 5.2782457e7*fcs_pow(u,12)*(11.*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 6.1611281e7*fcs_pow(v,12)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 8.473905e6*fcs_pow(v,8)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 1122.*v2*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 2.2220462e7*fcs_pow(v,10)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) + 49555.*fcs_pow(v,4)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 594660.*fcs_pow(v,6)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 8.2943861e7*fcs_pow(u,10)*(11.*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 1386.*v2*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 7.908978e6*fcs_pow(v,10)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 19635.*fcs_pow(v,4)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,12)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 277508.*fcs_pow(v,6)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 1.337985e6*fcs_pow(v,8)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 2.6471445e7*fcs_pow(u,8)*(11.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 3102.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 54285.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 5.632151e6*fcs_pow(v,12)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 246092.*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 1.397451e6*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 2.794902e6*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 1.2353341e7*fcs_pow(u,6)*(-33. + 2475.*w2 - 46530.*fcs_pow(w,4) + 325710.*fcs_pow(w,6) - 922845.*fcs_pow(w,8) + 889287.*fcs_pow(w,10) - 990.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 38775.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 217140.*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 395505.*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 931634.*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 635205.*fcs_pow(v,12)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 9.860459e6*fcs_pow(u,14)*(-2.39196738e8*fcs_pow(v,10)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) - 4.65234e6*fcs_pow(v,6)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 11.*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 2.9409435e7*fcs_pow(v,8)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 3498.*v2*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 131175.*fcs_pow(v,4)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10)) + 3.624193e6*fcs_pow(v,12)*(-693. + 75075.*w2 - 2.01201e6*fcs_pow(w,4) + 1.983267e7*fcs_pow(w,6) - 7.8228865e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10))));
        case 12 : return 4.712681604887011e31*fcs_pow(t,39)*(-11.*(-2.1177156e7*fcs_pow(v,6)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 3.4034715e7*fcs_pow(v,8)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 7.1094738e7*fcs_pow(v,10)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 167895.*fcs_pow(v,4)*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) + 3.*(77. - 18018.*w2 + 615615.*fcs_pow(w,4) - 7.059052e6*fcs_pow(w,6) + 3.4034715e7*fcs_pow(w,8) - 7.1094738e7*fcs_pow(w,10) + 5.2782457e7*fcs_pow(w,12)) + 7.540351e6*fcs_pow(v,12)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 702.*v2*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12))) - 7.9414335e7*fcs_pow(u,8)*(11.*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 3102.*v2*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 4.192353e6*fcs_pow(v,8)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 6.521438e6*fcs_pow(v,10)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 18095.*fcs_pow(v,4)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 5.632151e6*fcs_pow(v,12)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 246092.*fcs_pow(v,6)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 3.7060023e7*fcs_pow(u,6)*(11.*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 990.*v2*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 38775.*fcs_pow(v,4)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 2.794902e6*fcs_pow(v,10)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 1.482145e6*fcs_pow(v,12)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 72380.*fcs_pow(v,6)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 395505.*fcs_pow(v,8)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 2.585583e6*fcs_pow(u,4)*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12) - 2838.*v2*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 35475.*fcs_pow(v,4)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 444620.*fcs_pow(v,6)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 1.820921e6*fcs_pow(v,12)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 333465.*fcs_pow(v,8)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 755854.*fcs_pow(v,10)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 3003.*u2*(2.036265e6*fcs_pow(v,4)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 8.14506e6*fcs_pow(v,6)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 4.1016195e7*fcs_pow(v,8)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 5166.*v2*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) - 1.2760594e7*fcs_pow(v,10)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3.*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12)) + 9.860459e6*fcs_pow(v,12)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 2.48831583e8*fcs_pow(u,10)*(11.*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 832524.*fcs_pow(v,6)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 3.121965e6*fcs_pow(v,8)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 462.*v2*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 7.908978e6*fcs_pow(v,10)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 19635.*fcs_pow(v,4)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 7.070147e6*fcs_pow(v,12)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12))) - 5.2782457e7*fcs_pow(u,12)*(445995.*fcs_pow(v,4)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 4.16262e6*fcs_pow(v,6)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 11.*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 2.5421715e7*fcs_pow(v,8)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 3366.*v2*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) - 6.6661386e7*fcs_pow(v,10)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) + 6.1611281e7*fcs_pow(v,12)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12))) + 2.9581377e7*fcs_pow(u,14)*(-10494.*v2*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 306075.*fcs_pow(v,4)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 4.65234e6*fcs_pow(v,6)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 11.*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 2.9409435e7*fcs_pow(v,8)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) - 7.9732246e7*fcs_pow(v,10)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12)) + 3.624193e6*fcs_pow(v,12)*(231. - 90090.*w2 + 5.030025e6*fcs_pow(w,4) - 9.255246e7*fcs_pow(w,6) + 7.04059785e8*fcs_pow(w,8) - 2.284282858e9*fcs_pow(w,10) + 2.595775975e9*fcs_pow(w,12))));
      }
      case 13 : switch(k){
        case 0 : return -1.298054391195578e21*fcs_pow(t,28)*v*(203203.*u2*(3. - 186.*v2 + 3069.*fcs_pow(v,4) - 20460.*fcs_pow(v,6) + 63085.*fcs_pow(v,8) - 89466.*fcs_pow(v,10) + 47027.*fcs_pow(v,12)) - 6.299293e6*fcs_pow(u,4)*(3. - 198.*v2 + 3465.*fcs_pow(v,4) - 24420.*fcs_pow(v,6) + 79365.*fcs_pow(v,8) - 118326.*fcs_pow(v,10) + 65231.*fcs_pow(v,12)) - 143.*(21. - 1218.*v2 + 18879.*fcs_pow(v,4) - 118668.*fcs_pow(v,6) + 346115.*fcs_pow(v,8) - 465682.*fcs_pow(v,10) + 232841.*fcs_pow(v,12)) + 6.299293e6*fcs_pow(u,6)*(33. - 2310.*v2 + 42735.*fcs_pow(v,4) - 317460.*fcs_pow(v,6) + 1.084655e6*fcs_pow(v,8) - 1.696006e6*fcs_pow(v,10) + 978465.*fcs_pow(v,12)) + 3.3296263e7*fcs_pow(u,10)*(77. - 6006.*v2 + 123123.*fcs_pow(v,4) - 1.008436e6*fcs_pow(v,6) + 3.781635e6*fcs_pow(v,8) - 6.463158e6*fcs_pow(v,10) + 4.060189e6*fcs_pow(v,12)) - 4.499495e6*fcs_pow(u,8)*(231. - 17094.*v2 + 333333.*fcs_pow(v,4) - 2.603172e6*fcs_pow(v,6) + 9.328033e6*fcs_pow(v,8) - 1.5264054e7*fcs_pow(v,10) + 9.197571e6*fcs_pow(v,12)) - 3.026933e6*fcs_pow(u,12)*(1001. - 82082.*v2 + 1.764763e6*fcs_pow(v,4) - 1.512654e7*fcs_pow(v,6) + 5.9245615e7*fcs_pow(v,8) - 1.05564914e8*fcs_pow(v,10) + 6.9023213e7*fcs_pow(v,12)) + 1.363783e6*fcs_pow(u,14)*(1001. - 86086.*v2 + 1.936935e6*fcs_pow(v,4) - 1.734018e7*fcs_pow(v,6) + 7.0805735e7*fcs_pow(v,8) - 1.31312454e8*fcs_pow(v,10) + 8.9225129e7*fcs_pow(v,12)));
        case 1 : return 2.635050414127023e23*fcs_pow(t,29)*v*(-143.*(3. - 186.*v2 + 3069.*fcs_pow(v,4) - 20460.*fcs_pow(v,6) + 63085.*fcs_pow(v,8) - 89466.*fcs_pow(v,10) + 47027.*fcs_pow(v,12)) + 31031.*u2*(3. - 198.*v2 + 3465.*fcs_pow(v,4) - 24420.*fcs_pow(v,6) + 79365.*fcs_pow(v,8) - 118326.*fcs_pow(v,10) + 65231.*fcs_pow(v,12)) - 93093.*fcs_pow(u,4)*(33. - 2310.*v2 + 42735.*fcs_pow(v,4) - 317460.*fcs_pow(v,6) + 1.084655e6*fcs_pow(v,8) - 1.696006e6*fcs_pow(v,10) + 978465.*fcs_pow(v,12)) - 2.460315e6*fcs_pow(u,8)*(77. - 6006.*v2 + 123123.*fcs_pow(v,4) - 1.008436e6*fcs_pow(v,6) + 3.781635e6*fcs_pow(v,8) - 6.463158e6*fcs_pow(v,10) + 4.060189e6*fcs_pow(v,12)) + 155155.*fcs_pow(u,6)*(231. - 17094.*v2 + 333333.*fcs_pow(v,4) - 2.603172e6*fcs_pow(v,6) + 9.328033e6*fcs_pow(v,8) - 1.5264054e7*fcs_pow(v,10) + 9.197571e6*fcs_pow(v,12)) + 2.022161e6*fcs_pow(u,14)*(143. - 12870.*v2 + 302445.*fcs_pow(v,4) - 2.82282e6*fcs_pow(v,6) + 1.1996985e7*fcs_pow(v,8) - 2.3121462e7*fcs_pow(v,10) + 1.6303595e7*fcs_pow(v,12)) + 492063.*fcs_pow(u,10)*(1001. - 82082.*v2 + 1.764763e6*fcs_pow(v,4) - 1.512654e7*fcs_pow(v,6) + 5.9245615e7*fcs_pow(v,8) - 1.05564914e8*fcs_pow(v,10) + 6.9023213e7*fcs_pow(v,12)) - 611351.*fcs_pow(u,12)*(1001. - 86086.*v2 + 1.936935e6*fcs_pow(v,4) - 1.734018e7*fcs_pow(v,6) + 7.0805735e7*fcs_pow(v,8) - 1.31312454e8*fcs_pow(v,10) + 8.9225129e7*fcs_pow(v,12)))*w;
        case 2 : return -2.635050414127023e23*fcs_pow(t,30)*v*(-143.*(-3. + 93.*w2 - 186.*v2*(-1. + 33.*w2) + 3069.*fcs_pow(v,4)*(-1. + 35.*w2) - 20460.*fcs_pow(v,6)*(-1. + 37.*w2) + 63085.*fcs_pow(v,8)*(-1. + 39.*w2) - 89466.*fcs_pow(v,10)*(-1. + 41.*w2) + 47027.*fcs_pow(v,12)*(-1. + 43.*w2)) + 31031.*u2*(-3. + 99.*w2 - 198.*v2*(-1. + 35.*w2) + 3465.*fcs_pow(v,4)*(-1. + 37.*w2) - 24420.*fcs_pow(v,6)*(-1. + 39.*w2) + 79365.*fcs_pow(v,8)*(-1. + 41.*w2) - 118326.*fcs_pow(v,10)*(-1. + 43.*w2) + 65231.*fcs_pow(v,12)*(-1. + 45.*w2)) - 93093.*fcs_pow(u,4)*(33.*(-1. + 35.*w2) - 2310.*v2*(-1. + 37.*w2) + 42735.*fcs_pow(v,4)*(-1. + 39.*w2) - 317460.*fcs_pow(v,6)*(-1. + 41.*w2) + 1.084655e6*fcs_pow(v,8)*(-1. + 43.*w2) - 1.696006e6*fcs_pow(v,10)*(-1. + 45.*w2) + 978465.*fcs_pow(v,12)*(-1. + 47.*w2)) + 155155.*fcs_pow(u,6)*(231.*(-1. + 37.*w2) - 17094.*v2*(-1. + 39.*w2) + 333333.*fcs_pow(v,4)*(-1. + 41.*w2) - 2.603172e6*fcs_pow(v,6)*(-1. + 43.*w2) + 9.328033e6*fcs_pow(v,8)*(-1. + 45.*w2) - 1.5264054e7*fcs_pow(v,10)*(-1. + 47.*w2) + 9.197571e6*fcs_pow(v,12)*(-1. + 49.*w2)) - 2.460315e6*fcs_pow(u,8)*(77.*(-1. + 39.*w2) - 6006.*v2*(-1. + 41.*w2) + 123123.*fcs_pow(v,4)*(-1. + 43.*w2) - 1.008436e6*fcs_pow(v,6)*(-1. + 45.*w2) + 3.781635e6*fcs_pow(v,8)*(-1. + 47.*w2) - 6.463158e6*fcs_pow(v,10)*(-1. + 49.*w2) + 4.060189e6*fcs_pow(v,12)*(-1. + 51.*w2)) + 492063.*fcs_pow(u,10)*(1001.*(-1. + 41.*w2) - 82082.*v2*(-1. + 43.*w2) + 1.764763e6*fcs_pow(v,4)*(-1. + 45.*w2) - 1.512654e7*fcs_pow(v,6)*(-1. + 47.*w2) + 5.9245615e7*fcs_pow(v,8)*(-1. + 49.*w2) - 1.05564914e8*fcs_pow(v,10)*(-1. + 51.*w2) + 6.9023213e7*fcs_pow(v,12)*(-1. + 53.*w2)) - 611351.*fcs_pow(u,12)*(1001.*(-1. + 43.*w2) - 86086.*v2*(-1. + 45.*w2) + 1.936935e6*fcs_pow(v,4)*(-1. + 47.*w2) - 1.734018e7*fcs_pow(v,6)*(-1. + 49.*w2) + 7.0805735e7*fcs_pow(v,8)*(-1. + 51.*w2) - 1.31312454e8*fcs_pow(v,10)*(-1. + 53.*w2) + 8.9225129e7*fcs_pow(v,12)*(-1. + 55.*w2)) + 2.022161e6*fcs_pow(u,14)*(143.*(-1. + 45.*w2) - 12870.*v2*(-1. + 47.*w2) + 302445.*fcs_pow(v,4)*(-1. + 49.*w2) - 2.82282e6*fcs_pow(v,6)*(-1. + 51.*w2) + 1.1996985e7*fcs_pow(v,8)*(-1. + 53.*w2) - 2.3121462e7*fcs_pow(v,10)*(-1. + 55.*w2) + 1.6303595e7*fcs_pow(v,12)*(-1. + 57.*w2)));
        case 3 : return 2.450596885138131e25*fcs_pow(t,31)*v*w*(-143.*(-3. + 33.*w2 - 24420.*fcs_pow(v,6)*(-1. + 13.*w2) + 65231.*fcs_pow(v,12)*(-1. + 15.*w2) - 66.*v2*(-3. + 35.*w2) + 1155.*fcs_pow(v,4)*(-3. + 37.*w2) + 26455.*fcs_pow(v,8)*(-3. + 41.*w2) - 39442.*fcs_pow(v,10)*(-3. + 43.*w2)) + 1001.*u2*(128205.*fcs_pow(v,4)*(-1. + 13.*w2) - 5.088018e6*fcs_pow(v,10)*(-1. + 15.*w2) + 33.*(-3. + 35.*w2) - 2310.*v2*(-3. + 37.*w2) - 317460.*fcs_pow(v,6)*(-3. + 41.*w2) + 1.084655e6*fcs_pow(v,8)*(-3. + 43.*w2) + 978465.*fcs_pow(v,12)*(-3. + 47.*w2)) + 555555.*fcs_pow(u,6)*(77.*(-1. + 13.*w2) - 1.008436e6*fcs_pow(v,6)*(-1. + 15.*w2) + 4.060189e6*fcs_pow(v,12)*(-1. + 17.*w2) - 2002.*v2*(-3. + 41.*w2) + 41041.*fcs_pow(v,4)*(-3. + 43.*w2) + 1.260545e6*fcs_pow(v,8)*(-3. + 47.*w2) - 2.154386e6*fcs_pow(v,10)*(-3. + 49.*w2)) - 15015.*fcs_pow(u,4)*(-17094.*v2*(-1. + 13.*w2) + 9.328033e6*fcs_pow(v,8)*(-1. + 15.*w2) + 77.*(-3. + 37.*w2) + 111111.*fcs_pow(v,4)*(-3. + 41.*w2) - 867724.*fcs_pow(v,6)*(-3. + 43.*w2) - 5.088018e6*fcs_pow(v,10)*(-3. + 47.*w2) + 3.065857e6*fcs_pow(v,12)*(-3. + 49.*w2)) - 79365.*fcs_pow(u,8)*(5.294289e6*fcs_pow(v,4)*(-1. + 15.*w2) - 3.16694742e8*fcs_pow(v,10)*(-1. + 17.*w2) + 1001.*(-3. + 41.*w2) - 82082.*v2*(-3. + 43.*w2) - 1.512654e7*fcs_pow(v,6)*(-3. + 47.*w2) + 5.9245615e7*fcs_pow(v,8)*(-3. + 49.*w2) + 6.9023213e7*fcs_pow(v,12)*(-3. + 53.*w2)) - 5.936021e6*fcs_pow(u,12)*(143.*(-1. + 15.*w2) - 2.82282e6*fcs_pow(v,6)*(-1. + 17.*w2) + 1.6303595e7*fcs_pow(v,12)*(-1. + 19.*w2) - 4290.*v2*(-3. + 47.*w2) + 100815.*fcs_pow(v,4)*(-3. + 49.*w2) + 3.998995e6*fcs_pow(v,8)*(-3. + 53.*w2) - 7.707154e6*fcs_pow(v,10)*(-3. + 55.*w2)) + 216931.*fcs_pow(u,10)*(-258258.*v2*(-1. + 15.*w2) + 2.12417205e8*fcs_pow(v,8)*(-1. + 17.*w2) + 1001.*(-3. + 43.*w2) + 1.936935e6*fcs_pow(v,4)*(-3. + 47.*w2) - 1.734018e7*fcs_pow(v,6)*(-3. + 49.*w2) - 1.31312454e8*fcs_pow(v,10)*(-3. + 53.*w2) + 8.9225129e7*fcs_pow(v,12)*(-3. + 55.*w2)) + 3.587705e6*fcs_pow(u,14)*(269451.*fcs_pow(v,4)*(-1. + 17.*w2) - 2.3121462e7*fcs_pow(v,10)*(-1. + 19.*w2) + 39.*(-3. + 47.*w2) - 3666.*v2*(-3. + 49.*w2) - 872508.*fcs_pow(v,6)*(-3. + 53.*w2) + 3.853577e6*fcs_pow(v,8)*(-3. + 55.*w2) + 5.632151e6*fcs_pow(v,12)*(-3. + 59.*w2)));
        case 4 : return -2.450596885138131e25*fcs_pow(t,32)*v*(-143.*(3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) + 3465.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 24420.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 118326.*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 65231.*fcs_pow(v,12)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 66.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 26455.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4))) + 555555.*fcs_pow(u,6)*(77.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 123123.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 1.008436e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 6.463158e6*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,12)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 2002.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 1.260545e6*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 15015.*fcs_pow(u,4)*(231.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 17094.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 2.603172e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 9.328033e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 9.197571e6*fcs_pow(v,12)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 111111.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 5.088018e6*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 3003.*u2*(-2310.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 42735.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 1.084655e6*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 1.696006e6*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 11.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 105820.*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 326155.*fcs_pow(v,12)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 5.936021e6*fcs_pow(u,12)*(143.*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 302445.*fcs_pow(v,4)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 2.82282e6*fcs_pow(v,6)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 2.3121462e7*fcs_pow(v,10)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 1.6303595e7*fcs_pow(v,12)*(1. - 114.*w2 + 1121.*fcs_pow(w,4)) - 4290.*v2*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 3.998995e6*fcs_pow(v,8)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) + 650793.*fcs_pow(u,10)*(1001.*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 86086.*v2*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 1.734018e7*fcs_pow(v,6)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 7.0805735e7*fcs_pow(v,8)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 8.9225129e7*fcs_pow(v,12)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 645645.*fcs_pow(v,4)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) - 4.3770818e7*fcs_pow(v,10)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) - 79365.*fcs_pow(u,8)*(-246246.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 5.294289e6*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 1.77736845e8*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 3.16694742e8*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 1001.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 1.512654e7*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 6.9023213e7*fcs_pow(v,12)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) + 3.587705e6*fcs_pow(u,14)*(-10998.*v2*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 269451.*fcs_pow(v,4)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 1.1560731e7*fcs_pow(v,8)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) - 2.3121462e7*fcs_pow(v,10)*(1. - 114.*w2 + 1121.*fcs_pow(w,4)) + 39.*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) - 872508.*fcs_pow(v,6)*(3. - 318.*w2 + 2915.*fcs_pow(w,4)) + 5.632151e6*fcs_pow(v,12)*(3. - 354.*w2 + 3599.*fcs_pow(w,4))));
        case 5 : return 1.347828286825972e27*fcs_pow(t,33)*v*w*(-13.*(-5.088018e6*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 33.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) + 1.084655e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 25641.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) - 462.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 63492.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 195693.*fcs_pow(v,12)*(15. - 470.*w2 + 2303.*fcs_pow(w,4))) + 1.7808063e7*fcs_pow(u,10)*(143.*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 3.998995e6*fcs_pow(v,8)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 7.707154e6*fcs_pow(v,10)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) - 564564.*fcs_pow(v,6)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 3.260719e6*fcs_pow(v,12)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) - 858.*v2*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 20163.*fcs_pow(v,4)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 295815.*fcs_pow(u,8)*(-258258.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 1001.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 1.31312454e8*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 8.9225129e7*fcs_pow(v,12)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 4.2483441e7*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 387387.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 3.468036e6*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 10101.*fcs_pow(u,6)*(2.6471445e7*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 410410.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 3.45116065e8*fcs_pow(v,12)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 3.16694742e8*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 1001.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 1.512654e7*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 5.9245615e7*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 30303.*fcs_pow(u,4)*(-5.04218e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 205205.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 77.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 4.060189e6*fcs_pow(v,12)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 2002.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 1.260545e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 2.154386e6*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 273.*u2*(4.6640165e7*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 4.33862e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 17094.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 77.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 111111.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 5.088018e6*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 3.065857e6*fcs_pow(v,12)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 5.936021e6*fcs_pow(u,12)*(-4.36254e6*fcs_pow(v,6)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 1.9267885e7*fcs_pow(v,8)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 269451.*fcs_pow(v,4)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 2.3121462e7*fcs_pow(v,10)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) + 39.*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 3666.*v2*(15. - 490.*w2 + 2499.*fcs_pow(w,4)) + 5.632151e6*fcs_pow(v,12)*(15. - 590.*w2 + 3599.*fcs_pow(w,4))) + 3.065857e6*fcs_pow(u,14)*(487305.*fcs_pow(v,4)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 4.91946e6*fcs_pow(v,6)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) - 11466.*v2*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 1.3528515e7*fcs_pow(v,8)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) + 39.*(15. - 490.*w2 + 2499.*fcs_pow(w,4)) - 9.346974e6*fcs_pow(v,10)*(15. - 590.*w2 + 3599.*fcs_pow(w,4)) + 7.070147e6*fcs_pow(v,12)*(15. - 610.*w2 + 3843.*fcs_pow(w,4))));
        case 6 : return -4.043484860477916e27*fcs_pow(t,34)*v*(-13.*(33.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) - 317460.*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 1.084655e6*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.696006e6*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 462.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) + 8547.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 195693.*fcs_pow(v,12)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6))) - 295815.*fcs_pow(u,8)*(1001.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 86086.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 7.0805735e7*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 1.31312454e8*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 8.9225129e7*fcs_pow(v,12)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 387387.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 3.468036e6*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 50505.*fcs_pow(u,6)*(1001.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 82082.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 1.764763e6*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 1.05564914e8*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 6.9023213e7*fcs_pow(v,12)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 3.025308e6*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 1.1849123e7*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 10101.*fcs_pow(u,4)*(-30030.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 615615.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 5.04218e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 2.0300945e7*fcs_pow(v,12)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 77.*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 3.781635e6*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 6.463158e6*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 91.*u2*(1.666665e6*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 1.301586e7*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 4.6640165e7*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 231.*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.5264054e7*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 9.197571e6*fcs_pow(v,12)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 5.936021e6*fcs_pow(u,10)*(143.*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 2.82282e6*fcs_pow(v,6)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 1.1996985e7*fcs_pow(v,8)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 2.3121462e7*fcs_pow(v,10)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 2574.*v2*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 60489.*fcs_pow(v,4)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) + 3.260719e6*fcs_pow(v,12)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) + 3.065857e6*fcs_pow(u,14)*(-19110.*v2*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 487305.*fcs_pow(v,4)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 4.91946e6*fcs_pow(v,6)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 3.5350735e7*fcs_pow(v,12)*(-1. + 183.*w2 - 3843.*fcs_pow(w,4) + 16653.*fcs_pow(w,6)) + 39.*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) + 4.509505e6*fcs_pow(v,8)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6)) - 9.346974e6*fcs_pow(v,10)*(-5. + 885.*w2 - 17995.*fcs_pow(w,4) + 75579.*fcs_pow(w,6))) - 5.936021e6*fcs_pow(u,12)*(449085.*fcs_pow(v,4)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 4.36254e6*fcs_pow(v,6)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.9267885e7*fcs_pow(v,8)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 39.*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 3666.*v2*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) - 7.707154e6*fcs_pow(v,10)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6)) + 5.632151e6*fcs_pow(v,12)*(-5. + 885.*w2 - 17995.*fcs_pow(w,4) + 75579.*fcs_pow(w,6))));
        case 7 : return 4.043484860477916e27*fcs_pow(t,35)*v*w*(-13.*(3.26481155e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.06848378e8*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 1.301586e7*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 6.4382997e7*fcs_pow(v,12)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 231.*(-35. + 1295.*w2 - 10101.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 17094.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 333333.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6))) + 1.4494935e7*fcs_pow(u,6)*(-86086.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 387387.*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 143.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 3.468036e6*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 1.8758922e7*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.2746447e7*fcs_pow(v,12)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 2.023021e6*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 212121.*fcs_pow(u,4)*(8.823815e6*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 1.512654e7*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 58630.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 5.9245615e7*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 4.9302295e7*fcs_pow(v,12)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 143.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 1.5080702e7*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 70707.*u2*(-5.04218e6*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 3.781635e6*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 87945.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 6.463158e6*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 11.*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 858.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 580027.*fcs_pow(v,12)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 2.1460999e7*fcs_pow(u,14)*(2.156394835e9*fcs_pow(v,12)*(-1. + 63.*w2 - 819.*fcs_pow(w,4) + 2613.*fcs_pow(w,6)) + 2.57041785e8*fcs_pow(v,8)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) - 139230.*v2*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 5.51471466e8*fcs_pow(v,10)*(-5. + 305.*w2 - 3843.*fcs_pow(w,4) + 11895.*fcs_pow(w,6)) + 3.689595e6*fcs_pow(v,4)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 273.*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) - 7.73058e6*fcs_pow(v,6)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) - 1.952950909e9*fcs_pow(u,12)*(39.*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 9.346974e6*fcs_pow(v,10)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) + 69615.*fcs_pow(v,4)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 7.070147e6*fcs_pow(v,12)*(-5. + 305.*w2 - 3843.*fcs_pow(w,4) + 11895.*fcs_pow(w,6)) - 702780.*fcs_pow(v,6)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 546.*v2*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 644215.*fcs_pow(v,8)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) + 1.371220851e9*fcs_pow(u,10)*(39.*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 3666.*v2*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 5.632151e6*fcs_pow(v,12)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) - 623220.*fcs_pow(v,6)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 2.752555e6*fcs_pow(v,8)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 12831.*fcs_pow(v,4)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) - 1.101022e6*fcs_pow(v,10)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) - 6.23282205e8*fcs_pow(u,8)*(143.*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 2574.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 60489.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 1.713855e6*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 3.303066e6*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 80652.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 465817.*fcs_pow(v,12)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 8 : return -2.830439402334541e28*fcs_pow(t,36)*v*(-151515.*fcs_pow(u,4)*(-82082.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 1.764763e6*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 143.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 3.025308e6*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 1.1849123e7*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.5080702e7*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 9.860459e6*fcs_pow(v,12)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8))) + 1.07304995e8*fcs_pow(u,14)*(-7.8781638e7*fcs_pow(v,10)*(1. - 244.*w2 + 7686.*fcs_pow(w,4) - 66612.*fcs_pow(w,6) + 159393.*fcs_pow(w,8)) + 6.1611281e7*fcs_pow(v,12)*(1. - 252.*w2 + 8190.*fcs_pow(w,4) - 73164.*fcs_pow(w,6) + 180297.*fcs_pow(w,8)) + 39.*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 1.546116e6*fcs_pow(v,6)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) - 3978.*v2*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 7.344051e6*fcs_pow(v,8)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) + 105417.*fcs_pow(v,4)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 2.78992987e8*fcs_pow(u,12)*(3.5350735e7*fcs_pow(v,12)*(1. - 244.*w2 + 7686.*fcs_pow(w,4) - 66612.*fcs_pow(w,6) + 159393.*fcs_pow(w,8)) + 39.*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2730.*v2*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 4.509505e6*fcs_pow(v,8)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 69615.*fcs_pow(v,4)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 9.346974e6*fcs_pow(v,10)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) - 702780.*fcs_pow(v,6)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 1.95888693e8*fcs_pow(u,10)*(39.*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3666.*v2*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 64155.*fcs_pow(v,4)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 7.707154e6*fcs_pow(v,10)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) - 623220.*fcs_pow(v,6)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 5.632151e6*fcs_pow(v,12)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) + 2.752555e6*fcs_pow(v,8)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 8.9040315e7*fcs_pow(u,8)*(143.*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 2574.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 60489.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 403260.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 3.260719e6*fcs_pow(v,12)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 1.713855e6*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 3.303066e6*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 2.070705e6*fcs_pow(u,6)*(1001.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 86086.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 387387.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3.468036e6*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.0115105e7*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 1.8758922e7*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.2746447e7*fcs_pow(v,12)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 10101.*u2*(615615.*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 5.04218e6*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 4290.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 3.781635e6*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 6.463158e6*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 2.900135e6*fcs_pow(v,12)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 13.*(-1.301586e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 4.6640165e7*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 238095.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 1.5264054e7*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 9.197571e6*fcs_pow(v,12)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 33.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) - 2442.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))));
        case 9 : return 9.425363209774023e30*fcs_pow(t,37)*v*w*(-7.540351e6*fcs_pow(u,12)*(2.156394835e9*fcs_pow(v,12)*(1. - 84.*w2 + 1638.*fcs_pow(w,4) - 10452.*fcs_pow(w,6) + 20033.*fcs_pow(w,8)) + 8.5680595e7*fcs_pow(v,8)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 1.83823822e8*fcs_pow(v,10)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) - 7.73058e6*fcs_pow(v,6)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 15470.*v2*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 409955.*fcs_pow(v,4)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 91.*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 8.2943861e7*fcs_pow(u,10)*(-9.346974e6*fcs_pow(v,10)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 7.070147e6*fcs_pow(v,12)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) + 13.*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.932645e6*fcs_pow(v,8)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 23205.*fcs_pow(v,4)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 234260.*fcs_pow(v,6)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 546.*v2*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 8.823815e6*fcs_pow(u,8)*(1.6896453e7*fcs_pow(v,12)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 39.*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3666.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 9.909198e6*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 623220.*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 2.752555e6*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 38493.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 5.615155e6*fcs_pow(u,6)*(429.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 858.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 20163.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.397451e6*fcs_pow(v,12)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 571285.*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1.101022e6*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 80652.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 18655.*fcs_pow(u,4)*(-774774.*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 429.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 387387.*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 3.468036e6*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.8758922e7*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.2746447e7*fcs_pow(v,12)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 6.069063e6*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 91.*u2*(7.9414335e7*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 175890.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 1.512654e7*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 5.9245615e7*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 429.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 4.9302295e7*fcs_pow(v,12)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 4.5242106e7*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 13.*(-1.512654e7*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 87945.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 33.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 1.260545e6*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 2.154386e6*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 858.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) + 580027.*fcs_pow(v,12)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 4.9302295e7*fcs_pow(u,14)*(-2.82687054e8*fcs_pow(v,10)*(1. - 84.*w2 + 1638.*fcs_pow(w,4) - 10452.*fcs_pow(w,6) + 20033.*fcs_pow(w,8)) - 1.728012e6*fcs_pow(v,6)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 8.496059e6*fcs_pow(v,8)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) + 68211.*fcs_pow(v,4)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 13.*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1378.*v2*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 3.624193e6*fcs_pow(v,12)*(63. - 5460.*w2 + 109746.*fcs_pow(w,4) - 721188.*fcs_pow(w,6) + 1.422343e6*fcs_pow(w,8))));
        case 10 : return -4.712681604887011e31*fcs_pow(t,38)*v*(-13.*(369369.*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 3.025308e6*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 33.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 6.463158e6*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 1.260545e6*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 858.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 580027.*fcs_pow(v,12)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10))) - 5.2782457e7*fcs_pow(u,12)*(-7.73058e6*fcs_pow(v,6)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 7.8781638e7*fcs_pow(v,10)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) + 6.1611281e7*fcs_pow(v,12)*(-1. + 315.*w2 - 13650.*fcs_pow(w,4) + 182910.*fcs_pow(w,6) - 901485.*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 1.2240085e7*fcs_pow(v,8)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 105417.*fcs_pow(v,4)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 13.*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 442.*v2*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 2.48831583e8*fcs_pow(u,10)*(4.509505e6*fcs_pow(v,8)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,12)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) + 13.*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 3.115658e6*fcs_pow(v,10)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 140556.*fcs_pow(v,6)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 182.*v2*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 1547.*fcs_pow(v,4)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 2.6471445e7*fcs_pow(u,8)*(-2.3121462e7*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 3666.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 5.632151e6*fcs_pow(v,12)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 13.*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 1.651533e6*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 12831.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 41548.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 1.123031e6*fcs_pow(u,6)*(429.*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 4.8910785e7*fcs_pow(v,12)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 302445.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 4290.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 9.909198e6*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 403260.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 571285.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 11193.*fcs_pow(u,4)*(3003.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 258258.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.734018e7*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 645645.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 3.8239341e7*fcs_pow(v,12)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 1.0115105e7*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 6.252974e6*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 91.*u2*(-738738.*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 1.5882867e7*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.77736845e8*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 1.512654e7*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 429.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 4.5242106e7*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 9.860459e6*fcs_pow(u,14)*(2.387385e6*fcs_pow(v,4)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 1.27440885e8*fcs_pow(v,8)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) - 2.82687054e8*fcs_pow(v,10)*(-1. + 315.*w2 - 13650.*fcs_pow(w,4) + 182910.*fcs_pow(w,6) - 901485.*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) - 8.64006e6*fcs_pow(v,6)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 12402.*v2*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 13.*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10)) + 3.624193e6*fcs_pow(v,12)*(-63. + 20475.*w2 - 914550.*fcs_pow(w,4) + 1.262079e7*fcs_pow(w,6) - 6.4005435e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10))));
        case 11 : return 6.126486086353115e32*fcs_pow(t,39)*v*w*(3157.*u2*(-70434.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.936935e6*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.734018e7*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 3.0345315e7*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 1.8758922e7*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 1.2746447e7*fcs_pow(v,12)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 39.*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) - 11.*(1.443897e6*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.512654e7*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 5.9245615e7*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 4.5242106e7*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 9.860459e6*fcs_pow(v,12)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 39.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) - 3198.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) + 1.9140891e7*fcs_pow(u,10)*(1001.*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 4.31278967e8*fcs_pow(v,12)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 8.5680595e7*fcs_pow(v,8)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 34034.*v2*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 1.83823822e8*fcs_pow(v,10)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) + 901901.*fcs_pow(v,4)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 7.73058e6*fcs_pow(v,6)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 9.5704455e7*fcs_pow(u,8)*(143.*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 6006.*v2*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 9.346974e6*fcs_pow(v,10)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 51051.*fcs_pow(v,4)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 7.070147e6*fcs_pow(v,12)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 515372.*fcs_pow(v,6)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 1.932645e6*fcs_pow(v,8)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 4.751285e6*fcs_pow(u,6)*(429.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 40326.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 423423.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 1.6896453e7*fcs_pow(v,12)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 1.371084e6*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 6.055621e6*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 9.909198e6*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 2.850771e6*fcs_pow(u,4)*(39.*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 4290.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 100815.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 403260.*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 571285.*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 1.101022e6*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 635205.*fcs_pow(v,12)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 6.9023213e7*fcs_pow(u,12)*(-2.82687054e8*fcs_pow(v,10)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) - 8.64006e6*fcs_pow(v,6)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 143.*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 4.2480295e7*fcs_pow(v,8)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 15158.*v2*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 341055.*fcs_pow(v,4)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10)) + 3.624193e6*fcs_pow(v,12)*(-693. + 75075.*w2 - 2.01201e6*fcs_pow(w,4) + 1.983267e7*fcs_pow(w,6) - 7.8228865e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10))) + 5.22604327e8*fcs_pow(u,14)*(1.1282865e7*fcs_pow(v,8)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 65835.*fcs_pow(v,4)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 739860.*fcs_pow(v,6)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) + 11.*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 990.*v2*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10)) - 410286.*fcs_pow(v,10)*(-693. + 75075.*w2 - 2.01201e6*fcs_pow(w,4) + 1.983267e7*fcs_pow(w,6) - 7.8228865e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10)) + 341905.*fcs_pow(v,12)*(-693. + 77385.*w2 - 2.135826e6*fcs_pow(w,4) + 2.1663378e7*fcs_pow(w,6) - 8.7857033e7*fcs_pow(w,8) + 1.19805045e8*fcs_pow(w,10))));
        case 12 : return -1.837945825905934e33*fcs_pow(t,40)*v*(1.4253855e7*fcs_pow(u,6)*(143.*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 13442.*v2*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 6.055621e6*fcs_pow(v,8)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 7.707154e6*fcs_pow(v,10)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 47047.*fcs_pow(v,4)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 5.632151e6*fcs_pow(v,12)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 457028.*fcs_pow(v,6)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 2.850771e6*fcs_pow(u,4)*(143.*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 4290.*v2*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 100815.*fcs_pow(v,4)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 3.303066e6*fcs_pow(v,10)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 1.482145e6*fcs_pow(v,12)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 134420.*fcs_pow(v,6)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 571285.*fcs_pow(v,8)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 3157.*u2*(-258258.*v2*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 1.936935e6*fcs_pow(v,4)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 1.734018e7*fcs_pow(v,6)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 273.*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) + 3.8239341e7*fcs_pow(v,12)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 1.0115105e7*fcs_pow(v,8)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 1.8758922e7*fcs_pow(v,10)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 11.*(5.294289e6*fcs_pow(v,4)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 1.512654e7*fcs_pow(v,6)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 5.9245615e7*fcs_pow(v,8)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 22386.*v2*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) - 1.5080702e7*fcs_pow(v,10)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 39.*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12)) + 9.860459e6*fcs_pow(v,12)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 9.5704455e7*fcs_pow(u,8)*(143.*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 1.546116e6*fcs_pow(v,6)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 4.509505e6*fcs_pow(v,8)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 2002.*v2*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 9.346974e6*fcs_pow(v,10)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 51051.*fcs_pow(v,4)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 7.070147e6*fcs_pow(v,12)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12))) + 4.4662079e7*fcs_pow(u,10)*(1.159587e6*fcs_pow(v,4)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 7.73058e6*fcs_pow(v,6)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 143.*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3.6720255e7*fcs_pow(v,8)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 14586.*v2*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) - 7.8781638e7*fcs_pow(v,10)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) + 6.1611281e7*fcs_pow(v,12)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12))) + 5.22604327e8*fcs_pow(u,14)*(33.*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 2310.*v2*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 65835.*fcs_pow(v,4)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 739860.*fcs_pow(v,6)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) + 3.760955e6*fcs_pow(v,8)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12)) + 3.760955e6*fcs_pow(v,12)*(21. - 8442.*w2 + 485415.*fcs_pow(w,4) - 9.190524e6*fcs_pow(w,6) + 7.1883027e7*fcs_pow(w,8) - 2.3961009e8*fcs_pow(w,10) + 2.79545105e8*fcs_pow(w,12)) - 410286.*fcs_pow(v,10)*(231. - 90090.*w2 + 5.030025e6*fcs_pow(w,4) - 9.255246e7*fcs_pow(w,6) + 7.04059785e8*fcs_pow(w,8) - 2.284282858e9*fcs_pow(w,10) + 2.595775975e9*fcs_pow(w,12))) - 6.9023213e7*fcs_pow(u,12)*(-45474.*v2*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 795795.*fcs_pow(v,4)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 8.64006e6*fcs_pow(v,6)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 143.*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 4.2480295e7*fcs_pow(v,8)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) - 9.4229018e7*fcs_pow(v,10)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12)) + 3.624193e6*fcs_pow(v,12)*(231. - 90090.*w2 + 5.030025e6*fcs_pow(w,4) - 9.255246e7*fcs_pow(w,6) + 7.04059785e8*fcs_pow(w,8) - 2.284282858e9*fcs_pow(w,10) + 2.595775975e9*fcs_pow(w,12))));
        case 13 : return 7.535577886214331e34*fcs_pow(t,41)*v*w*(-3.0345315e7*fcs_pow(u,8)*(2.57041785e8*fcs_pow(v,8)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) - 5.51471466e8*fcs_pow(v,10)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) + 2.705703e6*fcs_pow(v,4)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 4.31278967e8*fcs_pow(v,12)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) - 7.73058e6*fcs_pow(v,6)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 77.*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 7854.*v2*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 2.12417205e8*fcs_pow(u,6)*(-9.346974e6*fcs_pow(v,10)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 7.070147e6*fcs_pow(v,12)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) + 11.*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 515372.*fcs_pow(v,6)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 644215.*fcs_pow(v,8)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 154.*v2*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3927.*fcs_pow(v,4)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 4.519515e6*fcs_pow(u,4)*(1.6896453e7*fcs_pow(v,12)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 33.*(39. - 3666.*w2 + 89817.*fcs_pow(w,4) - 872508.*fcs_pow(w,6) + 3.853577e6*fcs_pow(w,8) - 7.707154e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 3102.*v2*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 6.055621e6*fcs_pow(v,8)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) - 3.303066e6*fcs_pow(v,10)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 10857.*fcs_pow(v,4)*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 105468.*fcs_pow(v,6)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 903903.*u2*(143. - 12870.*w2 + 302445.*fcs_pow(w,4) - 2.82282e6*fcs_pow(w,6) + 1.1996985e7*fcs_pow(w,8) - 2.3121462e7*fcs_pow(w,10) + 1.6303595e7*fcs_pow(w,12) - 330.*v2*(39. - 3666.*w2 + 89817.*fcs_pow(w,4) - 872508.*fcs_pow(w,6) + 3.853577e6*fcs_pow(w,8) - 7.707154e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 7755.*fcs_pow(v,4)*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 1.101022e6*fcs_pow(v,10)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 211735.*fcs_pow(v,12)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 10340.*fcs_pow(v,6)*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 43945.*fcs_pow(v,8)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 143.*(148995.*fcs_pow(v,4)*(39. - 3666.*w2 + 89817.*fcs_pow(w,4) - 872508.*fcs_pow(w,6) + 3.853577e6*fcs_pow(w,8) - 7.707154e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 1.33386e6*fcs_pow(v,6)*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 1.2746447e7*fcs_pow(v,12)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) - 1806.*v2*(143. - 12870.*w2 + 302445.*fcs_pow(w,4) - 2.82282e6*fcs_pow(w,6) + 1.1996985e7*fcs_pow(w,8) - 2.3121462e7*fcs_pow(w,10) + 1.6303595e7*fcs_pow(w,12)) + 778085.*fcs_pow(v,8)*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 1.442994e6*fcs_pow(v,10)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 3.*(1001. - 86086.*w2 + 1.936935e6*fcs_pow(w,4) - 1.734018e7*fcs_pow(w,6) + 7.0805735e7*fcs_pow(w,8) - 1.31312454e8*fcs_pow(w,10) + 8.9225129e7*fcs_pow(w,12))) + 7.22218497e8*fcs_pow(u,10)*(-8.64006e6*fcs_pow(v,6)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 4.2480295e7*fcs_pow(v,8)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) - 15158.*v2*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) - 9.4229018e7*fcs_pow(v,10)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) + 113685.*fcs_pow(v,4)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 11.*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 3.624193e6*fcs_pow(v,12)*(231. - 30030.*w2 + 1.006005e6*fcs_pow(w,4) - 1.322178e7*fcs_pow(w,6) + 7.8228865e7*fcs_pow(w,8) - 2.07662078e8*fcs_pow(w,10) + 1.99675075e8*fcs_pow(w,12))) + 6.3732235e7*fcs_pow(u,14)*(-342342.*v2*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 1.0099089e7*fcs_pow(v,4)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) - 1.17341796e8*fcs_pow(v,6)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) + 429.*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 2.9335449e7*fcs_pow(v,8)*(231. - 30030.*w2 + 1.006005e6*fcs_pow(w,4) - 1.322178e7*fcs_pow(w,6) + 7.8228865e7*fcs_pow(w,8) - 2.07662078e8*fcs_pow(w,10) + 1.99675075e8*fcs_pow(w,12)) + 5.0396797e7*fcs_pow(v,12)*(273. - 37674.*w2 + 1.337427e6*fcs_pow(w,4) - 1.8596604e7*fcs_pow(w,6) + 1.16228775e8*fcs_pow(w,8) - 3.2544057e8*fcs_pow(w,10) + 3.29612885e8*fcs_pow(w,12)) - 5.333718e6*fcs_pow(v,10)*(3003. - 402402.*w2 + 1.3882869e7*fcs_pow(w,4) - 1.87749276e8*fcs_pow(w,6) + 1.142141429e9*fcs_pow(w,8) - 3.11493117e9*fcs_pow(w,10) + 3.074996155e9*fcs_pow(w,12))) - 1.159926677e9*fcs_pow(u,12)*(855855.*fcs_pow(v,4)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) - 9.61818e6*fcs_pow(v,6)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) + 143.*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 4.8892415e7*fcs_pow(v,8)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) - 4290.*v2*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 5.333718e6*fcs_pow(v,10)*(231. - 30030.*w2 + 1.006005e6*fcs_pow(w,4) - 1.322178e7*fcs_pow(w,6) + 7.8228865e7*fcs_pow(w,8) - 2.07662078e8*fcs_pow(w,10) + 1.99675075e8*fcs_pow(w,12)) + 341905.*fcs_pow(v,12)*(3003. - 402402.*w2 + 1.3882869e7*fcs_pow(w,4) - 1.87749276e8*fcs_pow(w,6) + 1.142141429e9*fcs_pow(w,8) - 3.11493117e9*fcs_pow(w,10) + 3.074996155e9*fcs_pow(w,12))));
      }
      case 14 : switch(k){
        case 0 : return 9.086380738369043e21*fcs_pow(t,29)*(-143.*(-3. + 609.*v2 - 18879.*fcs_pow(v,4) + 207669.*fcs_pow(v,6) - 1.038345e6*fcs_pow(v,8) + 2.561251e6*fcs_pow(v,10) - 3.026933e6*fcs_pow(v,12) + 1.363783e6*fcs_pow(v,14)) + 29029.*u2*(-3. + 651.*v2 - 21483.*fcs_pow(v,4) + 250635.*fcs_pow(v,6) - 1.324785e6*fcs_pow(v,8) + 3.444441e6*fcs_pow(v,10) - 4.279457e6*fcs_pow(v,12) + 2.022161e6*fcs_pow(v,14)) - 899899.*fcs_pow(u,4)*(-3. + 693.*v2 - 24255.*fcs_pow(v,4) + 299145.*fcs_pow(v,6) - 1.666665e6*fcs_pow(v,8) + 4.555551e6*fcs_pow(v,10) - 5.936021e6*fcs_pow(v,12) + 2.935395e6*fcs_pow(v,14)) - 1.3498485e7*fcs_pow(u,8)*(-11. + 2849.*v2 - 111111.*fcs_pow(v,4) + 1.518517e6*fcs_pow(v,6) - 9.328033e6*fcs_pow(v,8) + 2.7984099e7*fcs_pow(v,10) - 3.9856141e7*fcs_pow(v,12) + 2.1460999e7*fcs_pow(v,14)) + 3.3296263e7*fcs_pow(u,10)*(-11. + 3003.*v2 - 123123.*fcs_pow(v,4) + 1.764763e6*fcs_pow(v,6) - 1.1344905e7*fcs_pow(v,8) + 3.5547369e7*fcs_pow(v,10) - 5.2782457e7*fcs_pow(v,12) + 2.9581377e7*fcs_pow(v,14)) + 899899.*fcs_pow(u,6)*(-33. + 8085.*v2 - 299145.*fcs_pow(v,4) + 3.888885e6*fcs_pow(v,6) - 2.2777755e7*fcs_pow(v,8) + 6.5296231e7*fcs_pow(v,10) - 8.9040315e7*fcs_pow(v,12) + 4.5987855e7*fcs_pow(v,14)) - 3.026933e6*fcs_pow(u,12)*(-143. + 41041.*v2 - 1.764763e6*fcs_pow(v,4) + 2.6471445e7*fcs_pow(v,6) - 1.77736845e8*fcs_pow(v,8) + 5.80607027e8*fcs_pow(v,10) - 8.97301769e8*fcs_pow(v,12) + 5.22604327e8*fcs_pow(v,14)) + 1.363783e6*fcs_pow(u,14)*(-143. + 43043.*v2 - 1.936935e6*fcs_pow(v,4) + 3.0345315e7*fcs_pow(v,6) - 2.12417205e8*fcs_pow(v,8) + 7.22218497e8*fcs_pow(v,10) - 1.159926677e9*fcs_pow(v,12) + 7.01054585e8*fcs_pow(v,14)));
        case 1 : return -2.635050414127023e23*fcs_pow(t,30)*(-143.*(-3. + 651.*v2 - 21483.*fcs_pow(v,4) + 250635.*fcs_pow(v,6) - 1.324785e6*fcs_pow(v,8) + 3.444441e6*fcs_pow(v,10) - 4.279457e6*fcs_pow(v,12) + 2.022161e6*fcs_pow(v,14)) + 31031.*u2*(-3. + 693.*v2 - 24255.*fcs_pow(v,4) + 299145.*fcs_pow(v,6) - 1.666665e6*fcs_pow(v,8) + 4.555551e6*fcs_pow(v,10) - 5.936021e6*fcs_pow(v,12) + 2.935395e6*fcs_pow(v,14)) + 3.258255e6*fcs_pow(u,6)*(-11. + 2849.*v2 - 111111.*fcs_pow(v,4) + 1.518517e6*fcs_pow(v,6) - 9.328033e6*fcs_pow(v,8) + 2.7984099e7*fcs_pow(v,10) - 3.9856141e7*fcs_pow(v,12) + 2.1460999e7*fcs_pow(v,14)) - 1.7222205e7*fcs_pow(u,8)*(-11. + 3003.*v2 - 123123.*fcs_pow(v,4) + 1.764763e6*fcs_pow(v,6) - 1.1344905e7*fcs_pow(v,8) + 3.5547369e7*fcs_pow(v,10) - 5.2782457e7*fcs_pow(v,12) + 2.9581377e7*fcs_pow(v,14)) - 93093.*fcs_pow(u,4)*(-33. + 8085.*v2 - 299145.*fcs_pow(v,4) + 3.888885e6*fcs_pow(v,6) - 2.2777755e7*fcs_pow(v,8) + 6.5296231e7*fcs_pow(v,10) - 8.9040315e7*fcs_pow(v,12) + 4.5987855e7*fcs_pow(v,14)) + 2.2243771e7*fcs_pow(u,14)*(-13. + 4095.*v2 - 192465.*fcs_pow(v,4) + 3.143595e6*fcs_pow(v,6) - 2.2903335e7*fcs_pow(v,8) + 8.0925117e7*fcs_pow(v,10) - 1.34875195e8*fcs_pow(v,12) + 8.4482265e7*fcs_pow(v,14)) + 3.444441e6*fcs_pow(u,10)*(-143. + 41041.*v2 - 1.764763e6*fcs_pow(v,4) + 2.6471445e7*fcs_pow(v,6) - 1.77736845e8*fcs_pow(v,8) + 5.80607027e8*fcs_pow(v,10) - 8.97301769e8*fcs_pow(v,12) + 5.22604327e8*fcs_pow(v,14)) - 4.279457e6*fcs_pow(u,12)*(-143. + 43043.*v2 - 1.936935e6*fcs_pow(v,4) + 3.0345315e7*fcs_pow(v,6) - 2.12417205e8*fcs_pow(v,8) + 7.22218497e8*fcs_pow(v,10) - 1.159926677e9*fcs_pow(v,12) + 7.01054585e8*fcs_pow(v,14)))*w;
        case 2 : return 2.635050414127023e23*fcs_pow(t,31)*(-143.*(3. - 93.*w2 + 651.*v2*(-1. + 33.*w2) - 21483.*fcs_pow(v,4)*(-1. + 35.*w2) + 250635.*fcs_pow(v,6)*(-1. + 37.*w2) - 1.324785e6*fcs_pow(v,8)*(-1. + 39.*w2) + 3.444441e6*fcs_pow(v,10)*(-1. + 41.*w2) - 4.279457e6*fcs_pow(v,12)*(-1. + 43.*w2) + 2.022161e6*fcs_pow(v,14)*(-1. + 45.*w2)) + 31031.*u2*(3. - 99.*w2 + 693.*v2*(-1. + 35.*w2) - 24255.*fcs_pow(v,4)*(-1. + 37.*w2) + 299145.*fcs_pow(v,6)*(-1. + 39.*w2) - 1.666665e6*fcs_pow(v,8)*(-1. + 41.*w2) + 4.555551e6*fcs_pow(v,10)*(-1. + 43.*w2) - 5.936021e6*fcs_pow(v,12)*(-1. + 45.*w2) + 2.935395e6*fcs_pow(v,14)*(-1. + 47.*w2)) - 93093.*fcs_pow(u,4)*(33. - 1155.*w2 + 8085.*v2*(-1. + 37.*w2) - 299145.*fcs_pow(v,4)*(-1. + 39.*w2) + 3.888885e6*fcs_pow(v,6)*(-1. + 41.*w2) - 2.2777755e7*fcs_pow(v,8)*(-1. + 43.*w2) + 6.5296231e7*fcs_pow(v,10)*(-1. + 45.*w2) - 8.9040315e7*fcs_pow(v,12)*(-1. + 47.*w2) + 4.5987855e7*fcs_pow(v,14)*(-1. + 49.*w2)) + 3.258255e6*fcs_pow(u,6)*(11. - 407.*w2 + 2849.*v2*(-1. + 39.*w2) - 111111.*fcs_pow(v,4)*(-1. + 41.*w2) + 1.518517e6*fcs_pow(v,6)*(-1. + 43.*w2) - 9.328033e6*fcs_pow(v,8)*(-1. + 45.*w2) + 2.7984099e7*fcs_pow(v,10)*(-1. + 47.*w2) - 3.9856141e7*fcs_pow(v,12)*(-1. + 49.*w2) + 2.1460999e7*fcs_pow(v,14)*(-1. + 51.*w2)) - 1.7222205e7*fcs_pow(u,8)*(11. - 429.*w2 + 3003.*v2*(-1. + 41.*w2) - 123123.*fcs_pow(v,4)*(-1. + 43.*w2) + 1.764763e6*fcs_pow(v,6)*(-1. + 45.*w2) - 1.1344905e7*fcs_pow(v,8)*(-1. + 47.*w2) + 3.5547369e7*fcs_pow(v,10)*(-1. + 49.*w2) - 5.2782457e7*fcs_pow(v,12)*(-1. + 51.*w2) + 2.9581377e7*fcs_pow(v,14)*(-1. + 53.*w2)) + 3.444441e6*fcs_pow(u,10)*(143. - 5863.*w2 + 41041.*v2*(-1. + 43.*w2) - 1.764763e6*fcs_pow(v,4)*(-1. + 45.*w2) + 2.6471445e7*fcs_pow(v,6)*(-1. + 47.*w2) - 1.77736845e8*fcs_pow(v,8)*(-1. + 49.*w2) + 5.80607027e8*fcs_pow(v,10)*(-1. + 51.*w2) - 8.97301769e8*fcs_pow(v,12)*(-1. + 53.*w2) + 5.22604327e8*fcs_pow(v,14)*(-1. + 55.*w2)) - 4.279457e6*fcs_pow(u,12)*(143. - 6149.*w2 + 43043.*v2*(-1. + 45.*w2) - 1.936935e6*fcs_pow(v,4)*(-1. + 47.*w2) + 3.0345315e7*fcs_pow(v,6)*(-1. + 49.*w2) - 2.12417205e8*fcs_pow(v,8)*(-1. + 51.*w2) + 7.22218497e8*fcs_pow(v,10)*(-1. + 53.*w2) - 1.159926677e9*fcs_pow(v,12)*(-1. + 55.*w2) + 7.01054585e8*fcs_pow(v,14)*(-1. + 57.*w2)) + 2.2243771e7*fcs_pow(u,14)*(13. - 585.*w2 + 4095.*v2*(-1. + 47.*w2) - 192465.*fcs_pow(v,4)*(-1. + 49.*w2) + 3.143595e6*fcs_pow(v,6)*(-1. + 51.*w2) - 2.2903335e7*fcs_pow(v,8)*(-1. + 53.*w2) + 8.0925117e7*fcs_pow(v,10)*(-1. + 55.*w2) - 1.34875195e8*fcs_pow(v,12)*(-1. + 57.*w2) + 8.4482265e7*fcs_pow(v,14)*(-1. + 59.*w2)));
        case 3 : return -2.695656573651944e26*fcs_pow(t,32)*w*(-13.*(3. - 33.*w2 + 299145.*fcs_pow(v,6)*(-1. + 13.*w2) - 5.936021e6*fcs_pow(v,12)*(-1. + 15.*w2) + 231.*v2*(-3. + 35.*w2) - 8085.*fcs_pow(v,4)*(-3. + 37.*w2) - 555555.*fcs_pow(v,8)*(-3. + 41.*w2) + 1.518517e6*fcs_pow(v,10)*(-3. + 43.*w2) + 978465.*fcs_pow(v,14)*(-3. + 47.*w2)) - 9555.*fcs_pow(u,4)*(33. - 407.*w2 + 8547.*v2*(-1. + 13.*w2) - 2.7984099e7*fcs_pow(v,8)*(-1. + 15.*w2) + 6.4382997e7*fcs_pow(v,14)*(-1. + 17.*w2) - 111111.*fcs_pow(v,4)*(-3. + 41.*w2) + 1.518517e6*fcs_pow(v,6)*(-3. + 43.*w2) + 2.7984099e7*fcs_pow(v,10)*(-3. + 47.*w2) - 3.9856141e7*fcs_pow(v,12)*(-3. + 49.*w2)) + 273.*u2*(33. - 385.*w2 - 299145.*fcs_pow(v,4)*(-1. + 13.*w2) + 6.5296231e7*fcs_pow(v,10)*(-1. + 15.*w2) + 2695.*v2*(-3. + 37.*w2) + 1.296295e6*fcs_pow(v,6)*(-3. + 41.*w2) - 7.592585e6*fcs_pow(v,8)*(-3. + 43.*w2) - 2.9680105e7*fcs_pow(v,12)*(-3. + 47.*w2) + 1.5329285e7*fcs_pow(v,14)*(-3. + 49.*w2)) + 353535.*fcs_pow(u,6)*(11. - 143.*w2 + 1.764763e6*fcs_pow(v,6)*(-1. + 15.*w2) - 5.2782457e7*fcs_pow(v,12)*(-1. + 17.*w2) + 1001.*v2*(-3. + 41.*w2) - 41041.*fcs_pow(v,4)*(-3. + 43.*w2) - 3.781635e6*fcs_pow(v,8)*(-3. + 47.*w2) + 1.1849123e7*fcs_pow(v,10)*(-3. + 49.*w2) + 9.860459e6*fcs_pow(v,14)*(-3. + 53.*w2)) + 138047.*fcs_pow(u,10)*(429. - 6149.*w2 + 129129.*v2*(-1. + 15.*w2) - 6.37251615e8*fcs_pow(v,8)*(-1. + 17.*w2) + 2.103163755e9*fcs_pow(v,14)*(-1. + 19.*w2) - 1.936935e6*fcs_pow(v,4)*(-3. + 47.*w2) + 3.0345315e7*fcs_pow(v,6)*(-3. + 49.*w2) + 7.22218497e8*fcs_pow(v,10)*(-3. + 53.*w2) - 1.159926677e9*fcs_pow(v,12)*(-3. + 55.*w2)) - 50505.*fcs_pow(u,8)*(429. - 5863.*w2 - 5.294289e6*fcs_pow(v,4)*(-1. + 15.*w2) + 1.741821081e9*fcs_pow(v,10)*(-1. + 17.*w2) + 41041.*v2*(-3. + 43.*w2) + 2.6471445e7*fcs_pow(v,6)*(-3. + 47.*w2) - 1.77736845e8*fcs_pow(v,8)*(-3. + 49.*w2) - 8.97301769e8*fcs_pow(v,12)*(-3. + 53.*w2) + 5.22604327e8*fcs_pow(v,14)*(-3. + 55.*w2)) - 5.936021e6*fcs_pow(u,12)*(13. - 195.*w2 + 3.143595e6*fcs_pow(v,6)*(-1. + 17.*w2) - 1.34875195e8*fcs_pow(v,12)*(-1. + 19.*w2) + 1365.*v2*(-3. + 47.*w2) - 64155.*fcs_pow(v,4)*(-3. + 49.*w2) - 7.634445e6*fcs_pow(v,8)*(-3. + 53.*w2) + 2.6975039e7*fcs_pow(v,10)*(-3. + 55.*w2) + 2.8160755e7*fcs_pow(v,14)*(-3. + 59.*w2)) + 326155.*fcs_pow(u,14)*(39.*(3. - 47.*w2) - 1.886157e6*fcs_pow(v,4)*(-1. + 17.*w2) + 8.90176287e8*fcs_pow(v,10)*(-1. + 19.*w2) + 12831.*v2*(-3. + 49.*w2) + 1.0688223e7*fcs_pow(v,6)*(-3. + 53.*w2) - 8.0925117e7*fcs_pow(v,8)*(-3. + 55.*w2) - 5.12525741e8*fcs_pow(v,12)*(-3. + 59.*w2) + 3.32296909e8*fcs_pow(v,14)*(-3. + 61.*w2)));
        case 4 : return 2.695656573651944e26*fcs_pow(t,33)*(-28665.*fcs_pow(u,4)*(-11.*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 2849.*v2*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 1.518517e6*fcs_pow(v,6)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 9.328033e6*fcs_pow(v,8)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 3.9856141e7*fcs_pow(v,12)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 2.1460999e7*fcs_pow(v,14)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 37037.*fcs_pow(v,4)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 9.328033e6*fcs_pow(v,10)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 273.*u2*(8085.*v2*(1. - 74.*w2 + 481.*fcs_pow(w,4)) - 299145.*fcs_pow(v,4)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 2.2777755e7*fcs_pow(v,8)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 6.5296231e7*fcs_pow(v,10)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 4.5987855e7*fcs_pow(v,14)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 11.*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) + 1.296295e6*fcs_pow(v,6)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 2.9680105e7*fcs_pow(v,12)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) - 13.*(-3.*(1. - 66.*w2 + 385.*fcs_pow(w,4)) - 24255.*fcs_pow(v,4)*(1. - 74.*w2 + 481.*fcs_pow(w,4)) + 299145.*fcs_pow(v,6)*(1. - 78.*w2 + 533.*fcs_pow(w,4)) + 4.555551e6*fcs_pow(v,10)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 5.936021e6*fcs_pow(v,12)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 231.*v2*(3. - 210.*w2 + 1295.*fcs_pow(w,4)) - 555555.*fcs_pow(v,8)*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 978465.*fcs_pow(v,14)*(3. - 282.*w2 + 2303.*fcs_pow(w,4))) + 414141.*fcs_pow(u,10)*(-143.*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 43043.*v2*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 3.0345315e7*fcs_pow(v,6)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 2.12417205e8*fcs_pow(v,8)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 1.159926677e9*fcs_pow(v,12)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 7.01054585e8*fcs_pow(v,14)*(1. - 114.*w2 + 1121.*fcs_pow(w,4)) - 645645.*fcs_pow(v,4)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 2.40739499e8*fcs_pow(v,10)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) - 50505.*fcs_pow(u,8)*(123123.*v2*(1. - 86.*w2 + 645.*fcs_pow(w,4)) - 5.294289e6*fcs_pow(v,4)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 5.33210535e8*fcs_pow(v,8)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 1.741821081e9*fcs_pow(v,10)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 1.567812981e9*fcs_pow(v,14)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) - 143.*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) + 2.6471445e7*fcs_pow(v,6)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) - 8.97301769e8*fcs_pow(v,12)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) + 353535.*fcs_pow(u,6)*(-11.*(1. - 78.*w2 + 533.*fcs_pow(w,4)) - 123123.*fcs_pow(v,4)*(1. - 86.*w2 + 645.*fcs_pow(w,4)) + 1.764763e6*fcs_pow(v,6)*(1. - 90.*w2 + 705.*fcs_pow(w,4)) + 3.5547369e7*fcs_pow(v,10)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 5.2782457e7*fcs_pow(v,12)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 1001.*v2*(3. - 246.*w2 + 1763.*fcs_pow(w,4)) - 3.781635e6*fcs_pow(v,8)*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 9.860459e6*fcs_pow(v,14)*(3. - 318.*w2 + 2915.*fcs_pow(w,4))) + 326155.*fcs_pow(u,14)*(38493.*v2*(1. - 98.*w2 + 833.*fcs_pow(w,4)) - 1.886157e6*fcs_pow(v,4)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) - 2.42775351e8*fcs_pow(v,8)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) + 8.90176287e8*fcs_pow(v,10)*(1. - 114.*w2 + 1121.*fcs_pow(w,4)) + 9.96890727e8*fcs_pow(v,14)*(1. - 122.*w2 + 1281.*fcs_pow(w,4)) - 39.*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) + 1.0688223e7*fcs_pow(v,6)*(3. - 318.*w2 + 2915.*fcs_pow(w,4)) - 5.12525741e8*fcs_pow(v,12)*(3. - 354.*w2 + 3599.*fcs_pow(w,4))) - 5.936021e6*fcs_pow(u,12)*(-13.*(1. - 90.*w2 + 705.*fcs_pow(w,4)) - 192465.*fcs_pow(v,4)*(1. - 98.*w2 + 833.*fcs_pow(w,4)) + 3.143595e6*fcs_pow(v,6)*(1. - 102.*w2 + 901.*fcs_pow(w,4)) + 8.0925117e7*fcs_pow(v,10)*(1. - 110.*w2 + 1045.*fcs_pow(w,4)) - 1.34875195e8*fcs_pow(v,12)*(1. - 114.*w2 + 1121.*fcs_pow(w,4)) + 1365.*v2*(3. - 282.*w2 + 2303.*fcs_pow(w,4)) - 7.634445e6*fcs_pow(v,8)*(3. - 318.*w2 + 2915.*fcs_pow(w,4)) + 2.8160755e7*fcs_pow(v,14)*(3. - 354.*w2 + 3599.*fcs_pow(w,4))));
        case 5 : return -1.347828286825972e27*fcs_pow(t,34)*w*(-2.070705e6*fcs_pow(u,8)*(129129.*v2*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 143.*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 7.22218497e8*fcs_pow(v,10)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 1.159926677e9*fcs_pow(v,12)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) - 1.27450323e8*fcs_pow(v,8)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 4.20632751e8*fcs_pow(v,14)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) - 387387.*fcs_pow(v,4)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 6.069063e6*fcs_pow(v,6)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 70707.*fcs_pow(u,6)*(-2.6471445e7*fcs_pow(v,4)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 205205.*v2*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 4.486508845e9*fcs_pow(v,12)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 2.613021635e9*fcs_pow(v,14)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 1.741821081e9*fcs_pow(v,10)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 143.*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 2.6471445e7*fcs_pow(v,6)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 1.77736845e8*fcs_pow(v,8)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 212121.*fcs_pow(u,4)*(8.823815e6*fcs_pow(v,6)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 205205.*fcs_pow(v,4)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 11.*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 4.9302295e7*fcs_pow(v,14)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 5.2782457e7*fcs_pow(v,12)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 1001.*v2*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 3.781635e6*fcs_pow(v,8)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 1.1849123e7*fcs_pow(v,10)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 1911.*u2*(-1.39920495e8*fcs_pow(v,8)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) + 7.592585e6*fcs_pow(v,6)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) + 8547.*v2*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 6.4382997e7*fcs_pow(v,14)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 11.*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) - 111111.*fcs_pow(v,4)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) + 2.7984099e7*fcs_pow(v,10)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 3.9856141e7*fcs_pow(v,12)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) - 39.*(6.5296231e7*fcs_pow(v,10)*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 11.*(3. - 70.*w2 + 259.*fcs_pow(w,4)) - 7.592585e6*fcs_pow(v,8)*(3. - 86.*w2 + 387.*fcs_pow(w,4)) - 59829.*fcs_pow(v,4)*(5. - 130.*w2 + 533.*fcs_pow(w,4)) + 539.*v2*(15. - 370.*w2 + 1443.*fcs_pow(w,4)) + 259259.*fcs_pow(v,6)*(15. - 410.*w2 + 1763.*fcs_pow(w,4)) - 5.936021e6*fcs_pow(v,12)*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 3.065857e6*fcs_pow(v,14)*(15. - 490.*w2 + 2499.*fcs_pow(w,4))) + 1.95888693e8*fcs_pow(u,10)*(-13.*(1. - 30.*w2 + 141.*fcs_pow(w,4)) - 7.634445e6*fcs_pow(v,8)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 2.6975039e7*fcs_pow(v,10)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 628719.*fcs_pow(v,6)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 2.6975039e7*fcs_pow(v,12)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) + 273.*v2*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) - 12831.*fcs_pow(v,4)*(15. - 490.*w2 + 2499.*fcs_pow(w,4)) + 5.632151e6*fcs_pow(v,14)*(15. - 590.*w2 + 3599.*fcs_pow(w,4))) + 3.065857e6*fcs_pow(u,14)*(6.469184505e9*fcs_pow(v,14)*(1. - 42.*w2 + 273.*fcs_pow(w,4)) - 3.411135e6*fcs_pow(v,4)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) + 6.0263385e7*fcs_pow(v,6)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) + 40131.*v2*(5. - 170.*w2 + 901.*fcs_pow(w,4)) - 2.84098815e8*fcs_pow(v,8)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) - 39.*(15. - 490.*w2 + 2499.*fcs_pow(w,4)) + 3.59858499e8*fcs_pow(v,10)*(15. - 590.*w2 + 3599.*fcs_pow(w,4)) - 6.43383377e8*fcs_pow(v,12)*(15. - 610.*w2 + 3843.*fcs_pow(w,4))) - 5.936021e6*fcs_pow(u,12)*(5.3441115e7*fcs_pow(v,6)*(3. - 106.*w2 + 583.*fcs_pow(w,4)) - 4.04625585e8*fcs_pow(v,8)*(3. - 110.*w2 + 627.*fcs_pow(w,4)) - 1.886157e6*fcs_pow(v,4)*(5. - 170.*w2 + 901.*fcs_pow(w,4)) + 8.90176287e8*fcs_pow(v,10)*(5. - 190.*w2 + 1121.*fcs_pow(w,4)) - 39.*(15. - 470.*w2 + 2303.*fcs_pow(w,4)) + 12831.*v2*(15. - 490.*w2 + 2499.*fcs_pow(w,4)) - 5.12525741e8*fcs_pow(v,12)*(15. - 590.*w2 + 3599.*fcs_pow(w,4)) + 3.32296909e8*fcs_pow(v,14)*(15. - 610.*w2 + 3843.*fcs_pow(w,4))));
        case 6 : return 4.043484860477916e27*fcs_pow(t,35)*(353535.*fcs_pow(u,6)*(-143.*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 41041.*v2*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.764763e6*fcs_pow(v,4)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 5.80607027e8*fcs_pow(v,10)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 8.97301769e8*fcs_pow(v,12)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 5.22604327e8*fcs_pow(v,14)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 5.294289e6*fcs_pow(v,6)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 3.5547369e7*fcs_pow(v,8)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 70707.*fcs_pow(u,4)*(11.*(5. - 585.*w2 + 7995.*fcs_pow(w,4) - 22919.*fcs_pow(w,6)) + 15015.*v2*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 615615.*fcs_pow(v,4)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 8.823815e6*fcs_pow(v,6)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 2.63912285e8*fcs_pow(v,12)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 1.47906885e8*fcs_pow(v,14)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.1344905e7*fcs_pow(v,8)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 3.5547369e7*fcs_pow(v,10)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) + 1911.*u2*(11.*(5. - 555.*w2 + 7215.*fcs_pow(w,4) - 19721.*fcs_pow(w,6)) - 555555.*fcs_pow(v,4)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) + 7.592585e6*fcs_pow(v,6)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 4.6640165e7*fcs_pow(v,8)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 1.07304995e8*fcs_pow(v,14)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 2849.*v2*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) + 2.7984099e7*fcs_pow(v,10)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 3.9856141e7*fcs_pow(v,12)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 13.*(-33.*(-1. + 105.*w2 - 1295.*fcs_pow(w,4) + 3367.*fcs_pow(w,6)) + 3.888885e6*fcs_pow(v,6)*(-1. + 123.*w2 - 1763.*fcs_pow(w,4) + 5289.*fcs_pow(w,6)) - 2.2777755e7*fcs_pow(v,8)*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 6.5296231e7*fcs_pow(v,10)*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) + 1617.*v2*(-5. + 555.*w2 - 7215.*fcs_pow(w,4) + 19721.*fcs_pow(w,6)) - 59829.*fcs_pow(v,4)*(-5. + 585.*w2 - 7995.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 1.7808063e7*fcs_pow(v,12)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 9.197571e6*fcs_pow(v,14)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6))) - 2.070705e6*fcs_pow(u,8)*(-143.*(-1. + 129.*w2 - 1935.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 43043.*v2*(-1. + 135.*w2 - 2115.*fcs_pow(w,4) + 6909.*fcs_pow(w,6)) - 2.12417205e8*fcs_pow(v,8)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 7.22218497e8*fcs_pow(v,10)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.159926677e9*fcs_pow(v,12)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 387387.*fcs_pow(v,4)*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 6.069063e6*fcs_pow(v,6)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) + 1.40210917e8*fcs_pow(v,14)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) + 3.065857e6*fcs_pow(u,14)*(66885.*v2*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 3.411135e6*fcs_pow(v,4)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 6.0263385e7*fcs_pow(v,6)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 3.216916885e9*fcs_pow(v,12)*(-1. + 183.*w2 - 3843.*fcs_pow(w,4) + 16653.*fcs_pow(w,6)) + 2.156394835e9*fcs_pow(v,14)*(-1. + 189.*w2 - 4095.*fcs_pow(w,4) + 18291.*fcs_pow(w,6)) - 39.*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) - 9.4699605e7*fcs_pow(v,8)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6)) + 3.59858499e8*fcs_pow(v,10)*(-5. + 885.*w2 - 17995.*fcs_pow(w,4) + 75579.*fcs_pow(w,6))) - 5.936021e6*fcs_pow(u,12)*(-3.143595e6*fcs_pow(v,4)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) + 5.3441115e7*fcs_pow(v,6)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 4.04625585e8*fcs_pow(v,8)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 1.661484545e9*fcs_pow(v,14)*(-1. + 183.*w2 - 3843.*fcs_pow(w,4) + 16653.*fcs_pow(w,6)) - 39.*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) + 12831.*v2*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) + 2.96725429e8*fcs_pow(v,10)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6)) - 5.12525741e8*fcs_pow(v,12)*(-5. + 885.*w2 - 17995.*fcs_pow(w,4) + 75579.*fcs_pow(w,6))) + 6.5296231e7*fcs_pow(u,10)*(13.*(1. - 135.*w2 + 2115.*fcs_pow(w,4) - 6909.*fcs_pow(w,6)) + 3.143595e6*fcs_pow(v,6)*(-1. + 153.*w2 - 2703.*fcs_pow(w,4) + 9911.*fcs_pow(w,6)) - 2.2903335e7*fcs_pow(v,8)*(-1. + 159.*w2 - 2915.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 8.0925117e7*fcs_pow(v,10)*(-1. + 165.*w2 - 3135.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 819.*v2*(-5. + 705.*w2 - 11515.*fcs_pow(w,4) + 39151.*fcs_pow(w,6)) - 38493.*fcs_pow(v,4)*(-5. + 735.*w2 - 12495.*fcs_pow(w,4) + 44149.*fcs_pow(w,6)) - 2.6975039e7*fcs_pow(v,12)*(-5. + 855.*w2 - 16815.*fcs_pow(w,4) + 68381.*fcs_pow(w,6)) + 1.6896453e7*fcs_pow(v,14)*(-5. + 885.*w2 - 17995.*fcs_pow(w,4) + 75579.*fcs_pow(w,6))));
        case 7 : return -2.830439402334541e28*fcs_pow(t,36)*w*(-30303.*fcs_pow(u,4)*(-6.1766705e7*fcs_pow(v,4)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.85300115e8*fcs_pow(v,6)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 205205.*v2*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 1.244157915e9*fcs_pow(v,8)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 4.486508845e9*fcs_pow(v,12)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 2.613021635e9*fcs_pow(v,14)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 143.*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 5.80607027e8*fcs_pow(v,10)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 10101.*u2*(11.*(35. - 1365.*w2 + 11193.*fcs_pow(w,4) - 22919.*fcs_pow(w,6)) + 6.1766705e7*fcs_pow(v,6)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 7.9414335e7*fcs_pow(v,8)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 615615.*fcs_pow(v,4)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 2.48831583e8*fcs_pow(v,10)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 1.47906885e8*fcs_pow(v,14)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 3003.*v2*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) - 5.2782457e7*fcs_pow(v,12)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) - 39.*(11.*(35. - 1295.*w2 + 10101.*fcs_pow(w,4) - 19721.*fcs_pow(w,6)) - 3.26481155e8*fcs_pow(v,8)*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 1.95888693e8*fcs_pow(v,10)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 7.592585e6*fcs_pow(v,6)*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) - 2.78992987e8*fcs_pow(v,12)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 2849.*v2*(-35. + 1365.*w2 - 11193.*fcs_pow(w,4) + 22919.*fcs_pow(w,6)) - 111111.*fcs_pow(v,4)*(-35. + 1435.*w2 - 12341.*fcs_pow(w,4) + 26445.*fcs_pow(w,6)) + 2.1460999e7*fcs_pow(v,14)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6))) + 2.1460999e7*fcs_pow(u,14)*(-2.8033132855e10*fcs_pow(v,12)*(-1. + 63.*w2 - 819.*fcs_pow(w,4) + 2613.*fcs_pow(w,6)) - 7.71125355e8*fcs_pow(v,8)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) + 69615.*v2*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 3.033093063e9*fcs_pow(v,10)*(-5. + 305.*w2 - 3843.*fcs_pow(w,4) + 11895.*fcs_pow(w,6)) - 3.689595e6*fcs_pow(v,4)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 2.772507645e9*fcs_pow(v,14)*(-7. + 455.*w2 - 6097.*fcs_pow(w,4) + 20033.*fcs_pow(w,6)) - 39.*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 1.3528515e7*fcs_pow(v,6)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) - 2.78992987e8*fcs_pow(u,12)*(2.156394835e9*fcs_pow(v,14)*(-1. + 63.*w2 - 819.*fcs_pow(w,4) + 2613.*fcs_pow(w,6)) - 39.*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 3.59858499e8*fcs_pow(v,10)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) - 487305.*fcs_pow(v,4)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 6.43383377e8*fcs_pow(v,12)*(-5. + 305.*w2 - 3843.*fcs_pow(w,4) + 11895.*fcs_pow(w,6)) + 8.609055e6*fcs_pow(v,6)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 1911.*v2*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) - 1.3528515e7*fcs_pow(v,8)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) + 1.95888693e8*fcs_pow(u,10)*(-39.*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) + 12831.*v2*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) - 5.12525741e8*fcs_pow(v,12)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) + 7.634445e6*fcs_pow(v,6)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 3.32296909e8*fcs_pow(v,14)*(-5. + 305.*w2 - 3843.*fcs_pow(w,4) + 11895.*fcs_pow(w,6)) - 5.7803655e7*fcs_pow(v,8)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 89817.*fcs_pow(v,4)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 4.2389347e7*fcs_pow(v,10)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) - 9.79443465e8*fcs_pow(u,8)*(-13.*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) + 819.*v2*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 38493.*fcs_pow(v,4)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 1.6896453e7*fcs_pow(v,14)*(-5. + 295.*w2 - 3599.*fcs_pow(w,4) + 10797.*fcs_pow(w,6)) - 3.271905e6*fcs_pow(v,8)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) + 1.1560731e7*fcs_pow(v,10)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) + 89817.*fcs_pow(v,6)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) - 3.853577e6*fcs_pow(v,12)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))) + 2.070705e6*fcs_pow(u,6)*(301301.*v2*(-1. + 45.*w2 - 423.*fcs_pow(w,4) + 987.*fcs_pow(w,6)) - 2.711709e6*fcs_pow(v,4)*(-5. + 235.*w2 - 2303.*fcs_pow(w,4) + 5593.*fcs_pow(w,6)) - 143.*(-7. + 301.*w2 - 2709.*fcs_pow(w,4) + 6063.*fcs_pow(w,6)) + 4.2483441e7*fcs_pow(v,6)*(-5. + 245.*w2 - 2499.*fcs_pow(w,4) + 6307.*fcs_pow(w,6)) + 7.22218497e8*fcs_pow(v,10)*(-7. + 371.*w2 - 4081.*fcs_pow(w,4) + 11077.*fcs_pow(w,6)) - 1.159926677e9*fcs_pow(v,12)*(-7. + 385.*w2 - 4389.*fcs_pow(w,4) + 12331.*fcs_pow(w,6)) - 4.2483441e7*fcs_pow(v,8)*(-35. + 1785.*w2 - 18921.*fcs_pow(w,4) + 49555.*fcs_pow(w,6)) + 1.40210917e8*fcs_pow(v,14)*(-35. + 1995.*w2 - 23541.*fcs_pow(w,4) + 68381.*fcs_pow(w,6))));
        case 8 : return 2.830439402334541e28*fcs_pow(t,37)*(-2.78992987e8*fcs_pow(u,12)*(-3.216916885e9*fcs_pow(v,12)*(1. - 244.*w2 + 7686.*fcs_pow(w,4) - 66612.*fcs_pow(w,6) + 159393.*fcs_pow(w,8)) + 2.156394835e9*fcs_pow(v,14)*(1. - 252.*w2 + 8190.*fcs_pow(w,4) - 73164.*fcs_pow(w,6) + 180297.*fcs_pow(w,8)) - 39.*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 9555.*v2*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 9.4699605e7*fcs_pow(v,8)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) - 487305.*fcs_pow(v,4)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 3.59858499e8*fcs_pow(v,10)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) + 8.609055e6*fcs_pow(v,6)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 1.95888693e8*fcs_pow(u,10)*(1.661484545e9*fcs_pow(v,14)*(1. - 244.*w2 + 7686.*fcs_pow(w,4) - 66612.*fcs_pow(w,6) + 159393.*fcs_pow(w,8)) - 39.*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 12831.*v2*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 449085.*fcs_pow(v,4)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 2.96725429e8*fcs_pow(v,10)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 7.634445e6*fcs_pow(v,6)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 5.12525741e8*fcs_pow(v,12)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) - 5.7803655e7*fcs_pow(v,8)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 9.79443465e8*fcs_pow(u,8)*(-13.*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 819.*v2*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 38493.*fcs_pow(v,4)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 449085.*fcs_pow(v,6)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 2.6975039e7*fcs_pow(v,12)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) - 3.271905e6*fcs_pow(v,8)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 1.6896453e7*fcs_pow(v,14)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) + 1.1560731e7*fcs_pow(v,10)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 1.4494935e7*fcs_pow(u,6)*(-143.*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 43043.*v2*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 387387.*fcs_pow(v,4)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 6.069063e6*fcs_pow(v,6)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 3.0345315e7*fcs_pow(v,8)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 1.40210917e8*fcs_pow(v,14)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 1.03174071e8*fcs_pow(v,10)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1.65703811e8*fcs_pow(v,12)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) - 151515.*fcs_pow(u,4)*(287287.*v2*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 1.2353341e7*fcs_pow(v,4)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 143.*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 3.7060023e7*fcs_pow(v,6)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 2.48831583e8*fcs_pow(v,8)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 5.80607027e8*fcs_pow(v,10)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 8.97301769e8*fcs_pow(v,12)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 5.22604327e8*fcs_pow(v,14)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8))) + 10101.*u2*(-4.309305e6*fcs_pow(v,4)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) + 6.1766705e7*fcs_pow(v,6)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) + 15015.*v2*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) - 7.9414335e7*fcs_pow(v,8)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 2.48831583e8*fcs_pow(v,10)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 2.63912285e8*fcs_pow(v,12)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 1.47906885e8*fcs_pow(v,14)*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 11.*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) - 39.*(5.3148095e7*fcs_pow(v,6)*(1. - 172.*w2 + 3870.*fcs_pow(w,4) - 24252.*fcs_pow(w,6) + 42441.*fcs_pow(w,8)) - 3.26481155e8*fcs_pow(v,8)*(1. - 180.*w2 + 4230.*fcs_pow(w,4) - 27636.*fcs_pow(w,6) + 50337.*fcs_pow(w,8)) - 555555.*fcs_pow(v,4)*(7. - 1148.*w2 + 24682.*fcs_pow(w,4) - 148092.*fcs_pow(w,6) + 248583.*fcs_pow(w,8)) + 1.95888693e8*fcs_pow(v,10)*(5. - 940.*w2 + 23030.*fcs_pow(w,4) - 156604.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 2.78992987e8*fcs_pow(v,12)*(5. - 980.*w2 + 24990.*fcs_pow(w,4) - 176596.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.07304995e8*fcs_pow(v,14)*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) - 11.*(35. - 5180.*w2 + 101010.*fcs_pow(w,4) - 552188.*fcs_pow(w,6) + 848003.*fcs_pow(w,8)) + 2849.*v2*(35. - 5460.*w2 + 111930.*fcs_pow(w,4) - 641732.*fcs_pow(w,6) + 1.031355e6*fcs_pow(w,8))) + 1.07304995e8*fcs_pow(u,14)*(3.033093063e9*fcs_pow(v,10)*(1. - 244.*w2 + 7686.*fcs_pow(w,4) - 66612.*fcs_pow(w,6) + 159393.*fcs_pow(w,8)) - 5.606626571e9*fcs_pow(v,12)*(1. - 252.*w2 + 8190.*fcs_pow(w,4) - 73164.*fcs_pow(w,6) + 180297.*fcs_pow(w,8)) - 39.*(7. - 1428.*w2 + 37842.*fcs_pow(w,4) - 277508.*fcs_pow(w,6) + 564927.*fcs_pow(w,8)) + 1.8939921e7*fcs_pow(v,6)*(5. - 1140.*w2 + 33630.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 615429.*fcs_pow(w,8)) + 13923.*v2*(7. - 1484.*w2 + 40810.*fcs_pow(w,4) - 310156.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1.54225071e8*fcs_pow(v,8)*(5. - 1180.*w2 + 35990.*fcs_pow(w,4) - 302316.*fcs_pow(w,6) + 701805.*fcs_pow(w,8)) - 737919.*fcs_pow(v,4)*(7. - 1540.*w2 + 43890.*fcs_pow(w,4) - 345268.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 5.54501529e8*fcs_pow(v,14)*(7. - 1820.*w2 + 60970.*fcs_pow(w,4) - 560924.*fcs_pow(w,6) + 1.422343e6*fcs_pow(w,8))));
        case 9 : return -1.225297217270623e32*fcs_pow(t,38)*w*(-1.85300115e8*fcs_pow(v,6)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 615615.*fcs_pow(v,4)*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 33.*(35. - 1820.*w2 + 22386.*fcs_pow(w,4) - 91676.*fcs_pow(w,6) + 114595.*fcs_pow(w,8)) + 2.6471445e7*fcs_pow(v,8)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 8.2943861e7*fcs_pow(v,10)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 3003.*v2*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 4.9302295e7*fcs_pow(v,14)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 5.2782457e7*fcs_pow(v,12)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8)) + 6.380297e6*fcs_pow(u,10)*(6.469184505e9*fcs_pow(v,14)*(1. - 84.*w2 + 1638.*fcs_pow(w,4) - 10452.*fcs_pow(w,6) + 20033.*fcs_pow(w,8)) + 3.59858499e8*fcs_pow(v,10)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 6.43383377e8*fcs_pow(v,12)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) - 13.*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 4.0585545e7*fcs_pow(v,8)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 162435.*fcs_pow(v,4)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 2.869685e6*fcs_pow(v,6)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 1911.*v2*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 2.036265e6*fcs_pow(u,8)*(-5.12525741e8*fcs_pow(v,12)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 3.32296909e8*fcs_pow(v,14)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) - 13.*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 4277.*v2*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.27168041e8*fcs_pow(v,10)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 2.544815e6*fcs_pow(v,6)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1.9267885e7*fcs_pow(v,8)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 89817.*fcs_pow(v,4)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 4.751285e6*fcs_pow(u,6)*(-39.*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 1.6896453e7*fcs_pow(v,14)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 273.*v2*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 12831.*fcs_pow(v,4)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 1.1560731e7*fcs_pow(v,12)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) - 1.090635e6*fcs_pow(v,8)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 3.853577e6*fcs_pow(v,10)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 89817.*fcs_pow(v,6)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 1435.*fcs_pow(u,4)*(2.711709e6*v2*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) - 429.*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) - 2.711709e6*fcs_pow(v,4)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) + 4.2483441e7*fcs_pow(v,6)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) + 1.261898253e9*fcs_pow(v,14)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 7.22218497e8*fcs_pow(v,10)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 1.159926677e9*fcs_pow(v,12)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 1.27450323e8*fcs_pow(v,8)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) + 7.*u2*(-5.55900345e8*fcs_pow(v,4)*(1. - 60.*w2 + 846.*fcs_pow(w,4) - 3948.*fcs_pow(w,6) + 5593.*fcs_pow(w,8)) + 615615.*v2*(21. - 1204.*w2 + 16254.*fcs_pow(w,4) - 72756.*fcs_pow(w,6) + 99029.*fcs_pow(w,8)) + 1.85300115e8*fcs_pow(v,6)*(45. - 2820.*w2 + 41454.*fcs_pow(w,4) - 201348.*fcs_pow(w,6) + 296429.*fcs_pow(w,8)) - 1.244157915e9*fcs_pow(v,8)*(45. - 2940.*w2 + 44982.*fcs_pow(w,4) - 227052.*fcs_pow(w,6) + 346885.*fcs_pow(w,8)) - 429.*(105. - 5740.*w2 + 74046.*fcs_pow(w,4) - 317340.*fcs_pow(w,6) + 414305.*fcs_pow(w,8)) - 4.486508845e9*fcs_pow(v,12)*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) + 2.613021635e9*fcs_pow(v,14)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) + 1.741821081e9*fcs_pow(v,10)*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8))) - 4.060189e6*fcs_pow(u,12)*(-2.8033132855e10*fcs_pow(v,12)*(1. - 84.*w2 + 1638.*fcs_pow(w,4) - 10452.*fcs_pow(w,6) + 20033.*fcs_pow(w,8)) - 2.57041785e8*fcs_pow(v,8)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) + 1.011031021e9*fcs_pow(v,10)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) + 1.3528515e7*fcs_pow(v,6)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 7735.*v2*(63. - 4452.*w2 + 73458.*fcs_pow(w,4) - 398772.*fcs_pow(w,6) + 653543.*fcs_pow(w,8)) - 409955.*fcs_pow(v,4)*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 13.*(105. - 7140.*w2 + 113526.*fcs_pow(w,4) - 594660.*fcs_pow(w,6) + 941545.*fcs_pow(w,8)) + 3.08056405e8*fcs_pow(v,14)*(63. - 5460.*w2 + 109746.*fcs_pow(w,4) - 721188.*fcs_pow(w,6) + 1.422343e6*fcs_pow(w,8))) + 4.9302295e7*fcs_pow(u,14)*(-63. + 4452.*w2 - 73458.*fcs_pow(w,4) + 398772.*fcs_pow(w,6) - 653543.*fcs_pow(w,8) + 8.37188583e8*fcs_pow(v,10)*(1. - 84.*w2 + 1638.*fcs_pow(w,4) - 10452.*fcs_pow(w,6) + 20033.*fcs_pow(w,8)) + 1.628319e6*fcs_pow(v,6)*(15. - 1180.*w2 + 21594.*fcs_pow(w,4) - 129564.*fcs_pow(w,6) + 233935.*fcs_pow(w,8)) - 1.3724403e7*fcs_pow(v,8)*(15. - 1220.*w2 + 23058.*fcs_pow(w,4) - 142740.*fcs_pow(w,6) + 265655.*fcs_pow(w,8)) - 36729.*fcs_pow(v,4)*(35. - 2660.*w2 + 47082.*fcs_pow(w,4) - 273524.*fcs_pow(w,6) + 478667.*fcs_pow(w,8)) + 371.*v2*(63. - 4620.*w2 + 79002.*fcs_pow(w,4) - 443916.*fcs_pow(w,6) + 752191.*fcs_pow(w,8)) - 2.5369351e7*fcs_pow(v,12)*(63. - 5460.*w2 + 109746.*fcs_pow(w,4) - 721188.*fcs_pow(w,6) + 1.422343e6*fcs_pow(w,8)) + 3.624193e6*fcs_pow(v,14)*(315. - 28140.*w2 + 582498.*fcs_pow(w,4) - 3.938796e6*fcs_pow(w,6) + 7.987003e6*fcs_pow(w,8))));
        case 10 : return 6.126486086353115e32*fcs_pow(t,39)*(2.585583e6*fcs_pow(v,4)*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 3.7060023e7*fcs_pow(v,6)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 33.*(-7. + 1365.*w2 - 37310.*fcs_pow(w,4) + 320866.*fcs_pow(w,6) - 1.031355e6*fcs_pow(w,8) + 1.077193e6*fcs_pow(w,10)) - 2.48831583e8*fcs_pow(v,10)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 2.6471445e7*fcs_pow(v,8)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 3003.*v2*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 5.2782457e7*fcs_pow(v,12)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 9.860459e6*fcs_pow(v,14)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10)) + 1.9140891e7*fcs_pow(u,10)*(-9.4699605e7*fcs_pow(v,8)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 6.43383377e8*fcs_pow(v,12)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) - 13.*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 4.31278967e8*fcs_pow(v,14)*(-1. + 315.*w2 - 13650.*fcs_pow(w,4) + 182910.*fcs_pow(w,6) - 901485.*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 1.19952833e8*fcs_pow(v,10)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 1.721811e6*fcs_pow(v,6)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 637.*v2*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 10829.*fcs_pow(v,4)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 2.036265e6*fcs_pow(u,8)*(8.90176287e8*fcs_pow(v,10)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 9.96890727e8*fcs_pow(v,14)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) + 12831.*v2*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 5.12525741e8*fcs_pow(v,12)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 13.*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 3.4682193e7*fcs_pow(v,8)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 89817.*fcs_pow(v,4)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 508963.*fcs_pow(v,6)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 2.850771e6*fcs_pow(u,6)*(-13.*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.34875195e8*fcs_pow(v,12)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 64155.*fcs_pow(v,4)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 2.8160755e7*fcs_pow(v,14)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 455.*v2*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) + 1.1560731e7*fcs_pow(v,10)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 149695.*fcs_pow(v,6)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 363545.*fcs_pow(v,8)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) - 6027.*fcs_pow(u,4)*(-429.*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) + 129129.*v2*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 2.103163755e9*fcs_pow(v,14)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 3.0345315e7*fcs_pow(v,6)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) - 645645.*fcs_pow(v,4)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 4.97111433e8*fcs_pow(v,12)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 3.0345315e7*fcs_pow(v,8)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 3.4391357e7*fcs_pow(v,10)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 7.*u2*(2.585583e6*v2*(-1. + 215.*w2 - 6450.*fcs_pow(w,4) + 60630.*fcs_pow(w,6) - 212205.*fcs_pow(w,8) + 240499.*fcs_pow(w,10)) - 1.11180069e8*fcs_pow(v,4)*(-1. + 225.*w2 - 7050.*fcs_pow(w,4) + 69090.*fcs_pow(w,6) - 251685.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 3.732473745e9*fcs_pow(v,8)*(-3. + 735.*w2 - 24990.*fcs_pow(w,4) + 264894.*fcs_pow(w,6) - 1.040655e6*fcs_pow(w,8) + 1.318163e6*fcs_pow(w,10)) + 1.85300115e8*fcs_pow(v,6)*(-9. + 2115.*w2 - 69090.*fcs_pow(w,4) + 704718.*fcs_pow(w,6) - 2.667861e6*fcs_pow(w,8) + 3.260719e6*fcs_pow(w,10)) - 429.*(-21. + 4305.*w2 - 123410.*fcs_pow(w,4) + 1.11069e6*fcs_pow(w,6) - 3.728745e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 4.703438943e9*fcs_pow(v,14)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 1.741821081e9*fcs_pow(v,10)*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) - 8.97301769e8*fcs_pow(v,12)*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10))) + 9.860459e6*fcs_pow(u,14)*(63. - 16695.*w2 + 612150.*fcs_pow(w,4) - 6.97851e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 3.9866123e7*fcs_pow(w,10) - 1.285515e6*fcs_pow(v,4)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) - 2.05866045e8*fcs_pow(v,8)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) + 8.37188583e8*fcs_pow(v,10)*(-1. + 315.*w2 - 13650.*fcs_pow(w,4) + 182910.*fcs_pow(w,6) - 901485.*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 8.141595e6*fcs_pow(v,6)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 3339.*v2*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) + 5.4362895e7*fcs_pow(v,14)*(-21. + 7035.*w2 - 323610.*fcs_pow(w,4) + 4.595262e6*fcs_pow(w,6) - 2.3961009e7*fcs_pow(w,8) + 3.9935015e7*fcs_pow(w,10)) - 2.5369351e7*fcs_pow(v,12)*(-63. + 20475.*w2 - 914550.*fcs_pow(w,4) + 1.262079e7*fcs_pow(w,6) - 6.4005435e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10))) - 4.060189e6*fcs_pow(u,12)*(9.4699605e7*fcs_pow(v,6)*(-1. + 285.*w2 - 11210.*fcs_pow(w,4) + 136762.*fcs_pow(w,6) - 615429.*fcs_pow(w,8) + 888953.*fcs_pow(w,10)) + 3.033093063e9*fcs_pow(v,10)*(-1. + 305.*w2 - 12810.*fcs_pow(w,4) + 166530.*fcs_pow(w,6) - 796965.*fcs_pow(w,8) + 1.222013e6*fcs_pow(w,10)) - 5.606626571e9*fcs_pow(v,12)*(-1. + 315.*w2 - 13650.*fcs_pow(w,4) + 182910.*fcs_pow(w,6) - 901485.*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) - 2.57041785e8*fcs_pow(v,8)*(-3. + 885.*w2 - 35990.*fcs_pow(w,4) + 453474.*fcs_pow(w,6) - 2.105415e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 737919.*fcs_pow(v,4)*(-7. + 1925.*w2 - 73150.*fcs_pow(w,4) + 863170.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 5.265337e6*fcs_pow(w,10)) - 13.*(-21. + 5355.*w2 - 189210.*fcs_pow(w,4) + 2.08131e6*fcs_pow(w,6) - 8.473905e6*fcs_pow(w,8) + 1.1110231e7*fcs_pow(w,10)) + 1547.*v2*(-63. + 16695.*w2 - 612150.*fcs_pow(w,4) + 6.97851e6*fcs_pow(w,6) - 2.9409435e7*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10)) + 6.1611281e7*fcs_pow(v,14)*(-63. + 20475.*w2 - 914550.*fcs_pow(w,4) + 1.262079e7*fcs_pow(w,6) - 6.4005435e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10))));
        case 11 : return -6.126486086353115e32*fcs_pow(t,40)*w*(-11.*(-1.0107279e7*fcs_pow(v,4)*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 1.85300115e8*fcs_pow(v,6)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.244157915e9*fcs_pow(v,8)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 1.741821081e9*fcs_pow(v,10)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 8.97301769e8*fcs_pow(v,12)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 39.*(-231. + 15785.*w2 - 271502.*fcs_pow(w,4) + 1.74537e6*fcs_pow(w,6) - 4.557355e6*fcs_pow(w,8) + 4.060189e6*fcs_pow(w,10)) + 5.22604327e8*fcs_pow(v,14)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 11193.*v2*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10))) - 9.5704455e7*fcs_pow(u,8)*(-143.*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 21021.*v2*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 1.293836901e9*fcs_pow(v,14)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 3.59858499e8*fcs_pow(v,10)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 357357.*fcs_pow(v,4)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 6.43383377e8*fcs_pow(v,12)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) + 6.313307e6*fcs_pow(v,6)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 4.0585545e7*fcs_pow(v,8)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 1.4253855e7*fcs_pow(u,6)*(-143.*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 47047.*v2*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 987987.*fcs_pow(v,4)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 5.12525741e8*fcs_pow(v,12)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 5.598593e6*fcs_pow(v,6)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 3.32296909e8*fcs_pow(v,14)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 4.2389347e7*fcs_pow(v,8)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 1.27168041e8*fcs_pow(v,10)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) - 2.850771e6*fcs_pow(u,4)*(-39.*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 15015.*v2*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 705705.*fcs_pow(v,4)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) + 4.939935e6*fcs_pow(v,6)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 8.4482265e7*fcs_pow(v,14)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 1.1996985e7*fcs_pow(v,8)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 4.2389347e7*fcs_pow(v,10)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 5.7803655e7*fcs_pow(v,12)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 3157.*u2*(246519.*v2*(-11. + 825.*w2 - 15510.*fcs_pow(w,4) + 108570.*fcs_pow(w,6) - 307615.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) - 1.3558545e7*fcs_pow(v,4)*(-9. + 705.*w2 - 13818.*fcs_pow(w,4) + 100674.*fcs_pow(w,6) - 296429.*fcs_pow(w,8) + 296429.*fcs_pow(w,10)) + 2.12417205e8*fcs_pow(v,6)*(-9. + 735.*w2 - 14994.*fcs_pow(w,4) + 113526.*fcs_pow(w,6) - 346885.*fcs_pow(w,8) + 359499.*fcs_pow(w,10)) - 6.37251615e8*fcs_pow(v,8)*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) + 7.22218497e8*fcs_pow(v,10)*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 1.159926677e9*fcs_pow(v,12)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 39.*(-231. + 16555.*w2 - 297990.*fcs_pow(w,4) + 2.00079e6*fcs_pow(w,6) - 5.446595e6*fcs_pow(w,8) + 5.050479e6*fcs_pow(w,10)) + 5.73590115e8*fcs_pow(v,14)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10))) + 1.33986237e8*fcs_pow(u,10)*(-143.*(-21. + 1785.*w2 - 37842.*fcs_pow(w,4) + 297330.*fcs_pow(w,6) - 941545.*fcs_pow(w,8) + 1.010021e6*fcs_pow(w,10)) - 5.606626571e9*fcs_pow(v,12)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) - 2.57041785e8*fcs_pow(v,8)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 17017.*v2*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) + 1.011031021e9*fcs_pow(v,10)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) - 901901.*fcs_pow(v,4)*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 1.3528515e7*fcs_pow(v,6)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10)) + 6.1611281e7*fcs_pow(v,14)*(-693. + 75075.*w2 - 2.01201e6*fcs_pow(w,4) + 1.983267e7*fcs_pow(w,6) - 7.8228865e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10))) + 5.22604327e8*fcs_pow(u,14)*(-2.36940165e8*fcs_pow(v,8)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) - 460845.*fcs_pow(v,4)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) + 9.063285e6*fcs_pow(v,6)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) + 7.55951955e8*fcs_pow(v,14)*(-21. + 2415.*w2 - 68586.*fcs_pow(w,4) + 715254.*fcs_pow(w,6) - 2.980225e6*fcs_pow(w,8) + 4.172315e6*fcs_pow(w,10)) - 11.*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) + 3465.*v2*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10)) + 1.5796011e7*fcs_pow(v,10)*(-693. + 75075.*w2 - 2.01201e6*fcs_pow(w,4) + 1.983267e7*fcs_pow(w,6) - 7.8228865e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10)) - 3.1113355e7*fcs_pow(v,12)*(-693. + 77385.*w2 - 2.135826e6*fcs_pow(w,4) + 2.1663378e7*fcs_pow(w,6) - 8.7857033e7*fcs_pow(w,8) + 1.19805045e8*fcs_pow(w,10))) - 8.97301769e8*fcs_pow(u,12)*(8.37188583e8*fcs_pow(v,10)*(-11. + 1155.*w2 - 30030.*fcs_pow(w,4) + 287430.*fcs_pow(w,6) - 1.101815e6*fcs_pow(w,8) + 1.422343e6*fcs_pow(w,10)) + 8.141595e6*fcs_pow(v,6)*(-33. + 3245.*w2 - 79178.*fcs_pow(w,4) + 712602.*fcs_pow(w,6) - 2.573285e6*fcs_pow(w,8) + 3.134729e6*fcs_pow(w,10)) - 11.*(-63. + 5565.*w2 - 122430.*fcs_pow(w,4) + 996930.*fcs_pow(w,6) - 3.267715e6*fcs_pow(w,8) + 3.624193e6*fcs_pow(w,10)) - 6.8622015e7*fcs_pow(v,8)*(-33. + 3355.*w2 - 84546.*fcs_pow(w,4) + 785070.*fcs_pow(w,6) - 2.922205e6*fcs_pow(w,8) + 3.666039e6*fcs_pow(w,10)) + 4081.*v2*(-63. + 5775.*w2 - 131670.*fcs_pow(w,4) + 1.10979e6*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 4.308003e6*fcs_pow(w,10)) - 183645.*fcs_pow(v,4)*(-77. + 7315.*w2 - 172634.*fcs_pow(w,4) + 1.504382e6*fcs_pow(w,6) - 5.265337e6*fcs_pow(w,8) + 6.222671e6*fcs_pow(w,10)) - 2.5369351e7*fcs_pow(v,12)*(-693. + 75075.*w2 - 2.01201e6*fcs_pow(w,4) + 1.983267e7*fcs_pow(w,6) - 7.8228865e7*fcs_pow(w,8) + 1.03831039e8*fcs_pow(w,10)) + 1.8120965e7*fcs_pow(v,14)*(-693. + 77385.*w2 - 2.135826e6*fcs_pow(w,4) + 2.1663378e7*fcs_pow(w,6) - 8.7857033e7*fcs_pow(w,8) + 1.19805045e8*fcs_pow(w,10))));
        case 12 : return 1.837945825905934e33*fcs_pow(t,41)*(-2.850771e6*fcs_pow(u,4)*(-143.*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 15015.*v2*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 705705.*fcs_pow(v,4)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 1.27168041e8*fcs_pow(v,10)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 1.34875195e8*fcs_pow(v,12)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 1.646645e6*fcs_pow(v,6)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 8.4482265e7*fcs_pow(v,14)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 1.1996985e7*fcs_pow(v,8)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 66297.*u2*(43043.*v2*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) - 645645.*fcs_pow(v,4)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 1.0115105e7*fcs_pow(v,6)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 13.*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) - 1.65703811e8*fcs_pow(v,12)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 6.3732235e7*fcs_pow(v,14)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 1.0115105e7*fcs_pow(v,8)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3.4391357e7*fcs_pow(v,10)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 11.*(-3.7060023e7*fcs_pow(v,4)*(1. - 270.*w2 + 10575.*fcs_pow(w,4) - 138180.*fcs_pow(w,6) + 755055.*fcs_pow(w,8) - 1.778574e6*fcs_pow(w,10) + 1.482145e6*fcs_pow(w,12)) + 1.85300115e8*fcs_pow(v,6)*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 1.244157915e9*fcs_pow(v,8)*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 78351.*v2*(11. - 2838.*w2 + 106425.*fcs_pow(w,4) - 1.33386e6*fcs_pow(w,6) + 7.002765e6*fcs_pow(w,8) - 1.5872934e7*fcs_pow(w,10) + 1.2746447e7*fcs_pow(w,12)) + 1.567812981e9*fcs_pow(v,14)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 5.80607027e8*fcs_pow(v,10)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 39.*(77. - 18942.*w2 + 678755.*fcs_pow(w,4) - 8.14506e6*fcs_pow(w,6) + 4.1016195e7*fcs_pow(w,8) - 8.9324158e7*fcs_pow(w,10) + 6.9023213e7*fcs_pow(w,12)) - 8.97301769e8*fcs_pow(v,12)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 1.4253855e7*fcs_pow(u,6)*(-143.*(3. - 846.*w2 + 34545.*fcs_pow(w,4) - 469812.*fcs_pow(w,6) + 2.667861e6*fcs_pow(w,8) - 6.521438e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 47047.*v2*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 1.27168041e8*fcs_pow(v,8)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 2.96725429e8*fcs_pow(v,10)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 329329.*fcs_pow(v,4)*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 5.12525741e8*fcs_pow(v,12)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 5.598593e6*fcs_pow(v,6)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 3.32296909e8*fcs_pow(v,14)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12))) - 9.5704455e7*fcs_pow(u,8)*(-143.*(3. - 882.*w2 + 37485.*fcs_pow(w,4) - 529788.*fcs_pow(w,6) + 3.121965e6*fcs_pow(w,8) - 7.908978e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 1.8939921e7*fcs_pow(v,6)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 9.4699605e7*fcs_pow(v,8)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 7007.*v2*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3.59858499e8*fcs_pow(v,10)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 357357.*fcs_pow(v,4)*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) - 6.43383377e8*fcs_pow(v,12)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) + 4.31278967e8*fcs_pow(v,14)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12))) + 5.748647597e9*fcs_pow(u,14)*(-21. + 6930.*w2 - 329175.*fcs_pow(w,4) + 5.17902e6*fcs_pow(w,6) - 3.3848595e7*fcs_pow(w,8) + 9.4776066e7*fcs_pow(w,10) - 9.3340065e7*fcs_pow(w,12) + 735.*v2*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 41895.*fcs_pow(v,4)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 823935.*fcs_pow(v,6)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) - 7.180005e6*fcs_pow(v,8)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12)) - 3.1113355e7*fcs_pow(v,12)*(21. - 8442.*w2 + 485415.*fcs_pow(w,4) - 9.190524e6*fcs_pow(w,6) + 7.1883027e7*fcs_pow(w,8) - 2.3961009e8*fcs_pow(w,10) + 2.79545105e8*fcs_pow(w,12)) + 2.2907635e7*fcs_pow(v,14)*(21. - 8694.*w2 + 514395.*fcs_pow(w,4) - 1.0013556e7*fcs_pow(w,6) + 8.0466075e7*fcs_pow(w,8) - 2.7537279e8*fcs_pow(w,10) + 3.29612885e8*fcs_pow(w,12)) + 1.436001e6*fcs_pow(v,10)*(231. - 90090.*w2 + 5.030025e6*fcs_pow(w,4) - 9.255246e7*fcs_pow(w,6) + 7.04059785e8*fcs_pow(w,8) - 2.284282858e9*fcs_pow(w,10) + 2.595775975e9*fcs_pow(w,12))) - 8.97301769e8*fcs_pow(u,12)*(12243.*v2*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) - 428505.*fcs_pow(v,4)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) + 8.141595e6*fcs_pow(v,6)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) - 11.*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) - 6.8622015e7*fcs_pow(v,8)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) + 2.79062861e8*fcs_pow(v,10)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12)) + 1.99330615e8*fcs_pow(v,14)*(21. - 8442.*w2 + 485415.*fcs_pow(w,4) - 9.190524e6*fcs_pow(w,6) + 7.1883027e7*fcs_pow(w,8) - 2.3961009e8*fcs_pow(w,10) + 2.79545105e8*fcs_pow(w,12)) - 2.5369351e7*fcs_pow(v,12)*(231. - 90090.*w2 + 5.030025e6*fcs_pow(w,4) - 9.255246e7*fcs_pow(w,6) + 7.04059785e8*fcs_pow(w,8) - 2.284282858e9*fcs_pow(w,10) + 2.595775975e9*fcs_pow(w,12))) + 4.4662079e7*fcs_pow(u,10)*(-8.117109e6*fcs_pow(v,4)*(7. - 2310.*w2 + 109725.*fcs_pow(w,4) - 1.72634e6*fcs_pow(w,6) + 1.1282865e7*fcs_pow(w,8) - 3.1592022e7*fcs_pow(w,10) + 3.1113355e7*fcs_pow(w,12)) + 9.4699605e7*fcs_pow(v,6)*(11. - 3762.*w2 + 184965.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 2.0309157e7*fcs_pow(w,8) - 5.8670898e7*fcs_pow(w,10) + 5.9559851e7*fcs_pow(w,12)) - 143.*(21. - 6426.*w2 + 283815.*fcs_pow(w,4) - 4.16262e6*fcs_pow(w,6) + 2.5421715e7*fcs_pow(w,8) - 6.6661386e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 7.71125355e8*fcs_pow(v,8)*(11. - 3894.*w2 + 197945.*fcs_pow(w,4) - 3.325476e6*fcs_pow(w,6) + 2.3159565e7*fcs_pow(w,8) - 6.8964038e7*fcs_pow(w,10) + 7.2098767e7*fcs_pow(w,12)) + 51051.*v2*(21. - 6678.*w2 + 306075.*fcs_pow(w,4) - 4.65234e6*fcs_pow(w,6) + 2.9409435e7*fcs_pow(w,8) - 7.9732246e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 3.033093063e9*fcs_pow(v,10)*(11. - 4026.*w2 + 211365.*fcs_pow(w,4) - 3.66366e6*fcs_pow(w,6) + 2.6299845e7*fcs_pow(w,8) - 8.0652858e7*fcs_pow(w,10) + 8.6762923e7*fcs_pow(w,12)) - 5.606626571e9*fcs_pow(v,12)*(11. - 4158.*w2 + 225225.*fcs_pow(w,4) - 4.02402e6*fcs_pow(w,6) + 2.9749005e7*fcs_pow(w,8) - 9.3874638e7*fcs_pow(w,10) + 1.03831039e8*fcs_pow(w,12)) + 1.84833843e8*fcs_pow(v,14)*(231. - 90090.*w2 + 5.030025e6*fcs_pow(w,4) - 9.255246e7*fcs_pow(w,6) + 7.04059785e8*fcs_pow(w,8) - 2.284282858e9*fcs_pow(w,10) + 2.595775975e9*fcs_pow(w,12))));
        case 13 : return -7.535577886214331e34*fcs_pow(t,42)*w*(2.12417205e8*fcs_pow(u,6)*(3.59858499e8*fcs_pow(v,10)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) - 6.43383377e8*fcs_pow(v,12)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) - 11.*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 6.313307e6*fcs_pow(v,6)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 4.31278967e8*fcs_pow(v,14)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) - 1.3528515e7*fcs_pow(v,8)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 539.*v2*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 27489.*fcs_pow(v,4)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 1.3558545e7*fcs_pow(u,4)*(-5.12525741e8*fcs_pow(v,12)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) - 11.*(39. - 3666.*w2 + 89817.*fcs_pow(w,4) - 872508.*fcs_pow(w,6) + 3.853577e6*fcs_pow(w,8) - 7.707154e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 3.32296909e8*fcs_pow(v,14)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) + 3619.*v2*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 4.2389347e7*fcs_pow(v,8)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 4.2389347e7*fcs_pow(v,10)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 25333.*fcs_pow(v,4)*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 430661.*fcs_pow(v,6)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) + 903903.*u2*(-143. + 12870.*w2 - 302445.*fcs_pow(w,4) + 2.82282e6*fcs_pow(w,6) - 1.1996985e7*fcs_pow(w,8) + 2.3121462e7*fcs_pow(w,10) - 1.6303595e7*fcs_pow(w,12) + 8.4482265e7*fcs_pow(v,14)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 1155.*v2*(39. - 3666.*w2 + 89817.*fcs_pow(w,4) - 872508.*fcs_pow(w,6) + 3.853577e6*fcs_pow(w,8) - 7.707154e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) - 54285.*fcs_pow(v,4)*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) + 4.2389347e7*fcs_pow(v,10)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) - 1.9267885e7*fcs_pow(v,12)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 126665.*fcs_pow(v,6)*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) - 922845.*fcs_pow(v,8)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12))) - 143.*(-1.042965e6*fcs_pow(v,4)*(39. - 3666.*w2 + 89817.*fcs_pow(w,4) - 872508.*fcs_pow(w,6) + 3.853577e6*fcs_pow(w,8) - 7.707154e6*fcs_pow(w,10) + 5.632151e6*fcs_pow(w,12)) + 1.6339785e7*fcs_pow(v,6)*(39. - 3822.*w2 + 97461.*fcs_pow(w,4) - 983892.*fcs_pow(w,6) + 4.509505e6*fcs_pow(w,8) - 9.346974e6*fcs_pow(w,10) + 7.070147e6*fcs_pow(w,12)) - 1.159926677e9*fcs_pow(v,12)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 6321.*v2*(143. - 12870.*w2 + 302445.*fcs_pow(w,4) - 2.82282e6*fcs_pow(w,6) + 1.1996985e7*fcs_pow(w,8) - 2.3121462e7*fcs_pow(w,10) + 1.6303595e7*fcs_pow(w,12)) + 1.91196705e8*fcs_pow(v,14)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 1.6339785e7*fcs_pow(v,8)*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 5.5555269e7*fcs_pow(v,10)*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) - 3.*(1001. - 86086.*w2 + 1.936935e6*fcs_pow(w,4) - 1.734018e7*fcs_pow(w,6) + 7.0805735e7*fcs_pow(w,8) - 1.31312454e8*fcs_pow(w,10) + 8.9225129e7*fcs_pow(w,12))) - 2.12417205e8*fcs_pow(u,8)*(-7.71125355e8*fcs_pow(v,8)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 3.033093063e9*fcs_pow(v,10)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) - 2.705703e6*fcs_pow(v,4)*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) - 5.606626571e9*fcs_pow(v,12)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) + 1.3528515e7*fcs_pow(v,6)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 11.*(273. - 27846.*w2 + 737919.*fcs_pow(w,4) - 7.73058e6*fcs_pow(w,6) + 3.6720255e7*fcs_pow(w,8) - 7.8781638e7*fcs_pow(w,10) + 6.1611281e7*fcs_pow(w,12)) + 3927.*v2*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) + 1.84833843e8*fcs_pow(v,14)*(231. - 30030.*w2 + 1.006005e6*fcs_pow(w,4) - 1.322178e7*fcs_pow(w,6) + 7.8228865e7*fcs_pow(w,8) - 2.07662078e8*fcs_pow(w,10) + 1.99675075e8*fcs_pow(w,12))) + 7.01054585e8*fcs_pow(u,14)*(108927.*v2*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) - 6.426693e6*fcs_pow(v,4)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) + 1.30676091e8*fcs_pow(v,6)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) - 39.*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 5.6004039e7*fcs_pow(v,8)*(231. - 30030.*w2 + 1.006005e6*fcs_pow(w,4) - 1.322178e7*fcs_pow(w,6) + 7.8228865e7*fcs_pow(w,8) - 2.07662078e8*fcs_pow(w,10) + 1.99675075e8*fcs_pow(w,12)) - 4.16918957e8*fcs_pow(v,12)*(273. - 37674.*w2 + 1.337427e6*fcs_pow(w,4) - 1.8596604e7*fcs_pow(w,6) + 1.16228775e8*fcs_pow(w,8) - 3.2544057e8*fcs_pow(w,10) + 3.29612885e8*fcs_pow(w,12)) + 3.16125363e8*fcs_pow(v,14)*(273. - 38766.*w2 + 1.414959e6*fcs_pow(w,4) - 2.02137e7*fcs_pow(w,6) + 1.29704575e8*fcs_pow(w,8) - 3.7260587e8*fcs_pow(w,10) + 3.86936865e8*fcs_pow(w,12)) + 1.8668013e7*fcs_pow(v,10)*(3003. - 402402.*w2 + 1.3882869e7*fcs_pow(w,4) - 1.87749276e8*fcs_pow(w,6) + 1.142141429e9*fcs_pow(w,8) - 3.11493117e9*fcs_pow(w,10) + 3.074996155e9*fcs_pow(w,12))) - 1.159926677e9*fcs_pow(u,12)*(-5.990985e6*fcs_pow(v,4)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) + 1.17822705e8*fcs_pow(v,6)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) - 143.*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) - 1.026740715e9*fcs_pow(v,8)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) + 15015.*v2*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) + 2.05348143e8*fcs_pow(v,10)*(231. - 30030.*w2 + 1.006005e6*fcs_pow(w,4) - 1.322178e7*fcs_pow(w,6) + 7.8228865e7*fcs_pow(w,8) - 2.07662078e8*fcs_pow(w,10) + 1.99675075e8*fcs_pow(w,12)) + 2.51983985e8*fcs_pow(v,14)*(273. - 37674.*w2 + 1.337427e6*fcs_pow(w,4) - 1.8596604e7*fcs_pow(w,6) + 1.16228775e8*fcs_pow(w,8) - 3.2544057e8*fcs_pow(w,10) + 3.29612885e8*fcs_pow(w,12)) - 3.1113355e7*fcs_pow(v,12)*(3003. - 402402.*w2 + 1.3882869e7*fcs_pow(w,4) - 1.87749276e8*fcs_pow(w,6) + 1.142141429e9*fcs_pow(w,8) - 3.11493117e9*fcs_pow(w,10) + 3.074996155e9*fcs_pow(w,12))) + 7.22218497e8*fcs_pow(u,10)*(1.05840735e8*fcs_pow(v,6)*(11. - 1298.*w2 + 39589.*fcs_pow(w,4) - 475068.*fcs_pow(w,6) + 2.573285e6*fcs_pow(w,8) - 6.269458e6*fcs_pow(w,10) + 5.546059e6*fcs_pow(w,12)) - 8.92086195e8*fcs_pow(v,8)*(11. - 1342.*w2 + 42273.*fcs_pow(w,4) - 523380.*fcs_pow(w,6) + 2.922205e6*fcs_pow(w,8) - 7.332078e6*fcs_pow(w,10) + 6.674071e6*fcs_pow(w,12)) + 53053.*v2*(21. - 2310.*w2 + 65835.*fcs_pow(w,4) - 739860.*fcs_pow(w,6) + 3.760955e6*fcs_pow(w,8) - 8.616006e6*fcs_pow(w,10) + 7.180005e6*fcs_pow(w,12)) + 3.627817193e9*fcs_pow(v,10)*(11. - 1386.*w2 + 45045.*fcs_pow(w,4) - 574860.*fcs_pow(w,6) + 3.305445e6*fcs_pow(w,8) - 8.534058e6*fcs_pow(w,10) + 7.987003e6*fcs_pow(w,12)) - 795795.*fcs_pow(v,4)*(77. - 8778.*w2 + 258951.*fcs_pow(w,4) - 3.008764e6*fcs_pow(w,6) + 1.5796011e7*fcs_pow(w,8) - 3.7336026e7*fcs_pow(w,10) + 3.2070689e7*fcs_pow(w,12)) - 11.*(273. - 28938.*w2 + 795795.*fcs_pow(w,4) - 8.64006e6*fcs_pow(w,6) + 4.2480295e7*fcs_pow(w,8) - 9.4229018e7*fcs_pow(w,10) + 7.6108053e7*fcs_pow(w,12)) - 3.29801563e8*fcs_pow(v,12)*(231. - 30030.*w2 + 1.006005e6*fcs_pow(w,4) - 1.322178e7*fcs_pow(w,6) + 7.8228865e7*fcs_pow(w,8) - 2.07662078e8*fcs_pow(w,10) + 1.99675075e8*fcs_pow(w,12)) + 1.8120965e7*fcs_pow(v,14)*(3003. - 402402.*w2 + 1.3882869e7*fcs_pow(w,4) - 1.87749276e8*fcs_pow(w,6) + 1.142141429e9*fcs_pow(w,8) - 3.11493117e9*fcs_pow(w,10) + 3.074996155e9*fcs_pow(w,12))));
        case 14 : return 5.802394972385035e36*fcs_pow(t,43)*(-176085.*fcs_pow(u,4)*(-8.90176287e8*fcs_pow(v,8)*(-1. + 385.*w2 - 21945.*fcs_pow(w,4) + 431585.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 1.5796011e7*fcs_pow(w,10) - 3.1113355e7*fcs_pow(w,12) + 2.2907635e7*fcs_pow(w,14)) + 3.9190151e7*fcs_pow(v,6)*(-3. + 1113.*w2 - 61215.*fcs_pow(w,4) + 1.163085e6*fcs_pow(w,6) - 9.803145e6*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10) - 7.6108053e7*fcs_pow(w,12) + 5.4362895e7*fcs_pow(w,14)) + 2.96725429e8*fcs_pow(v,10)*(-11. + 4389.*w2 - 258951.*fcs_pow(w,4) + 5.265337e6*fcs_pow(w,6) - 4.7388033e7*fcs_pow(w,8) + 2.05348143e8*fcs_pow(w,10) - 4.16918957e8*fcs_pow(w,12) + 3.16125363e8*fcs_pow(w,14)) - 11.*(-39. + 12831.*w2 - 628719.*fcs_pow(w,4) + 1.0688223e7*fcs_pow(w,6) - 8.0925117e7*fcs_pow(w,8) + 2.96725429e8*fcs_pow(w,10) - 5.12525741e8*fcs_pow(w,12) + 3.32296909e8*fcs_pow(w,14)) - 5.12525741e8*fcs_pow(v,12)*(-11. + 4543.*w2 - 277123.*fcs_pow(w,4) + 5.819583e6*fcs_pow(w,6) - 5.4038985e7*fcs_pow(w,8) + 2.41374133e8*fcs_pow(w,10) - 5.04691369e8*fcs_pow(w,12) + 3.93770189e8*fcs_pow(w,14)) + 3619.*v2*(-39. + 13377.*w2 - 682227.*fcs_pow(w,4) + 1.2052677e7*fcs_pow(w,6) - 9.4699605e7*fcs_pow(w,8) + 3.59858499e8*fcs_pow(w,10) - 6.43383377e8*fcs_pow(w,12) + 4.31278967e8*fcs_pow(w,14)) + 3.32296909e8*fcs_pow(v,14)*(-11. + 4697.*w2 - 295911.*fcs_pow(w,4) + 6.411405e6*fcs_pow(w,6) - 6.1366305e7*fcs_pow(w,8) + 2.82285003e8*fcs_pow(w,10) - 6.07340461e8*fcs_pow(w,12) + 4.87207183e8*fcs_pow(w,14)) - 177331.*fcs_pow(v,4)*(-39. + 13923.*w2 - 737919.*fcs_pow(w,4) + 1.3528515e7*fcs_pow(w,6) - 1.10160765e8*fcs_pow(w,8) + 4.33299009e8*fcs_pow(w,10) - 8.00946653e8*fcs_pow(w,12) + 5.54501529e8*fcs_pow(w,14))) + 11739.*u2*(8.90176287e8*fcs_pow(v,10)*(-1. + 385.*w2 - 21945.*fcs_pow(w,4) + 431585.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 1.5796011e7*fcs_pow(w,10) - 3.1113355e7*fcs_pow(w,12) + 2.2907635e7*fcs_pow(w,14)) - 8.3978895e7*fcs_pow(v,8)*(-3. + 1113.*w2 - 61215.*fcs_pow(w,4) + 1.163085e6*fcs_pow(w,6) - 9.803145e6*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10) - 7.6108053e7*fcs_pow(w,12) + 5.4362895e7*fcs_pow(w,14)) - 11.*(-13. + 4095.*w2 - 192465.*fcs_pow(w,4) + 3.143595e6*fcs_pow(w,6) - 2.2903335e7*fcs_pow(w,8) + 8.0925117e7*fcs_pow(w,10) - 1.34875195e8*fcs_pow(w,12) + 8.4482265e7*fcs_pow(w,14)) - 1.34875195e8*fcs_pow(v,12)*(-11. + 4389.*w2 - 258951.*fcs_pow(w,4) + 5.265337e6*fcs_pow(w,6) - 4.7388033e7*fcs_pow(w,8) + 2.05348143e8*fcs_pow(w,10) - 4.16918957e8*fcs_pow(w,12) + 3.16125363e8*fcs_pow(w,14)) + 1155.*v2*(-39. + 12831.*w2 - 628719.*fcs_pow(w,4) + 1.0688223e7*fcs_pow(w,6) - 8.0925117e7*fcs_pow(w,8) + 2.96725429e8*fcs_pow(w,10) - 5.12525741e8*fcs_pow(w,12) + 3.32296909e8*fcs_pow(w,14)) + 8.4482265e7*fcs_pow(v,14)*(-11. + 4543.*w2 - 277123.*fcs_pow(w,4) + 5.819583e6*fcs_pow(w,6) - 5.4038985e7*fcs_pow(w,8) + 2.41374133e8*fcs_pow(w,10) - 5.04691369e8*fcs_pow(w,12) + 3.93770189e8*fcs_pow(w,14)) - 54285.*fcs_pow(v,4)*(-39. + 13377.*w2 - 682227.*fcs_pow(w,4) + 1.2052677e7*fcs_pow(w,6) - 9.4699605e7*fcs_pow(w,8) + 3.59858499e8*fcs_pow(w,10) - 6.43383377e8*fcs_pow(w,12) + 4.31278967e8*fcs_pow(w,14)) + 886655.*fcs_pow(v,6)*(-39. + 13923.*w2 - 737919.*fcs_pow(w,4) + 1.3528515e7*fcs_pow(w,6) - 1.10160765e8*fcs_pow(w,8) + 4.33299009e8*fcs_pow(w,10) - 8.00946653e8*fcs_pow(w,12) + 5.54501529e8*fcs_pow(w,14))) - 39.*(143. - 43043.*w2 + 1.936935e6*fcs_pow(w,4) - 3.0345315e7*fcs_pow(w,6) + 2.12417205e8*fcs_pow(w,8) - 7.22218497e8*fcs_pow(w,10) + 1.159926677e9*fcs_pow(w,12) - 7.01054585e8*fcs_pow(w,14) - 1.159926677e9*fcs_pow(v,12)*(-1. + 385.*w2 - 21945.*fcs_pow(w,4) + 431585.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 1.5796011e7*fcs_pow(w,10) - 3.1113355e7*fcs_pow(w,12) + 2.2907635e7*fcs_pow(w,14)) + 2.40739499e8*fcs_pow(v,10)*(-3. + 1113.*w2 - 61215.*fcs_pow(w,4) + 1.163085e6*fcs_pow(w,6) - 9.803145e6*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10) - 7.6108053e7*fcs_pow(w,12) + 5.4362895e7*fcs_pow(w,14)) + 3311.*v2*(-13. + 4095.*w2 - 192465.*fcs_pow(w,4) + 3.143595e6*fcs_pow(w,6) - 2.2903335e7*fcs_pow(w,8) + 8.0925117e7*fcs_pow(w,10) - 1.34875195e8*fcs_pow(w,12) + 8.4482265e7*fcs_pow(w,14)) + 6.3732235e7*fcs_pow(v,14)*(-11. + 4389.*w2 - 258951.*fcs_pow(w,4) + 5.265337e6*fcs_pow(w,6) - 4.7388033e7*fcs_pow(w,8) + 2.05348143e8*fcs_pow(w,10) - 4.16918957e8*fcs_pow(w,12) + 3.16125363e8*fcs_pow(w,14)) - 49665.*fcs_pow(v,4)*(-39. + 12831.*w2 - 628719.*fcs_pow(w,4) + 1.0688223e7*fcs_pow(w,6) - 8.0925117e7*fcs_pow(w,8) + 2.96725429e8*fcs_pow(w,10) - 5.12525741e8*fcs_pow(w,12) + 3.32296909e8*fcs_pow(w,14)) + 778085.*fcs_pow(v,6)*(-39. + 13377.*w2 - 682227.*fcs_pow(w,4) + 1.2052677e7*fcs_pow(w,6) - 9.4699605e7*fcs_pow(w,8) + 3.59858499e8*fcs_pow(w,10) - 6.43383377e8*fcs_pow(w,12) + 4.31278967e8*fcs_pow(w,14)) - 5.446595e6*fcs_pow(v,8)*(-39. + 13923.*w2 - 737919.*fcs_pow(w,4) + 1.3528515e7*fcs_pow(w,6) - 1.10160765e8*fcs_pow(w,8) + 4.33299009e8*fcs_pow(w,10) - 8.00946653e8*fcs_pow(w,12) + 5.54501529e8*fcs_pow(w,14))) + 2.758665e6*fcs_pow(u,6)*(1.32579447e8*fcs_pow(v,6)*(-1. + 385.*w2 - 21945.*fcs_pow(w,4) + 431585.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 1.5796011e7*fcs_pow(w,10) - 3.1113355e7*fcs_pow(w,12) + 2.2907635e7*fcs_pow(w,14)) - 2.501499e6*fcs_pow(v,4)*(-3. + 1113.*w2 - 61215.*fcs_pow(w,4) + 1.163085e6*fcs_pow(w,6) - 9.803145e6*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10) - 7.6108053e7*fcs_pow(w,12) + 5.4362895e7*fcs_pow(w,14)) - 9.4699605e7*fcs_pow(v,8)*(-11. + 4389.*w2 - 258951.*fcs_pow(w,4) + 5.265337e6*fcs_pow(w,6) - 4.7388033e7*fcs_pow(w,8) + 2.05348143e8*fcs_pow(w,10) - 4.16918957e8*fcs_pow(w,12) + 3.16125363e8*fcs_pow(w,14)) + 3.59858499e8*fcs_pow(v,10)*(-11. + 4543.*w2 - 277123.*fcs_pow(w,4) + 5.819583e6*fcs_pow(w,6) - 5.4038985e7*fcs_pow(w,8) + 2.41374133e8*fcs_pow(w,10) - 5.04691369e8*fcs_pow(w,12) + 3.93770189e8*fcs_pow(w,14)) - 11.*(-39. + 13377.*w2 - 682227.*fcs_pow(w,4) + 1.2052677e7*fcs_pow(w,6) - 9.4699605e7*fcs_pow(w,8) + 3.59858499e8*fcs_pow(w,10) - 6.43383377e8*fcs_pow(w,12) + 4.31278967e8*fcs_pow(w,14)) - 6.43383377e8*fcs_pow(v,12)*(-11. + 4697.*w2 - 295911.*fcs_pow(w,4) + 6.411405e6*fcs_pow(w,6) - 6.1366305e7*fcs_pow(w,8) + 2.82285003e8*fcs_pow(w,10) - 6.07340461e8*fcs_pow(w,12) + 4.87207183e8*fcs_pow(w,14)) + 3773.*v2*(-39. + 13923.*w2 - 737919.*fcs_pow(w,4) + 1.3528515e7*fcs_pow(w,6) - 1.10160765e8*fcs_pow(w,8) + 4.33299009e8*fcs_pow(w,10) - 8.00946653e8*fcs_pow(w,12) + 5.54501529e8*fcs_pow(w,14)) + 4.31278967e8*fcs_pow(v,14)*(-11. + 4851.*w2 - 315315.*fcs_pow(w,4) + 7.042035e6*fcs_pow(w,6) - 6.9414345e7*fcs_pow(w,8) + 3.28561233e8*fcs_pow(w,10) - 7.26817273e8*fcs_pow(w,12) + 5.99025225e8*fcs_pow(w,14))) - 1.9310655e7*fcs_pow(u,8)*(-8.117109e6*fcs_pow(v,4)*(-1. + 385.*w2 - 21945.*fcs_pow(w,4) + 431585.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 1.5796011e7*fcs_pow(w,10) - 3.1113355e7*fcs_pow(w,12) + 2.2907635e7*fcs_pow(w,14)) + 51051.*v2*(-3. + 1113.*w2 - 61215.*fcs_pow(w,4) + 1.163085e6*fcs_pow(w,6) - 9.803145e6*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10) - 7.6108053e7*fcs_pow(w,12) + 5.4362895e7*fcs_pow(w,14)) + 1.3528515e7*fcs_pow(v,6)*(-11. + 4389.*w2 - 258951.*fcs_pow(w,4) + 5.265337e6*fcs_pow(w,6) - 4.7388033e7*fcs_pow(w,8) + 2.05348143e8*fcs_pow(w,10) - 4.16918957e8*fcs_pow(w,12) + 3.16125363e8*fcs_pow(w,14)) - 1.10160765e8*fcs_pow(v,8)*(-11. + 4543.*w2 - 277123.*fcs_pow(w,4) + 5.819583e6*fcs_pow(w,6) - 5.4038985e7*fcs_pow(w,8) + 2.41374133e8*fcs_pow(w,10) - 5.04691369e8*fcs_pow(w,12) + 3.93770189e8*fcs_pow(w,14)) + 4.33299009e8*fcs_pow(v,10)*(-11. + 4697.*w2 - 295911.*fcs_pow(w,4) + 6.411405e6*fcs_pow(w,6) - 6.1366305e7*fcs_pow(w,8) + 2.82285003e8*fcs_pow(w,10) - 6.07340461e8*fcs_pow(w,12) + 4.87207183e8*fcs_pow(w,14)) - 11.*(-39. + 13923.*w2 - 737919.*fcs_pow(w,4) + 1.3528515e7*fcs_pow(w,6) - 1.10160765e8*fcs_pow(w,8) + 4.33299009e8*fcs_pow(w,10) - 8.00946653e8*fcs_pow(w,12) + 5.54501529e8*fcs_pow(w,14)) - 8.00946653e8*fcs_pow(v,12)*(-11. + 4851.*w2 - 315315.*fcs_pow(w,4) + 7.042035e6*fcs_pow(w,6) - 6.9414345e7*fcs_pow(w,8) + 3.28561233e8*fcs_pow(w,10) - 7.26817273e8*fcs_pow(w,12) + 5.99025225e8*fcs_pow(w,14)) + 1.84833843e8*fcs_pow(v,14)*(-33. + 15015.*w2 - 1.006005e6*fcs_pow(w,4) + 2.3138115e7*fcs_pow(w,6) - 2.34686595e8*fcs_pow(w,8) + 1.142141429e9*fcs_pow(w,10) - 2.595775975e9*fcs_pow(w,12) + 2.196425825e9*fcs_pow(w,14))) + 6.5656227e7*fcs_pow(u,10)*(159159.*v2*(-1. + 385.*w2 - 21945.*fcs_pow(w,4) + 431585.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 1.5796011e7*fcs_pow(w,10) - 3.1113355e7*fcs_pow(w,12) + 2.2907635e7*fcs_pow(w,14)) - 143.*(-3. + 1113.*w2 - 61215.*fcs_pow(w,4) + 1.163085e6*fcs_pow(w,6) - 9.803145e6*fcs_pow(w,8) + 3.9866123e7*fcs_pow(w,10) - 7.6108053e7*fcs_pow(w,12) + 5.4362895e7*fcs_pow(w,14)) - 795795.*fcs_pow(v,4)*(-11. + 4389.*w2 - 258951.*fcs_pow(w,4) + 5.265337e6*fcs_pow(w,6) - 4.7388033e7*fcs_pow(w,8) + 2.05348143e8*fcs_pow(w,10) - 4.16918957e8*fcs_pow(w,12) + 3.16125363e8*fcs_pow(w,14)) + 1.5120105e7*fcs_pow(v,6)*(-11. + 4543.*w2 - 277123.*fcs_pow(w,4) + 5.819583e6*fcs_pow(w,6) - 5.4038985e7*fcs_pow(w,8) + 2.41374133e8*fcs_pow(w,10) - 5.04691369e8*fcs_pow(w,12) + 3.93770189e8*fcs_pow(w,14)) - 1.27440885e8*fcs_pow(v,8)*(-11. + 4697.*w2 - 295911.*fcs_pow(w,4) + 6.411405e6*fcs_pow(w,6) - 6.1366305e7*fcs_pow(w,8) + 2.82285003e8*fcs_pow(w,10) - 6.07340461e8*fcs_pow(w,12) + 4.87207183e8*fcs_pow(w,14)) + 5.18259599e8*fcs_pow(v,10)*(-11. + 4851.*w2 - 315315.*fcs_pow(w,4) + 7.042035e6*fcs_pow(w,6) - 6.9414345e7*fcs_pow(w,8) + 3.28561233e8*fcs_pow(w,10) - 7.26817273e8*fcs_pow(w,12) + 5.99025225e8*fcs_pow(w,14)) - 3.29801563e8*fcs_pow(v,12)*(-33. + 15015.*w2 - 1.006005e6*fcs_pow(w,4) + 2.3138115e7*fcs_pow(w,6) - 2.34686595e8*fcs_pow(w,8) + 1.142141429e9*fcs_pow(w,10) - 2.595775975e9*fcs_pow(w,12) + 2.196425825e9*fcs_pow(w,14)) + 1.99330615e8*fcs_pow(v,14)*(-39. + 18291.*w2 - 1.262079e6*fcs_pow(w,4) + 2.9869203e7*fcs_pow(w,6) - 3.11493117e8*fcs_pow(w,8) + 1.557465585e9*fcs_pow(w,10) - 3.634086365e9*fcs_pow(w,12) + 3.154866185e9*fcs_pow(w,14))) - 1.159926677e9*fcs_pow(u,12)*(-39.*(-1. + 385.*w2 - 21945.*fcs_pow(w,4) + 431585.*fcs_pow(w,6) - 3.760955e6*fcs_pow(w,8) + 1.5796011e7*fcs_pow(w,10) - 3.1113355e7*fcs_pow(w,12) + 2.2907635e7*fcs_pow(w,14)) + 1365.*v2*(-11. + 4389.*w2 - 258951.*fcs_pow(w,4) + 5.265337e6*fcs_pow(w,6) - 4.7388033e7*fcs_pow(w,8) + 2.05348143e8*fcs_pow(w,10) - 4.16918957e8*fcs_pow(w,12) + 3.16125363e8*fcs_pow(w,14)) - 77805.*fcs_pow(v,4)*(-11. + 4543.*w2 - 277123.*fcs_pow(w,4) + 5.819583e6*fcs_pow(w,6) - 5.4038985e7*fcs_pow(w,8) + 2.41374133e8*fcs_pow(w,10) - 5.04691369e8*fcs_pow(w,12) + 3.93770189e8*fcs_pow(w,14)) + 1.530165e6*fcs_pow(v,6)*(-11. + 4697.*w2 - 295911.*fcs_pow(w,4) + 6.411405e6*fcs_pow(w,6) - 6.1366305e7*fcs_pow(w,8) + 2.82285003e8*fcs_pow(w,10) - 6.07340461e8*fcs_pow(w,12) + 4.87207183e8*fcs_pow(w,14)) - 1.3334295e7*fcs_pow(v,8)*(-11. + 4851.*w2 - 315315.*fcs_pow(w,4) + 7.042035e6*fcs_pow(w,6) - 6.9414345e7*fcs_pow(w,8) + 3.28561233e8*fcs_pow(w,10) - 7.26817273e8*fcs_pow(w,12) + 5.99025225e8*fcs_pow(w,14)) + 1.8668013e7*fcs_pow(v,10)*(-33. + 15015.*w2 - 1.006005e6*fcs_pow(w,4) + 2.3138115e7*fcs_pow(w,6) - 2.34686595e8*fcs_pow(w,8) + 1.142141429e9*fcs_pow(w,10) - 2.595775975e9*fcs_pow(w,12) + 2.196425825e9*fcs_pow(w,14)) - 3.1113355e7*fcs_pow(v,12)*(-39. + 18291.*w2 - 1.262079e6*fcs_pow(w,4) + 2.9869203e7*fcs_pow(w,6) - 3.11493117e8*fcs_pow(w,8) + 1.557465585e9*fcs_pow(w,10) - 3.634086365e9*fcs_pow(w,12) + 3.154866185e9*fcs_pow(w,14)) + 2.2907635e7*fcs_pow(v,14)*(-39. + 18837.*w2 - 1.337427e6*fcs_pow(w,4) + 3.2544057e7*fcs_pow(w,6) - 3.48686325e8*fcs_pow(w,8) + 1.789923135e9*fcs_pow(w,10) - 4.284967505e9*fcs_pow(w,12) + 3.814091955e9*fcs_pow(w,14))) + 6.3732235e7*fcs_pow(u,14)*(-39.*(-11. + 4389.*w2 - 258951.*fcs_pow(w,4) + 5.265337e6*fcs_pow(w,6) - 4.7388033e7*fcs_pow(w,8) + 2.05348143e8*fcs_pow(w,10) - 4.16918957e8*fcs_pow(w,12) + 3.16125363e8*fcs_pow(w,14)) + 15561.*v2*(-11. + 4543.*w2 - 277123.*fcs_pow(w,4) + 5.819583e6*fcs_pow(w,6) - 5.4038985e7*fcs_pow(w,8) + 2.41374133e8*fcs_pow(w,10) - 5.04691369e8*fcs_pow(w,12) + 3.93770189e8*fcs_pow(w,14)) - 918099.*fcs_pow(v,4)*(-11. + 4697.*w2 - 295911.*fcs_pow(w,4) + 6.411405e6*fcs_pow(w,6) - 6.1366305e7*fcs_pow(w,8) + 2.82285003e8*fcs_pow(w,10) - 6.07340461e8*fcs_pow(w,12) + 4.87207183e8*fcs_pow(w,14)) + 1.8668013e7*fcs_pow(v,6)*(-11. + 4851.*w2 - 315315.*fcs_pow(w,4) + 7.042035e6*fcs_pow(w,6) - 6.9414345e7*fcs_pow(w,8) + 3.28561233e8*fcs_pow(w,10) - 7.26817273e8*fcs_pow(w,12) + 5.99025225e8*fcs_pow(w,14)) - 5.6004039e7*fcs_pow(v,8)*(-33. + 15015.*w2 - 1.006005e6*fcs_pow(w,4) + 2.3138115e7*fcs_pow(w,6) - 2.34686595e8*fcs_pow(w,8) + 1.142141429e9*fcs_pow(w,10) - 2.595775975e9*fcs_pow(w,12) + 2.196425825e9*fcs_pow(w,14)) + 2.05348143e8*fcs_pow(v,10)*(-39. + 18291.*w2 - 1.262079e6*fcs_pow(w,4) + 2.9869203e7*fcs_pow(w,6) - 3.11493117e8*fcs_pow(w,8) + 1.557465585e9*fcs_pow(w,10) - 3.634086365e9*fcs_pow(w,12) + 3.154866185e9*fcs_pow(w,14)) - 4.16918957e8*fcs_pow(v,12)*(-39. + 18837.*w2 - 1.337427e6*fcs_pow(w,4) + 3.2544057e7*fcs_pow(w,6) - 3.48686325e8*fcs_pow(w,8) + 1.789923135e9*fcs_pow(w,10) - 4.284967505e9*fcs_pow(w,12) + 3.814091955e9*fcs_pow(w,14)) + 3.16125363e8*fcs_pow(v,14)*(-39. + 19383.*w2 - 1.414959e6*fcs_pow(w,4) + 3.5373975e7*fcs_pow(w,6) - 3.89113725e8*fcs_pow(w,8) + 2.049332285e9*fcs_pow(w,10) - 5.030179245e9*fcs_pow(w,12) + 4.587965685e9*fcs_pow(w,14))));
      }
    }
    default: return 0.0;
  }
}


